#ifdef __cplusplus
extern "C" {
#endif

#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include "3d_collider.h"

/* prototype c3d_list */
typedef struct c3d_list c3d_list;
struct c3d_list {
   #define IMPLEMENTS_C3D_LIST \
   c3d_list* next;\
   c3d_list* back;\
   c3d_list* up;\
   void (*erase)(void* self);
   IMPLEMENTS_C3D_LIST
};

/* Private creation functions. */
static void* create_item(void* with_what);

/* Prototype c3d_list functions. */
static void* c3d_list_append(void* in, void* what);
static void* c3d_list_isolate(void* in, void* what);
static void* c3d_list_remove(void* in, void* what);
static void  c3d_list_remove_all(void* in);

/* Prototype c3d_item list manipulation functions. */
static void* c3d_item_append(void* parent, void* start, void* which_one);
static void* c3d_item_locate(void* start, void* which_item);

/* c3d_domain functions. */
static void domain_erase(void* domain);
static void domain_display(void* domain);
static void domain_update(void* domain, c3d_update* info);
static void domain_add_object(c3d_domain* domain, void* what);
static void domain_add_objects(c3d_domain* domain, int number, ...);
static void domain_remove_object(c3d_domain* domain, void* what);
static void domain_add_collidable(c3d_domain* domain, void* what);
static void domain_remove_collidable(c3d_domain* domain, void* what);


/************************* creators and populators ****************************/

c3d_volume in_volume(float x, float y, float z, float r)
{
   c3d_volume v = {x,y,z,r};
   return v;
}

c3d_domain *c3d_create_domain()
{
   c3d_domain *d = (c3d_domain*)calloc(1,sizeof(c3d_domain));

   if (d) {
      d->erase             = &domain_erase;
      d->display           = &domain_display;
      d->update            = &domain_update;
      d->add_object        = &domain_add_object;
      d->add_objects       = &domain_add_objects;
      d->remove_object     = &domain_remove_object;
      d->add_collidable    = &domain_add_collidable;
      d->remove_collidable = &domain_remove_collidable;
   }
   return d;
}

c3d_element *c3d_create_element(c3d_volume at)
{
   c3d_element *e = (c3d_element*)calloc(1,sizeof(c3d_element));

   if (e) {
      e->erase = &free;
      e->at    = at;
   }
   return e;
}

void* create_item(void *with_what)
{
   c3d_item *i = (c3d_item*)calloc(1,sizeof(c3d_item));

   if (i) {
      i->erase = &free;
      i->what  = with_what;
   }
   return i;
}

/************************** non-generic erase functions************************/

void domain_erase(void *domain)
{
   c3d_domain *d = (c3d_domain*)domain;

   if (!d)
      return;

   if (d->contains)
      c3d_list_remove_all(d->contains);
   if (d->collidable)
      c3d_list_remove_all(d->collidable);
   free(d);
}

/**************************** c3d_list ****************************************/

void* c3d_list_append(void *start, void *which_one)
{
   c3d_list *at   = (c3d_list*)start;
   c3d_list *what = (c3d_list*)which_one;

   if (!what)
      return at;
   if (!at)
      return what;

   while (at->next) {
      if (at == what)   /* This item is already here... */
         return start;
      at = at->next;
   }

   at->next   = what;
   what->back = at;
   what->next = 0;
   what->up   = at->up;

   return start;
}

void* c3d_list_remove(void *start, void *which_one)
{
   c3d_list *in   = (c3d_list*)start;
   c3d_list *what = (c3d_list*)which_one;
   c3d_list *head = in;

   if (!in)
      return 0;
   if (!what)
      return head;

   if (what->back)
      (what->back)->next = what->next;
   if (what->next)
      (what->next)->back = what->back;
   if (what == head)
      head = what->next;

   what->erase(what);
   return head;
}

void* c3d_list_isolate(void *start, void *which_one)
{
   c3d_list *in   = (c3d_list*)start;
   c3d_list *what = (c3d_list*)which_one;
   c3d_list *head = in;

   if (!in)
      return 0;
   if (!what)
      return head;

   if (what->back)
      (what->back)->next = what->next;
   if (what->next)
      (what->next)->back = what->back;
   if (what == head)
      head = what->next;

   what->next = 0;
   what->back = 0;
   what->up   = 0;

   return head;
}

void c3d_list_remove_all(void *start)
{
   c3d_list *in     = (c3d_list*)start;
   c3d_list *target = in;
   c3d_list *save   = 0;

   while (target) {
      save = target->next;
      if (target->erase)
         target->erase(target);
      target = save;
   }
}

/***************************** c3d_item ***************************************/

void* c3d_item_append(void *parent, void *start, void *which_one)
{
   c3d_item *in      = (c3d_item*)start;
   c3d_element *what = (c3d_element*)which_one;
   c3d_item *create;

   if (!what)
      return in;
   if (!in ) {
      create     = create_item(what);
      create->up = parent;
      return create;
   }

   while (in->next && (in->what != what))
      in = in->next;
   if (in->what == what)
      return start;

   create       = create_item(what);
   in->next     = create;
   create->back = in;
   create->next = 0;
   create->up   = in->up;
   return start;
}

void* c3d_item_locate(void *start, void *which_item)
{
   c3d_item *in = (c3d_item*)start;
   c3d_element *what = (c3d_element*)which_item;

   if (!what || !in)
      return 0;

   while (in->next && (in->what != what))
      in = in->next;
   if (in->what == what)
      return in;
   else return 0;
}

/**************************** c3d_domain **************************************/

void domain_remove_object(c3d_domain *self, void *item)
{
   if (!self || !item)
      return;
   self->contains = c3d_list_remove(self->contains, item);
}

void domain_add_object(c3d_domain *self,void *what)
{
   c3d_element *item = (c3d_element*)what;
   if (!self || !item)
      return;

   item->container = self;
   self->contains = c3d_list_append(self->contains,item);
}

void domain_add_objects(c3d_domain *self,int how_many, ...)
{
   va_list args;
   c3d_element *item;
   int count = 0;

   if (!self || !how_many)
      return;

   va_start(args,how_many);
   while (count < how_many) {
      item = va_arg(args,c3d_element*);
      item->container = self;
      self->contains = c3d_list_append(self->contains,item);
      count++;
   }
   va_end(args);
}

void domain_remove_collidable(c3d_domain *self, void *item)
{
   c3d_item *target;
   if (!self || !item)
      return;

   target = c3d_item_locate(self->collidable,item);
   self->collidable = c3d_list_remove(self->collidable, target);
}

void domain_add_collidable(c3d_domain *self,void *what)
{
   c3d_element *item = what;
   if (!self || !item)
      return;

   self->collidable = c3d_item_append(self, self->collidable, item);
}

void domain_display(void *domain)
{
   c3d_domain *self = (c3d_domain*)domain;
   c3d_element *runner;

   if (!self || !self->contains)
      return;

   runner = self->contains;
   while (runner) {
      if (runner->display)
         runner->display(runner);
      runner = runner->more;
   }
}

void domain_update(void *domain,c3d_update *info)
{
   c3d_domain *self = (c3d_domain*)domain;
   c3d_element *update_runner;
   c3d_item *item_runner, *collision_runner;
   float vec_mag,x_item,y_item,z_item,plane,radius;
   int hit;

   if (!self || !self->contains)
      return;

   /* First, we inform all entities of current frame information */
   update_runner = self->contains;
   while (update_runner) {
      if (update_runner->update)
         update_runner->update(update_runner,info);
      update_runner = update_runner->more;
   }

   /* Next, we determine any collisions amongst the collision list items. */
   item_runner = self->collidable;
   while (item_runner) {
      if (!item_runner->what->initiator) {
         item_runner = item_runner->next;
         continue;
      }

      collision_runner = self->collidable;
      while (collision_runner) {

         if (collision_runner->what == item_runner->what) {
             collision_runner = collision_runner->next;
             continue;
         }
         hit    = 0;
         radius = item_runner->what->at.r;

         if (!collision_runner->what->at.r) {

            if (plane = collision_runner->what->at.x) {
               x_item = plane - ((plane > 0) ? radius : -radius);
               if (item_runner->what->at.x < x_item)
                  hit = 1;
            }
            if (plane = collision_runner->what->at.y) {
               y_item = plane - ((plane > 0) ? radius : -radius);
               if (item_runner->what->at.y < y_item)
                  hit = 1;
            }
            if (plane = collision_runner->what->at.z) {
               z_item = plane - ((plane > 0) ? radius : -radius);
               if (item_runner->what->at.z > z_item)
                  hit = 1;
            }
         }
         else {
            x_item  = item_runner->what->at.x - collision_runner->what->at.x;
            y_item  = item_runner->what->at.y - collision_runner->what->at.y;
            z_item  = item_runner->what->at.z - collision_runner->what->at.z;
            vec_mag = sqrt(x_item * x_item + y_item * y_item + z_item * z_item);
            if (vec_mag <= (radius + collision_runner->what->at.r))
               hit = 1;
         }

         if (hit) {
            if (item_runner->what->on_hit)
               item_runner->what->
                  on_hit(item_runner->what,collision_runner->what);
            if (collision_runner->what->on_hit)
               collision_runner->what->
                  on_hit(collision_runner->what,item_runner->what);
         }
         else {
            if (item_runner->what->no_hit)
               item_runner->what->
                  no_hit(item_runner->what,collision_runner->what);
            if (collision_runner->what->no_hit)
               collision_runner->what->
                  no_hit(collision_runner->what,item_runner->what);
         }
         collision_runner = collision_runner->next;
      }
      item_runner = item_runner->next;
   }
}

#ifdef __cplusplus
}
#endif

