#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include "aluf.h"

/* Wave file header byte offsets. */
#define WAV_POS_FORMAT     20
#define WAV_POS_CHANNELS   22
#define WAV_POS_RATE       24
#define WAV_POS_BYTE_RATE  28
#define WAV_POS_DATA_SIZE  34
#define WAV_POS_SIZE       40
#define WAV_POS_DATA       44

/* For oggvorbis and mpg123 loaders... */
#define LOCAL_BUFF_SIZE    512000
#define INIT_ALLOC_SIZE    1000000

#define MAX_THREADS        100

static t_func_t stream_run(void* src);
static unsigned setup_wav(stream_src *src);
static unsigned setup_ogg(stream_src *src);
static unsigned setup_mp3(stream_src *src);
static void read_wav_data(stream_src *src);
static void read_ogg_data(stream_src *src);
static void read_mp3_data(stream_src *src);
static void erase_wav(stream_src *src);
static void erase_ogg(stream_src *src);
static void erase_mp3(stream_src *src);

static int load_wav_file(const char* ,char** , int* ,unsigned* , int*);
static int load_ogg_file(const char* ,char** , int* ,unsigned* , int*);
static int load_mp3_file(const char* ,char** , int* ,unsigned* , int*);

static int initialized     = 0;
static int ogg_lib_loaded  = -1;
static int mpg_lib_loaded  = -1;
static unsigned active_thread[MAX_THREADS];

/****************************** openal runtime ********************************/

void aluf_init(void)
{

   ALCdevice* device;
   ALCcontext* context;

   if (initialized)
      return;

   #ifdef USING_OPENAL_RUNTIME_LD
      if (!ext_load_OpenAL()) {
         printf("Failed to load library!.\n");
         return;
      }
   #endif

   device = alcOpenDevice(NULL);
   if (!device) {
      printf("Failed to open audio device.\n");
      return;
   }

   context = alcCreateContext(device,NULL);
   if (!context) {
      printf("Failed to create audio context\n.");
      alcCloseDevice(device);
      return;
   }

   alGetError(); /* Essentially, clear the error buffer... */

   alcMakeContextCurrent(context);
   if (alGetError() != AL_NO_ERROR) {
      printf("Failed to associate audio context.\n");
      alcDestroyContext(context);
      alcCloseDevice(device);
      return;
   }

   initialized = 1;
}

int aluf_register_active_thread(void)
{
   unsigned n = 0;

   while (n < MAX_THREADS) {
      if (!active_thread[n]) {
         active_thread[n] = 1;
         break;
      }
      n++;
   }

   if (n == MAX_THREADS)
      return -1;
   return n;
}

void aluf_unregister_active_thread(int t)
{
   if (t < 0) {
      printf("Trying to unregister undefined thread serial number...\n");
      return;
   }

   active_thread[t] = 0;
}

void aluf_destroy(void)
{
   ALCcontext* context;
   ALCdevice* device;
   unsigned done = 0, n;

   if (!initialized)
      return;

   /* Be sure all of our registered audio threads have terminated... */
   while (!done) {
      n = 0;
      while (n < MAX_THREADS && !active_thread[n])
         n++;

      if (n == MAX_THREADS)
         done = 1;
      else Sleep(10);
   }

   context = alcGetCurrentContext();
   if (context) {
      device = alcGetContextsDevice(context);
      alcMakeContextCurrent(NULL);
      alcDestroyContext(context);
      if (device)
         alcCloseDevice(device);
   }

   initialized = 0;
}

unsigned int aluf_verify(ALuint source, ALuint buffer)
{
   return alIsSource(source) && buffer != AL_NONE;
}

/****************************** stream source *********************************/
#define MAX_QUEUED 3
#define QUEUE_BUFF_SZ 32000


stream_src *create_stream_source(const char *file)
{
   stream_src *obj;
   unsigned str_len, ok = 0;

   if (!initialized) {
      aluf_init();
      if (!initialized)
         return 0;
   }

   if (!file)
      return 0;

   alGetError();

   obj = (stream_src*)calloc(1,sizeof(stream_src));

   if (obj) {
      str_len = strlen(file);
      obj->file_name = (char*)calloc(str_len + 1, sizeof(char));

      if (obj->file_name)
         memcpy(obj->file_name,file,str_len * sizeof(char));

      if (strstr(file,".wav") || strstr(file,".WAV")) {
         if (ok = setup_wav(obj)) {
            obj->read  = &read_wav_data;
            obj->erase = &erase_wav;
         }
      }
      if (strstr(file,".ogg") || strstr(file,".OGG")) {
         if (ok = setup_ogg(obj)) {
            obj->read  = &read_ogg_data;
            obj->erase = &erase_ogg;
         }
      }
      if (strstr(file,".mp3") || strstr(file,".MP3")) {
         if (ok = setup_mp3(obj)) {
            obj->read  = &read_mp3_data;
            obj->erase = &erase_mp3;
         }
      }

      if (ok) {
         obj->serial_no = aluf_register_active_thread();
         alGenSources(1,&obj->src);
         obj->run = 1;
         obj->poll_time = 100;
         platform_new_thread(&stream_run,obj,&obj->tid);

         /* We need to wait for some buffers to be queued, or the call to
         alSourcePlay failes immediately, which causes this source not
         to play on scene changes... */
         while (!obj->ready && obj->run)
            Sleep(3);

         if (!obj->run)
            return 0;
         else return obj;
      }
      else {
         if (obj->file_name)
            free(obj->file_name);
         free(obj);
      }
   }

   return 0;
}

void erase_stream_source(stream_src *src)
{
   if (!src)
      return;

   src->run = 0;
}

static t_func_t stream_run(void* src)
{
   stream_src *s = (stream_src*)src;
   ALuint buffer_prop = 0, buffer[MAX_QUEUED];
   unsigned err = 0;

   while (!err && s->run) {

      alGetSourcei(s->src,AL_BUFFERS_PROCESSED,(ALint*)&buffer_prop);

      if (buffer_prop > 0) {
         alSourceUnqueueBuffers(s->src,buffer_prop,buffer);
         alDeleteBuffers(buffer_prop,buffer);
         buffer_prop = 0;
      }

      alGetSourcei(s->src,AL_BUFFERS_QUEUED,(ALint*)&buffer_prop);

      while (buffer_prop < MAX_QUEUED) {

         s->read(s);

         if (!s->buff_data || !s->buff_data_read) {
            printf("Error in stream run: no data!\n");
            err = 1;
            break;
         }

         alGetError();
         alGenBuffers(1,buffer);
         alBufferData(*buffer,s->format,s->buff_data,s->buff_data_read,s->rate);
         free(s->buff_data);
         s->buff_data = 0;
         s->buff_data_read = 0;

         if (alGetError() != AL_NO_ERROR) {
            alDeleteBuffers(1,buffer);
            printf("Problems with data from %s.\n",s->file_name);
            err = 1;
            break;
         }
         alSourceQueueBuffers(s->src,1,buffer);
         s->ready = 1;

         alGetSourcei(s->src,AL_BUFFERS_QUEUED,(ALint*)&buffer_prop);
      }
      Sleep(s->poll_time);
   }
   s->run   = 0;
   s->ready = 0;

   alSourceStop(s->src);
   alGetSourcei(s->src,AL_BUFFERS_PROCESSED,(ALint*)&buffer_prop);
   if (buffer_prop) {
      alSourceUnqueueBuffers(s->src,buffer_prop,buffer);
      alDeleteBuffers(buffer_prop,buffer);
   }
   alDeleteSources(1,&s->src);

   s->erase(s);

   platform_end_thread(0);
}

static unsigned setup_wav(stream_src *ss)
{
   FILE* wave_file_ptr;
   short audio_format = 0, bits_per_sample = 0, channels = 0;
   int byte_rate = 0;
   size_t ok;

   if (!ss)
      return 0;

   wave_file_ptr = fopen(ss->file_name,"rb");
   if (!wave_file_ptr)
      return 0;

   fseek(wave_file_ptr,WAV_POS_FORMAT,SEEK_SET);
   ok = fread(&audio_format,sizeof(char),2,wave_file_ptr);
   if (ok != 2)
      return 0;

   if (audio_format != 1) {
      fclose(wave_file_ptr);
      printf("Error: file %s unsupported format.\n", ss->file_name);
      return 0;
   }

   ok  = 0;
   ok += fread(&channels,sizeof(char),2,wave_file_ptr);
   ok += fread(&ss->rate,sizeof(char),4,wave_file_ptr);
   ok += fread(&byte_rate,sizeof(char),4,wave_file_ptr);
   fseek(wave_file_ptr,WAV_POS_DATA_SIZE,SEEK_SET);
   ok += fread(&bits_per_sample,sizeof(char),2,wave_file_ptr);

   if (ok != 12)
      return 0;

   if (!bits_per_sample && ss->rate && channels)
      bits_per_sample = byte_rate / channels / ss->rate * 8;

   switch (bits_per_sample) {
      case 8:
         ss->format = (channels == 2) ? AL_FORMAT_STEREO8 : AL_FORMAT_MONO8;
         break;
      case 16:
         ss->format = (channels == 2) ? AL_FORMAT_STEREO16 : AL_FORMAT_MONO16;
         break;
      default:
         fclose(wave_file_ptr);
         printf(
            "Error: file %s bits per sample is %d.\n",
            ss->file_name,
            bits_per_sample);
         return 0;
   }

   fseek(wave_file_ptr,WAV_POS_SIZE,SEEK_SET);
   ok = fread(&ss->data_size,sizeof(char),4,wave_file_ptr);
   if (ok != 4)
      return 0;

   ss->file_ptr = (void*)wave_file_ptr;
   return 1;
}

static unsigned setup_ogg(stream_src *ss)
{
   FILE* file_ptr;
   vorbis_info *info;
   OggVorbis_File *ogg_file;

   #ifdef USING_OGGVORBIS_RUNTIME_LD
      if (ogg_lib_loaded < 0)
         ogg_lib_loaded = ext_load_ogg_vorbis();
      if (!ogg_lib_loaded)
         return 0;
   #endif

   if (!ss || !ss->file_name)
      return 0;

   file_ptr = fopen(ss->file_name,"rb");
   if (!file_ptr) {
      printf("ogg: filename problems\n");
      return 0;
   }

   ogg_file = (OggVorbis_File*)calloc(1,sizeof(OggVorbis_File));
   if (!ogg_file) {
      printf("ogg: ogg file pointer problems\n");
      return 0;
   }

   ov_open_callbacks(file_ptr,ogg_file,NULL,0,OV_CALLBACKS_DEFAULT);
   info = ov_info(ogg_file, -1);

   if (!info) {
      printf("ogg: info struct problems\n");
      ov_clear(ogg_file);
      return 0;
   }

   ss->format = (info->channels == 1) ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16;
   ss->rate = info->rate;
   ss->file_ptr = (void*)ogg_file;
   ss->data_size = ov_pcm_total(ogg_file, -1);
   ss->total_time = ov_time_total(ogg_file, -1);
   return 1;
}

static unsigned setup_mp3(stream_src *ss)
{
   int err, channels = 0, encoding = 0;
   long this_rate = 0;
   mpg123_handle *mh;

   if (!ss || !ss->file_name)
      return 0;

   #ifdef USING_MPG123_RUNTIME_LD
      if (mpg_lib_loaded < 0)
         mpg_lib_loaded = ext_load_mpg123();
      if (!mpg_lib_loaded)
         return 0;
   #endif

   err = mpg123_init();
   if (err != MPG123_OK) {
      printf("Mpg123 error: %s\n",mpg123_plain_strerror(err));
      return 0;
   }

   mh = mpg123_new(NULL, &err);
   if (!mh) {
       printf("Mpg123 error: %s\n",mpg123_plain_strerror(err));
       mpg123_exit();
       return 0;
   }

   err = mpg123_open(mh, ss->file_name);
   if (err != MPG123_OK) {
      printf("Mpg123 error: %s\n",mpg123_strerror(mh));
      mpg123_delete(mh);
      mpg123_exit();
      return 0;
   }

   err = mpg123_getformat(mh, &this_rate, &channels, &encoding);
   if (err != MPG123_OK) {
      printf("Mpg123 error: %s\n",mpg123_strerror(mh));
      mpg123_close(mh);
      mpg123_delete(mh);
      mpg123_exit();
      return 0;
   }

   if (encoding != MPG123_ENC_SIGNED_16) {
      printf("Bad encoding: 0x%x!\n", encoding);
      mpg123_close(mh);
      mpg123_delete(mh);
      mpg123_exit();
      return 0;
   }

   mpg123_format_none(mh);
   mpg123_format(mh,this_rate,channels,encoding);

   ss->format    = (channels == 1) ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16;
   ss->rate      = this_rate;
   ss->data_size = mpg123_length(mh);
   ss->file_ptr  = (void*)mh;

   return 1;
}


static void read_wav_data(stream_src *ss)
{
   if (!ss || !ss->file_ptr || !ss->format || !ss->rate)
      return;

   /* Unprocessed data in the pipeline?? */
   if (ss->buff_data || ss->buff_data_read)
      return;

   ss->buff_data = (unsigned char*)malloc(QUEUE_BUFF_SZ * sizeof(char));
   ss->buff_data_read = fread(
      ss->buff_data,
      sizeof(char),
      QUEUE_BUFF_SZ,
      (FILE*)ss->file_ptr);

   /* Probably an EOF... */
   if (ss->buff_data_read < QUEUE_BUFF_SZ) {
      fseek((FILE*)ss->file_ptr,WAV_POS_DATA,SEEK_SET);
      ss->buff_data_read += fread(
         ss->buff_data + ss->buff_data_read,
         sizeof(char),
         QUEUE_BUFF_SZ - ss->buff_data_read,
         (FILE*)ss->file_ptr);

      if (ss->buff_data_read < QUEUE_BUFF_SZ)
         printf("Errors encountered in wav stream loader...\n");
   }
}

static void read_ogg_data(stream_src *ss)
{
   int stream, read = 0, err = 0, misses = 0;

   if (!ss || !ss->file_ptr || !ss->format || !ss->rate) {
      printf("Object problems...\n");
      return;
   }

   /* Unprocessed data in the pipeline?? */
   if (ss->buff_data || ss->buff_data_read) {
      printf("Object problems...unprocessed data.\n");
      return;
   }

   ss->buff_data = (unsigned char*)malloc(QUEUE_BUFF_SZ * sizeof(char));

   while (!err && ss->buff_data_read < QUEUE_BUFF_SZ) {

      read = ov_read(
         (OggVorbis_File*)ss->file_ptr,
         ss->buff_data + ss->buff_data_read,
         QUEUE_BUFF_SZ - ss->buff_data_read,
         0,2,1,
         &stream);

      ss->buff_data_read += read;

      if (!read) {
         ov_time_seek((OggVorbis_File*)ss->file_ptr, 0);
         misses++;
      }

      if (misses > 1)
         err = 1;
   }

   if (err)
      printf("Errors encountered in ogg stream loader...\n");
}

static void read_mp3_data(stream_src *ss)
{
   size_t read = 0;
   int err = 0, misses = 0;

   if (!ss || !ss->file_ptr || !ss->format || !ss->rate)
      return;

   /* Unprocessed data in the pipeline?? */
   if (ss->buff_data || ss->buff_data_read)
      return;

   ss->buff_data = (unsigned char*)malloc(QUEUE_BUFF_SZ * sizeof(char));

   while (ss->buff_data_read < QUEUE_BUFF_SZ) {

      err = mpg123_read(
         (mpg123_handle*)ss->file_ptr,
         ss->buff_data + ss->buff_data_read,
         QUEUE_BUFF_SZ - ss->buff_data_read,
         &read);

      ss->buff_data_read += read;

      if (err != MPG123_OK) {
         if (!read) {
            mpg123_seek((mpg123_handle*)ss->file_ptr,0,SEEK_SET);
            misses++;
         }
      }

      if (misses > 1)
         break;
   }

   if (err != MPG123_OK)
      printf("Errors encountered in mp3 stream loader...\n");
}

static void erase_wav(stream_src *ss)
{
   fclose((FILE*)ss->file_ptr);
   if (ss->file_name)
      free(ss->file_name);

   aluf_unregister_active_thread(ss->serial_no);
   free(ss);
}

static void erase_ogg(stream_src *ss)
{
   if (ss->file_ptr)
      ov_clear((OggVorbis_File*)ss->file_ptr);
   if (ss->file_name)
      free(ss->file_name);

   aluf_unregister_active_thread(ss->serial_no);
   free(ss);
}

static void erase_mp3(stream_src *ss)
{
   if (ss->file_ptr) {
      mpg123_close((mpg123_handle*)ss->file_ptr);
      mpg123_delete((mpg123_handle*)ss->file_ptr);
      mpg123_exit();
   }

   if (ss->file_name)
      free(ss->file_name);

   aluf_unregister_active_thread(ss->serial_no);
   free(ss);

}

/****************************** openal loaders ********************************/

ALuint aluf_buffer_from(const char *snd_file)
{
   char *data = 0;
   int data_size = 0, rate = 0, ok = 0;
   unsigned format = 0;
   ALuint buffer;

   if (!initialized) {
      aluf_init();
      if (!initialized)
         return AL_NONE;
   }
   if (!snd_file)
      return AL_NONE;

   alGetError();
   alGenBuffers(1,&buffer);

   if (buffer == AL_NONE) {
      alGetError();
      printf("Could not create AL buffer for file %s.\n",snd_file);
      return AL_NONE;
   }

   if (strstr(snd_file,".wav") || strstr(snd_file,".WAV"))
      ok = load_wav_file(snd_file,&data,&data_size,&format,&rate);
   if (strstr(snd_file,".ogg") || strstr(snd_file,".OGG"))
      ok = load_ogg_file(snd_file,&data,&data_size,&format,&rate);
   if (strstr(snd_file,".mp3") || strstr(snd_file,".MP3"))
      ok = load_mp3_file(snd_file,&data,&data_size,&format,&rate);

   if (ok)
      alBufferData(buffer,format,data,data_size,rate);

   if (data)
      free(data);

   if (!ok || alGetError() != AL_NO_ERROR) {
      alDeleteBuffers(1,&buffer);
      printf("Problems with data from %s.\n",snd_file);
      return AL_NONE;
   }

   return buffer;
}

/*************************** raw format loaders *******************************/

int load_wav_file(
   const char *filename,char **buff, int *buff_size,unsigned *format, int *rate)
{
   FILE* wave_file_ptr;
   unsigned int total_bytes = 0;
   short audio_format = 0;
   short bits_per_sample = 0;
   int byte_rate = 0;
   short channels = 0;
   size_t ok;

   if (!filename || !buff || !format || !rate)
      return 0;

   wave_file_ptr = fopen(filename,"rb");
   if (!wave_file_ptr)
      return 0;

   fseek(wave_file_ptr,WAV_POS_FORMAT,SEEK_SET);
   ok = fread(&audio_format,sizeof(char),2,wave_file_ptr);
   if (ok != 2)
      return 0;

   if (audio_format != 1) {
      fclose(wave_file_ptr);
      printf("Error: file %s unsupported format.\n", filename);
      return 0;
   }

   ok  = 0;
   ok += fread(&channels,sizeof(char),2,wave_file_ptr);
   ok += fread(rate,sizeof(char),4,wave_file_ptr);
   ok += fread(&byte_rate,sizeof(char),4,wave_file_ptr);

   fseek(wave_file_ptr,WAV_POS_DATA_SIZE,SEEK_SET);
   ok += fread(&bits_per_sample,sizeof(char),2,wave_file_ptr);

   if (ok != 12)
      return 0;

   if (!bits_per_sample && *rate && channels)
      bits_per_sample = byte_rate / channels / *rate * 8;

   switch (bits_per_sample) {
      case 8:
         *format = (channels == 2) ? AL_FORMAT_STEREO8 : AL_FORMAT_MONO8;
         break;
      case 16:
         *format = (channels == 2) ? AL_FORMAT_STEREO16 : AL_FORMAT_MONO16;
         break;
      default:
         fclose(wave_file_ptr);
         printf(
            "Error: file %s bits per sample is %d.\n",
            filename,
            bits_per_sample);
         return 0;
   }

   fseek(wave_file_ptr,WAV_POS_SIZE,SEEK_SET);
   ok = fread(&total_bytes,sizeof(char),4,wave_file_ptr);
   if (ok != 4)
      return 0;
   *buff = (char*)calloc(total_bytes,sizeof(char));
   *buff_size = (int)fread(*buff,sizeof(char),total_bytes,wave_file_ptr);
   fclose(wave_file_ptr);
   return *buff_size;
}

/*
  The ogg vorbis specific portions of this code were adapted from:
    "Introduction to Ogg Vorbis"
    by Anthony "TangentZ" Yuen
    at http://www.gamedev.net/reference/articles/article2031.asp
*/
int load_ogg_file(
   const char *filename,char **buff, int *buff_size,unsigned *format, int *rate)
{
   int stream, done, total_bytes;
   long read_bytes;
   char local_buff[LOCAL_BUFF_SIZE];
   FILE* ogg_file_ptr;
   vorbis_info *info;
   OggVorbis_File ogg_file;

   #ifdef USING_OGGVORBIS_RUNTIME_LD
      if (ogg_lib_loaded < 0)
         ogg_lib_loaded = ext_load_ogg_vorbis();
      if (!ogg_lib_loaded)
         return 0;
   #endif

   if (!filename || !buff || !format || !rate)
      return 0;

   ogg_file_ptr = fopen(filename,"rb");
   if (!ogg_file_ptr)
      return 0;

   *buff_size = INIT_ALLOC_SIZE;
   *buff      = (char*)calloc(*buff_size,sizeof(char));

   ov_open_callbacks(ogg_file_ptr,&ogg_file,NULL,0,OV_CALLBACKS_DEFAULT);
   info = ov_info(&ogg_file, -1);

   if (!info) {
      ov_clear(&ogg_file);
      return 0;
   }

   *format     = (info->channels == 1) ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16;
   *rate       = info->rate;
   total_bytes = 0;
   done        = 0;

   while (!done) {

      read_bytes = ov_read(&ogg_file,local_buff,LOCAL_BUFF_SIZE,0,2,1,&stream);

      if (read_bytes <= 0) {
         done = 1;
         continue;
      }
      if (read_bytes + total_bytes > *buff_size) {
         *buff_size *= 2;
         *buff = (char*)realloc(*buff, *buff_size * sizeof(char));
      }

      memcpy(*buff + total_bytes,local_buff,read_bytes * sizeof(char));
      total_bytes += read_bytes;
   }

   *buff = (char*)realloc(*buff,total_bytes * sizeof(char));
   *buff_size = total_bytes;
   ov_clear(&ogg_file);
   return total_bytes;
}

/*
  The mpg123 specific portions of this code were adapted from:
    "mgp123_to_wav.c" example written by Nicholas Humfrey from
    the mpg123 source release package;
    available at http://www.mpg123.de/download.shtml
 */
int load_mp3_file(
   const char *filename,char **buff, int *buff_size,unsigned *format, int *rate)
{
   int err, done, channels = 0, encoding = 0;
   size_t read_bytes = 0, total_bytes = 0;
   long this_rate = 0;
   unsigned char local_buff[LOCAL_BUFF_SIZE];
   mpg123_handle *mh;

   #ifdef USING_MPG123_RUNTIME_LD
      if (mpg_lib_loaded < 0)
         mpg_lib_loaded = ext_load_mpg123();
      if (!mpg_lib_loaded)
         return 0;
   #endif

   if (!filename || !buff || !format || !rate)
      return 0;

   *buff_size = INIT_ALLOC_SIZE;
   *buff = (char*)calloc(*buff_size,sizeof(char));

   err = mpg123_init();
   if (err != MPG123_OK) {
      printf("Mpg123 error: %s\n",mpg123_plain_strerror(err));
      return 0;
   }

   mh = mpg123_new(NULL, &err);
   if (!mh) {
       printf("Mpg123 error: %s\n",mpg123_plain_strerror(err));
       mpg123_exit();
       return 0;
   }

   err = mpg123_open(mh, filename);
   if (err != MPG123_OK) {
      printf("Mpg123 error: %s\n",mpg123_strerror(mh));
      mpg123_delete(mh);
      mpg123_exit();
      return 0;
   }

   err = mpg123_getformat(mh, &this_rate, &channels, &encoding);
   if (err != MPG123_OK) {
      printf("Mpg123 error: %s\n",mpg123_strerror(mh));
      mpg123_close(mh);
      mpg123_delete(mh);
      mpg123_exit();
      return 0;
   }

   if (encoding != MPG123_ENC_SIGNED_16) {
      printf("Bad encoding: 0x%x!\n", encoding);
      mpg123_close(mh);
      mpg123_delete(mh);
      mpg123_exit();
      return 0;
   }

   mpg123_format_none(mh);
   mpg123_format(mh,this_rate,channels,encoding);

   *format = (channels == 1) ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16;
   *rate   = this_rate;
   done    = 0;

   while (!done) {

      err = mpg123_read(mh,local_buff,LOCAL_BUFF_SIZE,&read_bytes);

      if (err != MPG123_OK) {
         done = 1;
         continue;
      }
      if (read_bytes + total_bytes > (size_t)*buff_size) {
         *buff_size *= 2;
         *buff = (char*)realloc(*buff, *buff_size * sizeof(char));
      }

      memcpy(*buff + total_bytes,local_buff,read_bytes * sizeof(char));
      total_bytes += read_bytes;
   }

   if (err != MPG123_DONE) {
      printf(
         "Warning: Decoding ended prematurely because: %s\n",
         err == MPG123_ERR ? mpg123_strerror(mh) : mpg123_plain_strerror(err));
   }

   *buff = (char*)realloc(*buff,total_bytes * sizeof(char));
   *buff_size = (int)total_bytes;
   mpg123_close(mh);
   mpg123_delete(mh);
   mpg123_exit();
   return (int)total_bytes;
}

#ifdef __cplusplus
}
#endif
