#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include <ctype.h>
#include "config_loader.h"

#define MAX_ITEM_BUFFER 128

typedef struct {
	char *buffer;
	unsigned sz;
	unsigned at;
}
stream_buffer;


static void get_value(config_record *cr, char *input);
static void config_apply(config_loader *cl, char *buffer);
static int sb_append(stream_buffer *sb, char *string);
static char *get_file_contents(char *file_name);
static char *get_config_string(char *buffer, char *item);
static void erase_config_annotations(config_annotation *ann);


/********************************* config_loader ******************************/

config_loader *create_config_loader(char *name)
{
	config_loader *cl = (config_loader*)calloc(1,sizeof(config_loader));

	if (cl && name)
		cl->name = strdup(name);

	return cl;
}


void erase_config_loader(config_loader *cl)
{
	config_record *cr, *tmp;

	if (!cl)
		return;

	if (cl->name)
		free(cl->name);

	erase_config_annotations(cl->doc);

	cr = cl->list;

	while (cr) {
		if (cr->name)
			free(cr->name);
		erase_config_annotations(cr->doc);
		tmp = cr->next;
		free(cr);
		cr = tmp;
	}

	free(cl);
}


void config_load(config_loader *cl,char *file_name)
{
  	FILE *file;
  	unsigned n, ret;
  	struct stat info;
  	char *buffer;


  	if (!cl) {
	  	printf("Bad config_loader object parameter.\n");
	  	return;
  	}

  	if	(!file_name) {
    	printf("Invalid file loading parameters.\n");
    	return;
  	}

  	/* Read the entire text file into memory... */
   if (file_name) {

      ret = stat(file_name,&info);

      if (!ret && (file = fopen(file_name,"rb"))) {

         buffer = (char*)calloc(info.st_size + 1,sizeof(char));

         if (buffer)
            ret = fread(buffer,sizeof(char),info.st_size,file);
         if (ret != info.st_size)
            printf("fread size does not match...\n");
         fclose(file);
      }
      else {
	      printf("could not open file %s\n",file_name);
	      return;
      }
   }

	config_apply(cl,buffer);
	free(buffer);
}


void config_read(config_loader *cl, char *string)
{
	char *buffer;
	unsigned sz;

	if (!cl || !string)
		return;

	sz = strlen(string);
	if (!sz)
		return;

	buffer = (char*)calloc(sz + 1,sizeof(char));
	if (buffer) {
  		strncpy(buffer,string,sz);
  		config_apply(cl,buffer);
		free(buffer);
	}
}


static void config_apply(config_loader *cl, char *buffer)
{
  	char *buff_end, *read_name, *read_pos = buffer;
  	unsigned done;
  	config_record *cr;

  	if (!cl || !buffer)
  		return;

  	if (!cl->list)
  		return;

  	buff_end = buffer + strlen(buffer);


  	while (read_pos != buff_end) {

    	/* Find the first valid character and note the location... */
    	while (read_pos != buff_end
      		&& *read_pos != '#'
          	&& (*read_pos < 'A' || *read_pos > 'z'
             	&& (*read_pos == '[' || *read_pos == '\\' || *read_pos == ']'
                	|| *read_pos == '^' || *read_pos == '`')))
      	read_pos++;

    	if	(read_pos == buff_end)
    		continue;

   	/* We allow script-like comments...so skip this line...*/
    	if	(*read_pos == '#') {
      	while (read_pos != buff_end && *read_pos != '\n')
      		read_pos++;
     		continue;
    	}

    	read_name = read_pos;
    	done = 0;

    	/* Find the full variable name string by finding the seperator ... */
    	while	(read_pos != buff_end
      		&& !done
          	&& !(*read_pos == ' ' || *read_pos == '=')) {

	      if (*read_pos == '\r' || *read_pos == '\n') {
    			printf("Empty value field for '%s', skipping.\n",read_name);
       		done = 1;
       		continue;
    		}
     		read_pos++;
  		}

  		if (done)
  			continue;

    	if (read_pos == buff_end)
    		continue;

    	*read_pos = 0;

    	/* Find the assignment value... */
    	read_pos++;
    	while (read_pos != buff_end  && (*read_pos == ' ' || *read_pos == '='
    			|| *read_pos == '\r' || *read_pos == '\n'))
      	read_pos++;

    	if	(read_pos == buff_end)
    		continue;

    	/* See if this name matches a var_name table entry. */
    	done = 0;
    	cr = cl->list;

    	while (!done && cr) {
	    	if	(!strcmp(read_name,cr->name)) {
       		get_value(cr,read_pos);
        		while (read_pos != buff_end
        				&& !(*read_pos == '\r' || *read_pos == '\n'))
          		read_pos++;
        		done = 1;
        		continue;
      	}
      	cr = cr->next;
   	}
  	}
}


static void get_value(config_record *cr, char *ptr)
{
  	char lsn, msn;
  	unsigned n = 0, count;

  	if (!cr || !ptr)
  		return;

  	switch (cr->format) {

    	case CONFIG_STRING:
      	memcpy((char*)cr->loc,ptr,cr->bytes - 1);
      	*(((char*)cr->loc) + cr->bytes) = 0;
      	break;

      case CONFIG_BYTE:
     		while (n < cr->bytes && *ptr != '\n' && *ptr != '\r' && *ptr != 0) {
	        	msn = *ptr - '0';
	        	if	(msn > 15)
	        		msn -= 7;
	        	lsn = *(ptr + 1) - '0';
	        	if	(lsn > 15)
	        		lsn -= 7;
	        	*(((char*)cr->loc) + n) = ((msn << 4) & 240) | (lsn & 15);
	        	ptr += 3;
	        	n++;
      	}
      	break;

    	case CONFIG_UINT:
    		count = cr->bytes/sizeof(unsigned);
    		while (n < count) {
      		sscanf(ptr,"%u",(unsigned int*)cr->loc + n++);
      		while (*ptr != ' ' && *ptr != '\n')
      			ptr++;
      		ptr++;
   		}
      	break;

    	case CONFIG_INT:
    		count = cr->bytes/sizeof(int);
    		while (n < count) {
      		sscanf(ptr,"%d",(int*)cr->loc + n++);
      		while (*ptr != ' ' && *ptr != '\n')
      			ptr++;
      		ptr++;
   		}
      	break;

    	case CONFIG_FLOAT:
    		count = cr->bytes/sizeof(float);
    		while (n < count) {
      		sscanf(ptr,"%f",(float*)cr->loc + n++);
      		while (*ptr != ' ' && *ptr != '\n')
      			ptr++;
      		ptr++;
   		}
      	break;

    	case CONFIG_DOUBLE:
    		count = cr->bytes/sizeof(double);
    		while (n < count) {
      		sscanf(ptr,"%lf",(double*)cr->loc + n++);
      		while (*ptr != ' ' && *ptr != '\n')
      			ptr++;
      		ptr++;
   		}
      	break;

    	default:
      	printf("Format type unknown for '%s': skipping.\n", cr->name);
      	break;
	}
}


config_record *config_bind(config_loader *cl, char *name, int format, int bytes,
																						 void *target)
{
	config_record *cr, *tmp;
	char *new_name;

  	if	(!cl || !name || !(*name) || format < 1 || bytes < 1 || !target) {
    	printf("Invalid variable map parameters.\n");
    	return 0;
  	}

  	tmp = (config_record*)calloc(1,sizeof(config_record));
  	if (!tmp) {
	  	printf("New variable record allocation failure for %s\n",name);
	  	return 0;
  	}

  	new_name = strdup(name);
  	if (!new_name) {
	  	printf("New name allocation failure for %s\n",name);
	  	free(tmp);
	  	return 0;
  	}

  	tmp->name   = new_name;
  	tmp->format = format;
  	tmp->bytes  = bytes;
  	tmp->loc    = target;

  	if (!cl->list)
  		cl->list = tmp;
  	else {
	  	cr = cl->list;
	  	while (cr->next)
	  		cr = cr->next;
	  	cr->next = tmp;
  	}

  	return tmp;
}


char *config_to_new_string(config_loader *cl)
{
	#define MAX_ITEM_LEN		1024
	stream_buffer buff = {0,0,0}, *sb = &buff;
  	config_record *cr;
  	config_annotation *ann;
  	unsigned count, n;
  	char item[MAX_ITEM_LEN], msn, lsn;

  	if	(!cl) {
     	printf("Invalid save cfg parameters.\n");
     	return 0;
  	}

  	if (!cl->list) {
	  	printf("No values to save.\n");
  		return 0;
	}

	ann = cl->doc;
  	while (ann) {
	  	snprintf(item,MAX_ITEM_LEN,"\t# %s\n",ann->item);
	  	sb_append(sb,item);
	  	ann = ann->next;
	  	if (!ann)
	  		sb_append(sb,"\n");
  	}

	cr = cl->list;

	while (cr) {

	  	n   = 0;
	  	ann = cr->doc;

	  	if (ann)
	 		sb_append(sb,"\n");

	  	while (ann) {
		  	snprintf(item,MAX_ITEM_LEN,"\t# %s\n",ann->item);
		  	sb_append(sb,item);
		  	ann = ann->next;
	  	}

	 	switch (cr->format) {

	 		case CONFIG_STRING:
	 			snprintf(item,MAX_ITEM_LEN,"\t%s = ",cr->name);
	 			sb_append(sb,item);
	 			sb_append(sb,(char*)cr->loc);
	 			sb_append(sb,"\n");
	 			if (cr->doc)
	 				sb_append(sb,"\n");
	    		break;

	    	case CONFIG_BYTE:
	    		snprintf(item,MAX_ITEM_LEN,"\t%s = ",cr->name);
	 			sb_append(sb,item);
	    		while	(n < cr->bytes) {
	      		msn = '0' + (((*(((char*)cr->loc) + n)) >> 4) & 15);
	      		if	(msn > '9')
	      			msn += 7;
	      		lsn = '0' + ((*(((char*)cr->loc) + n)) & 15);
	      		if	(lsn > '9')
	      			lsn += 7;
	      		sprintf(item,"%c%c ",msn,lsn);
	 				sb_append(sb,item);
	      		n++;
	    		}
	    		sb_append(sb,"\n");
	    		if (cr->doc)
	 				sb_append(sb,"\n");
	    		break;

	 		case CONFIG_INT:
	 			count = cr->bytes/sizeof(int);
	 			snprintf(item,MAX_ITEM_LEN,"\t%s =",cr->name);
	 			sb_append(sb,item);
	 			while (n < count) {
		 			sprintf(item," %d",*(((int*)cr->loc) + n++));
		 			sb_append(sb,item);
	    		}
	      	sb_append(sb,"\n");
	      	if (cr->doc)
	 				sb_append(sb,"\n");
	    		break;

	 		case CONFIG_UINT:
	 			count = cr->bytes/sizeof(unsigned);
	 			snprintf(item,MAX_ITEM_LEN,"\t%s =",cr->name);
	 			sb_append(sb,item);
	 			while (n < count) {
		 			sprintf(item," %u",*(((unsigned*)cr->loc) + n++));
		 			sb_append(sb,item);
	    		}
	      	sb_append(sb,"\n");
	      	if (cr->doc)
	 				sb_append(sb,"\n");
	    		break;

	 		case CONFIG_FLOAT:
	 			count = cr->bytes/sizeof(float);
	 			snprintf(item,MAX_ITEM_LEN,"\t%s =",cr->name);
	 			sb_append(sb,item);
	 			while (n < count) {
		 			sprintf(item," %f",*(((float*)cr->loc) + n++));
		 			sb_append(sb,item);
	    		}
	      	sb_append(sb,"\n");
	      	if (cr->doc)
	 				sb_append(sb,"\n");
	    		break;

	 		case CONFIG_DOUBLE:
	 			count = cr->bytes/sizeof(double);
	 			snprintf(item,MAX_ITEM_LEN,"\t%s =",cr->name);
	 			sb_append(sb,item);
	 			while (n < count) {
		 			sprintf(item," %f",*(((double*)cr->loc) + n++));
		 			sb_append(sb,item);
	    		}
	      	sb_append(sb,"\n");
	      	if (cr->doc)
	 				sb_append(sb,"\n");
	    		break;

	 		default:
	    		printf("Variable %s, unknown format: skipping.\n",cr->name);
	    		break;
		}
		cr = cr->next;
	}

  	return sb->buffer;
}


static int sb_append(stream_buffer *sb, char *string)
{
	unsigned str_len, total_len;

	if (!sb || !string || !(*string))
		return 0;

	str_len   = strlen(string);
	total_len = sb->at + str_len + 1;

	if (total_len > sb->sz) {

		sb->buffer = (char*)realloc(sb->buffer,2 * total_len);
		sb->sz 	  = 2 * total_len;

		if (!sb->buffer) {
			printf("Buffer resize failed.\n");
			return 0;
		}
	}

	memcpy(sb->buffer + sb->at,string,str_len);
	sb->at += str_len;
	*(sb->buffer + sb->at) = 0;

	return str_len;
}


void config_save(config_loader *cl, char *file_name)
{
	FILE *save_file;
	char *string;

	if	(!file_name || !cl) {
    	printf("Invalid save cfg parameters.\n");
    	return;
  	}

  	string = config_to_new_string(cl);

  	if (!string)
  		return;

  	if	((save_file = fopen(file_name,"w")) != 0) {
	  	fprintf(save_file,"%s",string);
	  	fclose(save_file);
  	}
  	else
  		printf("Could not write config file %s.\n",file_name);

  	free(string);
}

/**************************** config_aggregate ********************************/


static char *get_file_contents(char *file_name)
{
	FILE *file;
  	unsigned ret;
  	struct stat info;
  	char *buffer = 0;

  	/* Read the entire text file into memory... */
   if (file_name) {

      ret = stat(file_name,&info);

      if (!ret && (file = fopen(file_name,"rb"))) {

         buffer = (char*)calloc(info.st_size + 1,sizeof(char));

         if (buffer)
            ret = fread(buffer,sizeof(char),info.st_size,file);
         else {
	         printf("Aggregate file buffer allocation failed...\n");
	         fclose(file);
	         return 0;
         }
         if (ret != info.st_size)
            printf("fread size does not match...\n");
         fclose(file);
      }
      else
	      printf("File %s does not exist...\n",file_name);
   }
   else
	   printf("Missing aggregate file name.\n");

	return buffer;
}


static char *get_config_string(char *buffer, char *item)
{

	char tmp[MAX_ITEM_BUFFER], *begin, *end, *ret;


	if (!buffer || !(*buffer) || !item || !(*item))
		return 0;

	snprintf(tmp,MAX_ITEM_BUFFER,"[%s]",item);

	begin = strstr(buffer,tmp);

	if (begin) {

		while (!isspace(*begin))
			begin++;
		while (*begin == '\r' || *begin == '\n')
			begin++;

		snprintf(tmp,MAX_ITEM_BUFFER,"[/%s]",item);
		end = strstr(begin,tmp);

		if (end) {
			ret = (char*)calloc(end - begin + 1,sizeof(char));
			if (ret)
				memcpy(ret,begin,end - begin);
			return ret;
		}
		else
			return 0;
	}
	else return 0;
}


config_aggregate *create_config_aggregate(char *name)
{
	config_aggregate *ca = (config_aggregate*)calloc(1,sizeof(config_aggregate));

	if (ca && name)
		ca->name = strdup(name);

	return ca;
}


void erase_config_aggregate(config_aggregate *ca)
{
	config_loader *cl, *tmp;

	if (!ca)
		return;

	if (ca->name)
		free(ca->name);

	erase_config_annotations(ca->doc);

	cl = ca->list;
	while (cl) {
		tmp = cl->link;
		erase_config_loader(cl);
		cl = tmp;
	}

	free(ca);
}


void aggregate_register_config(config_aggregate *ca, config_loader *cl)
{
	config_loader *ptr;

	if (!ca || !cl)
		return;

	if (!ca->list)
		ca->list = cl;
	else {
		ptr = ca->list;
		while (ptr->link)
			ptr = ptr->link;

		ptr->link = cl;
	}
}


void aggregate_save_config(config_aggregate *ca, config_loader *cl)
{
	config_loader *ptr;
	char *buffer, *config = 0;
	FILE *out;
	config_annotation *ann;

	if (!ca)
		return;

	ptr = ca->list;

	if (!ptr)
		return;

	buffer = get_file_contents(ca->name);

	if	(out = fopen(ca->name,"wb")) {

		ann = ca->doc;
		while (ann) {
			fprintf(out,"# %s\n",ann->item);
			ann = ann->next;
			if (!ann)
				fprintf(out,"\n");
		}

  		while (ptr) {

	  		if (!cl || ptr == cl)
		  		config = config_to_new_string(ptr);
		  	else if (buffer)
				config = get_config_string(buffer,ptr->name);
			else
				config = 0;

			if (!config)
 				config = config_to_new_string(ptr);

			if (config) {
				fprintf(out,"[%s]\n%s[/%s]\n\n",ptr->name,config,ptr->name);
				free(config);
			}

			ptr = ptr->link;
		}
  		fclose(out);
	}

  	else
  			printf("Could not write config file %s.\n",ca->name);

  	if (buffer)
  		free(buffer);
}


void aggregate_load_config(config_aggregate *ca, config_loader *cl)
{
	char *buffer, *config = 0;

	if (!ca || !cl)
		return;

	buffer = get_file_contents(ca->name);

	if (buffer) {
		config = get_config_string(buffer,cl->name);
		free(buffer);
	}

	if (config) {
		config_read(cl,config);
		free(config);
	}
}


void config_doc(void *obj,char *line_comment)
{
	config_item *item = (config_item*)obj;
	config_annotation *ann, *tmp;

	if (!item || !line_comment || !(*line_comment))
		return;

	tmp = calloc(1,sizeof(config_annotation));
	if (!tmp)
		return;

	tmp->item = strdup(line_comment);
	if (!tmp->item) {
		free(tmp);
		return;
	}

	if (!item->doc)
		item->doc = tmp;
	else {
		ann = item->doc;
		while (ann->next)
			ann = ann->next;
		ann->next = tmp;
	}
}

static void erase_config_annotations(config_annotation *ann)
{
	config_annotation *tmp;

	while (ann) {
		tmp = ann->next;
		free(ann->item);
		free(ann);
		ann = tmp;
	}
}


#ifdef __cplusplus
}
#endif
