#ifndef __DISPLAY_H__
#define __DISPLAY_H__

#include "matrix44.h"

#ifdef __cplusplus
extern "C" {
#endif

/* All display systems will implement this api... */

typedef void void_callback(void);

typedef struct {

   /* Capabilities which may not be supported by all display systems... */
   unsigned int   stereo_modes;
   unsigned int   current_mode;
   char           (*stereo_mode_names)[30]; /* Don't like this... change */
   float          eye_seperation;
   float          focal_plane;
   float          clipping_plane;

   /* Real buffer swapped frames per second. */
   unsigned int   fps;

   /* Elapsed time between each frame update call */
   float          dt;

   /* Current projection matrix */
   mat44          projection_matrix;

   /* overlay_area[4] format (in gl world coordinates/dimensions):
         [0] lower left x,
         [1] lower left y,
         [2] width,
         [3] height,
         [4] z depth  */
   float          overlay_area[5];

   /* current window dimensions in pixels:
         [0] width;
         [1] height  */
   unsigned       window_pixels[2];

   /* rotation[3] format (in absolute radians):
         [0] gl x axis rotation,
         [1] gl y axis rotation,
         [2] gl z axis rotation  */
   float          *rotation;

   /* position[3] format (in absolute gl world coordinates):
         [0] gl x value,
         [1] gl y value,
         [2] gl z value  */
   float          *position;

   /* pointer[2] format (in screen relative pixels):

         [0] dx,
         [1] dy  */
   int            pointer[2];

   /* A bitfield variable manipulated with the "ctl" macros below. */
   unsigned int   button;

   /* The raw data receieved by the display character input device. */
   int            key;
}
display_properties;

/* Exposed state/capability structure which is intended to be READ ONLY. */
extern display_properties frame_state;

/* Basic button bitset encoding/decoding macro functions... */
#define ctl_button(num,input) ((1<<(num)) & (input))
#define ctl_button_set(num,input) ((input) |= (1<<(num)))
#define ctl_button_unset(num,input) ((input) &= ~(1<<(num)))
#define ctl_mask_button(num,input) ((input) & ~(1<<(num)))

/* Optional display variable setter function pointers. A null function pointer
   indicates that this functionality is not supported... */
extern void (*display_set_stereo_mode)(unsigned int mode_number);
extern void (*display_set_clip_plane)(float gl_z_value);
extern void (*display_set_eye_seperation)(float value);
extern void (*display_set_focal_plane)(float gl_z_value);
extern void (*display_set_movement_acceleration)(float value);

/* Callback function asignment fuctions... */
void display_set_context_renew_function(void_callback *ptr);
void display_set_preframe_function(void_callback *ptr);
void display_set_display_function(void_callback *ptr);

/* Pre-run function which will define configurable values for "config_loader",
   load any external display configuration files. */
void display_prepare(int argc, char **argv);

/* Program main run loop which will invoke defined callback functions; will not
   terminate unless "display_shutdown" is called. */
void display_run_loop(void);

/* Shutdown method */
void display_shutdown(void);

#ifdef __cplusplus
}
#endif

#endif /* __DISPLAY_H__ */