/*
   This file is intended to be a common area to "hide" dll loading
   code for all used libraries. It is expected to grow as more
   functions are added with time.
*/


#ifdef __cplusplus
   extern "C" {
#endif




/***************************** END PLATFORMS **********************************//*************************** BEGIN PLATFORMS **********************************/
#ifndef _CRT_SECURE_NO_WARNINGS
   #define _CRT_SECURE_NO_WARNINGS
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if defined(WIN32) || defined(_WIN32) || defined(WIN64)
   /* Use Widows API... */
   #define _USE_MSVC
   #define WIN32_LEAN_AND_MEAN
   #include <windows.h>
   #define AL_LIB_NAME           "OpenAL32.dll"
   #define MPG123_LIB_NAME       "libmpg123-0.dll"
   #define OGG_VORB_LIB_NAME     "vorbisfile.dll"
   #define FREEIMAGE_LIB_NAME    "FreeImage.dll"
   #define REVEL_LIB_NAME        "revel.dll"
   #define GLFW_LIB_NAME         "glfw.dll"
   #define VAR_TO_USE            "PATH"

   #define snprintf              _snprintf
   #define putenv                _putenv
   #define load_lib              LoadLibrary
   #define free_lib              FreeLibrary
   #define lib_ref_t             HINSTANCE
   #define load_func             GetProcAddress

#else
   /* ...use Linux/POSIX API. */
   #include <dlfcn.h>
   #define AL_LIB_NAME           "libopenal.so"
   #define MPG123_LIB_NAME       "libmpg123.so"
   #define OGG_VORB_LIB_NAME     "libvorbisfile.so"
   #define FREEIMAGE_LIB_NAME    "libfreeimage.so"
   #define REVEL_LIB_NAME        "librevel.so"
   #define GLFW_LIB_NAME         "libglfw.so"
   #define VAR_TO_USE            "LD_LIBRARY_PATH"

   #define load_lib(x)           dlopen(x,RTLD_LAZY)
   #define free_lib              !dlclose
   #define lib_ref_t             void*
   #define load_func             dlsym
#endif

#define USING_OPENAL_RUNTIME_LD
#define USING_MPG123_RUNTIME_LD
#define USING_OGGVORBIS_RUNTIME_LD
#define USING_FREEIMAGE_RUNTIME_LD
#define USING_REVEL_RUNTIME_LD
#define USING_GLFW_RUNTIME_LD
#include "ext_libs.h"

/***************************** BEGIN COMMON ***********************************/
static   unsigned int         path_added   = 0;
static   lib_ref_t            glfw_lib     = 0;
static   lib_ref_t            al_lib       = 0;
static   lib_ref_t            mpg123_lib   = 0;
static   lib_ref_t            ogg_vorb_lib = 0;
static   lib_ref_t            free_img_lib = 0;
static   lib_ref_t            revel_lib    = 0;

static lib_ref_t get_lib(char* name)
{
   lib_ref_t ret = 0;

   #ifndef _USE_MSVC
      if (name) {
         size_t length = strlen(DLL_PATH_ADD) + strlen(name) + 2;
         char *path    = (char*)malloc(length * sizeof(char));

         snprintf(path,length,"%s/%s",DLL_PATH_ADD,name);
         //printf("load path: --%s--\n",path)
         ret = load_lib(path);
         free(path);
      }
   #else
      ret = load_lib(name);
   #endif

   if (!ret) printf("**** Could not load library %s ****\n",name);
   return ret;
}

/* We are not expecting any null symbols, so this error check is OK... */
static void* get_func(lib_ref_t lib, char* libname, char* fname)
{
   void* ret = load_func(lib,fname);
   if (!ret) printf("**** Populating %s for %s failed. ****\n",fname,libname);
   return ret;
}

static void path_add(void)
{
   #ifdef DLL_PATH_ADD
      char *path, *new_val;
      size_t length, p_length = 0;

      if (path_added) return;

      path = getenv(VAR_TO_USE);
      if (path)
         p_length = strlen(path);
      length  = p_length + strlen(DLL_PATH_ADD) + strlen(VAR_TO_USE) + 2;
      new_val = (char*)malloc(length * sizeof(char));

      snprintf(new_val,length,"%s=%s%s",VAR_TO_USE,DLL_PATH_ADD,path);
      putenv(new_val);
      free(new_val);
      path_added = 1;
   #endif
}

void ext_libs_unload(void)
{
   if (glfw_lib) {
      printf("GLFW is loaded...");
      if (free_lib(glfw_lib))
         printf("unloaded.\n");
      else printf("unload failed.\n");
   }
   if (al_lib) {
      printf("Open AL is loaded...");
      if (free_lib(al_lib))
         printf("unloaded.\n");
      else printf("unload failed.\n");
   }
   if (mpg123_lib) {
      printf("Mpg123 is loaded...");
      if (free_lib(mpg123_lib))
         printf("unloaded.\n");
      else printf("unload failed.\n");
   }
   if (ogg_vorb_lib) {
      printf("Ogg Vorbis is loaded...");
      if (free_lib(ogg_vorb_lib))
         printf("unloaded.\n");
      else printf("unload failed.\n");
   }
   if (free_img_lib) {
      printf("FreeImage is loaded...");
      if (free_lib(free_img_lib))
         printf("unloaded.\n");
      else printf("unload failed.\n");
   }
   if (revel_lib) {
      printf("Revel is loaded...");
      if (free_lib(revel_lib))
         printf("unloaded.\n");
      else printf("unload failed.\n");
   }

   glfw_lib     = 0;
   al_lib       = 0;
   mpg123_lib   = 0;
   ogg_vorb_lib = 0;
   free_img_lib = 0;
   revel_lib    = 0;
   path_added   = 0;

   printf("Done unloading libraries...\n");
}
/****************************** END COMMON ************************************/


/******************************* BEGIN GLFW ***********************************/
PGLFWINIT                              glfwInit;
PGLFWTERMINATE                         glfwTerminate;
PGLFWGETVIDEOMODES                     glfwGetVideoModes;
PGLFWGETDESKTOPMODE                    glfwGetDesktopMode;
PGLFWOPENWINDOWHINT                    glfwOpenWindowHint;
PGLFWOPENWINDOW                        glfwOpenWindow;
PGLFWSWAPBUFFERS                       glfwSwapBuffers;
PGLFWENABLE                            glfwEnable;
PGLFWDISABLE                           glfwDisable;
PGLFWSETWINDOWPOS                      glfwSetWindowPos;
PGLFWSETWINDOWTITLE                    glfwSetWindowTitle;
PGLFWSETWINDOWCLOSECALLBACK            glfwSetWindowCloseCallback;
PGLFWSETWINDOWSIZECALLBACK             glfwSetWindowSizeCallback;
PGLFWSETKEYCALLBACK                    glfwSetKeyCallback;
PGLFWSETMOUSEBUTTONCALLBACK            glfwSetMouseButtonCallback;
PGLFWSETMOUSEWHEELCALLBACK             glfwSetMouseWheelCallback;
PGLFWGETWINDOWSIZE                     glfwGetWindowSize;
PGLFWGETMOUSEPOS                       glfwGetMousePos;
PGLFWGETKEY                            glfwGetKey;
PGLFWGETMOUSEBUTTON                    glfwGetMouseButton;

unsigned int ext_load_glfw(void)
{
   unsigned int ok = 1;

   if (glfw_lib)
      return 1;

   path_add();
   glfw_lib = get_lib(GLFW_LIB_NAME);
   if (!glfw_lib) return 0;

   #define GB_GW(t,v) ok &= (v =(t)get_func(glfw_lib,GLFW_LIB_NAME,#v))? 1 : 0
   GB_GW(PGLFWINIT,glfwInit);
   GB_GW(PGLFWTERMINATE,glfwTerminate);
   GB_GW(PGLFWGETVIDEOMODES,glfwGetVideoModes);
   GB_GW(PGLFWGETDESKTOPMODE,glfwGetDesktopMode);
   GB_GW(PGLFWOPENWINDOWHINT,glfwOpenWindowHint);
   GB_GW(PGLFWOPENWINDOW,glfwOpenWindow);
   GB_GW(PGLFWSWAPBUFFERS,glfwSwapBuffers);
   GB_GW(PGLFWENABLE,glfwEnable);
   GB_GW(PGLFWDISABLE,glfwDisable);
   GB_GW(PGLFWSETWINDOWPOS,glfwSetWindowPos);
   GB_GW(PGLFWSETWINDOWTITLE,glfwSetWindowTitle);
   GB_GW(PGLFWSETWINDOWCLOSECALLBACK,glfwSetWindowCloseCallback);
   GB_GW(PGLFWSETWINDOWSIZECALLBACK,glfwSetWindowSizeCallback);
   GB_GW(PGLFWSETKEYCALLBACK,glfwSetKeyCallback);
   GB_GW(PGLFWSETMOUSEBUTTONCALLBACK,glfwSetMouseButtonCallback);
   GB_GW(PGLFWSETMOUSEWHEELCALLBACK,glfwSetMouseWheelCallback);
   GB_GW(PGLFWGETWINDOWSIZE,glfwGetWindowSize);
   GB_GW(PGLFWGETMOUSEPOS,glfwGetMousePos);
   GB_GW(PGLFWGETKEY,glfwGetKey);
   GB_GW(PGLFWGETMOUSEBUTTON,glfwGetMouseButton);

   return ok;
}
/****************************** END GLFW **************************************/



/***************************** BEGIN OPENAL ***********************************/
/* Nice dynamic load setup in OpenAL, except for alc.h...                     */
LPALCOPENDEVICE                           alcOpenDevice;
LPALCCREATECONTEXT                        alcCreateContext;
LPALCMAKECONTEXTCURRENT                   alcMakeContextCurrent;
LPALCDESTROYCONTEXT                       alcDestroyContext;
LPALCCLOSEDEVICE                          alcCloseDevice;
LPALCGETCURRENTCONTEXT                    alcGetCurrentContext;
LPALCGETCONTEXTSDEVICE                    alcGetContextsDevice;
LPALCGETSTRING                            alcGetString;
LPALCCAPTUREOPENDEVICE                    alcCaptureOpenDevice;
LPALCCAPTURESTART                         alcCaptureStart;
LPALCCAPTURESTOP                          alcCaptureStop;
LPALCGETINTEGERV                          alcGetIntegerv;
LPALCCAPTURESAMPLES                       alcCaptureSamples;
LPALCCAPTURECLOSEDEVICE                   alcCaptureCloseDevice;
LPALGETERROR                              alGetError;
LPALGENSOURCES                            alGenSources;
LPALGENBUFFERS                            alGenBuffers;
LPALSOURCEF                               alSourcef;
LPALSOURCEI                               alSourcei;
LPALGETSOURCEI                            alGetSourcei;
LPALISSOURCE                              alIsSource;
LPALSOURCESTOP                            alSourceStop;
LPALSOURCEPLAY                            alSourcePlay;
LPALSOURCEPAUSE                           alSourcePause;
LPALBUFFERDATA                            alBufferData;
LPALDELETEBUFFERS                         alDeleteBuffers;
LPALDELETESOURCES                         alDeleteSources;
LPALSOURCEQUEUEBUFFERS                    alSourceQueueBuffers;
LPALSOURCEUNQUEUEBUFFERS                  alSourceUnqueueBuffers;

unsigned int ext_load_OpenAL(void)
{
   unsigned int ok = 1;

   if (al_lib)
      return 1;

   path_add();

   al_lib = get_lib(AL_LIB_NAME);
   if (!al_lib) return 0;

   #define GB_AL(t,v) ok &= (v = (t)get_func(al_lib,AL_LIB_NAME,#v))? 1 :0
   GB_AL(LPALCOPENDEVICE,alcOpenDevice);
   GB_AL(LPALCCREATECONTEXT,alcCreateContext);
   GB_AL(LPALCMAKECONTEXTCURRENT,alcMakeContextCurrent);
   GB_AL(LPALCDESTROYCONTEXT,alcDestroyContext);
   GB_AL(LPALCCLOSEDEVICE,alcCloseDevice);
   GB_AL(LPALCGETCURRENTCONTEXT,alcGetCurrentContext);
   GB_AL(LPALCGETCONTEXTSDEVICE,alcGetContextsDevice);
   GB_AL(LPALCGETSTRING,alcGetString);
   GB_AL(LPALCCAPTUREOPENDEVICE,alcCaptureOpenDevice);
   GB_AL(LPALCCAPTURESTART,alcCaptureStart);
   GB_AL(LPALCCAPTURESTOP,alcCaptureStop);
   GB_AL(LPALCGETINTEGERV,alcGetIntegerv);
   GB_AL(LPALCCAPTURESAMPLES,alcCaptureSamples);
   GB_AL(LPALCCAPTURECLOSEDEVICE,alcCaptureCloseDevice);
   GB_AL(LPALGETERROR,alGetError);
   GB_AL(LPALGENSOURCES,alGenSources);
   GB_AL(LPALGENBUFFERS,alGenBuffers);
   GB_AL(LPALSOURCEF,alSourcef);
   GB_AL(LPALSOURCEI,alSourcei);
   GB_AL(LPALGETSOURCEI,alGetSourcei);
   GB_AL(LPALISSOURCE,alIsSource);
   GB_AL(LPALSOURCESTOP,alSourceStop);
   GB_AL(LPALSOURCEPLAY,alSourcePlay);
   GB_AL(LPALSOURCEPAUSE,alSourcePause);
   GB_AL(LPALBUFFERDATA,alBufferData);
   GB_AL(LPALDELETEBUFFERS,alDeleteBuffers);
   GB_AL(LPALDELETESOURCES,alDeleteSources);
   GB_AL(LPALSOURCEQUEUEBUFFERS,alSourceQueueBuffers);
   GB_AL(LPALSOURCEUNQUEUEBUFFERS,alSourceUnqueueBuffers);

   return ok;
}
/******************************** END OPENAL **********************************/



/******************************* BEGIN MPG123 *********************************/
PMPG123_INIT                              mpg123_init;
PMPG123_EXIT                              mpg123_exit;
PMPG123_PLAIN_STRERROR                    mpg123_plain_strerror;
PMPG123_STRERROR                          mpg123_strerror;
PMPG123_NEW                               mpg123_new;
PMPG123_DELETE                            mpg123_delete;
PMPG123_OPEN                              mpg123_open;
PMPG123_CLOSE                             mpg123_close;
PMPG123_FORMAT_NONE                       mpg123_format_none;
PMPG123_GETFORMAT                         mpg123_getformat;
PMPG123_FORMAT                            mpg123_format;
PMPG123_READ                              mpg123_read;
PMPG123_LENGTH                            mpg123_length;
PMPG123_SEEK                              mpg123_seek;

unsigned int ext_load_mpg123(void)
{
   unsigned int ok = 1;

   if (mpg123_lib)
      return 1;

   path_add();
   mpg123_lib = get_lib(MPG123_LIB_NAME);
   if (!mpg123_lib) return 0;

   #define GB_MP(t,v) ok &= (v =(t)get_func(mpg123_lib,MPG123_LIB_NAME,#v))?1:0;
   GB_MP(PMPG123_INIT,mpg123_init);
   GB_MP(PMPG123_EXIT,mpg123_exit);
   GB_MP(PMPG123_PLAIN_STRERROR,mpg123_plain_strerror);
   GB_MP(PMPG123_STRERROR,mpg123_strerror);
   GB_MP(PMPG123_NEW,mpg123_new);
   GB_MP(PMPG123_DELETE,mpg123_delete);
   GB_MP(PMPG123_OPEN,mpg123_open);
   GB_MP(PMPG123_CLOSE,mpg123_close);
   GB_MP(PMPG123_GETFORMAT,mpg123_getformat);
   GB_MP(PMPG123_FORMAT_NONE,mpg123_format_none);
   GB_MP(PMPG123_FORMAT,mpg123_format);
   GB_MP(PMPG123_READ,mpg123_read);
   GB_MP(PMPG123_LENGTH,mpg123_length);
   GB_MP(PMPG123_SEEK,mpg123_seek);

   return ok;
}
/******************************** END MPG123 **********************************/



/******************************* BEGIN OGGVORBIS ******************************/
POV_OPEN_CALLBACKS                        ov_open_callbacks;
POV_INFO                                  ov_info;
POV_CLEAR                                 ov_clear;
POV_READ                                  ov_read;
POV_TIME_TOTAL                            ov_time_total;
POV_PCM_TOTAL                             ov_pcm_total;
POV_TIME_SEEK                             ov_time_seek;

unsigned int ext_load_ogg_vorbis(void)
{
   unsigned int ok = 1;

   if (ogg_vorb_lib)
      return 1;

   path_add();
   ogg_vorb_lib = get_lib(OGG_VORB_LIB_NAME);
   if (!ogg_vorb_lib) return 0;

   #define GRAB_OV(t,v) ok &= (v =(t)get_func(ogg_vorb_lib,OGG_VORB_LIB_NAME,#v))?1:0
   GRAB_OV(POV_OPEN_CALLBACKS,ov_open_callbacks);
   GRAB_OV(POV_INFO,ov_info);
   GRAB_OV(POV_CLEAR,ov_clear);
   GRAB_OV(POV_READ,ov_read);
   GRAB_OV(POV_TIME_TOTAL,ov_time_total);
   GRAB_OV(POV_PCM_TOTAL,ov_pcm_total);
   GRAB_OV(POV_TIME_SEEK,ov_time_seek);

   return ok;
}
/******************************** END OGGVORBIS *******************************/



/******************************** BEGIN FREEIMAGE *****************************/
PFREEIMAGE_GETFILETYPE                    FreeImage_GetFileType;
PFREEIMAGE_GETFIFFROMFILENAME             FreeImage_GetFIFFromFilename;
PFREEIMAGE_FIFSUPPORTSREADING             FreeImage_FIFSupportsReading;
PFREEIMAGE_LOAD                           FreeImage_Load;
PFREEIMAGE_GETBITS                        FreeImage_GetBits;
PFREEIMAGE_GETWIDTH                       FreeImage_GetWidth;
PFREEIMAGE_GETHEIGHT                      FreeImage_GetHeight;
PFREEIMAGE_GETBPP                         FreeImage_GetBPP;
PFREEIMAGE_UNLOAD                         FreeImage_Unload;
PFREEIMAGE_RESCALE                        FreeImage_Rescale;
PFREEIMAGE_ALLOCATE                       FreeImage_Allocate;
PFREEIMAGE_SAVE                           FreeImage_Save;
PFREEIMAGE_CONVERTFROMRAWBITS             FreeImage_ConvertFromRawBits;
PFREEIMAGE_FIFSUPPORTSWRITING             FreeImage_FIFSupportsWriting;


unsigned int ext_load_freeimage(void)
{
   unsigned int ok = 1;

   if (free_img_lib)
      return 1;

   path_add();
   free_img_lib = get_lib(FREEIMAGE_LIB_NAME);
   if (!free_img_lib) return 0;

   #define GB_FI(t,v) ok &= (v =(t)get_func(free_img_lib,FREEIMAGE_LIB_NAME,#v))?1:0
   //#define GB_FI(t,v,n) ok &= (v=(t)get_func(free_img_lib,FREEIMAGE_LIB_NAME,n))?1:0
   GB_FI(PFREEIMAGE_GETFILETYPE,FreeImage_GetFileType);
   GB_FI(PFREEIMAGE_GETFIFFROMFILENAME,FreeImage_GetFIFFromFilename);
   GB_FI(PFREEIMAGE_FIFSUPPORTSREADING,FreeImage_FIFSupportsReading);
   GB_FI(PFREEIMAGE_LOAD,FreeImage_Load);
   GB_FI(PFREEIMAGE_GETBITS,FreeImage_GetBits);
   GB_FI(PFREEIMAGE_GETWIDTH,FreeImage_GetWidth);
   GB_FI(PFREEIMAGE_GETHEIGHT,FreeImage_GetHeight);
   GB_FI(PFREEIMAGE_GETBPP,FreeImage_GetBPP);
   GB_FI(PFREEIMAGE_UNLOAD,FreeImage_Unload);
   GB_FI(PFREEIMAGE_RESCALE,FreeImage_Rescale);
   GB_FI(PFREEIMAGE_ALLOCATE,FreeImage_Allocate);
   GB_FI(PFREEIMAGE_SAVE,FreeImage_Save);
   GB_FI(PFREEIMAGE_CONVERTFROMRAWBITS,FreeImage_ConvertFromRawBits);
   GB_FI(PFREEIMAGE_FIFSUPPORTSWRITING,FreeImage_FIFSupportsWriting);

   return ok;
}
/********************************* END FREEIMAGE ******************************/



/********************************** BEGIN REVEL *******************************/
PREVEL_GETVERSION                      Revel_GetVersion;
PREVEL_GETAPIVERSION                   Revel_GetApiVersion;
PREVEL_CREATEENCODER                   Revel_CreateEncoder;
PREVEL_INITIALIZEPARAMS                Revel_InitializeParams;
PREVEL_ENCODESTART                     Revel_EncodeStart;
PREVEL_ENCODEFRAME                     Revel_EncodeFrame;
PREVEL_ENCODEAUDIO                     Revel_EncodeAudio;
PREVEL_ENCODEEND                       Revel_EncodeEnd;
PREVEL_DESTROYENCODER                  Revel_DestroyEncoder;

unsigned int ext_load_revel(void)
{
   unsigned int ok = 1;

   if (revel_lib)
      return 1;

   path_add();
   revel_lib = get_lib(REVEL_LIB_NAME);
   if (!revel_lib) return 0;

   #define GB_RV(t,v) ok &= (v =(t)get_func(revel_lib,REVEL_LIB_NAME,#v))? 1 : 0
   GB_RV(PREVEL_GETVERSION,Revel_GetVersion);
   GB_RV(PREVEL_GETAPIVERSION,Revel_GetApiVersion);
   GB_RV(PREVEL_CREATEENCODER,Revel_CreateEncoder);
   GB_RV(PREVEL_INITIALIZEPARAMS,Revel_InitializeParams);
   GB_RV(PREVEL_ENCODESTART,Revel_EncodeStart);
   GB_RV(PREVEL_ENCODEFRAME,Revel_EncodeFrame);
   GB_RV(PREVEL_ENCODEAUDIO,Revel_EncodeAudio);
   GB_RV(PREVEL_ENCODEEND,Revel_EncodeEnd);
   GB_RV(PREVEL_DESTROYENCODER,Revel_DestroyEncoder);

   return ok;
}
/*********************************** END REVEL ********************************/



#ifdef __cplusplus
}
#endif
