#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include "file_parser.h"

#define MAX_LINE 1024

/* Document tree level.. */
enum { t_doc, t_def };

/* Def branch... */
enum {
   d_undef,
   d_at, d_clr, d_hdr, d_msg, d_nar, d_pic,
   d_at_bdy, d_clr_bdy, d_hdr_bdy, d_msg_bdy, d_nar_bdy, d_pic_bdy,
   d_close
};

/* Parser line state... */
enum  { p_err = -1, p_ok = 0, p_more = 1 };

typedef char *reader_func(char *buffer, unsigned buf_sz, void *str);

static char *read_str(char *buffer, unsigned buf_sz, void *str);
static char *read_file(char *buffer, unsigned buf_sz, void *file_ptr);

static int parse_eat_whitespace(char **parse);
static int parse_eat_bracketspace(char **parse);
static int parse_find_open_bracket(char **parse);
static int parse_find_close_bracket(char **parse);
static int parse_get_string_line(char **parse, char* ln);
static int parse_concat_text(char **text, char *text_line, unsigned *sz);
static int parse_text_trim(char **text);

static color_t default_clr = {0,0,0,0};


void file_parser_erase_list(parse_list *list)
{
   parse_list *p_sav;
   loc_list *a_ptr, *a_sav;

   while (list) {

      p_sav = list->more;

      if (list->p->at) {
         a_ptr = list->p->at;
         while (a_ptr) {
            a_sav = a_ptr->more;
            free(a_ptr);
            a_ptr = a_sav;
         }
      }
      if (list->p->hdr)
         free(list->p->hdr);
      if (list->p->msg)
         free(list->p->msg);
      if (list->p->nar)
         free(list->p->nar);
      if (list->p->pic)
         free(list->p->pic);
      free(list->p);
      free(list);
      list = p_sav;
   }
}


static char *read_str(char *buffer, unsigned buf_sz, void *str)
{
   char **blk = (char**)str;
   char *cpy = buffer;
   unsigned n = 0;

   if (!(**blk))
      return 0;

   while (**blk && n < buf_sz - 1 && **blk != '\n') {
      *cpy = **blk;
      (*blk)++;
      cpy++;
      n++;
   }

   if (**blk == '\n') {
      (*blk)++;
      *cpy = '\n';
      cpy++;
   }

   *cpy = 0;
   return buffer;
}

static char *read_file(char *buffer, unsigned buf_sz, void *file_ptr)
{
   return fgets(buffer,buf_sz,(FILE*)file_ptr);
}

parse_list *file_parser_get_objects(const char* src, unsigned mode)
{
   void *document;
   unsigned level = t_doc, def = d_undef, sv_sz;
   int ret = p_ok;
   char line_buff[MAX_LINE], text_buff[MAX_LINE];
   char *p_at, *head = 0, *text = 0, *audio_file = 0, *pic_file = 0;
   color_t clr = default_clr;
   color_t head_clr, msg_clr;
   loc_list *loc = 0, *loc_ptr, *sv_nxt;
   parse_list *obj_ret = 0, *obj_ptr;
   parse_obj *p;
   reader_func *reader;

   if (!src)
      return 0;

   switch (mode) {
      case P_FILE:
         document = (void*)fopen(src,"r");
         reader = &read_file;
         break;
      case P_STRING:
         document = (void*)&src;
         reader = &read_str;
         break;
      default:
         document = 0;
         break;
    }

   if (!document)
      return 0;

   while (ret != p_err && (p_at = reader(line_buff,MAX_LINE,document))) {

      ret = p_ok;

      while (p_at && ret == p_ok) {

         if (level == t_doc) {

            ret = parse_find_open_bracket(&p_at);
            if (ret)
               continue;

            p_at++;
            level = t_def;

         }

         if (level == t_def) {

            if (def == d_undef) {

               ret = parse_eat_whitespace(&p_at);
               if (ret)
                  continue;

               switch (*p_at) {

                  case '}':

                     if (!obj_ret) {
                        obj_ret = (parse_list*)calloc(1,sizeof(parse_list));
                        obj_ptr = obj_ret;
                     }
                     else {
                        obj_ptr = obj_ret;
                        while (obj_ptr->more)
                           obj_ptr = obj_ptr->more;
                        obj_ptr->more = (parse_list*)calloc(1,sizeof(parse_list));
                        obj_ptr = obj_ptr->more;
                     }

                     p = (parse_obj*)calloc(1,sizeof(parse_obj));

                     if (!obj_ptr || !p) {
                         printf(
                           "error: allocation failed for return object(s).\n");
                         ret = p_err;
                         continue;
                     }

                     obj_ptr->p = p;

                     if (head) {
                        p->hdr = head;
                        head = 0;
                        p->hdr_clr = head_clr;
                     }

                     if (text) {
                        p->msg = text;
                        text = 0;
                        p->msg_clr = msg_clr;
                     }

                     if (audio_file) {
                        p->nar = audio_file;
                        audio_file = 0;
                     }

                     if (pic_file) {
                        p->pic = pic_file;
                        pic_file = 0;
                     }

                     if (loc) {
                        p->at = loc;
                        loc = 0;
                     }

                     clr   = default_clr;
                     level = t_doc;
                     p_at++;
                     break;

                  case '{':
                     printf("error at:\n\t%s\nUnidentified body.\n",p_at);
                     ret = p_err;
                     continue;

                  default:
                     if (strstr(p_at,"at ") == p_at ||
                         strstr(p_at,"at{") == p_at) {
                        def = d_at;
                        break;
                     }
                     else if (strstr(p_at,"clr ") == p_at ||
                              strstr(p_at,"clr{") == p_at ) {
                        def = d_clr;
                        break;
                     }
                     else if (strstr(p_at,"hdr ") == p_at ||
                              strstr(p_at,"hdr{") == p_at) {
                        def = d_hdr;
                        break;
                     }
                     else if (strstr(p_at,"msg ") == p_at ||
                              strstr(p_at,"msg{") == p_at) {
                        def = d_msg;
                        break;
                     }
                     else if (strstr(p_at,"nar ") == p_at ||
                              strstr(p_at,"nar{") == p_at) {
                        def = d_nar;
                        break;
                     }
                     else if (strstr(p_at,"pic ") == p_at ||
                              strstr(p_at,"pic{") == p_at) {
                        def = d_pic;
                        break;
                     }
                     else {
                        printf("error at:\n\t%s\nUnknown body.\n",p_at);
                        ret = p_err;
                        continue;
                     }
                     break;
               }
            }

            if (def == d_at) {

               ret = parse_find_open_bracket(&p_at);
               if (ret)
                  continue;

               def = d_at_bdy;
               p_at++;
            }

            if (def == d_at_bdy) {

               float x,y,z,r;

               ret = parse_eat_bracketspace(&p_at);
               if (ret)
                  continue;

               if (sscanf(p_at,"%f,%f,%f,%f",&x,&y,&z,&r) == 4) {
                  loc_ptr = loc;
                  if (!loc) {
                     loc = (loc_list*)calloc(1,sizeof(loc_list));
                     loc_ptr = loc;
                  }
                  else {
                     while (loc_ptr->more)
                        loc_ptr = loc_ptr->more;
                     loc_ptr->more = (loc_list*)calloc(1,sizeof(loc_list));
                     if (loc_ptr->more)
                        loc_ptr = loc_ptr->more;
                     else loc_ptr = 0;
                  }

                  if (!loc_ptr) {
                     ret = p_err;
                     continue;
                  }

                  loc_ptr->at.v.x = x;
                  loc_ptr->at.v.y = y;
                  loc_ptr->at.v.z = z;
                  loc_ptr->at.v.r = r;

                  def = d_close;
               }
               else {
                  printf("error at:\n\t%s\nInvalid \"at\" property.\n",p_at);
                  ret = p_err;
                  continue;
               }
            }

            if (def == d_clr) {

               ret = parse_find_open_bracket(&p_at);
               if (ret)
                  continue;

               def = d_clr_bdy;
               p_at++;
            }

            if (def == d_clr_bdy) {

               ret = parse_eat_bracketspace(&p_at);
               if (ret)
                  continue;

               if (sscanf(p_at,"%f,%f,%f,%f",&clr.r,&clr.g,&clr.b,&clr.a) == 4)
                  def = d_close;
               else {
                  printf("error at:\n\t%s\nInvalid \"clr\" property.\n",p_at);
                  ret = p_err;
                  continue;
               }
            }

            if (def == d_hdr) {

               if (head) {
                  printf("error: multiple header bodies.\n");
                  ret = p_err;
                  continue;
               }

               ret = parse_find_open_bracket(&p_at);
               if (ret)
                  continue;

               def = d_hdr_bdy;
               sv_sz = 0;
               p_at++;
            }

            if (def == d_hdr_bdy) {

               if ((ret = parse_get_string_line(&p_at,text_buff)) < 0)
                  continue;

               if (parse_concat_text(&head,text_buff,&sv_sz) < 0) {
                  ret = p_err;
                  continue;
               }

               if (ret)
                  continue;

               head_clr = clr;
               def = d_close;
            }

            if (def == d_msg) {

               if (text) {
                  printf("error: multiple message bodies.\n");
                  ret = p_err;
                  continue;
               }

               ret = parse_find_open_bracket(&p_at);
               if (ret)
                  continue;

               def = d_msg_bdy;
               sv_sz = 0;
               p_at++;
            }

            if (def == d_msg_bdy) {

               if ((ret = parse_get_string_line(&p_at,text_buff)) < 0)
                  continue;

               if (parse_concat_text(&text,text_buff,&sv_sz) < 0) {
                  ret = p_err;
                  continue;
               }

               if (ret)
                  continue;

               msg_clr = clr;
               def = d_close;
            }

            if (def == d_nar) {

               if (audio_file) {
                  printf("error: multiple narration bodies.\n");
                  ret = p_err;
                  continue;
               }

               ret = parse_find_open_bracket(&p_at);
               if (ret)
                  continue;

               def = d_nar_bdy;
               sv_sz = 0;
               p_at++;
            }

            if (def == d_nar_bdy) {

               if ((ret = parse_get_string_line(&p_at,text_buff)) < 0)
                  continue;

               if (parse_concat_text(&audio_file,text_buff,&sv_sz) < 0) {
                  ret = p_err;
                  continue;
               }

               if (ret)
                  continue;

               if (parse_text_trim(&audio_file) < 0) {
                  ret = p_err;
                  continue;
               }
               def = d_close;
            }

            if (def == d_pic) {

               if (pic_file) {
                  printf("error: multiple picture bodies.\n");
                  ret = p_err;
                  continue;
               }

               ret = parse_find_open_bracket(&p_at);
               if (ret)
                  continue;

               def = d_pic_bdy;
               sv_sz = 0;
               p_at++;
            }

            if (def == d_pic_bdy) {

               if ((ret = parse_get_string_line(&p_at,text_buff)) < 0)
                  continue;

               if (parse_concat_text(&pic_file,text_buff,&sv_sz) < 0) {
                  ret = p_err;
                  continue;
               }

               if (ret)
                  continue;

               if (parse_text_trim(&pic_file) < 0) {
                  ret = p_err;
                  continue;
               }
               def = d_close;
            }

            if (def == d_close) {

               ret = parse_find_close_bracket(&p_at);
               if (ret)
                  continue;

               def = d_undef;
               p_at++;
            }
         }
      }
   }

   if (mode == P_FILE)
      fclose((FILE*)document);

   if (ret == p_err) {
      while (loc) {
         sv_nxt = loc->more;
         free(loc);
         loc = sv_nxt;
      }
      if (head)
         free(head);
      if (text)
         free(text);
      if (audio_file)
         free(audio_file);
      if (pic_file)
         free(pic_file);
      file_parser_erase_list(obj_ret);
      return 0;
   }

   return obj_ret;
}


static int parse_eat_whitespace(char **p_at)
{
   while (**p_at == ' ' || **p_at == '\t')
      (*p_at)++;

   if (!(**p_at) || **p_at == '\n' || **p_at == '#')
      return 1;

   return 0;
}

static int parse_eat_bracketspace(char **p_at)
{
   while (**p_at == ' ' || **p_at == '\t')
      (*p_at)++;

   if (!(**p_at) || **p_at == '\n' || **p_at == '#')
      return 1;

   if (**p_at == '{' || **p_at == '}') {
      printf("error at:\n\t%s\nUnexpected token.\n",*p_at);
      return -1;
   }

   return 0;
}

static int parse_find_open_bracket(char **p_at)
{
   while (**p_at != '{') {
      if (!(**p_at) || **p_at == '\n' || **p_at == '#')
         return 1;
      if (**p_at == '}') {
         printf("error at:\n\t%s\nUnexpected token.\n",*p_at);
         return -1;
      }
      (*p_at)++;
   }
   return 0;
}

static int parse_find_close_bracket(char **p_at)
{
   while (**p_at != '}') {
      if (!(**p_at) || **p_at == '\n' || **p_at == '#')
         return 1;
      if (**p_at == '{') {
         printf("error at:\n\t%s\nUnexpected token.\n",*p_at);
         return -1;
      }
      (*p_at)++;
   }
   return 0;
}

static int parse_get_string_line(char **p_at, char* line_buff)
{
   unsigned sz = 0;

   while (**p_at != '}') {

      switch (**p_at) {

         case 0:
            line_buff[sz] = 0;
            return 1;

         case '#':
            if (!sz)
               line_buff[sz] = 0;
            else {
               line_buff[sz] = '\n';
               sz++;
               line_buff[sz] = 0;
            }
            return 1;

         case '\n':
            line_buff[sz] = **p_at;
            sz++;
            line_buff[sz] = 0;
            return 1;

         case '{':
            printf("error at:\n\t%s\nUnexpected token.\n",*p_at);
            return -1;

         case '\\':
            (*p_at)++;

         default:
            line_buff[sz] = **p_at;
            break;
      }
      sz++;
      (*p_at)++;
   }

   line_buff[sz] = 0;
   return 0;
}

static int parse_concat_text(char **text, char *text_buff, unsigned *sv_sz)
{
   unsigned sz;

   if (!text || !text_buff || !sv_sz) {
      printf("concat_text error: missing parameters.\n");
      return -1;
   }

   sz = strlen(text_buff);

   if (!sz)
      return 0;

   *text = (char*)realloc(*text, (*sv_sz + sz + 1) * sizeof(char));
   if (*text) {
      strncpy(*text + *sv_sz,text_buff,sz);
      *sv_sz += sz;
      *(*text + *sv_sz) = 0;
      return 0;
   }
   else return -1;
}

static int parse_text_trim(char **text)
{
   int sz, n = 0;
   char *nw_str, *ptr, *ptr_end;

   if (!text) {
      printf("concat_text error: missing parameters.\n");
      return -1;
   }

   if (!*text)
      return 0;

   sz = strlen(*text);
   if (!sz)
      return 0;

   ptr = *text;
   ptr_end = *text + sz - 1;

   while ( *ptr &&
         ( *ptr == ' ' || *ptr == '\t' ||
           *ptr == '\n' || *ptr == '\r')) {
      ptr++;
      sz--;
   }

   while ( ptr_end != *text &&
         ( *ptr_end == ' ' || *ptr_end == '\t' ||
           *ptr_end == '\n' || *ptr_end == '\r')) {
      ptr_end--;
      sz--;
   }

   /* was all whitespace... */
   if (sz < 1) {
      *text = realloc(*text,1 * sizeof(char));
      if (*text) {
         *text = 0;
         return 0;
      }
      else return -1;
   }

   nw_str = calloc(sz + 1, sizeof(char));

   if (nw_str) {
      while (n < sz) {
         *(nw_str + n) = *(ptr + n);
         n++;
      }
      *(nw_str + sz) = 0;
      free(*text);
      *text = nw_str;
      return 0;
   }
   else return -1;

   return 0;
}

#ifdef __cplusplus
}
#endif
