#define USING_OPENGL
#include "platform.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "geom_object.h"

typedef struct  { float x, y; }        coord_2d;
typedef struct  { float x, y, z; }     coord_3d;
typedef struct  { float r, g, b, a; }  coord_4d;

/* x = r cos(v) cos(u)
   y = r cos(v) sin(u)     u = [0, 2*Pi)
   z = r sin(v)            v = [-Pi/2, Pi/2] */

#define set_3d(cd,x0,y0,z0) { (cd).x  = (x0); (cd).y  = (y0); (cd).z  = (z0); }
#define set_2d(cd,x0,y0)    { (cd).x  = (x0); (cd).y  = (y0); }
#define negate_3d(cd)       { (cd).x *= -1;   (cd).y *= -1;   (cd).z *= -1; }
#define scale_3d(cd,s)      { (cd).x *= (s);  (cd).y *= (s);  (cd).z *= (s); }

#ifndef PI
   #define PI 3.14159265359
#endif

#define CUBE_FLOATS 288
static float basic_cube[CUBE_FLOATS] = {
/* VERTICES@0  V11        V12        V13        V21        V22        V23    */
/* Front  */  -1,-1, 1,   1,-1, 1,   1, 1, 1,   1, 1, 1,  -1, 1, 1,  -1,-1, 1,
/* Right  */   1,-1, 1,   1,-1,-1,   1, 1,-1,   1, 1,-1,   1, 1, 1,   1,-1, 1,
/* Top    */  -1, 1, 1,   1, 1, 1,   1, 1,-1,   1, 1,-1,  -1, 1,-1,  -1, 1, 1,
/* Back   */   1,-1,-1,  -1,-1,-1,  -1, 1,-1,  -1, 1,-1,   1, 1,-1,   1,-1,-1,
/* Left   */  -1,-1,-1,  -1,-1, 1,  -1, 1, 1,  -1, 1, 1,  -1, 1,-1,  -1,-1,-1,
/* Bottom */  -1,-1,-1,   1,-1,-1,   1,-1, 1,   1,-1, 1,  -1,-1, 1,  -1,-1,-1,

/* NORMALS@108 N11        N12        N13        N21        N22        N23    */
/* Front  */   0, 0, 1,   0, 0, 1,   0, 0, 1,   0, 0, 1,   0, 0, 1,   0, 0, 1,
/* Right  */   1, 0, 0,   1, 0, 0,   1, 0, 0,   1, 0, 0,   1, 0, 0,   1, 0, 0,
/* Top    */   0, 1, 0,   0, 1, 0,   0, 1, 0,   0, 1, 0,   0, 1, 0,   0, 1, 0,
/* Back   */   0, 0,-1,   0, 0,-1,   0, 0,-1,   0, 0,-1,   0, 0,-1,   0, 0,-1,
/* Left   */  -1, 0, 0,  -1, 0, 0,  -1, 0, 0,  -1, 0, 0,  -1, 0, 0,  -1, 0, 0,
/* Bottom */   0,-1, 0,   0,-1, 0,   0,-1, 0,   0,-1, 0,   0,-1, 0,   0,-1, 0,

/* TEXTURE@216 T11        T12        T13        T21        T22        T23    */
/* Front  */  .25,.333,  .50,.333,  .50,.666,  .50,.666,  .25,.666,  .25,.333,
/* Right  */  .50,.333,  .75,.333,  .75,.666,  .75,.666,  .50,.666,  .50,.333,
/* Top    */  .25,.666,  .50,.666,  .50,.999,  .50,.999,  .25,.999,  .25,.666,
/* Back   */  .75,.333,  .99,.333,  .99,.666,  .99,.666,  .75,.666,  .75,.333,
/* Left   */  .00,.333,  .25,.333,  .25,.666,  .25,.666,  .00,.666,  .00,.333,
/* Bottom */  .25,.000,  .50,.000,  .50,.333,  .50,.333,  .25,.333,  .25,.000
};

#define BILLBOARD_FLOATS  48
static float billboard[BILLBOARD_FLOATS] = {
/* VERTICES@0  V11        V12        V13        V21        V22        V23    */
/* Front  */  -1,-1, 0,   1,-1, 0,   1, 1, 0,   1, 1, 0,  -1, 1, 0,  -1,-1, 0,

/* NORMALS@18  N11        N12        N13        N21        N22        N23    */
/* Front  */   0, 0, 1,   0, 0, 1,   0, 0, 1,   0, 0, 1,   0, 0, 1,   0, 0, 1,

/* TEXTURE@36  T11        T12        T13        T21        T22        T23    */
/* Front  */   0, 0,      1, 0,      1, 1,      1, 1,      0, 1,      0, 0
};


geom_object *geom_object_create_sphere(int divisions, int sectors, unsigned inside)
{
   coord_3d *v, *n, *r1, *r2, *swap;
   coord_2d *t;
   geom_object *ret;
   float *data, tu, tv, tu1, tv1, pu, pv;
   unsigned total, s, d, cnt = 0;

   /* Each triangle is three vertices; two triangles per sector for 'divisions'
   of sectors plus two rows of sectors with single triangles at the
   sphere top/bottom.  */

   total = 2 * sectors * divisions;

   data = (float*)malloc(total * (6 * sizeof(coord_3d) + 3 * sizeof(coord_2d)));
   if (!data) {
      printf("Data block allocation error.\n");
      return 0;
   }
   r1 = (coord_3d*)malloc((sectors + 1) * sizeof(coord_3d));
   if (!r1) {
      printf("Row 1 allocation error.\n");
      free(data);
      return 0;
   }
   r2 = (coord_3d*)malloc((sectors + 1) * sizeof(coord_3d));
   if (!r2) {
      printf("Row 2 allocation error.\n");
      free(data);
      free(r1);
      return 0;
   }

   v = (coord_3d*) data;
   n = (coord_3d*)(data + total * 3 * 3);
   t = (coord_2d*)(data + 2 * total * 3 * 3);

   d = 0;
   while (d <= divisions) {

      swap = r1;
      r1   = r2;
      r2   = swap;

      /* generate new coordinates in r2... */

      pv = ((float)d + 1)/(divisions + 1) * PI - PI/2;
      s  = 0;
      while (s <= sectors) {
         pu = ((float)s)/sectors * 2 * PI;
         if (d == 0)
            set_3d(r1[s],cos(-PI/2) * cos(pu),cos(-PI/2) * sin(pu),sin(-PI/2));
         set_3d(r2[s], cos(pv) * cos(pu), cos(pv) * sin(pu), sin(pv));
         s++;
      }

      /* write out triangles using r1 and r2... */

      tv  = ((float)d)/(divisions + 1);
      tv1 = ((float)d + 1)/(divisions + 1);
      s   = 0;
      while (s < sectors) {

         tu  = ((float)s)/sectors;
         tu1 = ((float)s + 1)/sectors;

         /* "top-left" triangles */
         if (d != divisions) {

            v[cnt] = r2[s + 1];
            n[cnt] = v[cnt];
            if (inside)
               negate_3d(n[cnt]);
            set_2d(t[cnt],tu1,tv1);
            cnt++;

            v[cnt] = r2[s];
            n[cnt] = v[cnt];
            if (inside)
               negate_3d(n[cnt]);
            set_2d(t[cnt],tu,tv1);
            cnt++;

            v[cnt] = r1[s];
            n[cnt] = v[cnt];
            if (inside)
               negate_3d(n[cnt]);
            set_2d(t[cnt],tu,tv);
            cnt++;
         }

         /* "bottom-right" triangles */
         if (d != 0) {

            v[cnt] = r1[s];
            n[cnt] = v[cnt];
            if (inside)
               negate_3d(n[cnt]);
            set_2d(t[cnt],tu,tv);
            cnt++;

            v[cnt] = r1[s + 1];
            n[cnt] = v[cnt];
            if (inside)
               negate_3d(n[cnt]);
            set_2d(t[cnt],tu1,tv);
            cnt++;

            v[cnt] = r2[s + 1];
            n[cnt] = v[cnt];
            if (inside)
               negate_3d(n[cnt]);
            set_2d(t[cnt],tu1,tv1);
            cnt++;
         }
         s++;
      }
      d++;
   }



   free(r1);
   free(r2);
   ret = (geom_object*)calloc(1,sizeof(geom_object));
   if (!ret) {
      printf("Error allocating return object.\n");
      free(data);
      return 0;
   }

   ret->array          = (void*)data;
   ret->array_bytes    = total * (6 * sizeof(coord_3d) + 3 * sizeof(coord_2d));
   ret->vertex_count   = total * 3;
   ret->normal_offset  = total * 3 * sizeof(coord_3d);
   ret->texture_offset = total * 2 * 3 * sizeof(coord_3d);

   return ret;
}

geom_object *geom_object_create_washer(int divisions, int sectors, float center)
{
   coord_3d *v, *n, *r1, *r2, *swap;
   coord_2d *t;
   geom_object *ret;
   float *data, u, r;
   unsigned total, s, d, cnt = 0;

   /* Each triangle is three vertices; two triangles per sector for 'divisions'
   of sectors.  */

   total = 2 * sectors * (divisions + 1);

   data = (float*)malloc(total * (6 * sizeof(coord_3d) + 3 * sizeof(coord_2d)));
   if (!data) {
      printf("Data block allocation error.\n");
      return 0;
   }
   r1 = (coord_3d*)malloc((sectors + 1) * sizeof(coord_3d));
   if (!r1) {
      printf("Row 1 allocation error.\n");
      free(data);
      return 0;
   }
   r2 = (coord_3d*)malloc((sectors + 1) * sizeof(coord_3d));
   if (!r2) {
      printf("Row 2 allocation error.\n");
      free(data);
      free(r1);
      return 0;
   }

   v = (coord_3d*) data;
   n = (coord_3d*)(data + total * 3 * 3);
   t = (coord_2d*)(data + 2 * total * 3 * 3);

   d = 0;
   while (d <= divisions) {

      swap = r1;
      r1   = r2;
      r2   = swap;

      /* generate new coordinates in r2... */

      s = 0;
      r = (1 - ((float)(d + 1))/(divisions + 1)) * (1 - center) + center;
      while (s <= sectors) {
         u = ((float)s)/sectors * 2 * PI;
         if (d == 0)
            set_3d(r1[s],cos(u),sin(u),0);
         set_3d(r2[s],r * cos(u),r * sin(u),0);
         s++;
      }

      /* write out triangles using r1 and r2... */

      s   = 0;
      while (s < sectors) {

         v[cnt] = r2[s + 1];
         set_3d(n[cnt],0,0,1);
         set_2d(t[cnt],(v[cnt].x + 1)/2,(v[cnt].y + 1)/2);
         cnt++;

         v[cnt] = r2[s];
         set_3d(n[cnt],0,0,1);
         set_2d(t[cnt],(v[cnt].x + 1)/2,(v[cnt].y + 1)/2);
         cnt++;

         v[cnt] = r1[s];
         set_3d(n[cnt],0,0,1);
         set_2d(t[cnt],(v[cnt].x + 1)/2,(v[cnt].y + 1)/2);
         cnt++;

         v[cnt] = r1[s];
         set_3d(n[cnt],0,0,1);
         set_2d(t[cnt],(v[cnt].x + 1)/2,(v[cnt].y + 1)/2);
         cnt++;

         v[cnt] = r1[s + 1];
         set_3d(n[cnt],0,0,1);
         set_2d(t[cnt],(v[cnt].x + 1)/2,(v[cnt].y + 1)/2);
         cnt++;

         v[cnt] = r2[s + 1];
         set_3d(n[cnt],0,0,1);
         set_2d(t[cnt],(v[cnt].x + 1)/2,(v[cnt].y + 1)/2);
         cnt++;

         s++;
      }
      d++;
   }

   free(r1);
   free(r2);
   ret = (geom_object*)calloc(1,sizeof(geom_object));
   if (!ret) {
      printf("Error allocating return object.\n");
      free(data);
      return 0;
   }

   ret->array          = (void*)data;
   ret->array_bytes    = total * (6 * sizeof(coord_3d) + 3 * sizeof(coord_2d));
   ret->vertex_count   = total * 3;
   ret->normal_offset  = total * 3 * sizeof(coord_3d);
   ret->texture_offset = total * 2 * 3 * sizeof(coord_3d);

   return ret;
}

geom_object *geom_object_create_basic_cube(void)
{
   geom_object *ret = calloc(1,sizeof(geom_object));
   if (ret) {
      ret->array_bytes = CUBE_FLOATS * sizeof(float);
      ret->array = malloc(ret->array_bytes);
      if (ret->array) {
         memcpy(ret->array,&basic_cube,ret->array_bytes);
         ret->vertex_count   = 36;
         ret->normal_offset  = 108 * sizeof(float);
         ret->texture_offset = 216 * sizeof(float);

         return ret;
      }
      free(ret);
   }
   return 0;
}

geom_object *geom_object_create_billboard(float aspect)
{
   int n = 0;
   geom_object *ret = calloc(1,sizeof(geom_object));
   if (ret) {
      ret->array_bytes = BILLBOARD_FLOATS * sizeof(float);
      ret->array = malloc(ret->array_bytes);
      if (ret->array) {
         memcpy(ret->array,&billboard,ret->array_bytes);
         ret->vertex_count   = 6;
         ret->normal_offset  = 18 * sizeof(float);
         ret->texture_offset = 36 * sizeof(float);

         while (n < 18) {
            ((float*)ret->array)[n] *= aspect;
            n += 3;
         }

         return ret;
      }
      free(ret);
   }
   return 0;
}

void geom_object_draw(geom_object *v)
{
   if (!v || !v->array)
      return;

   glEnableClientState(GL_VERTEX_ARRAY);
   glEnableClientState(GL_TEXTURE_COORD_ARRAY);
   glEnableClientState(GL_NORMAL_ARRAY);


   glTexCoordPointer(2,GL_FLOAT,0,((char*)v->array) + v->texture_offset);
   glNormalPointer(GL_FLOAT,0,((char*)v->array) + v->normal_offset);
   glVertexPointer(3,GL_FLOAT,0,v->array);

   glDrawArrays(GL_TRIANGLES,0,v->vertex_count);

   glDisableClientState(GL_VERTEX_ARRAY);
   glDisableClientState(GL_TEXTURE_COORD_ARRAY);
   glDisableClientState(GL_NORMAL_ARRAY);
}

void geom_object_free(geom_object *v)
{
   if (!v)
      return;

   if (v->array)
      free(v->array);

   free(v);
}

geom_list *convert_to_geom_list(geom_object *g)
{
	vec4 Kd = {1,1,1,1};
	geom_list *l = create_geom_list();
	geom_part_t *p;

	if (!l)
		return 0;

	p = (geom_part_t*)calloc(1,sizeof(geom_part_t));
	if (!p) {
		geom_list_free(l);
		return 0;
	}

	l->array 			= g->array;
	l->array_bytes 	= g->array_bytes;
	l->part 				= p;
	l->part_count 		= 1;

	p->vertex_count 	= g->vertex_count;
	p->vertex_start 	= 0;
	p->normal_start 	= g->normal_offset;
	p->texture_start 	= g->texture_offset;
	p->color_start 	= g->color_offset;
	p->material.Kd 	= Kd;

	free(g);

	return l;
}