#include "renderer_api.h"
#include "gl_extensions.h"
#include "gluf.h"


typedef struct tex_context tex_context;
struct tex_context {
   tex_context *next;
   GLuint gl_id;
   char *file;
   renderer_bitmap *bm;
   renderer_context_function fn;
   void *ref;
   unsigned group;
   unsigned alpha_map;
};

typedef struct list_context list_context;
struct list_context {
   list_context *next;
   GLuint gl_id;
   renderer_context_function fn;
   void *ref;
   unsigned group;
};

typedef struct generic_context generic_context;
struct generic_context {
   generic_context *next;
   renderer_generic_function begin;
   renderer_generic_function end;
   void *ref;
   unsigned group;
};

typedef struct vbo_context vbo_context;
struct vbo_context {
   vbo_context *next;
   GLuint gl_id;
   renderer_context_function fn;
   void *ref;
   unsigned group;
};

static void context_initialize(unsigned int id);
static tex_context *new_texture_context(void);
static list_context *new_list_context(void);
static generic_context *new_generic_context(void);
static vbo_context *new_vbo_context(void);

static unsigned get_texture_id(unsigned w,unsigned h,unsigned bpp, unsigned wrap,
                                        unsigned static_id, unsigned char *data);

static void map_alpha_tex(renderer_bitmap *bm, unsigned tex_id);

/* Context managment lists... */
static tex_context      *renderer_textures;
static list_context     *renderer_lists;
static generic_context  *renderer_generics;
static vbo_context      *renderer_vbos;
static unsigned         current_group = 0;

void renderer_init(void)
{
   gl_extensions_init();
}


void renderer_generate_stencil(int width, int height, unsigned type)
{
   /* Apparently, must be aligned on 4 byte boundries for glBitmap... */
   unsigned row_bits = width/32 * 32 + 32;

   unsigned char *bm;
   int scanline_count;

   /* Set up the raster position. */
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(0,width,0,height,0,1);
   glViewport(0,0,width,height);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glRasterPos2i(0,0);

   /* This should be the ONLY place the stencil buffer is cleared, as we want
   to reuse the raster pattern for each frame. */
   glStencilMask(~0);
   glClearStencil(0);
   glClear(GL_STENCIL_BUFFER_BIT);
   glStencilFunc(GL_ALWAYS,1,1);
   glEnable(GL_STENCIL_TEST);

   bm = (unsigned char*)malloc(row_bits/8 * height * sizeof(unsigned char));
   if (!bm) {
      glDisable(GL_STENCIL_TEST);
      glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
      return;
   }

   memset(bm,0xFF,row_bits/8 * height);

   if (type == RENDERER_HORIZONTAL || type == RENDERER_CHECKER) {
      scanline_count = 0;
      while (scanline_count < height) {
         if (scanline_count%2)
            memset(bm + scanline_count * (row_bits/8),0xFF,row_bits/8);
         else
            memset(bm + scanline_count * (row_bits/8),0,row_bits/8);
         scanline_count++;
      }
      glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
      glBitmap(row_bits,height,0,0,0,0,bm);
   }

   if (type == RENDERER_VERTICAL || type == RENDERER_CHECKER) {
      memset(bm,0xAA,row_bits/8 * height);
      glStencilOp(GL_INVERT,GL_INVERT,GL_INVERT);
      glBitmap(row_bits,height,0,0,0,0,bm);
   }

   free(bm);
   glDisable(GL_STENCIL_TEST);
   glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
}

void renderer_use_buffer(unsigned buffer)
{
   switch (buffer) {
      case RENDERER_LEFT:
         glDrawBuffer(GL_BACK_LEFT);
         break;
      case RENDERER_RIGHT:
         glDrawBuffer(GL_BACK_RIGHT);
         break;
      case RENDERER_FRONT:
         glDrawBuffer(GL_FRONT);
         break;
      case RENDERER_BACK:
      default:
         glDrawBuffer(GL_BACK);
         break;
   }
}

void renderer_use_stencil(unsigned stencil)
{
   switch (stencil) {
      case RENDERER_LEFT:
         glEnable(GL_STENCIL_TEST);
         glStencilFunc(GL_NOTEQUAL,1,1);
         break;
      case RENDERER_RIGHT:
         glEnable(GL_STENCIL_TEST);
         glStencilFunc(GL_EQUAL,1,1);
         break;
      case RENDERER_NONE:
      default:
         glDisable(GL_STENCIL_TEST);
         break;
   }
}

void renderer_use_colormask(unsigned colormask)
{
   switch (colormask) {
      case RENDERER_LEFT:
         glColorMask(GL_TRUE,GL_FALSE,GL_FALSE,GL_FALSE);
         break;
      case RENDERER_RIGHT:
         glColorMask(GL_FALSE,GL_TRUE,GL_TRUE,GL_FALSE);
         break;
      case RENDERER_NONE:
      default:
         glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
         break;
   }
}

void renderer_clear_buffer(unsigned buffer)
{
   switch (buffer) {
      case RENDERER_COLOR:
         glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
         glClear(GL_COLOR_BUFFER_BIT);
         break;
      case RENDERER_DEPTH:
         glClear(GL_DEPTH_BUFFER_BIT);
         break;
      default:
         break;
   }
}

void renderer_clear_color(float r, float g, float b, float a)
{
   glClearColor(r,g,b,a);
}

void renderer_use_depth_buffer(void)
{
   glEnable(GL_DEPTH_TEST);
}

void renderer_disable_depth_buffer(void)
{
   glDisable(GL_DEPTH_TEST);
}

void renderer_print_error(void)
{
   unsigned error_check = glGetError();

   switch (error_check) {
      case GL_NO_ERROR:
         break;
      case GL_INVALID_ENUM:
         printf("Last GL_ERROR: invalid enumerator.\n");
         break;
      case GL_INVALID_VALUE:
         printf("Last GL_ERROR: invalid value.\n");
         break;
      case GL_INVALID_OPERATION:
         printf("Last GL_ERROR: invalid operation.\n");
         break;
      case GL_STACK_OVERFLOW:
         printf("Last GL_ERROR: stack overflow.\n");
         break;
      case GL_STACK_UNDERFLOW:
         printf("Last GL_ERROR: stack underflow.\n");
         break;
      case GL_OUT_OF_MEMORY:
         printf("Last GL_ERROR: out of memory.\n");
         break;
      default:
         printf("Last GL_ERROR: unknown.\n");
   }
}

void renderer_viewport(int x, int y, int width, int height)
{
   glViewport(x,y,width,height);
}

void renderer_polygon_fill(void)
{
   glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
}
void renderer_polygon_wireframe(void)
{
   glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
}

void renderer_cull_back(void)
{
   glCullFace(GL_BACK);
   glEnable(GL_CULL_FACE);
}

void renderer_cull_none(void)
{
   glDisable(GL_CULL_FACE);
}

void renderer_high_quality_lines(void)
{
   glEnable(GL_LINE_SMOOTH);
   glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
   glEnable(GL_DITHER);
}

void renderer_high_quality_polygons(void)
{
   //glEnable(GL_NORMALIZE);
   glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
   glEnable(GL_BLEND);
   glEnable(GL_ALPHA_TEST);
   glAlphaFunc(GL_NOTEQUAL,0);
   glShadeModel(GL_SMOOTH);
}

void renderer_set_modelview(mat44 *m)
{
   glMatrixMode(GL_MODELVIEW);
   glLoadMatrixf((float*)m);
}

void renderer_set_projection(mat44 *m)
{
   glMatrixMode(GL_PROJECTION);
   glLoadMatrixf((float*)m);
}

void renderer_use_built_in_lighting(float level)
{
   static float light_position[4] = {0,0,0,1};
   float light_vector[4] = {level,level,level,1};

   glEnable(GL_LIGHTING);
   glEnable(GL_LIGHT0);
   glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,1);
   glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,1);
   glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL,GL_SEPARATE_SPECULAR_COLOR);
   glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);

   glLightfv(GL_LIGHT0,GL_AMBIENT,light_vector);
   glLightfv(GL_LIGHT0,GL_DIFFUSE,light_vector);
   glLightfv(GL_LIGHT0,GL_SPECULAR,light_vector);
   glLightfv(GL_LIGHT0,GL_POSITION,light_position);
}

/*********************** context management ***********************************/
unsigned renderer_generate_unique_id(void)
{
   static unsigned id = 0;
   return ++id;
}

void renderer_set_context_group(unsigned id)
{
   current_group = id;
}

void renderer_activate_context_group(unsigned id)
{
   current_group = id;
   context_initialize(id);
   //printf("init %d\n",id);
}

void renderer_suppress_context_group(unsigned id)
{
   tex_context *tex = renderer_textures;
   list_context *lst = renderer_lists;
   generic_context *gen = renderer_generics;
   vbo_context *vbo = renderer_vbos;

   current_group = id;

   while (tex) {
      if (glIsTexture(tex->gl_id) && (!id || tex->group == id))
         glDeleteTextures(1,&tex->gl_id);
      tex = tex->next;
   }

   while (lst) {
      if (glIsList(lst->gl_id) && (!id || lst->group == id))
         glDeleteLists(lst->gl_id,1);
      lst = lst->next;
   }

   while (gen) {
      if ((!id || gen->group == id) && gen->end)
         gen->end(gen->ref);
      gen = gen->next;
   }

   while (vbo) {
      if (glIsBuffer(vbo->gl_id) && (!id || vbo->group == id))
         glDeleteBuffers(1,&vbo->gl_id);
      vbo = vbo->next;
   }

   //printf("suppress %d\n",id);
}

void renderer_context_initialize(void)
{
   context_initialize(0);
}

static void context_initialize(unsigned int id)
{
   tex_context *tex = renderer_textures;
   renderer_bitmap *bm;
   list_context *lst = renderer_lists;
   generic_context *gen = renderer_generics;
   vbo_context *vbo = renderer_vbos;

   while (tex) {

      if (!tex->gl_id || glIsTexture(tex->gl_id)) {
         tex = tex->next;
         continue;
      }

      if (tex->file && (!id || tex->group == id)) {
         bm = gluf_get_bitmap(tex->file);
         renderer_map_texture(bm,1,tex->gl_id);
         gluf_erase_bitmap(bm);
         tex = tex->next;
         continue;
      }

      if (tex->bm && (!id || tex->group == id)) {
         if (tex->alpha_map)
            map_alpha_tex(tex->bm,tex->gl_id);
         else
            renderer_map_texture(tex->bm,1,tex->gl_id);
         tex = tex->next;
         continue;
      }

      if (tex->fn && (!id || tex->group == id)) {
         tex->fn(tex->ref,tex->gl_id);
         tex = tex->next;
         continue;
      }

      tex = tex->next;
   }

   while (lst) {

      if (!lst->gl_id || glIsList(lst->gl_id)) {
         lst = lst->next;
         continue;
      }

      if (lst->fn && (!id || lst->group == id)) {

         glNewList(lst->gl_id,GL_COMPILE);
            lst->fn(lst->ref,lst->gl_id);
         glEndList();
      }

      lst = lst->next;
   }

   while (gen) {

      if (gen->begin && (!id || gen->group == id))
         gen->begin(gen->ref);

      gen = gen->next;
   }

   while (vbo) {

      if (!vbo->gl_id || glIsBuffer(vbo->gl_id)) {
         vbo = vbo->next;
         continue;
      }

      if (vbo->fn && (!id || vbo->group == id)) {

         glBindBuffer(GL_ARRAY_BUFFER,vbo->gl_id);
            vbo->fn(vbo->ref,vbo->gl_id);
         glBindBuffer(GL_ARRAY_BUFFER,0);
      }

      vbo = vbo->next;
   }
}

void renderer_erase_registry(void)
{
   tex_context *tex = renderer_textures, *tex_next;
   list_context *lst = renderer_lists, *lst_next;
   generic_context *gen = renderer_generics, *gen_next;
   vbo_context *vbo = renderer_vbos, *vbo_next;

   while (tex) {
      if (glIsTexture(tex->gl_id))
         glDeleteTextures(1,&tex->gl_id);
      tex_next = tex->next;
      if (tex->file)
         free(tex->file);
      free(tex);
      tex = tex_next;
   }
   renderer_textures = 0;

   while (lst) {
      /* if (glIsList(lst->gl_id)) crashes on VRJuggler 64 bit... */
      if (glIsList(lst->gl_id))
         glDeleteLists(lst->gl_id,1);
      lst_next = lst->next;
      free(lst);
      lst = lst_next;
   }
   renderer_lists = 0;

   while (gen) {
      gen_next = gen->next;
      free(gen);
      gen = gen_next;
   }
   renderer_generics = 0;

   while (vbo) {
      if (glIsBuffer(vbo->gl_id))
         glDeleteBuffers(1,&vbo->gl_id);
      vbo_next = vbo->next;
      free(vbo);
      vbo = vbo_next;
   }
   renderer_vbos = 0;
}

static tex_context *new_texture_context(void)
{
   tex_context *ptr = renderer_textures;

   if (!ptr) {
      ptr = (tex_context*)calloc(1,sizeof(tex_context));
      if (!ptr)
         return 0;
      renderer_textures = ptr;
      ptr->gl_id = 1;
      ptr->group = current_group;
   }
   else {
      while (ptr->next)
         ptr = ptr->next;

      ptr->next = (tex_context*)calloc(1,sizeof(tex_context));
      if (!ptr->next)
         return 0;
      ptr->next->gl_id = ptr->gl_id + 1;
      ptr->next->group = current_group;
      ptr = ptr->next;
   }

   //printf("New texture context...\n");

   return ptr;
}

static list_context *new_list_context(void)
{
   list_context *ptr = renderer_lists;

   if (!ptr) {
      ptr = (list_context*)calloc(1,sizeof(list_context));
      if (!ptr)
         return 0;
      renderer_lists = ptr;
      ptr->gl_id = 1;
      ptr->group = current_group;
   }
   else {
      while (ptr->next)
         ptr = ptr->next;

      ptr->next = (list_context*)calloc(1,sizeof(list_context));
      if (!ptr->next)
         return 0;
      ptr->next->gl_id = ptr->gl_id + 1;
      ptr->next->group = current_group;
      ptr = ptr->next;
   }

   return ptr;
}

static generic_context *new_generic_context(void)
{
   generic_context *ptr = renderer_generics;

   if (!ptr) {
      ptr = (generic_context*)calloc(1,sizeof(generic_context));
      if (!ptr)
         return 0;
      renderer_generics = ptr;
      ptr->group = current_group;
   }
   else {
      while (ptr->next)
         ptr = ptr->next;

      ptr->next = (generic_context*)calloc(1,sizeof(generic_context));
      if (!ptr->next)
         return 0;
      ptr->next->group = current_group;
      ptr = ptr->next;
   }

   return ptr;
}

static vbo_context *new_vbo_context(void)
{
   vbo_context *ptr = renderer_vbos;

   if (!ptr) {
      ptr = (vbo_context*)calloc(1,sizeof(vbo_context));
      if (!ptr)
         return 0;
      renderer_vbos = ptr;
      ptr->gl_id = 1;
      ptr->group = current_group;
   }
   else {
      while (ptr->next)
         ptr = ptr->next;

      ptr->next = (vbo_context*)calloc(1,sizeof(vbo_context));
      if (!ptr->next)
         return 0;
      ptr->next->gl_id = ptr->gl_id + 1;
      ptr->next->group = current_group;
      ptr = ptr->next;
   }

   return ptr;
}

unsigned renderer_register_texture_file(const char *file)
{
   tex_context *ptr = renderer_textures;

   if (!file || !*file)
      return 0;

   /* Find out if this file is registered in the current context group. */
   if (ptr) {
      while (ptr->next) {
         if (ptr->file)
            if (!strcmp(file,ptr->file) && ptr->group == current_group)
                return ptr->gl_id;
         ptr = ptr->next;
      }
   }

   ptr = new_texture_context();

   if (ptr) {
      ptr->file = (char*)calloc(strlen(file) + 1, sizeof(char));
      if (ptr->file)
         strcpy(ptr->file,file);
      return ptr->gl_id;
   }

   return 0;
}

unsigned renderer_register_texture_ptr(renderer_bitmap *obj)
{
   tex_context *ptr;

   if (!obj)
      return 0;

   ptr = new_texture_context();
   if (ptr) {
      ptr->bm = obj;
      return ptr->gl_id;
   }

   return 0;
}

unsigned renderer_register_alpha_map_ptr(renderer_bitmap *obj)
{
   tex_context *ptr;

   if (!obj)
      return 0;

   ptr = new_texture_context();
   if (ptr) {
      ptr->bm = obj;
      ptr->alpha_map = 1;
      return ptr->gl_id;
   }

   return 0;
}

unsigned renderer_register_texture_function(void *owner, renderer_context_function fn)
{
   tex_context *ptr;

   if (!fn)
      return 0;

   ptr = new_texture_context();
   if (ptr) {
      ptr->fn = fn;
      ptr->ref = owner;
      return ptr->gl_id;
   }

   return 0;
}

unsigned renderer_register_list_function(void *owner, renderer_context_function fn)
{
   list_context *ptr;

   if (!fn)
      return 0;

   ptr = new_list_context();
   if (ptr) {
      ptr->fn = fn;
      ptr->ref = owner;
      return ptr->gl_id;
   }

   return 0;
}

void renderer_register_generic(void *owner, renderer_generic_function begin, renderer_generic_function end)
{
   generic_context *ptr;

   ptr = new_generic_context();
   if (ptr) {
      ptr->begin = begin;
      ptr->end   = end;
      ptr->ref   = owner;
   }
}

unsigned renderer_register_vertex_buffer_function(void *owner, renderer_context_function fn)
{
   vbo_context *ptr;

   if (!fn)
      return 0;

   ptr = new_vbo_context();
   if (ptr) {
      ptr->fn = fn;
      ptr->ref = owner;
      return ptr->gl_id;
   }

   return 0;
}


/******************************* texture mapping ******************************/

unsigned renderer_map_texture(renderer_bitmap *bm, unsigned wrap, unsigned designate)
{
   if (!bm)
      return 0;

   return get_texture_id(bm->w,bm->h,bm->bpp,wrap,designate,bm->data);
}


static unsigned get_texture_id(unsigned w,unsigned h,unsigned bpp, unsigned wrap,
                                        unsigned static_id, unsigned char *data)
{
   unsigned fmt = 0, clr_comp = 0, d_type = 0;
   GLuint tex = static_id;

   if (!w || !h || !bpp || !data)
      return 0;

   switch (bpp) {
      case 1:
         clr_comp = GL_LUMINANCE;
         fmt = GL_LUMINANCE;
         d_type = GL_BITMAP;
      case 4:
         clr_comp = GL_LUMINANCE4;
         fmt = GL_LUMINANCE;
         d_type = GL_BITMAP; /* ...don't know if this is right... */
         break;
      case 8:
         clr_comp = GL_LUMINANCE8;
         fmt = GL_LUMINANCE;
         d_type = GL_UNSIGNED_BYTE;
         break;
      case 16:
         clr_comp = GL_LUMINANCE8_ALPHA8;
         fmt = GL_LUMINANCE_ALPHA;
         d_type = GL_UNSIGNED_BYTE;
         break;
      case 24:
         clr_comp = GL_RGB;
         /* clr_comp = GL_COMPRESSED_RGB; */
         fmt = GL_BGR_EXT;
         d_type = GL_UNSIGNED_BYTE;
         break;
      case 32:
         clr_comp = GL_RGBA;
         /* clr_comp = GL_COMPRESSED_RGBA; */
         fmt = GL_BGRA_EXT;
         d_type = GL_UNSIGNED_BYTE;
         break;
      default:
         return 0;
         break;
   }

   if (!static_id)
      glGenTextures(1,&tex);

   glBindTexture(GL_TEXTURE_2D,tex);
   glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR_MIPMAP_LINEAR);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,wrap ? GL_REPEAT : GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,wrap ? GL_REPEAT : GL_CLAMP);
   glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP,GL_TRUE);

   glTexImage2D(
      GL_TEXTURE_2D,
      0,
      clr_comp,
      w,
      h,
      0,
      fmt,
      d_type,
      data);

   return tex;
}


static void map_alpha_tex(renderer_bitmap *bm, unsigned tex_id)
{
   unsigned fmt = 0, clr_cmp = 0, total = 0;
   unsigned n = 0;
   unsigned char *data;

   if (!bm || !bm->w || !bm->h || !bm->bpp || !bm->data)
      return;

   data = malloc(bm->w * bm->h * sizeof(int));
   if (!data) {
      return;
   }

   total = bm->w * bm->h;

   switch (bm->bpp) {
      case 8:

         while (n < total) {
            memset(data + 4 * n,*(bm->data + n),sizeof(int));
            n++;
         }
         break;
      case 16:
         while (n < total) {
            memset(data + 4 * n, *(bm->data + 2 * n + 1),sizeof(int));
            n++;
         }
         break;
      case 24:
         while (n < total) {
            memset(data + 4 * n, *(bm->data + 3 * n + 2),sizeof(int));
            n++;
         }
         break;
      case 32:
         while (n < total) {
            memset(data + 4 * n, *(bm->data + 4 * n + 3),sizeof(int));
            n++;
         }
         break;
      default:
         return;
         break;
   }

   glBindTexture(GL_TEXTURE_2D,tex_id);
   glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR_MIPMAP_LINEAR);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
   glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
   glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP,GL_TRUE);
   glTexImage2D(
      GL_TEXTURE_2D,
      0,
      GL_RGBA,
      bm->w,
      bm->h,
      0,
      GL_RGBA,
      GL_UNSIGNED_BYTE,
      data);

   free(data);
}

/******************************** buffers *************************************/
void renderer_static_buffer_data(unsigned buffer,unsigned size,void *data)
{
   glBindBuffer(GL_ARRAY_BUFFER,buffer);
   glBufferData(
      GL_ARRAY_BUFFER,size,data,GL_STATIC_DRAW);
   glBindBuffer(GL_ARRAY_BUFFER,0);
}
