#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include "ui_2d_shader.h"
#include "gltf.h"

/******************** named memory maps (for "dump") **************************/

static unsigned char dump[256];

/************************* file header map ************************************/

struct bitmap_file_header {            /* 4 bytes */
   char              *file_id;
   unsigned char     *version;
}
static f_hdr = {
   (char*)dump,
   &dump[3]
};

/************************* block header map ***********************************/

struct bitmap_block_header {           /* 5 bytes */
   char              *block_id;
   int               *block_size;
}
static b_hdr = {
   dump,
   (int*)&dump[1]
};

/*************************** info block map ***********************************/

struct bitmap_info{                    /* Variable size... */
   short             *font_size;
   unsigned char     *bit_field;
   unsigned char     *char_set;
   unsigned short    *stretch_h;
   unsigned char     *aa;
   unsigned char     *padding_up;
   unsigned char     *padding_right;
   unsigned char     *padding_down;
   unsigned char     *padding_left;
   unsigned char     *padding_horiz;
   unsigned char     *padding_vert;
   unsigned char     *outline;         /* 14 bytes to this field... */
   char *font_name;
}
static inf = {
   (short*)dump,
   &dump[2],
   &dump[3],
   (unsigned short*)&dump[4],
   &dump[6],
   &dump[7],
   &dump[8],
   &dump[9],
   &dump[10],
   &dump[11],
   &dump[12],
   &dump[13],
   &dump[14]
};

/***************************** common block map *******************************/

struct bitmap_common {                 /* 15 bytes */
   unsigned short    *line_height;
   unsigned short    *base;
   unsigned short    *scale_w;
   unsigned short    *scale_h;
   unsigned short    *pages;
   unsigned char     *bit_field;
   unsigned char     *alpha_ch;
   unsigned char     *red_ch;
   unsigned char     *green_ch;
   unsigned char     *blue_ch;
}
static cmn = {
   (unsigned short*)dump,
   (unsigned short*)&dump[2],
   (unsigned short*)&dump[4],
   (unsigned short*)&dump[6],
   (unsigned short*)&dump[8],
   &dump[10],
   &dump[11],
   &dump[12],
   &dump[13],
   &dump[14],
};

/******************* pages block map: is array of strings *********************/


/************************** character block map *******************************/

struct bitmap_char {                   /* 20 bytes */
   unsigned int      *id;
   unsigned short    *x;
   unsigned short    *y;
   unsigned short    *width;
   unsigned short    *height;
   short             *x_offset;
   short             *y_offset;
   short             *x_advance;
   unsigned char     *page;
   unsigned char     *chnl;
}
static chr = {
   (unsigned int*)dump,
   (unsigned short*)&dump[4],
   (unsigned short*)&dump[6],
   (unsigned short*)&dump[8],
   (unsigned short*)&dump[10],
   (short*)&dump[12],
   (short*)&dump[14],
   (short*)&dump[16],
   &dump[18],
   &dump[19]
};

/************************** kerning pairs block map  **************************/

struct bitmap_kerning_pairs {          /* 10 bytes */
   unsigned int      *first;
   unsigned int      *second;
   short             *amount;
}
static krn = {
   (unsigned int*)&dump[0],
   (unsigned int*)&dump[4],
   (short*)&dump[8]
};

/***************** object and private function prototypes *********************/

static int load_font_def(gl_tex_font *f, const char *file);
static void erase(gl_tex_font *f);
static void draw(gl_tex_font *f, const char* txt, mat44 *projection, mat44 *modelview);

/************************** object implementation *****************************/

gl_tex_font *create_texture_font(const char *font_file)
{
   gl_tex_font *tx;
   char *temp;
   unsigned str_sz;

   if (!font_file)
      return 0;

   tx = (gl_tex_font*)calloc(1,sizeof(gl_tex_font));
   if (tx) {
      if (load_font_def(tx,font_file)) {

         str_sz = strlen(DEFAULT_FONT_DIR) + strlen(tx->fnt_tex_file) + 2;
         temp   = (char*)calloc(str_sz,sizeof(char));

         if (temp)
            snprintf(temp,str_sz,"%s/%s",DEFAULT_FONT_DIR,tx->fnt_tex_file);

         tx->fnt_image = gluf_get_bitmap(temp);
         tx->tex_id    = renderer_register_alpha_map_ptr(tx->fnt_image);
         tx->cpl       = 0;
         tx->tab_stop  = 3;
         tx->render    = 1;
         tx->draw      = &draw;
         tx->erase     = &erase;



         free(temp);
         return tx;
      }
      free(tx);
   }

   return 0;
}

static void erase(gl_tex_font *f)
{
   if (!f)
      return;

   if (f->fnt_tex_file)
      free(f->fnt_tex_file);
   if (f->fnt_image)
      gluf_erase_bitmap(f->fnt_image);
   if (f->vtx)
      free(f->vtx);
   if (f->tex)
      free(f->tex);

   free(f);
}

static int load_font_def(gl_tex_font *f, const char *file)
{
   FILE *fnt_file;
   unsigned blk_sz, n, err, str_sz;
   size_t got;

   if (!f || !file)
      return 0;

   fnt_file = fopen(file,"rb");
   if (!fnt_file) {
      printf("Could not open font file...\n");
      return 0;
   }
   else printf("\nRetrieving information from font file %s...\n",file);

   /* Get the file header... */

   if (fread(dump,sizeof(char),4,fnt_file) == 4) {
      if (*f_hdr.version != 3) {
         printf("Unsupported fnt file version %d, exit.\n",*f_hdr.version);
         fclose(fnt_file);
         return 0;
      }
   }
   else {
      fclose(fnt_file);
      printf("Read error on file header in  %s, exit.\n",file);
      return 0;
   }

   /* Skip the info block... */

   if (fread(dump,sizeof(char),5,fnt_file) == 5) {
      blk_sz = *b_hdr.block_size;
      fseek(fnt_file,blk_sz,SEEK_CUR);
   }
   else {
      fclose(fnt_file);
      printf("Read error on info block in %s, exit.\n",file);
      return 0;
   }

   /* In the common block... */

   got       = fread(dump,sizeof(char),5,fnt_file);
   blk_sz    = *b_hdr.block_size;
   got      += fread(dump,sizeof(char),15,fnt_file);
   f->line_h = *cmn.line_height;
   f->max_h  = *cmn.base;
   f->tex_w  = *cmn.scale_w;
   f->tex_h  = *cmn.scale_h;

   if (got != 20) {
      fclose(fnt_file);
      printf("Read error on common block in %s, exit.\n",file);
      return 0;
   }

   /* In the pages block: we only want the first page file texture.. */

   if (fread(dump,sizeof(char),5,fnt_file) == 5) {
      blk_sz = *b_hdr.block_size;
      if (fread(dump,sizeof(char),blk_sz,fnt_file) == blk_sz) {
         str_sz = strlen(dump);
         f->fnt_tex_file = (char*)calloc(str_sz + 1,sizeof(char));
         strncpy(f->fnt_tex_file,dump,str_sz);
         printf("Font texture file is: %s...\n",f->fnt_tex_file);
      }
      else printf("Problems with font texture file name...\n");
   }
   else {
      fclose(fnt_file);
      printf("Read error on pages block in %s, exit.\n",file);
      return 0;
   }

   /* In the character block: now, the serious work of grabbing all of the
   glyphs. These appear to assume an orgin at the top left corner of the
   image -- this doesn't suit GL texture mapping directly, so we do some
   conversions... */

   if (fread(dump,sizeof(char),5,fnt_file) == 5) {
      blk_sz = *b_hdr.block_size;
      n      = blk_sz/20;
      err    = 0;
      printf("Loading %d glyphs...",n);

      while (n) {
         if (fread(dump,sizeof(char),20,fnt_file) != 20) {
            n = 0;
            err++;
            continue;
         }

         if (*chr.id > 255) {
            err++;
            continue;
         }

         f->x[*chr.id]     = *chr.x;
         f->y[*chr.id]     = f->tex_h - *chr.y - *chr.height;
         f->w[*chr.id]     = *chr.width;
         f->h[*chr.id]     = *chr.height;
         f->pad[*chr.id]   = *chr.x_advance;
         f->x_off[*chr.id] = *chr.x_offset;
         f->y_off[*chr.id] = f->max_h - *chr.y_offset - *chr.height;
         n--;
      }
      if (err)
         printf("%d errors were encountered.\n",err);
      else printf("done.\n");
   }
   else {
      fclose(fnt_file);
      printf("Read error on character block in %s, exit.\n",file);
      return 0;
   }

   /* In the kerning pairs block... */

   if (fread(dump,sizeof(char),5,fnt_file) == 5) {
      blk_sz = *b_hdr.block_size;
      n      = blk_sz/10;
      err    = 0;
      printf("Loading %d kerning pairs...",n);

      while (n) {
         if (fread(dump,sizeof(char),10,fnt_file) == 10) {
            if (*krn.first > 255 || *krn.second > 255)
               continue;
            f->kern[*krn.first][*krn.second] = *krn.amount;
         }
         else err++;
         n--;
      }
      if (err)
         printf("%d errors were encountered.\n\n",err);
      else printf("done.\n\n");
   }
   else printf("No kerning pairs discovered.\n\n");

   fclose(fnt_file);
   return 1;
}

/* Converted to triangle vertex array on February 25, 2013 */
static void draw(gl_tex_font *f, const char *txt,mat44 *projection, mat44 *modelview)
{
   unsigned n = 0, max_w = 0;
   const char *at;
   int sx ,sy, tab, chr_cnt = 0, sv_max, mono_line = 0;
   mat44 mvp;

   if (!f || !txt)
      return;

   /* Find all character values which will result in vertices... */
   at = txt;
   while (*at) {
      if (*at > ' ' && *at <= '~') n++;
      at++;
   }

    /* ...upsize the coordinate arrays if needed... */
   if (f->vtx_sz < 6 * n) {
      f->vtx = realloc(f->vtx, 6 * n * sizeof(vec2));
      f->tex = realloc(f->tex, 6 * n * sizeof(vec2));
      if (!f->vtx || !f->tex) {
         printf("vertex arrays failed on allocation...\n");
         f->vtx_sz = 0;
         return;
      }
      f->vtx_sz = 6 * n;
   }

   /* ...set up the cursor, etc... */
   if (f->pseudo_mono && !f->use_w)
      f->use_w = f->pad['r'];

   tab = f->tab_stop * ((f->pseudo_mono) ? f->use_w : f->pad['r']);
   sx  = 0;
   sy  = -f->max_h;
   n   = 0;
   at  = txt;

   /* ...parse the character string... */
   while (*at) {
      switch (*at) {
         case '\n':
            sy -= f->line_h;
         case '\r':
            if (sx > max_w)
               max_w = sx;
            sx = 0;
            mono_line = 0;
            chr_cnt = 0;
            break;
         case '\t':
            if (f->pseudo_mono) {
               mono_line += f->use_w;
               sx = mono_line;
            }
            else sx++;

            while (sx%tab)
               sx++;
            chr_cnt += f->tab_stop;
            mono_line = sx;
            break;
         case ' ':
            if (f->pseudo_mono) {
               mono_line += f->use_w;
               sx = mono_line;
            }
            else sx += f->pad[' '] + f->kern[' '][(*at + 1)];

            chr_cnt++;
            break;
         default:

            /* T1: lower left... */
            f->vtx[n].x = sx + f->x_off[*at];
            f->vtx[n].y = sy + f->y_off[*at];
            f->tex[n].x = f->x[*at]/(double)f->tex_w;
            f->tex[n].y = f->y[*at]/(double)f->tex_h;

            /* T1: upper right... */
            f->vtx[n + 1].x = f->vtx[n].x;
            f->vtx[n + 1].y = f->vtx[n].y + f->h[*at];
            f->tex[n + 1].x = f->tex[n].x;
            f->tex[n + 1].y = f->tex[n].y + f->h[*at]/(double)f->tex_h;

            /* T1: upper left... */
            f->vtx[n + 2].x = f->vtx[n].x + f->w[*at];
            f->vtx[n + 2].y = f->vtx[n + 1].y;
            f->tex[n + 2].x = f->tex[n].x + f->w[*at]/(double)f->tex_w;
            f->tex[n + 2].y = f->tex[n + 1].y;

            /* T2: lower right... */
            f->vtx[n + 3].x = f->vtx[n + 2].x;
            f->vtx[n + 3].y = f->vtx[n].y;
            f->tex[n + 3].x = f->tex[n + 2].x;
            f->tex[n + 3].y = f->tex[n].y;

            /* T2: lower left... */
            f->vtx[n + 4].x = sx + f->x_off[*at];
            f->vtx[n + 4].y = sy + f->y_off[*at];
            f->tex[n + 4].x = f->x[*at]/(double)f->tex_w;
            f->tex[n + 4].y = f->y[*at]/(double)f->tex_h;

            /* T2: upper left... */
            f->vtx[n + 5].x = f->vtx[n].x + f->w[*at];
            f->vtx[n + 5].y = f->vtx[n + 1].y;
            f->tex[n + 5].x = f->tex[n].x + f->w[*at]/(double)f->tex_w;
            f->tex[n + 5].y = f->tex[n + 1].y;


            n += 6;

            mono_line += f->use_w;
            sx += f->pad[*at] + f->kern[*at][*(at + 1)];
            chr_cnt++;
            break;
      }
      at++;

      /* ...force a new line if for a fixed width... */
      if (*at && f->cpl && chr_cnt >= f->cpl && *at != '\n' && *at != '\r') {
         sy -= f->line_h;
         sx = 0;
         chr_cnt = 0;
      }
   }
   if (sx > max_w)
      max_w = sx;

   sv_max = max_w;

   /* ...determine the document scaling and centering factors... */
   if (f->cpl)
      max_w = f->cpl * f->pad['r'];
   if (!f->cpl && -sy > max_w)
      max_w = -sy;
   if (sv_max > max_w)
      sv_max = max_w;

   /* ...normalize the document... */
   if (!f->justify_top)
      gluf_translate_2d(f->vtx,n,0,-sy);
   if (f->centered)
      gluf_translate_2d(f->vtx,n,(max_w - sv_max)/2,0);
   gluf_scale_2d(f->vtx,n,1.0/max_w,1.0/max_w);
   if (f->justify_top)
      gluf_translate_2d(f->vtx,n,0,1);

   /* Make external calls aware of the number of used vertices... */
   f->use_vtx = n;

   /* ...and draw the document. */
   if (f->render && projection && modelview) {
      mvp = *projection;
      mat44_mult(&mvp,modelview);

      ui_2d_shader.um44mvp    = &mvp;
      ui_2d_shader.av2vertex  = f->vtx;
      ui_2d_shader.texture    = f->tex_id;
      ui_2d_shader.av2texture = f->tex;
      ui_2d_shader.uv4color   = f->color;
      ui_2d_shader.count      = n;

      shader_program_execute(&ui_2d_shader);
   }
}

#ifdef __cplusplus
}
#endif
