#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include "display.h"
#include "stage.h"
#include "ui_element.h"
#include "renderer_api.h"
#include "renderer_api.h"
#include "aluf.h"
#include "std_decor.h"
#include "recorder_controls.h"
#include "display_controls.h"
#include "file_parser.h"
#include "hotspot.h"
#include "instruments_scene.h"
#include "obj_model.h"
#include "model_shader_nice.h"
#include "model_shader_nicest.h"

#define SPHERE_RADIUS   4000
#define HOTSPOT_FILE    "./data/instrument_spots.txt"
#define BUBBLER_OUT     "./data/new_bubbles.txt"
#define VIEWER_RADIUS   .5
#define BUBBLER_RADIUS  .5

#define AUDIO_FILE      "./data/audio/scan.ogg"
#define CONFIG_ID       "INSTRUMENTS_SCENE"


enum {ctl_exit, ctl_menu, ctl_options, ctl_rec, ctl_text, ctl_nar, ctl_fade,
      ctl_zone, ctl_bub, ctl_build, ctl_bub_sz, ctl_bub_sav};

typedef struct {

   IMPLEMENTS_SCENE_API

   ui_viewer            *viewer;
   ui_balloon           *menu;
   ui_balloon           *more;
   ui_element           *cmn;

   stream_src           *stream;
   recorder_controls    *rc;
   c3d_ui_hotspot       *hs;
   c3d_element          *me;
   c3d_update           upd;

   obj_model            instrument_model;
   unsigned             out_tex;
   unsigned             planck_list;

   unsigned             show_text;
   unsigned             use_narration;
   float                cross_fade;

   unsigned             allow_3d_options;
   unsigned             allow_recording;

   /* Auto trail creator */
   unsigned             bubbler;
   loc_list             *bubbles;
   loc_list             *last_bubble;
   unsigned             show_zones;
   unsigned             show_builder;
   c3d_ui_hotspot       *builder;
   c3d_element          *me_too;
   float                bub_sz;
   float                bub_exp;
   char                 *bubble_def;
   unsigned             def_sz;
}
instruments_scene;

static void init(scene_api *object);
static void update(scene_api *object);
static void draw(scene_api *object);
static void enter(scene_api *object);
static void leave(scene_api *object);
static void erase(scene_api *object);
static void audio_level(scene_api *object,float lvl);
static void configuration(instruments_scene *i);

static void generic_context_begin(void* owner);
static void generic_context_end(void* owner);
static void vbo_context(void *owner, unsigned id);
static void switchboard(void *obj,ui_update *u);
static void on_menu(ui_balloon *b, ui_update *u);
static void no_menu(ui_balloon *b, ui_update *u);
static void collision_update(void* self,c3d_update* info);

static void dump_bubbles(instruments_scene *i);
static void bubbles_to_file(instruments_scene *i);
static void bubble_check(instruments_scene *i);
static void bubbles_add_all(instruments_scene *i);
static void bubbles_remove_all(instruments_scene *i);
static void erase_bubbles(loc_list *ls);

static void string_append(char **buff, unsigned *sz, const char *add);

static instruments_scene instruments_scene_proto = {
   &init,
   &update,
   &draw,
   &enter,
   &leave,
   &erase,
   &audio_level,
   .5,         /* Illumination */
   0,
   1.75,       /* Default eye separation */
   55.0,       /* Default focal plane */
   0.5,        /* Default front clipping plane */
   0,0,0,
   0,0,0,
   0,0,0,
   0,
               /* Begin instruments_scene variables */
   0,0,0,0,
   0,0,0,0,
   0,0,0,0,0,0,0,0,
   0,0,0,0,
   1,1,.5,1,1,

   0,0,0,0,0,0,0,BUBBLER_RADIUS, -1,0,0
};

static unsigned vbo;
static geom_list *model;
static model_shader_nicest_t model_shader1;
static model_shader_nice_t model_shader0;

static void string_append(char **buff, unsigned *sz, const char *add)
{
   unsigned str_sz;

   if (!buff || !sz || !add || !(*add)) {
      printf("returned right away...\n");
      return;
   }

   str_sz = strlen(add);

   if (!(*buff)) {
      *buff = (char*)calloc(str_sz + 1, sizeof(char));
      if (*buff) {
         strncpy(*buff,add,str_sz);
         *sz = str_sz + 1;
      }
      else  printf("string append failed on allocation\n");
      return;
   }

   *buff = (char*)realloc(*buff,str_sz + *sz);
   if (*buff) {
      strncpy(*buff + *sz - 1 ,add,str_sz);
      *sz += str_sz;
      *(*buff + *sz - 1) = 0;
   }
   else {
      printf("string append failed on reallocation\n");
      exit(1);
   }
}

static void configuration(instruments_scene *i)
{
   i->cl = create_config_loader(CONFIG_ID);
   aggregate_register_config(i->ca,i->cl);

   config_doc(i->cl,
      "Identical names/functions as Planck scene variables...");

   config_bind(
      i->cl,
      "eye_separation",
      CONFIG_FLOAT,
      sizeof(float),
      &i->eye_seperation);
   config_bind(
      i->cl,
      "focal_plane",
      CONFIG_FLOAT,
      sizeof(float),
      &i->focal_plane);
   config_bind(
      i->cl,
      "front_clip",
      CONFIG_FLOAT,
      sizeof(float),
      &i->clipping_plane);
   config_bind(
      i->cl,
      "allow_3d_options",
      CONFIG_UINT,
      sizeof(unsigned),
      &i->allow_3d_options);
   config_bind(
      i->cl,
      "allow_recording",
      CONFIG_UINT,
      sizeof(unsigned),
      &i->allow_recording);

      aggregate_load_config(i->ca,i->cl);
}

scene_api *create_instruments_scene(unsigned designer, config_aggregate *ca)
{
   instruments_scene *create;
   ui_icon *i;
   ui_container *s;
   ui_balloon *b;
   ui_button *bt;
   ui_slider *sl;

   create = (instruments_scene*)calloc(1, sizeof(instruments_scene));

   if (create) {
      *create = instruments_scene_proto;

      create->context = renderer_generate_unique_id();
      renderer_set_context_group(create->context);

      create->ca = ca;
      configuration(create);

      create->viewer = create_viewer(frame_state.overlay_area);
      create->viewer->pointer = 0;

      create->instrument_model = obj_model_get("planck_9_6",1);
      model = obj_model_to_geom_list(create->instrument_model);
      model->light.cutoff_distance = 300;
      model->light.minimum_level   = .5;
      vbo  = renderer_register_vertex_buffer_function(create,&vbo_context);
      renderer_register_generic(create,&generic_context_begin,&generic_context_end);

      create->viewer = create_viewer(frame_state.overlay_area);
      create->viewer->pointer = 0;
      s = create->viewer->screen;

      /* The collision text notifier... */
      create->hs = create_c3d_ui_hotspot(HOTSPOT_FILE,FROM_FILE);
      if (create->hs) {
         create->me = c3d_create_element(in_volume(0,0,0,VIEWER_RADIUS));
         create->me->update = &collision_update;
         create->me->initiator = 1;
         create->hs->spots->add_object(create->hs->spots,create->me);
         create->hs->spots->add_collidable(create->hs->spots,create->me);
         s->add_item(s,create->hs);

         create->hs->set(create->hs,HS_SHOW_VISUAL,create->show_text);
         create->hs->set(create->hs,HS_USE_AUDIO,create->use_narration);
      }

      /* The pop up menu... */
      b = create_balloon();
      b->owner = (void*)create;
      b->button_mask = 48;
      b->on_use = &switchboard;
      b->on_grow = &on_menu;
      b->on_shrink = &no_menu;
      b->id = ctl_menu;
      s->add_item(s,b);
      create->menu = b;

      /* Control board background... */
      i = create_custom_icon(.1);
      i->place(i,0,0,1,.5);
      i->look = material_is(0,0,0,.85,0);
      b->add_item(b,i);

      /* Exit button... */
      bt = create_button();
      bt->place(bt,.85,.22,.1,.1);
      *((ui_decor*)&bt->look) = std_rb_decor;
      bt->id = ctl_exit;
      bt->on_use = &switchboard;
      bt->owner = (void*)create;
      b->add_item(b,bt);
      b->add_item(b,create_item_label(bt,"Exit"));


      /* "More" button... */
      if (create->allow_3d_options) {
         bt = create_button();
         bt->place(bt,.05,.27,.1,.1);
         *((ui_decor*)&bt->look) = std_bl_decor;
         bt->id = ctl_options;
         bt->on_use = &switchboard;
         bt->owner = (void*)create;
         b->add_item(b,bt);
         b->add_item(b,create_item_label(bt,"More"));
      }

      /* Record button... */
      if (create->allow_recording) {
         bt = create_button();
         bt->place(bt,.05,.13,.1,.1);
         *((ui_decor*)&bt->look) = std_bl_decor;
         bt->id = ctl_rec;
         bt->on_use = &switchboard;
         bt->owner = (void*)create;
         b->add_item(b,bt);
         b->add_item(b,create_item_label(bt,"Record"));
      }

      if (designer) {

         /* Zones button... */
         bt = create_button();
         bt->place(bt,.5,0,.1,.1);
         *((ui_decor*)&bt->look) = std_gr_gn_decor;
         bt->id = ctl_zone;
         bt->latches = 1;
         bt->value = create->show_zones;
         bt->on_use = &switchboard;
         bt->owner = (void*)create;
         b->add_item(b,bt);
         b->add_item(b,create_item_label(bt,"Zones"));


         /* Builder button... */
         bt = create_button();
         bt->place(bt,.62,0,.1,.1);
         *((ui_decor*)&bt->look) = std_gr_gn_decor;
         bt->id = ctl_build;
         bt->latches = 1;
         bt->value = create->show_builder;
         bt->on_use = &switchboard;
         bt->owner = (void*)create;
         b->add_item(b,bt);
         b->add_item(b,create_item_label(bt,"Sequence"));

         /* Delete button... */
         bt = create_button();
         bt->place(bt,.74,0,.1,.1);
         *((ui_decor*)&bt->look) = std_rb_decor;
         bt->id = ctl_bub_sav;
         bt->on_use = &switchboard;
         bt->owner = (void*)create;
         b->add_item(b,bt);
         b->add_item(b,create_item_label(bt,"Save"));

         /* Bubbler button... */
         bt = create_button();
         bt->place(bt,.86,0,.1,.1);
         *((ui_decor*)&bt->look) = std_gr_gn_decor;
         bt->id = ctl_bub;
         bt->latches = 1;
         bt->value = create->bubbler;
         bt->on_use = &switchboard;
         bt->owner = (void*)create;
         b->add_item(b,bt);
         b->add_item(b,create_item_label(bt,"Build Sequence"));
      }

      /* Visuals button... */
      bt = create_button();
      bt->place(bt,.15,0,.1,.1);
      *((ui_decor*)&bt->look) = std_gr_gn_decor;
      bt->id = ctl_text;
      bt->latches = 1;
      bt->value = create->show_text;
      bt->on_use = &switchboard;
      bt->owner = (void*)create;
      b->add_item(b,bt);
      b->add_item(b,create_item_label(bt,"Captions"));

      /* Narration button... */
      bt = create_button();
      bt->place(bt,.3,0,.1,.1);
      *((ui_decor*)&bt->look) = std_gr_gn_decor;
      bt->id = ctl_nar;
      bt->latches = 1;
      bt->value = create->use_narration;
      bt->on_use = &switchboard;
      bt->owner = (void*)create;
      b->add_item(b,bt);
      b->add_item(b,create_item_label(bt,"Narration"));

      if (designer) {

          /* Bubble path increment slider... */
         sl = create_slider();
         sl->place(sl,.2,.15,.6,.05);
         *((ui_decor*)&sl->look) = std_sl_decor;
         ((ui_slider*)sl)->handle_width = .065;
         sl->min = -1;
         sl->span = 8;
         sl->detents = 7;
         sl->detent_width = .05;
         sl->value = create->bub_exp;
         sl->id = ctl_bub_sz;
         sl->on_use = &switchboard;
         sl->owner = (void*)create;
         b->add_item(b,create_item_label(sl,"Sequence Length"));
         b->add_item(b,create_slider_labels(sl,9,
             ".5","1","2","4","8","16","32","64","128"));
         b->add_item(b,sl);
      }

      /* Crossfade slider... */
      sl = create_slider();
      sl->place(sl,.2,.3,.6,.05);
      *((ui_decor*)&sl->look) = std_sl_decor;
      ((ui_slider*)sl)->handle_width = .065;
      sl->min = 0;
      sl->span = 1;
      sl->detents = 1;
      sl->detent_width = .1;
      sl->value = create->cross_fade;
      sl->id = ctl_fade;
      sl->on_use = &switchboard;
      sl->owner = (void*)create;
      b->add_item(b,create_item_label(sl,"Emphasis"));
      b->add_item(b,create_slider_labels(sl,7,
         "","Music","","","","Voice",""));
      b->add_item(b,sl);

      /* More pop up menu... */
      create->more = create_display_common_controls((scene_api*)create);
      b->add_item(b,create->more);

      /* A recorder_control implements ui_balloon, and so can be treated
      as such... */
      create->rc = create_recorder_controls("instruments");
      b->add_item(b,create->rc);
   }

   return (scene_api*)create;
}

static void erase(scene_api *object)
{
   instruments_scene *i = (instruments_scene*)object;

   if (!i)
      return;

   if (i->viewer)
      i->viewer->erase(i->viewer);

   if (i->stream)
      erase_stream_source(i->stream);

   erase_bubbles(i->bubbles);

   if (i->bubble_def)
      free(i->bubble_def);

   obj_model_delete(i->instrument_model);
   geom_list_free(model);

   free(i);
}

static void init(scene_api *object)
{
   instruments_scene *i = (instruments_scene*)object;

   if (!i)
      return;
}

static void vbo_context(void *owner, unsigned id)
{
   renderer_static_buffer_data(id,model->array_bytes,model->array);
}

static void generic_context_begin(void* owner)
{
   if (!shader_program_compile(&model_shader_nicest,1))
      exit(0);

   model_shader1 = model_shader_nicest;

   model_shader1.vbo  = vbo;
   model_shader1.list = model;

   if (!shader_program_compile(&model_shader_nice,1))
      exit(0);

   model_shader0 = model_shader_nice;

   model_shader0.vbo  = vbo;
   model_shader0.list = model;
}

static void generic_context_end(void* owner)
{
   shader_program_free(&model_shader_nicest);
   shader_program_free(&model_shader_nice);
}
static void vbo_context(void *owner, unsigned id);


static void bubbles_to_file(instruments_scene *i)
{
   FILE *out;

   if (!i || !i->bubble_def)
      return;

   out = fopen(BUBBLER_OUT,"a");
   if (!out) {
      printf("Could not open %s\n",BUBBLER_OUT);
      return;
   }

   fprintf(out,"%s",i->bubble_def);

   fclose(out);

   free(i->bubble_def);
   i->bubble_def = 0;
   i->def_sz = 0;
}



static void dump_bubbles(instruments_scene *i)
{
   loc_list *wrt, *sv;
   struct tm *loc_time;
   time_t now;
   char temp[128], title[128];

   const char *at_tmpl = "at { %f,%f,%f,%f }\n";
   const char *blk_strt = "\n%s {\nnar { }\npic { }\n";
   const char *blk_end  = "hdr { %s }\nmsg {\n\n\t }\n}\n";

   if (!i || !i->bubbles)
      return;

   time(&now);
   loc_time = localtime(&now);
   snprintf(
      title,
      128,
      "sequence_%d.%02d_%02d.%02d.%02d",
      loc_time->tm_mon + 1,
      loc_time->tm_mday,
      loc_time->tm_hour,
      loc_time->tm_min,
      loc_time->tm_sec);

   snprintf(temp,128,blk_strt,title);
   string_append(&i->bubble_def,&i->def_sz,temp);

   wrt = i->bubbles;
   while (wrt) {
      sv = wrt->more;
      snprintf(temp,128,at_tmpl,wrt->at.v.x,wrt->at.v.y,wrt->at.v.z,wrt->at.v.r);
      string_append(&i->bubble_def,&i->def_sz,temp);
      wrt = wrt->more;
   }

   snprintf(temp,128,blk_end,title);
   string_append(&i->bubble_def,&i->def_sz,temp);

   bubbles_add_all(i);
   erase_bubbles(i->bubbles);
   i->bubbles = 0;
}

static void bubble_check(instruments_scene *i)
{
   float vec_mag,x_item,y_item,z_item;
   loc_list *end;

   if (!i)
      return;

   end = i->last_bubble;

   x_item  = -i->position[0] - end->at.v.x;
   y_item  = -i->position[1] - end->at.v.y;
   z_item  = -i->position[2] - end->at.v.z;
   vec_mag = sqrt(x_item * x_item + y_item * y_item + z_item * z_item);

   /* if (vec_mag >= 2 * i->bub_sz) { */
   if (vec_mag >= i->bub_sz) {

      end->more = (loc_list*)calloc(1,sizeof(loc_list));
      if (end->more) {
         end->more->at.v.x = -i->position[0];
         end->more->at.v.y = -i->position[1];
         end->more->at.v.z = -i->position[2];
         /* end->more->at.v.r = vec_mag/2; */
         end->more->at.v.r = i->bub_sz;
      }
      i->last_bubble = end->more;
   }
}

static void bubbles_add_all(instruments_scene *i)
{
   if (!i )
      return;

   if (i->builder)
      bubbles_remove_all(i);

   if (!i->bubble_def)
      return;

   i->builder = create_c3d_ui_hotspot(i->bubble_def,FROM_STRING);
   if (i->builder) {
      i->me_too = c3d_create_element(in_volume(0,0,0,VIEWER_RADIUS));
      i->me_too->update = &collision_update;
      i->me_too->initiator = 1;
      i->builder->spots->add_object(i->builder->spots,i->me_too);
      i->builder->spots->add_collidable(i->builder->spots,i->me_too);
      if (i->show_builder)
         i->viewer->screen->to_list_head(i->viewer->screen, i->builder);

      i->builder->set(i->builder,HS_SHOW_VISUAL,1);
      i->builder->set(i->builder,HS_USE_AUDIO,0);
   }
}

static void bubbles_remove_all(instruments_scene *i)
{
   if (!i || !i->builder)
      return;

   i->viewer->screen->remove_item(i->viewer->screen,i->builder);
   i->builder = 0;
   i->me_too  = 0;
}

static void erase_bubbles(loc_list *ls)
{
   loc_list *sv;

   while (ls) {
      sv = ls->more;
      free(ls);
      ls = sv;
   }
}

static void switchboard(void *obj,ui_update *u)
{
   ui_element *e = (ui_element*)obj;
   instruments_scene *i;
   float lvl;

   if (!e)
      return;

   i = (instruments_scene*)e->owner;

   switch (e->id) {
      case ctl_exit:
         i->menu->shrink(i->menu);
         stage_pop_scene();
         break;
      case ctl_menu:
         if (ctl_button(4,u->button))
            i->menu->shrink(i->menu);
         break;
      case ctl_options:
         i->more->grow(i->more);
         break;
      case ctl_rec:
         if (i->rc)
            ((ui_balloon*)i->rc)->grow((ui_balloon*)i->rc);
         break;
      case ctl_text:
         i->show_text = ((ui_button*)e)->value;
         if (i->hs)
            i->hs->set(i->hs,HS_SHOW_VISUAL,i->show_text);
         break;
      case ctl_nar:
         i->use_narration = ((ui_button*)e)->value;
         if (i->hs)
            i->hs->set(i->hs,HS_USE_AUDIO,i->use_narration);
         break;
      case ctl_fade:
         lvl = stage_get_master_audio_level();
         i->cross_fade = ((ui_slider*)e)->value;

         if (i->hs)
            i->hs->set(i->hs,HS_AUDIO_LEVEL,lvl * i->cross_fade);
         if (i->stream)
            alSourcef(i->stream->src,AL_GAIN,lvl * (1 - i->cross_fade));
         break;
      case ctl_zone:
         i->show_zones = ((ui_button*)e)->value;
         break;
      case ctl_bub:
         i->bubbler = ((ui_button*)e)->value;

         if (!i->bubbler) {
            dump_bubbles(i);
            i->bubbles = 0;
            i->last_bubble = 0;
         }
         else {
            bubbles_remove_all(i);

            if (i->bubble_def) {
               free(i->bubble_def);
               i->bubble_def = 0;
               i->def_sz = 0;
            }

            i->bubbles = (loc_list*)calloc(1,sizeof(loc_list));
            if (i->bubbles) {
               i->bubbles->at.v.x = -i->position[0];
               i->bubbles->at.v.y = -i->position[1];
               i->bubbles->at.v.z = -i->position[2];
               i->bubbles->at.v.r = i->bub_sz;
            }
            i->last_bubble = i->bubbles;
         }
         break;
     case ctl_build:
         i->show_builder = ((ui_button*)e)->value;

         if (i->show_builder)
            bubbles_add_all(i);
         else bubbles_remove_all(i);
         break;
     case ctl_bub_sz:
         i->bub_exp = ((ui_slider*)e)->value;
         i->bub_sz = pow(2.0,i->bub_exp);
         break;
     case ctl_bub_sav:
         if (i->bubble_def)
            bubbles_to_file(i);
         break;
   }
}

static void on_menu(ui_balloon *b, ui_update *u)
{
   instruments_scene *i;

   if (!b)
      return;

   i = (instruments_scene*)b->owner;

   i->viewer->pointer = &ui_basic_cursor;
   i->viewer->px = .5;
   i->viewer->py = .17;
   frame_state.rotation = 0;
   frame_state.position = 0;
}

static void no_menu(ui_balloon *b, ui_update *u)
{
   instruments_scene *i;

   if (!b)
      return;

   i = (instruments_scene*)b->owner;

   i->viewer->pointer   = 0;
   frame_state.rotation = i->rotation;
   frame_state.position = i->position;
}


static void update(scene_api *object)
{
   instruments_scene *i = (instruments_scene*)object;

   if (!i)
      return;

   if (i->hs) {
      i->upd.dt    = frame_state.dt;
      i->upd.pos_x = -i->position[0];
      i->upd.pos_y = -i->position[1];
      i->upd.pos_z = -i->position[2];
      i->hs->spots->update(i->hs->spots,&i->upd);
   }

   if (i->show_builder && i->builder)
      i->builder->spots->update(i->builder->spots,&i->upd);

   i->viewer->update(
      i->viewer,
      frame_state.pointer[0],
      frame_state.pointer[1],
      frame_state.dt,
      frame_state.button,
      frame_state.key);

   if (ctl_button(4,frame_state.button))
       i->menu->grow(i->menu);

   if (i->bubbler)
      bubble_check(i);
}

static void collision_update(void* self,c3d_update* u)
{
   c3d_element *c = (c3d_element*)self;
   c->at = in_volume(u->pos_x,u->pos_y,u->pos_z,VIEWER_RADIUS);
}

static void enter(scene_api *object)
{
   instruments_scene *i = (instruments_scene*)object;

   i->position[0] = 0;
   i->position[1] = -100;
   i->position[2] = -600;
   i->rotation[0] = 0;
   i->rotation[1] = 0;

   display_set_movement_acceleration(.2);

   if (!i->stream)
      i->stream = create_stream_source(AUDIO_FILE);

   if (i->stream)
      alSourcePlay(i->stream->src);

   if (i->hs)
      i->hs->set(i->hs,HS_SUSPEND,HS_NO);

   if (i->rc)
      i->rc->resume(i->rc);

   i->cmn = retrieve_common_display();
   if (i->cmn)
      i->viewer->screen->add_item(i->viewer->screen,i->cmn);
}

static void leave(scene_api *object)
{
   instruments_scene *i = (instruments_scene*)object;

   if (i->stream) {
      erase_stream_source(i->stream);
      i->stream = 0;
   }

   if (i->hs)
      i->hs->set(i->hs,HS_SUSPEND,HS_YES);

   if (i->rc)
      i->rc->pause(i->rc);

   if (i->cmn) {
      i->viewer->screen->remove_item(i->viewer->screen,i->cmn);
      i->cmn = 0;
   }
}

static void audio_level(scene_api *object,float lvl)
{
   instruments_scene *i = (instruments_scene*)object;

   if (i->stream)
      alSourcef(i->stream->src,AL_GAIN,lvl * (1 - i->cross_fade));

   if (i->hs)
      i->hs->set(i->hs,HS_AUDIO_LEVEL,lvl * i->cross_fade);
}

static void draw(scene_api *object)
{
   instruments_scene *i = (instruments_scene*)object;
   mat44 tm, mv, mvp = *i->projection;
   vec4 light = {0,10,-1,1};

   if (!i)
      return;


   mat44_rotation(&mv,i->rotation[0],1,0,0);
   mat44_rotation(&tm,i->rotation[1],0,1,0);
   mat44_mult(&mv,&tm);
   mat44_translation(&tm,i->position[0],i->position[1],i->position[2]);
   mat44_mult(&mv,&tm);

   mat44_mult(&mvp,&mv);

  	model_shader1.um44mvp = &mvp;
   model_shader1.um44mv  = &mv;
   model_shader1.list->light.Lp = light;
   shader_program_execute(&model_shader1);

//    model_shader0.um44mvp = &mvp;
//    model_shader0.um44mv  = &mv;
//    model_shader0.list->light.Lp = light;
//    shader_program_execute(&model_shader0);


   if (i->show_zones && i->hs)
      i->hs->spots->display(i->hs->spots);

   if (i->show_builder && i->builder)
      i->builder->spots->display(i->builder->spots);

   i->viewer->display(i->viewer, i->projection);
}

#ifdef __cplusplus
}
#endif
