#include "matrix44.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>


#ifndef DEG
   #define DEG(x) ((x) * 57.295779513)
#endif

mat44 *mat44_create()
{
   return (mat44*)calloc(1,sizeof(mat44));
}

void mat44_destroy(mat44 *m)
{
   if (m)
      free(m);
}

void mat44_print(mat44 *m)
{
   if (!m)
      return;

   printf("%0.3f %0.3f %0.3f %0.3f\n",m->a11,m->a12,m->a13,m->a14);
   printf("%0.3f %0.3f %0.3f %0.3f\n",m->a21,m->a22,m->a23,m->a24);
   printf("%0.3f %0.3f %0.3f %0.3f\n",m->a31,m->a32,m->a33,m->a34);
   printf("%0.3f %0.3f %0.3f %0.3f\n\n",m->a41,m->a42,m->a43,m->a44);
}

void mat44_mult(mat44 *m, mat44 *n)
{
   mat44 r;

   #define EVAL44(i,j) \
      r.a##i##j =\
         m->a##i##1 * n->a1##j +\
         m->a##i##2 * n->a2##j +\
         m->a##i##3 * n->a3##j +\
         m->a##i##4 * n->a4##j

   if (!m || !n)
      return;

   EVAL44(1,1);
   EVAL44(1,2);
   EVAL44(1,3);
   EVAL44(1,4);

   EVAL44(2,1);
   EVAL44(2,2);
   EVAL44(2,3);
   EVAL44(2,4);

   EVAL44(3,1);
   EVAL44(3,2);
   EVAL44(3,3);
   EVAL44(3,4);

   EVAL44(4,1);
   EVAL44(4,2);
   EVAL44(4,3);
   EVAL44(4,4);

   memcpy(m,&r,sizeof(mat44));
}

void mat44_identity(mat44 *m)
{
   if (!m)
      return;

   memset(m,0,sizeof(mat44));

   m->a11 = 1.0;
   m->a22 = 1.0;
   m->a33 = 1.0;
   m->a44 = 1.0;
}

void mat44_translation(mat44 *m, FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   if (!m)
      return;

   memset(m,0,sizeof(mat44));

   m->a11 = 1.0;
   m->a22 = 1.0;
   m->a33 = 1.0;
   m->a44 = 1.0;

   m->a14 = x;
   m->a24 = y;
   m->a34 = z;
}

void mat44_rotation(mat44 *m, FLOAT_T rad, FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   /* Based on article at:
         http://www.fastgraph.com/makegames/3drotation/
      with unified rotation matrix formula (column-major order):

         txx + c    txy - sz   txz + sy   0
         tyx + sz   tyy + c    tyz - sx   0
         tzx - sy   tzy + sx   tzz + c    0
         0          0          0          1

         c = cos(rad)
         s = sin(rad)
         t = 1 - cos(rad)


   Using combined term variables for speed up... */
   FLOAT_T c, s, t, tx, ty, tz, sx, sy, sz;

   if (!m)
      return;

   c  = cos(rad);
   s  = sin(rad);
   t  = 1.0 - c;
   tx = t * x;
   ty = t * y;
   tz = t * z;
   sx = s * x;
   sy = s * y;
   sz = s * z;

   m->a11 = tx * x + c;
   m->a12 = tx * y - sz;
   m->a13 = tx * z + sy;
   m->a14 = 0;

   m->a21 = ty * x + sz;
   m->a22 = ty * y + c;
   m->a23 = ty * z - sx;
   m->a24 = 0;

   m->a31 = tz * x - sy;
   m->a32 = tz * y + sx;
   m->a33 = tz * z + c;
   m->a34 = 0;

   m->a41 = 0;
   m->a42 = 0;
   m->a43 = 0;
   m->a44 = 1.0;
}

void mat44_scaling(mat44 *m, FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   if (!m)
      return;

   memset(m,0,sizeof(mat44));

   m->a11 = x;
   m->a22 = y;
   m->a33 = z;
   m->a44 = 1.0;
}

void mat44_transpose(mat44 *m)
{
   FLOAT_T temp;

   if (!m)
      return;

   temp   = m->a21;
   m->a21 = m->a12;
   m->a12 = temp;

   temp   = m->a31;
   m->a31 = m->a13;
   m->a13 = temp;

   temp   = m->a41;
   m->a41 = m->a14;
   m->a14 = temp;

   temp   = m->a23;
   m->a23 = m->a32;
   m->a32 = temp;

   temp   = m->a24;
   m->a24 = m->a42;
   m->a42 = temp;

   temp   = m->a34;
   m->a34 = m->a43;
   m->a43 = temp;
}

INT_T mat44_invert(mat44 *m)
{

   /* Direct implementation of pseudocode from article at:
         http://en.wikipedia.org/wiki/Gaussian_elimination

      Formal Gauss elimination algorithm:

      for k = 1 ... m:

         Find pivot for column k:
            i_max := argmax (i = k ... m, abs(A[i, k]))
            if A[i_max, k] = 0
               error "Matrix is singular!"

         swap rows(k, i_max)

         Do for all rows below pivot:
            for i = k + 1 ... m:

               Do for all remaining elements in current row:
                  for j = k + 1 ... n:
                     A[i, j] := A[i, j] - A[k, j] * (A[i, k] / A[k, k])

               Fill lower triangular matrix with zeros:
                  A[i, k] := 0

   Verified to work as expected, i.e. invertable matrix * inverse = identity.

   TODO: Unroll loops similar to Mesa GL's implementation.. */


   if (m) {
      #define SQUARE_DIMENSION 4
      #define COL 8
      #define ROW 4

      float r1[COL] = {m->a11, m->a12, m->a13, m->a14, 1,0,0,0};
      float r2[COL] = {m->a21, m->a22, m->a23, m->a24, 0,1,0,0};
      float r3[COL] = {m->a31, m->a32, m->a33, m->a34, 0,0,1,0};
      float r4[COL] = {m->a41, m->a42, m->a43, m->a44, 0,0,0,1};
      float *A[ROW] = {r1,r2,r3,r4};

      float *swap, val, val_max;
      int k = 0, i, j, i_max;

      /* Begin Gauss reduction implementation... */

      while (k < ROW) {

         /* Find pivot in column k */
         i = k + 1;
         i_max = k;
         val_max = fabs(A[k][k]);
         while (i < ROW) {
            val = fabs(A[i][k]);
            if (val > val_max) {
               i_max = i;
               val_max = val;
            }
            i++;
         }

         if (!A[i_max][k]) {
            printf("Matrix is singular!\n");
            return 0;
         }

         /* Swap rows.. */
         swap     = A[k];
         A[k]     = A[i_max];
         A[i_max] = swap;

         /* Reduce... */
         i = k + 1;
         while (i < ROW) {
            j = k + 1;
            while (j < COL) {
               if (A[k][j] && A[i][k])
                  A[i][j] -= A[k][j] * A[i][k]/A[k][k];
               j++;
            }
            A[i][k] = 0;
            i++;
         }

         k++;
      }
      /* End Gauss reduction implementation. */

      /* Now, transform to reduced row echelon form... */

      k = ROW - 1;
      while (k >= 0) {
         i = k - 1;
         while (i >= 0) {
            j = k + 1;
            while (j < COL) {
               if (A[k][j] && A[i][k])
                  A[i][j] -= A[k][j] * A[i][k]/A[k][k];
               j++;
            }
            A[i][k] = 0;
            i--;
         }
         k--;
      }

      k = 0;
      while (k < ROW) {
         j = SQUARE_DIMENSION;
         while (j < COL) {
            if (A[k][j])
               A[k][j] /= A[k][k];
            j++;
         }
         A[k][k] = 1;
         k++;
      }

      /* Copy back to input matrix... */

      m->a11 = A[0][4];
      m->a12 = A[0][5];
      m->a13 = A[0][6];
      m->a14 = A[0][7];

      m->a21 = A[1][4];
      m->a22 = A[1][5];
      m->a23 = A[1][6];
      m->a24 = A[1][7];

      m->a31 = A[2][4];
      m->a32 = A[2][5];
      m->a33 = A[2][6];
      m->a34 = A[2][7];

      m->a41 = A[3][4];
      m->a42 = A[3][5];
      m->a43 = A[3][6];
      m->a44 = A[3][7];

      return 1;
   }
   return 0;
}

void mat44_perspective( mat44 *m, FLOAT_T left, FLOAT_T right, FLOAT_T bottom, FLOAT_T top, FLOAT_T near, FLOAT_T far)
{
   if (!m)
      return;

   m->a11 = 2 * near / (right - left);
   m->a12 = 0;
   m->a13 = (right + left) / (right - left);
   m->a14 = 0;
   m->a21 = 0;
   m->a22 = 2 * near / (top - bottom);
   m->a23 = (top + bottom) / (top - bottom);
   m->a24 = 0;
   m->a31 = 0;
   m->a32 = 0;
   m->a33 = -(far + near) / (far - near);
   m->a34 = -2 * far * near / (far - near);
   m->a41 = 0;
   m->a42 = 0;
   m->a43 = -1.0;
   m->a44 = 0;
}

void mat44_orthographic(mat44 *m,FLOAT_T left, FLOAT_T right, FLOAT_T bottom, FLOAT_T top, FLOAT_T near, FLOAT_T far)
{
   if (!m)
      return;

   m->a11 = 2 / (right - left);
   m->a12 = 0;
   m->a13 = 0;
   m->a14 = -(right + left) / (right - left);
   m->a21 = 0;
   m->a22 = 2 / (top - bottom);
   m->a23 = 0;
   m->a24 = -(top + bottom) / (top - bottom);
   m->a31 = 0;
   m->a32 = 0;
   m->a33 = -2 / (far - near);
   m->a34 = -(far + near) / (far - near);
   m->a41 = 0;
   m->a42 = 0;
   m->a43 = 0;
   m->a44 = -1;
}

mat33 *mat33_create(void)
{
   return (mat33*)calloc(1,sizeof(mat33));
}

void mat33_from_mat44(mat33 *n, mat44 *m)
{
   if (!n || !m)
      return;

   n->a11 = m->a11;
   n->a12 = m->a12;
   n->a13 = m->a13;
   n->a21 = m->a21;
   n->a22 = m->a22;
   n->a23 = m->a23;
   n->a31 = m->a31;
   n->a32 = m->a32;
   n->a33 = m->a33;
}

void mat33_destroy(mat33 *m)
{
   if (m)
      free(m);
}

void mat33_mult(mat33 *m, mat33 *n)
{
	mat33 r;

   #define EVAL33(i,j) \
      r.a##i##j =\
         m->a##i##1 * n->a1##j +\
         m->a##i##2 * n->a2##j +\
         m->a##i##3 * n->a3##j

   if (!m || !n)
      return;

   EVAL33(1,1);
   EVAL33(1,2);
   EVAL33(1,3);

   EVAL33(2,1);
   EVAL33(2,2);
   EVAL33(2,3);

   EVAL33(3,1);
   EVAL33(3,2);
   EVAL33(3,3);

   memcpy(m,&r,sizeof(mat33));
}

void mat33_identity(mat33 *m)
{
	if (!m)
      return;

   memset(m,0,sizeof(mat33));

   m->a11 = 1.0;
   m->a22 = 1.0;
   m->a33 = 1.0;
}

void mat33_scaling(mat33 *m, FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   if (!m)
      return;

   memset(m,0,sizeof(mat33));

   m->a11 = x;
   m->a22 = y;
   m->a33 = z;
}

void mat33_rotation(mat33 *m, FLOAT_T rad, FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   FLOAT_T c, s, t, tx, ty, tz, sx, sy, sz;

   if (!m)
      return;

   c  = cos(rad);
   s  = sin(rad);
   t  = 1.0 - c;
   tx = t * x;
   ty = t * y;
   tz = t * z;
   sx = s * x;
   sy = s * y;
   sz = s * z;

   m->a11 = tx * x + c;
   m->a12 = tx * y - sz;
   m->a13 = tx * z + sy;

   m->a21 = ty * x + sz;
   m->a22 = ty * y + c;
   m->a23 = ty * z - sx;

   m->a31 = tz * x - sy;
   m->a32 = tz * y + sx;
   m->a33 = tz * z + c;
}

void mat33_transpose(mat33 *m)
{
   FLOAT_T temp;

   if (!m)
      return;

   temp   = m->a21;
   m->a21 = m->a12;
   m->a12 = temp;

   temp   = m->a31;
   m->a31 = m->a13;
   m->a13 = temp;

   temp   = m->a23;
   m->a23 = m->a32;
   m->a32 = temp;
}

void mat33h_translation(mat33 *m, FLOAT_T x, FLOAT_T y)
{
   if (!m)
      return;

   memset(m,0,sizeof(mat33));

   m->a11 = 1.0;
   m->a22 = 1.0;
   m->a33 = 1.0;

   m->a13 = x;
   m->a23 = y;
}

void mat33h_rotation(mat33 *m, FLOAT_T rad)
{
	if (!m)
      return;

   memset(m,0,sizeof(mat33));

   m->a11 = cos(rad);
   m->a12 = -sin(rad);
   m->a21 = sin(rad);
   m->a22 = cos(rad);
   m->a33 = 1.0;
}

void mat33h_scaling(mat33 *m, FLOAT_T x, FLOAT_T y)
{
	if (!m)
      return;

   memset(m,0,sizeof(mat33));

   m->a11 = x;
   m->a22 = y;
   m->a33 = 1.0;
}


mat22 *mat22_create(void)
{
	return (mat22*)calloc(1,sizeof(mat22));
}

void mat22_from_mat33(mat22 *n, mat33 *m)
{
	if (!n || !m)
      return;

   n->a11 = m->a11;
   n->a12 = m->a12;
   n->a21 = m->a21;
   n->a22 = m->a22;
}

void mat22_destroy(mat22 *m)
{
	if (m)
      free(m);
}

void mat22_mult(mat22 *m, mat22 *n)
{
	mat22 r;

   #define EVAL22(i,j) \
      r.a##i##j =\
         m->a##i##1 * n->a1##j +\
         m->a##i##2 * n->a2##j

   if (!m || !n)
      return;

   EVAL22(1,1);
   EVAL22(1,2);

   EVAL22(2,1);
   EVAL22(2,2);

   memcpy(m,&r,sizeof(mat22));
}

void mat22_identity(mat22 *m)
{
	if (!m)
      return;

   m->a11 = 1.0;
   m->a12 = 0.0;
   m->a21 = 0.0;
   m->a22 = 1.0;
}

void mat22_rotation(mat22 *m, FLOAT_T rad)
{
	if (!m)
      return;

   m->a11 = cos(rad);
   m->a12 = -sin(rad);
   m->a21 = sin(rad);
   m->a22 = cos(rad);
}

void mat22_scaling(mat22 *m, FLOAT_T x, FLOAT_T y)
{
	if (!m)
      return;

   m->a11 = x;
   m->a12 = 0.0;
   m->a21 = 0.0;
   m->a22 = y;
}

void mat22_transpose(mat22 *m)
{
	FLOAT_T temp;

   if (!m)
      return;

   temp   = m->a21;
   m->a21 = m->a12;
   m->a12 = temp;
}

void vec4_set(vec4 *v, FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   if (!v)
      return;

   v->x = x;
   v->y = y;
   v->z = z;
   v->w = 1.0;
}

FLOAT_T vec4_magnitude(vec4 *v)
{
   if (!v)
      return 0;

   return sqrt(v->x * v->x + v->y * v->y + v->z * v->z);
}

void vec4_add(vec4 *v, vec4 *to_v)
{
   if (!v || !to_v)
      return;

   v->x = v->x + to_v->x;
   v->y = v->y + to_v->y;
   v->z = v->z + to_v->z;
}


void vec4_subtract(vec4 *v, vec4 *from_v)
{
   if (!v || !from_v)
      return;

   v->x = v->x - from_v->x;
   v->y = v->y - from_v->y;
   v->z = v->z - from_v->z;
}

void vec4_scale(vec4 *v, FLOAT_T scaler)
{
    if (!v)
      return;

   v->x *= scaler;
   v->y *= scaler;
   v->z *= scaler;
}

FLOAT_T vec4_dot(vec4 *v, vec4 *v2)
{
   if (!v || !v2)
      return 0;

   return v->x * v2->x + v->y * v2->y + v->z * v2->z;
}

vec4 vec4_cross(vec4 *v, vec4 *v2)
{
   vec4 ret = {0,0,0,1};

   if (!v || !v2)
      return ret;

   ret.x = v->y * v2->z - v->z * v2->y;
   ret.y = v->z * v2->x - v->x * v2->z;
   ret.z = v->x * v2->y - v->y * v2->x;
   ret.w = 1;

   return ret;
}

void vec4_normalize(vec4 *v)
{
   FLOAT_T mag;

   if (!v)
      return;

   mag = sqrt(v->x * v->x + v->y * v->y + v->z * v->z);

   v->x /= mag;
   v->y /= mag;
   v->z /= mag;
   v->w  = 1;
}

void vec4_transform(vec4 *v, mat44 *m)
{
   FLOAT_T r[4];

   #define EVAL_ROW4(i) \
      r[i - 1] =\
         m->a##i##1 * v->x +\
         m->a##i##2 * v->y +\
         m->a##i##3 * v->z +\
         m->a##i##4 * v->w

   if (!v || !m)
      return;

   EVAL_ROW4(1);
   EVAL_ROW4(2);
   EVAL_ROW4(3);
   EVAL_ROW4(4);

   *v = *((vec4*)r);
}



void vec3_set(vec3 *v, FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   if (!v)
      return;

   v->x = x;
   v->y = y;
   v->z = z;
}

FLOAT_T vec3_magnitude(vec3 *v)
{
	if (!v)
      return 0;

   return sqrt(v->x * v->x + v->y * v->y + v->z * v->z);
}

void vec3_add(vec3 *v, vec3 *to_v)
{
   if (!v || !to_v)
      return;

   v->x = v->x + to_v->x;
   v->y = v->y + to_v->y;
   v->z = v->z + to_v->z;
}


void vec3_subtract(vec3 *v, vec3 *from_v)
{
   if (!v || !from_v)
      return;

   v->x = v->x - from_v->x;
   v->y = v->y - from_v->y;
   v->z = v->z - from_v->z;
}

void vec3_scale(vec3 *v, FLOAT_T scaler)
{
    if (!v)
      return;

   v->x *= scaler;
   v->y *= scaler;
   v->z *= scaler;
}

FLOAT_T vec3_dot(vec3 *v, vec3 *v2)
{
   if (!v || !v2)
      return 0;

   return v->x * v2->x + v->y * v2->y + v->z * v2->z;
}

vec3 vec3_cross(vec3 *v, vec3 *v2)
{
   vec3 ret = {0,0,0};

   if (!v || !v2)
      return ret;

   ret.x = v->y * v2->z - v->z * v2->y;
   ret.y = v->z * v2->x - v->x * v2->z;
   ret.z = v->x * v2->y - v->y * v2->x;

   return ret;
}

void vec3_normalize(vec3 *v)
{
   FLOAT_T mag;

   if (!v)
      return;

   mag = sqrt(v->x * v->x + v->y * v->y + v->z * v->z);

   v->x /= mag;
   v->y /= mag;
   v->z /= mag;
}

void vec3_transform(vec3 *v, mat33 *m)
{
   FLOAT_T r[3];

   #define EVAL_ROW3(i) \
      r[i - 1] =\
         m->a##i##1 * v->x +\
         m->a##i##2 * v->y +\
         m->a##i##3 * v->z

   if (!v || !m)
      return;

   EVAL_ROW3(1);
   EVAL_ROW3(2);
   EVAL_ROW3(3);

   *v = *((vec3*)r);
}

void vec3_transform44(vec3 *v, mat44 *m)
{
	FLOAT_T r[3];

   #define EVAL_ROW34(i) \
      r[i - 1] =\
         m->a##i##1 * v->x +\
         m->a##i##2 * v->y +\
         m->a##i##3 * v->z +\
         m->a##i##4

   if (!v || !m)
      return;

   EVAL_ROW34(1);
   EVAL_ROW34(2);
   EVAL_ROW34(3);

   *v = *((vec3*)r);
}


void vec3h_set(vec3 *v,FLOAT_T x, FLOAT_T y)
{
	if (!v)
      return;

   v->x = x;
   v->y = y;
   v->w = 1.0;
}

FLOAT_T vech3_magnitude(vec3 *v)
{
	if (!v)
      return 0;

   return sqrt(v->x * v->x + v->y * v->y);
}

void vec3h_add(vec3 *v, vec3 *to_v)
{
	if (!v || !to_v)
      return;

   v->x = v->x + to_v->x;
   v->y = v->y + to_v->y;
   v->w = 1.0;
}

void vec3h_subtract(vec3 *v, vec3 *from_v)
{
	if (!v || !from_v)
      return;

   v->x = v->x - from_v->x;
   v->y = v->y - from_v->y;
   v->w = 1.0;
}

void vec3h_scale(vec3 *v, FLOAT_T scaler)
{
	if (!v)
      return;

   v->x *= scaler;
   v->y *= scaler;
   v->w = 1.0;
}
FLOAT_T vec3h_dot(vec3 *v, vec3 *v2)
{
	 if (!v || !v2)
      return 0;

   return v->x * v2->x + v->y * v2->y;
}

void vec3h_normalize(vec3 *v)
{
	FLOAT_T mag;

   if (!v)
      return;

   mag = sqrt(v->x * v->x + v->y * v->y);

   v->x /= mag;
   v->y /= mag;
   v->z = 1.0;
}



void vec2_set(vec2 *v,FLOAT_T x, FLOAT_T y)
{
	if (!v)
      return;

   v->x = x;
   v->y = y;
}

FLOAT_T vec2_magnitude(vec2 *v)
{
	if (!v)
      return 0;

   return sqrt(v->x * v->x + v->y * v->y);
}

void vec2_add(vec2 *v, vec2 *to_v)
{
	if (!v || !to_v)
      return;

   v->x = v->x + to_v->x;
   v->y = v->y + to_v->y;
}

void vec2_subtract(vec2 *v, vec2 *from_v)
{
	if (!v || !from_v)
      return;

   v->x = v->x - from_v->x;
   v->y = v->y - from_v->y;
}

void vec2_scale(vec2 *v, FLOAT_T scaler)
{
	if (!v)
      return;

   v->x *= scaler;
   v->y *= scaler;
}

FLOAT_T vec2_dot(vec2 *v, vec2 *v2)
{
	 if (!v || !v2)
      return 0;

   return v->x * v2->x + v->y * v2->y;
}

void vec2_normalize(vec2 *v)
{
	FLOAT_T mag;

   if (!v)
      return;

   mag = sqrt(v->x * v->x + v->y * v->y);

   v->x /= mag;
   v->y /= mag;
}

void vec2_transform(vec2 *v, mat22 *m)
{
	FLOAT_T r[2];

   #define EVAL_ROW2(i) \
      r[i - 1] =\
         m->a##i##1 * v->x +\
         m->a##i##2 * v->y

   if (!v || !m)
      return;

   EVAL_ROW2(1);
   EVAL_ROW2(2);

   *v = *((vec2*)r);
}

void vec2_transform33(vec2 *v, mat33 *m)
{
	FLOAT_T r[2];

   #define EVAL_ROW23(i) \
      r[i - 1] =\
         m->a##i##1 * v->x +\
         m->a##i##2 * v->y +\
         m->a##i##3

   if (!v || !m)
      return;

   EVAL_ROW23(1);
   EVAL_ROW23(2);

   *v = *((vec2*)r);
}

