#ifndef __MATRIX44__
#define __MATRIX44__

#include "types.h"

/*
   Matrix layout is column major order:
      m0 m4 m8  m12
      m1 m5 m9  m13
      m2 m6 m10 m14
      m3 m7 m11 m15

      or

      m11 m21 m31 m41
      m12 m22 m32 m42
      m13 m23 m33 m43
      m14 m24 m34 m44

      where m<i><j> is m<row><column>
*/

/* All types should result in a non-padded alignment for all platforms... */

typedef union {
	struct { FLOAT_T x, y; };
	struct { FLOAT_T s, t; };
	FLOAT_T fv[2];
}
vec2;

typedef union {
	struct { FLOAT_T  x,  y,  z; };
	struct { FLOAT_T _x, _y,  w; };
	struct { FLOAT_T  r,  g,  b; };
	vec2 xy;
	vec2 rg;
	FLOAT_T fv[3];
}
vec3;


typedef union {
	struct { FLOAT_T  x,  y,  z,  w; };
	struct {	FLOAT_T _x, _y,  h, _w; };
	struct { FLOAT_T  r,  g,  b,  a; };
	vec2 xy;
	vec3 xyz;
	vec3 rgb;
	FLOAT_T fv[4];
}
vec4;





typedef union {
	struct {
	   FLOAT_T a11, a21, a31, a41;
	   FLOAT_T a12, a22, a32, a42;
	   FLOAT_T a13, a23, a33, a43;
	   FLOAT_T a14, a24, a34, a44;
	};
	struct {
	   FLOAT_T a11, a12, a13, a14;
	   FLOAT_T a21, a22, a23, a24;
	   FLOAT_T a31, a32, a33, a34;
	   FLOAT_T a41, a42, a43, a44;
	} T;
	struct { vec4 v1, v2, v3, v4; };
	struct { FLOAT_T r1[4], r2[4], r3[4], r4[4]; };
	vec4 v[4];
	FLOAT_T fv[16];
}
mat44;

typedef union {
   struct {
	   FLOAT_T a11, a21, a31;
   	FLOAT_T a12, a22, a32;
   	FLOAT_T a13, a23, a33;
	};
	struct {
	   FLOAT_T a11, a12, a13;
   	FLOAT_T a21, a22, a23;
   	FLOAT_T a31, a32, a33;
	} T;
	struct { vec3 v1, v2, v3; };
	struct { FLOAT_T r1[3], r2[3], r3[3]; };
	vec3	v[3];
	FLOAT_T fv[9];
}
mat33;

typedef union {
   struct {
	   FLOAT_T a11, a21;
   	FLOAT_T a12, a22;
	};
	struct {
		FLOAT_T a11, a12;
   	FLOAT_T a21, a22;
	} T;
	struct { vec2 v1, v2; };
	struct { FLOAT_T r1[2], r2[2]; };
	vec2 v[2];
	FLOAT_T fv[4];
}
mat22;


#define MAT44_IDENTITY {\
	1,0,0,0,\
	0,1,0,0,\
	0,0,1,0,\
	0,0,0,1 \
}

#define MAT33_IDENTITY {\
	1,0,0,\
	0,1,0,\
	0,0,1 \
}

#define MAT22_IDENTITY {\
	1,0,\
	0,1 \
}


mat44 *mat44_create(void);
void mat44_destroy(mat44 *m);
void mat44_print(mat44 *m);

/* These functions REPLACE contents of input matrix 'm'. */
void mat44_mult(mat44 *m, mat44 *by_m);
void mat44_identity(mat44 *m);
void mat44_translation(mat44 *m, FLOAT_T x, FLOAT_T y, FLOAT_T z);
void mat44_rotation(mat44 *m, FLOAT_T rad, FLOAT_T x, FLOAT_T y, FLOAT_T z);
void mat44_scaling(mat44 *m, FLOAT_T x, FLOAT_T y, FLOAT_T z);
void mat44_transpose(mat44 *m);
INT_T mat44_invert(mat44 *m);

void mat44_perspective(mat44 *m,FLOAT_T left, FLOAT_T right, FLOAT_T bottom, FLOAT_T top, FLOAT_T znear, FLOAT_T zfar);
void mat44_orthographic(mat44 *m, FLOAT_T left, FLOAT_T right, FLOAT_T bottom, FLOAT_T top, FLOAT_T znear, FLOAT_T zfar);


/* Non-Homogeneous mat33 */
mat33 *mat33_create(void);
void mat33_from_mat44(mat33 *m, mat44 *n);
void mat33_mult(mat33 *m, mat33 *by_m);
void mat33_identity(mat33 *m);
void mat33_rotation(mat33 *m, FLOAT_T rad, FLOAT_T x, FLOAT_T y, FLOAT_T z);
void mat33_scaling(mat33 *m, FLOAT_T x, FLOAT_T y, FLOAT_T z);
void mat33_destroy(mat33 *m);
void mat33_transpose(mat33 *m);

/* Homogeneous mat33 */
void mat33h_translation(mat33 *m, FLOAT_T x, FLOAT_T y);
void mat33h_rotation(mat33 *m, FLOAT_T rad);
void mat33h_scaling(mat33 *m, FLOAT_T x, FLOAT_T y);

/* mat22 */
mat22 *mat22_create(void);
void mat22_from_mat33(mat22 *m, mat33 *n);
void mat22_destroy(mat22 *m);
void mat22_mult(mat22 *m, mat22 *by_m);
void mat22_identity(mat22 *m);
void mat22_rotation(mat22 *m, FLOAT_T rad);
void mat22_scaling(mat22 *m, FLOAT_T x, FLOAT_T y);
void mat22_transpose(mat22 *m);


/* These functions REPLACE contents of input vector 'v'.
A vec4 is a homogeneous vector; general operations IGNORE the 'w' component.
For all practical purposes, the 'w' component is always set to 1.0. */
void vec4_set(vec4 *v, FLOAT_T x, FLOAT_T y,FLOAT_T z);
FLOAT_T vec4_magnitude(vec4 *v);
void vec4_add(vec4 *v, vec4 *to_v);
void vec4_subtract(vec4 *v, vec4 *from_v);
void vec4_scale(vec4 *v, FLOAT_T scaler);
FLOAT_T vec4_dot(vec4 *v, vec4 *with_v);
vec4 vec4_cross(vec4 *v, vec4 *with_v);
void vec4_normalize(vec4 *v);
void vec4_transform(vec4 *v, mat44 *by_m);

/* vec3 generic */
void vec3_set(vec3 *v, FLOAT_T x, FLOAT_T y,FLOAT_T z);
FLOAT_T vec3_magnitude(vec3 *v);
void vec3_add(vec3 *v, vec3 *to_v);
void vec3_subtract(vec3 *v, vec3 *from_v);
void vec3_scale(vec3 *v, FLOAT_T scaler);
FLOAT_T vec3_dot(vec3 *v, vec3 *with_v);
vec3 vec3_cross(vec3 *v, vec3 *with_v);
void vec3_normalize(vec3 *v);
void vec3_transform(vec3 *v, mat33 *by_m);
void vec3_transform44(vec3 *v, mat44 *by_m);

/* vec3 homogenous */
void vec3h_set(vec3 *v,FLOAT_T x, FLOAT_T y);
FLOAT_T vec3h_magnitude(vec3 *v);
void vec3h_add(vec3 *v, vec3 *to_v);
void vec3h_subtract(vec3 *v, vec3 *from_v);
void vec3h_scale(vec3 *v, FLOAT_T scaler);
FLOAT_T vec3h_dot(vec3 *v, vec3 *with_v);
void vec3h_normalize(vec3 *v);

/* vec2 generic */
void vec2_set(vec2 *v, FLOAT_T x, FLOAT_T y);
FLOAT_T vec2_magnitude(vec2 *v);
void vec2_add(vec2 *v, vec2 *to_v);
void vec2_subtract(vec2 *v, vec2 *from_v);
void vec2_scale(vec2 *v, FLOAT_T scaler);
FLOAT_T vec2_dot(vec2 *v, vec2 *with_v);
void vec2_normalize(vec2 *v);
void vec2_transform(vec2 *v, mat22 *by_m);
void vec2_transform33(vec2 *v, mat33 *by_m);

/* Future...
void vec2_to_vec3(vec3 *v, vec2 *v2, FLOAT_T z);
void vec3_to_vec4(vec4 *v, vec3 *v3, FLOAT_T w);
void vec4_to_vec3(vec3 *v, vec4 *v4);
void vec3_to_vec2(vec2 *v, vec3 *v3); */



#endif /* __MATRIX44__ */