#include "matrix_library.h"
#include <string.h>

#define DEG(x) \
   ((x) * 57.295779513082320876798154814105)

#define RAD(x) \
   ((x) * 0.01745329251994329576923690768489)

static mat44 projection[MAX_PROJECTION_STACK + 1];
static mat44 modelview [MAX_MODELVIEW_STACK  + 1];
static int projection_idx;
static int modelview_idx;

mat44 *projection_matrix;
mat44 *modelview_matrix;

static mat44 mvp;

static mat44 *current_matrix;
static ENUM_T current_type;

void mlInit(void)
{
   projection_idx    = 0;
   projection_matrix = projection;
   mat44_identity(projection);

   modelview_idx     = 0;
   modelview_matrix  = modelview;
   mat44_identity(modelview);

   current_matrix    = modelview;
   current_type      = ML_MODELVIEW;
}

void mlMatrixMode(ENUM_T mode)
{
   switch (mode) {
      case ML_PROJECTION:
         current_matrix = projection_matrix;
         current_type   = mode;
         break;
      case ML_MODELVIEW:
         current_matrix = modelview_matrix;
         current_type   = mode;
         break;
      default:
         return;
   }
}

void mlPushMatrix(void)
{
   switch (current_type) {
      case ML_PROJECTION:
         if (projection_idx < MAX_PROJECTION_STACK) {
            projection_matrix++;
            projection_idx++;
            current_matrix = projection_matrix;
            mat44_identity(projection_matrix);
         }
         break;
      case ML_MODELVIEW:
         if (modelview_idx < MAX_MODELVIEW_STACK) {
            modelview_matrix++;
            modelview_idx++;
            current_matrix = modelview_matrix;
            mat44_identity(modelview_matrix);
         }
         break;
   }
}

void mlPopMatrix(void)
{
   switch (current_type) {
      case ML_PROJECTION:
         if (projection_idx > 0) {
            projection_matrix--;
            projection_idx--;
            current_matrix = projection_matrix;
         }
         break;
      case ML_MODELVIEW:
         if (modelview_idx > 0) {
            modelview_matrix--;
            modelview_idx--;
            current_matrix = modelview_matrix;
         }
         break;
   }
}

void mlLoadIdentity(void)
{
   mat44_identity(current_matrix);
}

void mlMultMatrixf(mat44 *m)
{
   mat44_mult(current_matrix,m);
}

void mlLoadMatrixf(mat44 *m)
{
   memcpy(current_matrix,m,sizeof(mat44));
}

void mlGetProjectionMatrix(FLOAT_T *m)
{
   if (m)
      memcpy(m,projection_matrix,sizeof(mat44));
}

void mlGetModelviewMatrix(FLOAT_T *m)
{
   if (m)
      memcpy(m,modelview_matrix,sizeof(mat44));
}

FLOAT_T *mlGetMVPMatrix(void)
{
   mat44_identity(&mvp);
   mat44_mult(&mvp,projection_matrix);
   mat44_mult(&mvp,modelview_matrix);

   return (float*)&mvp;
}

void mlFrustum(FLOAT_T left, FLOAT_T right, FLOAT_T bottom, FLOAT_T top, FLOAT_T near, FLOAT_T far)
{
   static mat44 p;

   mat44_perspective(&p,left,right,bottom,top,near,far);
   mat44_mult(current_matrix,&p);
}

void mlOrtho(FLOAT_T left, FLOAT_T right, FLOAT_T bottom, FLOAT_T top, FLOAT_T near, FLOAT_T far)
{
   static mat44 o;

   mat44_orthographic(&o,left,right,bottom,top,near,far);
   mat44_mult(current_matrix,&o);
}

void mlTranslatef(FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   static mat44 p;

   mat44_translation(&p,x,y,z);
   mat44_mult(current_matrix,&p);
}
void mlRotatef(FLOAT_T angle, FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   static mat44 p;

   mat44_rotation(&p,RAD(angle),x,y,z);
   mat44_mult(current_matrix,&p);
}

void mlScalef(FLOAT_T x, FLOAT_T y, FLOAT_T z)
{
   static mat44 p;

   mat44_scaling(&p,x,y,z);
   mat44_mult(current_matrix,&p);
}

INT_T mlUnProject(
   FLOAT_T winx, FLOAT_T winy, FLOAT_T winz,
   mat44 *modelMatrix, mat44 *projMatrix, INT_T viewport[4],
   FLOAT_T *objx, FLOAT_T *objy, FLOAT_T *objz)
{
	mat44 inv;
	vec4 coord;

	if (!modelMatrix || !projMatrix || !objx || !objy || !objz)
   	return 0;

  	mat44_identity(&inv);
  	mat44_mult(&inv,(mat44*)modelMatrix);
  	mat44_mult(&inv,(mat44*)projMatrix);

  	if (!mat44_invert(&inv))
  		return 0;

  	vec4_set(&coord,
  		(winx - viewport[0])/viewport[2] * 2 - 1,
  		(winy - viewport[1])/viewport[3] * 2 - 1,
  		winz * 2 - 1);
  	vec4_transform(&coord,&inv);

  	if (!coord.w)
		return 0;

	*objx = coord.x/coord.w;
	*objy = coord.y/coord.w;
	*objz = coord.z/coord.w;

  	return 1;
}
