#define USING_OPENGL
#include "platform.h"
#include "renderer_api.h"

#include "obj_model.h"

#define CRD_BLK      100000
#define MSH_BLK      200
#define MAT_BLK      20

#define SET_MATERIAL_PROPERTY(prop) \
{\
   glMaterialfv(GL_FRONT_AND_BACK,GL_AMBIENT,prop);\
   glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,&(prop[4]));\
   glMaterialfv(GL_FRONT_AND_BACK,GL_SPECULAR,&(prop[8]));\
   glMaterialfv(GL_FRONT_AND_BACK,GL_EMISSION,&(prop[12]));\
   glMaterialf(GL_FRONT_AND_BACK,GL_SHININESS,prop[16]);\
}

#define GET_MATERIAL_PROPERTY(prop) \
{\
   glGetMaterialfv(GL_FRONT,GL_AMBIENT,prop);\
   glGetMaterialfv(GL_FRONT,GL_DIFFUSE,&(prop[4]));\
   glGetMaterialfv(GL_FRONT,GL_SPECULAR,&(prop[8]));\
   glGetMaterialfv(GL_FRONT,GL_EMISSION,&(prop[12]));\
   glGetMaterialfv(GL_FRONT,GL_SHININESS,&prop[16]);\
}

#define GET_WORD(from,to) \
{\
   while ((*(from) != ' ') && (*(from) != '\n') && (*(from) != '\r')) {\
      *(to) = *(from);\
      (to)++;\
      (from)++;\
   }\
   *(to) = 0;\
}

#define SKIP_WHITESPACE(at) \
{\
   while ((*(at) == ' ') || (*(at) == '\n') || (*(at) == '\r')) at++;\
}

#define SCALE_COORD(coord,scale) \
{\
   (coord).x *= (scale);\
   (coord).y *= (scale);\
   (coord).z *= (scale);\
}

#define CHECK_SIZE(curr_sz,block_sz,ptr,type) \
{\
   if (!((curr_sz)%(block_sz))) {\
      (ptr) = (type*)realloc((ptr),sizeof(type) * (curr_sz + block_sz));\
      if (!ptr) {\
         printf(\
            "realloc failed at %d, size: %ld\n",\
            __LINE__,\
            sizeof(type) * (curr_sz + block_sz));\
         exit(1);\
      }\
   }\
}

#define TO_STR(x) STR(x)
#define STR(x) #x

typedef struct { float x,y,z; }                             coord_3d;
typedef struct { int mat, type, v_cnt, v_at, t_at, n_at; }  surface_3d;

typedef struct {
   char        name[80];
   float       clr_mat[17];
   unsigned    tex;
}
gl_mat;

typedef struct {
   char        name[80];
   float       scale;
   surface_3d  *sfc;
   coord_3d    *vtx;
   coord_3d    *tex;
   coord_3d    *nrm;
   coord_3d    max;
   coord_3d    min;
   int         sfc_cnt;
   int         vtx_cnt;
   int         tex_cnt;
   int         nrm_cnt;

   char        mtl_lib[80];
   gl_mat      *mtl;
   int         mtl_cnt;
}
model_3d;

/* Default template material. */
static float defaultMat[17] = {
   .2, .2, .2, 1,    /* Ambient  */
   .8, .8, .8, 1,    /* Diffuse  */
    0,  0,  0, 1,    /* Specular */
    0,  0,  0, 1,    /* Emission */
    0                /* Shininess */
};

static char *dynamic_fgets(char **var_buff,int *size,FILE *stream);
static void *new_mem(char*,int);
static void *chg_mem(char*,void*,int);

static model_3d *model_load_bin(char* name);
static void model_write_bin(model_3d *new_mdl, char *name);
static void model_report(model_3d *new_mdl, unsigned begin);
static void model_rescale(model_3d *new_mdl,float scale);
static void model_add_mtl(model_3d*);
static int  model_get_mtl(model_3d*,char*);
static void face_parser(model_3d*,char*,coord_3d*,coord_3d*,coord_3d*);

static void delete_model(model_3d*);

geom_list *obj_model_to_geom_list(obj_model model)
{
   unsigned char *arr;
   float *ptr;
   unsigned re_tex_total, total, sz, at, n;
   unsigned nrm_at, tex_at;
   geom_list *g;
   geom_part_t *p;
   model_3d *m = (model_3d*)model;


   if (!m)
      return 0;

   re_tex_total   = m->tex_cnt * 2 * sizeof(float);
   total          = (m->vtx_cnt + m->nrm_cnt) * sizeof(coord_3d) + re_tex_total;
   arr            = (unsigned char*)malloc(total);
   if (!arr)
      return 0;

   sz = m->vtx_cnt * sizeof(coord_3d);
   memcpy(arr,m->vtx,sz);
   at = sz;
   nrm_at = at;
   sz = m->nrm_cnt * sizeof(coord_3d);
   memcpy(arr + at,m->nrm,sz);
   at += sz;
   tex_at = at;

   n  = 0;
   sz = m->tex_cnt;
   ptr = (float*)(arr + tex_at);
   while (n < sz) {
      ptr[2 * n]     = m->tex[n].x;
      ptr[2 * n + 1] = m->tex[n].y;
      n++;
   }

   /* model vertex/normal/texture coordinate data is now packed in 'arr'... */

   sz = m->sfc_cnt;
   p = (geom_part_t*)calloc(sz,sizeof(geom_part_t));
   if (!p) {
      free(arr);
      return 0;
   }

   n = 0;
   while (n < sz) {
      p[n].vertex_count = (m->sfc[n]).v_cnt;
      if (m->sfc[n].v_at > 0)
         p[n].vertex_start = m->sfc[n].v_at *  sizeof(coord_3d); /* bytes */
      if (m->sfc[n].n_at > 0)
         p[n].normal_start = nrm_at + m->sfc[n].n_at * sizeof(coord_3d);
      if (m->sfc[n].t_at > 0)
         p[n].texture_start = tex_at + m->sfc[n].t_at * 2 * sizeof(float);

      if (m->sfc[n].mat > -1) {
         memcpy(
            &(p[n].material),
            &((m->mtl[m->sfc[n].mat]).clr_mat),
            17 * sizeof(float) + sizeof(int));
      }
      else {
         memcpy(
            &(p[n].material),
            defaultMat,
            17 * sizeof(float));
      }

      n++;
   }

   /* model mesh/material data is now packed in 'p'... */

   g = create_geom_list();
   if (!g) {
      free(arr);
      free(p);
      return 0;
   }

   g->array       = arr;
   g->array_bytes = total;
   g->part_count  = m->sfc_cnt;
   g->part        = p;

   return g;
}

int model_get_mtl(model_3d *m, char *mtl)
{
   int n = 0;

   if (!m->mtl_cnt || !mtl)
      return -1;
   while (n < m->mtl_cnt && strcmp((m->mtl[n]).name,mtl))
      n++;
   if (n == m->mtl_cnt)
      return -1;
   return n;

}

void model_new_mtl(model_3d* m)
{
   FILE *mtl_file = 0;
   char buff[256], temp[80], full_name[80], tex_path[256];
   char *rd_at = buff, *cpy_ptr = 0, *ok;
   float *new_mtl = 0, *comp;

   sprintf(full_name,"%s%s/%s",OBJ_MODEL_LIB,m->name,m->mtl_lib);

   mtl_file = fopen(full_name,"r");
   if (!mtl_file) {
      printf("%s : File not found\n\n",full_name);
      return;
   }

   while (!feof(mtl_file)) {

      ok = fgets(buff,256,mtl_file);
      if (!ok)
         continue;
      rd_at = buff;
      while ((*rd_at == ' ' || *rd_at == '\t') && *rd_at != '\n')
         rd_at++;

      if ((!new_mtl && !strstr(rd_at,"new")) || *rd_at == '#')
         continue;

      switch (*rd_at) {
         case 'n':
         case 'N':
            switch (*(rd_at + 1)) {
               case 's': /* shininess */
                  rd_at += 3;
                  sscanf(rd_at,"%f",(new_mtl + 16));
                  //new_mtl[16] /= 1000.0;
                  //new_mtl[16] *= 128.0;
                  break;
               case 'e': /* new material */
                  while (*rd_at != ' ' && *rd_at != '\n')
                     rd_at++;
                  if (*rd_at == '\n')
                     break;
                  rd_at++;
                  cpy_ptr = temp;
                  GET_WORD(rd_at,cpy_ptr);

                  CHECK_SIZE(m->mtl_cnt,MAT_BLK,m->mtl,gl_mat);
                  memset(&(m->mtl[m->mtl_cnt]),0,sizeof(gl_mat));
                  new_mtl = (m->mtl[m->mtl_cnt]).clr_mat;
                  strncpy((m->mtl[m->mtl_cnt]).name,temp,80);
                  memcpy(new_mtl,defaultMat,17 * sizeof(float));
                  m->mtl_cnt++;
                  break;
               default:
                  break;
            }
            break;
         case 'm':
         case 'M':
            while (*rd_at != ' ' && *rd_at != '\n')
               rd_at++;
            if (*rd_at == '\n')
               break;
            rd_at++;
            cpy_ptr = temp;
            GET_WORD(rd_at,cpy_ptr);
            sprintf(tex_path,"%s%s/%s",OBJ_MODEL_LIB,m->name,temp);
            (m->mtl[m->mtl_cnt - 1]).tex = renderer_register_texture_file(tex_path);
            break;
         case 'T':
         case 't':
            rd_at++;
            break;
         case 'd':
            rd_at += 2;
            sscanf(rd_at,"%f",(new_mtl + 3));
            new_mtl[7] = new_mtl[11] = new_mtl[15] = new_mtl[3];
            break;
         case 'K':
         case 'k':
            rd_at++;

            switch (*rd_at) {
               case 'a':
               case 'A':
                  comp = new_mtl;
                  break;
               case 'd':
               case 'D':
                  comp = new_mtl + 4;
                  break;
               case 's':
               case 'S':
                  comp = new_mtl + 8;
                  break;
               case 'e':
               case 'E':
                  comp = new_mtl + 12;
                  break;
               default:
                  comp = 0;
                  break;
            }

            if (comp) {
               rd_at += 2;
               sscanf(rd_at,"%f %f %f",comp,comp + 1,comp + 2);
            }
            break;
         case 'i':
         case 'I':
            break;
         default:
            break;
      }
   }
   fclose(mtl_file);
}


void obj_model_draw(obj_model m)
{
   int n = 0;
   unsigned int type = 0;
   float save_mat[17];
   model_3d *model = (model_3d*)m;
   surface_3d *sfc;

   if (!m)
      return;

   GET_MATERIAL_PROPERTY(save_mat);
   glDisable(GL_COLOR_MATERIAL);
   glEnableClientState(GL_VERTEX_ARRAY);

   while (n < model->sfc_cnt) {

      sfc = &model->sfc[n];

      if (sfc->mat > -1) {
         SET_MATERIAL_PROPERTY((model->mtl[sfc->mat]).clr_mat);

         if ((model->mtl[sfc->mat]).tex > 0) {
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D,(model->mtl[sfc->mat]).tex);
         }
         else glDisable(GL_TEXTURE_2D);
      }
      if (sfc->t_at > -1) {
         glEnableClientState(GL_TEXTURE_COORD_ARRAY);
         glTexCoordPointer(3,GL_FLOAT,0,&model->tex[sfc->t_at]);
      }
      if (sfc->n_at > -1) {
         glEnableClientState(GL_NORMAL_ARRAY);
         glNormalPointer(GL_FLOAT,0,&model->nrm[sfc->n_at]);
      }
      if (sfc->v_at > -1) {
         glVertexPointer(3,GL_FLOAT,0,&model->vtx[sfc->v_at]);
      }

      switch (sfc->type) {
         case 0:
         case 1:
            type = GL_POINTS;
            break;
         case 2:
            type = GL_LINES;
            break;
         case 3:
            type = GL_TRIANGLES;
            break;
         case 4:
            type = GL_QUADS;
            break;
         default:
            type = GL_POLYGON;
            break;
      }

      glDrawArrays(type,0,sfc->v_cnt);
      glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      glDisableClientState(GL_NORMAL_ARRAY);
      n++;
   }
   glDisableClientState(GL_VERTEX_ARRAY);
   SET_MATERIAL_PROPERTY(save_mat);
   glEnable(GL_COLOR_MATERIAL);
}

#define READ_COORD(s,c) sscanf((s),"%f %f %f",&((c).x),&((c).y),&((c).z))

model_3d *model_load_bin(char *name)
{
   FILE *mdl_file;
   model_3d* new_mdl;
   char temp[256];
   unsigned ret, sz, err = 0;

   sprintf(temp,"%s%s/%s.bof",OBJ_MODEL_LIB,name,name);

   mdl_file = fopen(temp,"rb");
   if (!mdl_file)
      return 0;

   new_mdl = (model_3d*)new_mem(TO_STR(__LINE__),sizeof(model_3d));
   memset(new_mdl,0,sizeof(model_3d));
   strncpy(new_mdl->name,name,80);

   /* Material Library... */
   ret = fread(&sz,sizeof(unsigned int),1,mdl_file);
   if (ret && sz) {
      ret = fread(new_mdl->mtl_lib,sizeof(char),sz,mdl_file);
      new_mdl->mtl_lib[sz] = 0;
      //printf("Mat lib >%s<\n",new_mdl->mt_lib);
      if (ret == sz)
         model_new_mtl(new_mdl);
      else
         err = 1;
   }

   while (!err) {

      /* Surfaces... */
      ret = fread(&sz,sizeof(unsigned int),1,mdl_file);
      if (ret && sz) {
         new_mdl->sfc = (surface_3d*)new_mem("bin:sfc",sz * sizeof(surface_3d));
         ret = fread(new_mdl->sfc,sizeof(surface_3d),sz,mdl_file);
         if (ret != sz)
            break;
      }

      /* Vertices... */
      ret = fread(&sz,sizeof(unsigned int),1,mdl_file);
      if (ret && sz) {
         new_mdl->vtx = (coord_3d*)new_mem("bin:vtx",sz * sizeof(coord_3d));
         ret = fread(new_mdl->vtx,sizeof(coord_3d),sz,mdl_file);
         if (ret != sz)
            break;
      }

      /* Texture Coords... */
      ret = fread(&sz,sizeof(unsigned int),1,mdl_file);
      if (ret && sz) {
         new_mdl->tex = (coord_3d*)new_mem("bin:tex",sz * sizeof(coord_3d));
         ret = fread(new_mdl->tex,sizeof(coord_3d),sz,mdl_file);
         if (ret != sz)
            break;
      }

      /* Normal Vectors... */
      ret = fread(&sz,sizeof(unsigned int),1,mdl_file);
      if (ret && sz) {
         new_mdl->nrm = (coord_3d*)new_mem("bin:nrm",sz * sizeof(coord_3d));
         ret = fread(new_mdl->nrm,sizeof(coord_3d),sz,mdl_file);
         if (ret != sz)
            break;
      }

      ret = fread(&(new_mdl->max),sizeof(coord_3d),1,mdl_file);
      if (!ret)
         break;

      ret = fread(&(new_mdl->min),sizeof(coord_3d),1,mdl_file);
      if (!ret)
         break;

      ret = fread(&(new_mdl->sfc_cnt),sizeof(int),1,mdl_file);
      if (!ret)
         break;

      ret = fread(&(new_mdl->vtx_cnt),sizeof(int),1,mdl_file);
      if (!ret)
         break;

      ret = fread(&(new_mdl->tex_cnt),sizeof(int),1,mdl_file);
      if (!ret)
         break;

      ret = fread(&(new_mdl->nrm_cnt),sizeof(int),1,mdl_file);
      if (!ret)
         break;

      fclose(mdl_file);

      new_mdl->scale = 1.0;

      return new_mdl;
   }

   fclose(mdl_file);
   delete_model(new_mdl);
   return 0;
}

static void model_write_bin(model_3d *m, char *name)
{
   FILE *mdl_file;
   char temp[256];
   unsigned ret, sz, err = 0;

   sprintf(temp,"%s%s/%s.bof",OBJ_MODEL_LIB,name,name);

   mdl_file = fopen(temp,"wb");
   if (!mdl_file) {
      printf("Binary file creation error...\n");
      return;
   }

   sz  = strlen(m->mtl_lib);
   ret = fwrite(&sz,sizeof(unsigned),1,mdl_file);
   ret = fwrite(m->mtl_lib,sizeof(char),sz,mdl_file);

   ret = fwrite(&m->sfc_cnt,sizeof(int),1,mdl_file);
   ret = fwrite(m->sfc,sizeof(surface_3d),m->sfc_cnt,mdl_file);

   ret = fwrite(&m->vtx_cnt,sizeof(int),1,mdl_file);
   ret = fwrite(m->vtx,sizeof(coord_3d),m->vtx_cnt,mdl_file);

   ret = fwrite(&m->tex_cnt,sizeof(int),1,mdl_file);
   ret = fwrite(m->tex,sizeof(coord_3d),m->tex_cnt,mdl_file);

   ret = fwrite(&m->nrm_cnt,sizeof(int),1,mdl_file);
   ret = fwrite(m->nrm,sizeof(coord_3d),m->nrm_cnt,mdl_file);

   ret = fwrite(&m->max,sizeof(coord_3d),1,mdl_file);
   ret = fwrite(&m->min,sizeof(coord_3d),1,mdl_file);

   ret = fwrite(&m->sfc_cnt,sizeof(int),1,mdl_file);
   ret = fwrite(&m->vtx_cnt,sizeof(int),1,mdl_file);
   ret = fwrite(&m->tex_cnt,sizeof(int),1,mdl_file);
   ret = fwrite(&m->nrm_cnt,sizeof(int),1,mdl_file);

   fclose(mdl_file);
}

obj_model obj_model_get(char *name, float scale)
{
   FILE *mdl_file = 0;
   char *line_buff = 0, temp[256], *rd_at = 0, *rd_nxt = 0, *cpy_ptr = 0;
   int line_buff_size = 0, first_sfc = 1;
   coord_3d *vtx_arr = 0, *tex_arr = 0, *nrm_arr = 0, *vtx = 0;
   int vtx_arr_cnt = 0, tex_arr_cnt = 0, nrm_arr_cnt = 0;
   unsigned long begin = clock();
   surface_3d empty_sfc = {-1,0,0,-1,-1,-1};
   model_3d* new_mdl = 0;

   if (!name || !scale)
      return 0;

   if (new_mdl = model_load_bin(name)) {
      model_rescale(new_mdl,scale);
      model_report(new_mdl,begin);
      return (obj_model)new_mdl;
   }

   sprintf(temp,"%s%s/%s.obj",OBJ_MODEL_LIB,name,name);

   mdl_file = fopen(temp,"r");
   if (!mdl_file) {
      printf("%s: File not found.\n\n",name);
      return 0;
   }

   new_mdl = (model_3d*)new_mem(TO_STR(__LINE__),sizeof(model_3d));
   memset(new_mdl,0,sizeof(model_3d));
   strncpy(new_mdl->name,name,80);

   new_mdl->scale = 1.0;
   new_mdl->sfc = (surface_3d*)new_mem(TO_STR(__LINE__),sizeof(surface_3d) * MSH_BLK);
   new_mdl->sfc[0] = empty_sfc;

   while(!feof(mdl_file)) {

      rd_at = dynamic_fgets(&line_buff,&line_buff_size,mdl_file);

      switch (*rd_at) {
         case 'm':
            if (!strncmp("mtllib ",rd_at,7)) {
               rd_nxt = rd_at + 7;
               cpy_ptr = new_mdl->mtl_lib;
               GET_WORD(rd_nxt,cpy_ptr);
               model_new_mtl(new_mdl);
            }
            break;
         case 'u':
            if (!strncmp("usemtl ",rd_at,7)) {
               rd_nxt = rd_at + 7;
               cpy_ptr = temp;
               GET_WORD(rd_nxt,cpy_ptr);

               if (!first_sfc) {
                  new_mdl->sfc_cnt++;
                  CHECK_SIZE(new_mdl->sfc_cnt,MSH_BLK,new_mdl->sfc,surface_3d);
                  new_mdl->sfc[new_mdl->sfc_cnt] = empty_sfc;
               }
               new_mdl->sfc[new_mdl->sfc_cnt].mat = model_get_mtl(new_mdl,temp);
               first_sfc = 0;
            }
            break;
         case 'v':
            rd_at++;
            switch (*rd_at) {
               case ' ':
                  rd_at++;
                  CHECK_SIZE(vtx_arr_cnt,CRD_BLK,vtx_arr,coord_3d);
                  READ_COORD(rd_at,vtx_arr[vtx_arr_cnt]);
                  vtx = &vtx_arr[vtx_arr_cnt];
                  if (vtx->x < new_mdl->min.x)
                     new_mdl->min.x = vtx->x;
                  if (vtx->y < new_mdl->min.y)
                     new_mdl->min.y = vtx->y;
                  if (vtx->z < new_mdl->min.z)
                     new_mdl->min.z = vtx->z;
                  if (vtx->x > new_mdl->max.x)
                     new_mdl->max.x = vtx->x;
                  if (vtx->y > new_mdl->max.y)
                     new_mdl->max.y = vtx->y;
                  if (vtx->z > new_mdl->max.z)
                     new_mdl->max.z = vtx->z;
                  vtx_arr_cnt++;
                  break;
               case 'n':
                  rd_at++;
                  CHECK_SIZE(nrm_arr_cnt,CRD_BLK,nrm_arr,coord_3d);
                  READ_COORD(rd_at,nrm_arr[nrm_arr_cnt]);
                  nrm_arr_cnt++;
                  break;
               case 't':
                  rd_at++;
                  CHECK_SIZE(tex_arr_cnt,CRD_BLK,tex_arr,coord_3d);
                  READ_COORD(rd_at,tex_arr[tex_arr_cnt]);
                  tex_arr_cnt++;
                  break;
               default:
                  printf("What?\n");
                  break;
            }
            break;
         case 'f':
            rd_at++;
            face_parser(new_mdl,rd_at,vtx_arr,nrm_arr,tex_arr);
            break;
         default:
            break;
      }
   }
   fclose(mdl_file);

   new_mdl->sfc_cnt++;
   free(line_buff);
   free(vtx_arr);
   free(nrm_arr);
   free(tex_arr);
   new_mdl->sfc = (surface_3d*)chg_mem(
      TO_STR(__LINE__),
      new_mdl->sfc,
      sizeof(surface_3d) * new_mdl->sfc_cnt);
   new_mdl->vtx = (coord_3d*)chg_mem(
      TO_STR(__LINE__),
      new_mdl->vtx,
      sizeof(coord_3d) * new_mdl->vtx_cnt);
   new_mdl->nrm = (coord_3d*)chg_mem(
      TO_STR(__LINE__),
      new_mdl->nrm,
      sizeof(coord_3d) * new_mdl->nrm_cnt);
   new_mdl->tex = (coord_3d*)chg_mem(
      TO_STR(__LINE__),
      new_mdl->tex,
      sizeof(coord_3d) * new_mdl->tex_cnt);
   new_mdl->mtl = (gl_mat*)chg_mem(
      TO_STR(__LINE__),
      new_mdl->mtl,
      sizeof(gl_mat) * new_mdl->mtl_cnt);

   model_write_bin(new_mdl,name);
   model_rescale(new_mdl,scale);
   model_report(new_mdl,begin);

   return (obj_model) new_mdl;
}

void model_report(model_3d *new_mdl, unsigned begin)
{
   printf(
      "\n"
      "Model File:             %s\n"
      "Base model size:        %ld bytes\n"
      "Load time:              %.2f sec.\n"
      "Coordinate minimums:    (%f, %f, %f)\n"
      "Coordinate maximums:    (%f, %f, %f)\n\n",
      new_mdl->name,
      sizeof(model_3d)
      + sizeof(surface_3d) * new_mdl->sfc_cnt
      + sizeof(coord_3d) * (new_mdl->vtx_cnt + new_mdl->nrm_cnt + new_mdl->tex_cnt),
      (clock() - begin)/((float)CLOCKS_PER_SEC),
      new_mdl->min.x, new_mdl->min.y, new_mdl->min.z,
      new_mdl->max.x, new_mdl->max.y, new_mdl->max.z);
}


void model_rescale(model_3d *new_mdl,float scale)
{
   int n = 0;

   if (scale != 1) {

      new_mdl->scale = scale;

      while (n < new_mdl->vtx_cnt) {
         SCALE_COORD(new_mdl->vtx[n],new_mdl->scale);
         n++;
      }

      SCALE_COORD(new_mdl->min,new_mdl->scale);
      SCALE_COORD(new_mdl->max,new_mdl->scale);
   }
}


void face_parser(model_3d *mdl,char *at,coord_3d *v_tab,coord_3d *n_tab,coord_3d *t_tab)
{
   /* write_state: 0 = vertex,1 = texture,2 = normal,2+ undefined. */
   int accum = 0, write_state = 0, *size, *cnt_ptr;
   /* new_cnt[0], or the number of vertices, also defines render type. */
   int new_cnt[3] = {0,0,0}, ini_cnt[3] = {0,0,0};
   coord_3d *read_from[3] = {v_tab,t_tab,n_tab}, **feed, **target;

   if (!mdl)
      return;

   target     = &mdl->vtx;
   feed       = read_from;
   size       = &mdl->vtx_cnt;
   ini_cnt[0] = mdl->sfc[mdl->sfc_cnt].v_at;
   ini_cnt[1] = mdl->sfc[mdl->sfc_cnt].t_at;
   ini_cnt[2] = mdl->sfc[mdl->sfc_cnt].n_at;
   cnt_ptr    = new_cnt;

   SKIP_WHITESPACE(at);

   while (*at) {
      switch (*at) {
         case ' ':
         case '\n':
         case '\r':
            if (write_state < 3) {
               if (accum && *feed) {
                  CHECK_SIZE(*size,CRD_BLK,*target,coord_3d);
                  (*target)[*size] = (*feed)[accum - 1];
                  (*size)++;
                  (*cnt_ptr)++;
               }
               SKIP_WHITESPACE(at);
               if (!*at)
                  continue;
            }
            write_state = 0; /* vertex */
            accum       = 0;
            target      = &mdl->vtx;
            feed        = read_from;
            size        = &mdl->vtx_cnt;
            cnt_ptr     = new_cnt;
            break;
         case '/':
            if (accum && *feed) {
               CHECK_SIZE(*size,CRD_BLK,*target,coord_3d);
               (*target)[*size] = (*feed)[accum - 1];
               (*size)++;
               (*cnt_ptr)++;
            }
            accum = 0;
            write_state++;
            if (write_state < 3) {
               target++;
               feed++;
               size++;
               cnt_ptr++;
            }
            else printf("Improperly formed vertex definition!\n");
            at++;
            break;
         case '-':
            printf("Sorry,no relative references implemented.\n");
            /* Just skip to end of line. */
            while (*++at);
            break;
         default:
            /* Simple non-negative atoi implementation. */
            accum = accum * 10 + *at - '0';
            at++;
            break;
      }
   }

   /* Face differs from current configuration (could be just an empty mesh). */
   if (new_cnt[0] > 4 /* each GL_POLYGON must be in a seperate group. */
         || mdl->sfc[mdl->sfc_cnt].type != new_cnt[0]
         || ini_cnt[1] < 0  &&  new_cnt[1]
         || ini_cnt[1] >= 0 && !new_cnt[1]
         || ini_cnt[2] < 0  &&  new_cnt[2]
         || ini_cnt[2] >= 0 && !new_cnt[2]) {

      if (ini_cnt[0] >= 0) {
         mdl->sfc_cnt++;
         CHECK_SIZE(mdl->sfc_cnt,MSH_BLK,mdl->sfc,surface_3d);
         mdl->sfc[mdl->sfc_cnt]       = mdl->sfc[mdl->sfc_cnt - 1];
         mdl->sfc[mdl->sfc_cnt].t_at  = -1;
         mdl->sfc[mdl->sfc_cnt].n_at  = -1;
         mdl->sfc[mdl->sfc_cnt].v_cnt = 0;
      }

      mdl->sfc[mdl->sfc_cnt].v_at = mdl->vtx_cnt - new_cnt[0];
      mdl->sfc[mdl->sfc_cnt].type = new_cnt[0];
      if (new_cnt[1])
         mdl->sfc[mdl->sfc_cnt].t_at = mdl->tex_cnt - new_cnt[1];
      if (new_cnt[2])
         mdl->sfc[mdl->sfc_cnt].n_at = mdl->nrm_cnt - new_cnt[2];
   }
   mdl->sfc[mdl->sfc_cnt].v_cnt += new_cnt[0];
}


void obj_model_get_bounds(obj_model mdl, float *x_min,float *y_min,float *z_min,
                                         float *x_max,float *y_max,float *z_max)
{
   model_3d *m = (model_3d*)mdl;

   if (!m)
      return;

   if (x_min)
      *x_min = m->min.x;
   if (y_min)
      *y_min = m->min.y;
   if (z_min)
      *z_min = m->min.z;
   if (x_max)
      *x_max = m->max.x;
   if (y_max)
      *y_max = m->max.y;
   if (z_max)
      *z_max = m->max.z;
}

void delete_model(model_3d *mdl)
{
   if (!mdl)
      return;

   if (mdl->vtx)
      free(mdl->vtx);

   if (mdl->nrm)
      free(mdl->nrm);

   if (mdl->tex)
      free(mdl->tex);

   if (mdl->sfc)
      free(mdl->sfc);

   if (mdl->mtl)
      free(mdl->mtl);

   free(mdl);
}


void obj_model_delete(obj_model m)
{
   delete_model((model_3d*)m);
}


void *new_mem(char *where,int size)
{
   void *temp = malloc(size);

   if (!temp) {
      printf("malloc failed at %s, size: %d\n",where,size);
      exit(1);
   }

   return temp;
}

void *chg_mem(char *where,void *who,int size)
{
   void *temp = realloc(who,size);

   if (!temp) {
      printf("realloc failed at %s, size: %d\n",where,size);
      exit(1);
   }

   return temp;
}

char* dynamic_fgets(char** var_buff, int* size, FILE* stream)
{
   long int rewind = ftell(stream);

   if (!*var_buff) {
      *var_buff = (char*)chg_mem(TO_STR(__LINE__),*var_buff,256);
      *size = 256;
   }

   while (*var_buff) {
      if (!fgets(*var_buff,*size,stream))
         return *var_buff;

      if (strlen(*var_buff) == (*size - 1)) {
         *size *= 2;
         *var_buff = (char*)chg_mem(TO_STR(__LINE__),*var_buff,*size);
         /* Back up and re-read the incomplete line of data from the file. */
         fseek(stream,rewind,SEEK_SET);
      }
      else return *var_buff;
   }

   return 0;
}
