#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include "planck_functions.h"


#define PLANCK_DATA_FILE            "./data/PlanckTraj.txt"
#define L2_DIST_KM                  1500000.0   /* km */

typedef struct { float x,y,z,t; } planck_point;
typedef struct { float x,y,z; } vtx_t;

/* Internal planck_object implementation... */

typedef struct _planck_object _planck_object;
struct _planck_object {

   IMPLEMENTS_PLANCK_OBJECT

   unsigned int at;
   double fx_clock;
   unsigned int trace_pulse;
};

static void update(planck_object *obj,double system_time,double frame_dt);
static void get_path(planck_object *obj, float **vtx, float **clr, unsigned *cnt);
static void erase(planck_object *obj);

static void planck_loader(void);
static void gen_full_strip_arrays(int strips,int divs);
static void draw_array(int strips,int divs,unsigned swap);

_planck_object planck_proto = {
   &update,
   &get_path,
   &erase,
   0,0,0,

   0,0,1
};

/* Static Planck object path data and scan arrays... */
static planck_point  p_path[2000];
static int           p_points = 0;
static float         vtx_clr[400];
static vtx_t			path_vtx[2000];

planck_object *create_planck_object(unsigned trace_fade)
{
   _planck_object *create = (_planck_object*)malloc(sizeof(_planck_object));
   unsigned n = 0;

   if (create) {
      *create = planck_proto;
      create->trace_pulse = trace_fade;
      planck_loader();

      while(n < 400) {
         vtx_clr[n] = 1;
         vtx_clr[n + 1] = 1;
         vtx_clr[n + 2] = 0;
         vtx_clr[n + 3] = 1;
         n += 4;

      }
   }
   return (planck_object*)create;
}

static void erase(planck_object *p)
{
   if (!p)
      return;

   free(p);
}

static void get_path(planck_object *obj, float **vtx, float **clr, unsigned *cnt)
{
	 _planck_object *p = (_planck_object*)obj;

   int start = 0, denom, n = 0;
   float intensity, glow_inc, luminosity = 0;
   static begin = 1;
   static vtx_t save;

	if (!p)
	    return;

	if (begin)
		begin = 0;
	else
		path_vtx[p->at] = save;


   /* We have three modes of trailing line display:

      * 0 to current point for points less than 100
      * current minus 50 to current for points between 100 and 700
      * current minus 30 to current for points above 700

   This gives a good non-distracting view of the orbital path, and
   mostly hides the path line jump when the repeating end loop is
   entered. */

   if (p->at > 99)
      start = p->at - 50;

   if (p->at > 700)
      start = p->at - 30;

   /* Control the way in which the path line color pulses and ramps... */
   denom = p->at - start;

   if (p->trace_pulse)
      intensity = (1 + cos(3 * p->fx_clock))/2;
   else
      intensity = 1;

   glow_inc = (denom > 0) ? (1.0/denom) : .01;

   /* Update the vertex color array... */
   while (start + n <= p->at) {
      vtx_clr[4 * n + 3] = luminosity * intensity;
      luminosity += glow_inc;
      n++;
   }

   /* We have to substitute the current interpolated point into the table. */
   save = path_vtx[p->at];
   path_vtx[p->at].x = p->x;
   path_vtx[p->at].y = p->y;
   path_vtx[p->at].z = p->z;

   if (vtx)
   	*vtx = (float*)&path_vtx[start];
   if (clr)
   	*clr = vtx_clr;
   if (cnt)
   	*cnt = n;
}


static void update(planck_object *obj, double systime, double dt)
{
   _planck_object *p = (_planck_object*)obj;
   float t_frac, t_denom;
   float t_max  = p_path[p_points - 3].t;
   float t_loop = 652;

   if (!p)
      return;

   /* Add some fakery for time points past actual file data:
   loops over the last L2 orbit cycle indefinitely. Controlled
   by t_max and t_loop above (which are "real" time points). */

   if (systime > t_max)
      systime = fmod(systime - t_loop,t_max - t_loop) + t_loop;

   /* Time can go backward and forward: update the point number, with
   sanity checks... */

   while (p->at < p_points - 2 && p_path[p->at + 1].t <= systime )
      p->at++;

   while (p->at > 0 && p_path[p->at].t > systime)
      p->at--;

   /* Find the interpolation coefficent... */

   t_denom = p_path[p->at + 1].t - p_path[p->at].t;
   t_frac  = (t_denom) ? (systime - p_path[p->at].t)/t_denom : 0;

   if (t_frac < 0)
      t_frac = -t_frac;

   if (t_frac > 1)
      t_frac = 1;

   /* ... and update the current position. */

   #define INTERPOLATE(dim) \
      (p_path[p->at].dim + (p_path[p->at + 1].dim - p_path[p->at].dim) * t_frac)

   p->x = INTERPOLATE(x);
   p->y = INTERPOLATE(y);
   p->z = INTERPOLATE(z);

   /* Update the "path glow pulse" timer. */

   p->fx_clock += dt *.45;;
}

static void planck_loader(void)
{
   FILE *file = fopen(PLANCK_DATA_FILE,"r");
   int data;

   if (!file) {
      printf("No Planck Data!!\n");
      return;
   }

   while (!feof(file)) {
      data = fscanf(
         file,
         "%f %f %f %f",
         &p_path[p_points].t,  /* Mission time from launch: Julian days */
         &p_path[p_points].x,  /* ER -x */
         &p_path[p_points].z,  /* ER z */
         &p_path[p_points].y); /* ER y */

         path_vtx[p_points].x = p_path[p_points].x;
         path_vtx[p_points].y = p_path[p_points].y;
         path_vtx[p_points].z = p_path[p_points].z;

      if (data == 4)
         p_points++;
   }

   fclose(file);
}

#ifdef __cplusplus
}
#endif

