#ifdef __cplusplus
extern "C" {
#endif

#define USING_OPENGL
#include "platform.h"
#include "renderer_api.h"

#include "solar_system.h"
#include "display.h"
#include "stage.h"
#include "ui_element.h"
#include "gluf.h"
#include "aluf.h"
#include "std_decor.h"
#include "display_controls.h"
#include "recorder_controls.h"
#include "planck_scene.h"

#if defined(WIN32) || defined(_WIN32) || defined(WIN64)
   #define snprintf _snprintf
#endif

#define CONFIG_ID "PLANCK_SCENE"

#define MUSIC_START_DELAY 2.0

typedef struct {
   IMPLEMENTS_SCENE_API
   ALuint               snd_src;
   ui_viewer            *viewer;
   ui_balloon           *menu;
   ui_balloon           *more;
   ui_balloon           *sky_menu;
   ui_balloon           *scan_menu;
   recorder_controls    *rc;
   ui_element           *cmn;
   ui_element           *hud;

   solar_system         *ss;

   unsigned             allow_3d_options;
   unsigned             allow_recording;
   unsigned             allow_date_control;
   unsigned             allow_skymap_control;
   unsigned             allow_orbit_control;

   unsigned             show_date;
   unsigned             show_skymap;
   unsigned             use_static_orbit_traces;
   unsigned             use_location_hotkeys;

   float                date_text_size;
   float                date_update_interval;

   unsigned             allow_object_scale_control;
   float                initial_object_scale;

   float                snd_init_delay;

}
planck_scene;

static void init(scene_api *object);
static void update(scene_api *object);
static void draw(scene_api *object);
static void enter(scene_api *object);
static void leave(scene_api *object);
static void erase(scene_api *object);
static void audio_level(scene_api *object,float lvl);



static void switchboard(void *obj,ui_update *u);
static void on_menu(ui_balloon *b, ui_update *u);
static void no_menu(ui_balloon *b, ui_update *u);
static void add_ss_hud(planck_scene *p);
static void configuration(planck_scene *p);

static planck_scene planck_scene_proto = {
   &init,
   &update,
   &draw,
   &enter,
   &leave,
   &erase,
   &audio_level,
   1,
   0,0.001800,30.0,0.05,
   0,0,0,
   0,0,0,
   0,0,0,
   0,

   0,          /* Begin planck_scene variables */
   0,0,0,0,0,0,0,0,
   0,
   1,1,1,1,1,
   1,0,0,1,
   .3,.3,
   1,15.0,0
};

enum {ctl_exit, ctl_menu, ctl_options, ctl_rec, ctl_date};

static void configuration(planck_scene *p)
{
   config_record *ret;

   p->cl = create_config_loader(CONFIG_ID);
   config_doc(p->cl,
      "Settings specific to the Planck Launch scene...");

   aggregate_register_config(p->ca,p->cl);

   config_bind(
      p->cl,
      "eye_separation",
      CONFIG_FLOAT,
      sizeof(float),
      &p->eye_seperation);
   config_bind(
      p->cl,
      "focal_plane",
      CONFIG_FLOAT,
      sizeof(float),
      &p->focal_plane);
   config_bind(
      p->cl,
      "front_clip",
      CONFIG_FLOAT,
      sizeof(float),
      &p->clipping_plane);
   ret = config_bind(
      p->cl,
      "allow_3d_options",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->allow_3d_options);
   config_doc(ret,
      "Controls display of 'More' options button");
   ret = config_bind(
      p->cl,
      "allow_recording",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->allow_recording);
   config_doc(ret,
      "Controls display of 'Record' button");
   ret = config_bind(
      p->cl,
      "allow_date_control",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->allow_date_control);
   config_doc(ret,
      "Governs display of on-screen text (date/location) control");
   ret = config_bind(
      p->cl,
      "allow_skymap_control",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->allow_skymap_control);
   config_doc(ret,
      "Governs display of star background selector");
   ret = config_bind(
      p->cl,
      "allow_orbit_control",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->allow_orbit_control);
   config_doc(ret,
      "Governs presence of planetary orbit trace display control");
   ret = config_bind(
      p->cl,
      "show_date",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->show_date);
   config_doc(ret,
      "Forces display of on-screen text (date/location):");
   config_doc(ret,
      "Use in conjunction with 'allow_date_control' to create a custom start-up");
   config_doc(ret,
      "configuration");
   ret = config_bind(
      p->cl,
      "show_skymap",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->show_skymap);
   config_doc(ret,
      "Forces display of star constellation map:");
   config_doc(ret,
      "Use in conjunction with 'allow_orbit_control' for custom configuration");
   ret = config_bind(
      p->cl,
      "use_static_orbit_traces",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->use_static_orbit_traces);
   config_doc(ret,
      "Forces display of static, non-fading orbits.");
   ret = config_bind(
      p->cl,
      "use_location_hotkeys",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->use_location_hotkeys);
   config_doc(ret,
      "Enables number key/numpad hotkeys for locations (keys '1' - '9')");
   ret = config_bind(
      p->cl,
      "date_text_size",
      CONFIG_FLOAT,
      sizeof(float),
      &p->date_text_size);
   config_doc(ret,
      "Controls the size of the date/location text, values 0 - 1.0");
   ret = config_bind(
      p->cl,
      "date_update_interval",
      CONFIG_FLOAT,
      sizeof(float),
      &p->date_update_interval);
   config_doc(ret,
      "Controls the rate at which the data/time text is updated in seconds");
   ret = config_bind(
      p->cl,
      "allow_object_scale_control",
      CONFIG_UINT,
      sizeof(unsigned),
      &p->allow_object_scale_control);
   config_doc(ret,
      "Controls availability of scaling slider in menu control panel");
   ret = config_bind(
      p->cl,
      "initial_object_scale",
      CONFIG_FLOAT,
      sizeof(float),
      &p->initial_object_scale);
   config_doc(ret,
      "Controls the initial size of orbiting bodies");

   aggregate_load_config(p->ca,p->cl);
}

scene_api *create_planck_scene(config_aggregate *ca)
{
   planck_scene *create = (planck_scene*)calloc(1,sizeof(planck_scene));
   ui_element *e;
   ui_icon *i;
   ui_container *s;
   ui_balloon *b;
   ui_button *bt;


   if (create) {
      *create = planck_scene_proto;

      create->context = renderer_generate_unique_id();
      renderer_set_context_group(create->context);

      create->ca = ca;
      configuration(create);

      create->viewer = create_viewer(frame_state.overlay_area);
      //create->viewer->pointer = &ui_busy_cursor;
      create->viewer->pointer = 0;
      s = create->viewer->screen;

      create->ss = create_solar_system(
         create->show_skymap,
         create->use_static_orbit_traces,
         !create->use_static_orbit_traces,
         create->initial_object_scale,
         &create->viewer->pointer,
         create->rotation);

      /* Hotkeys... */
      if (create->use_location_hotkeys) {
         e = create->ss->create_hotkeys(create->ss);
         s->add_item(s,e);
      }

      /* HUD */
      if (create->show_date)
         add_ss_hud(create);

      /* The pop up menu... */
      b = create_balloon();
      b->owner = (void*)create;
      b->button_mask = 48;
      b->on_use = &switchboard;
      b->on_grow = &on_menu;
      b->on_shrink = &no_menu;
      b->id = ctl_menu;
      s->add_item(s,b);
      create->menu = b;

      /* Control board background... */
      i = create_custom_icon(.1);
      i->place(i,0,0,1,.5);
      i->look = material_is(0,0,0,.75,0);
      b->add_item(b,i);

       /* Exit button... */
      bt = create_button();
      bt->place(bt,.85,.22,.1,.1);
      *((ui_decor*)&bt->look) = std_rb_decor;
      bt->id = ctl_exit;
      bt->on_use = &switchboard;
      bt->owner = (void*)create;
      b->add_item(b,bt);
      b->add_item(b,create_item_label(bt,"Exit"));

      /* "More" button... */
      if (create->allow_3d_options) {
         bt = create_button();
         bt->place(bt,.05,.27,.1,.1);
         *((ui_decor*)&bt->look) = std_bl_decor;
         bt->id = ctl_options;
         bt->on_use = &switchboard;
         bt->owner = (void*)create;
         b->add_item(b,bt);
         b->add_item(b,create_item_label(bt,"More"));
      }

      /* Record button... */
      if (create->allow_recording) {
         bt = create_button();
         bt->place(bt,.05,.13,.1,.1);
         *((ui_decor*)&bt->look) = std_bl_decor;
         bt->id = ctl_rec;
         bt->on_use = &switchboard;
         bt->owner = (void*)create;
         b->add_item(b,bt);
         b->add_item(b,create_item_label(bt,"Record"));
      }

      /* All solar system button controls... */

      e = create->ss->create_location_ctl(create->ss);
      e->place(e,.13,0,.1,.1);
      *((ui_decor*)&e->look) = std_gb_decor;
      b->add_item(b,e);
      b->add_item(b,create_item_label(e,"Location"));

      if (create->allow_orbit_control) {
         e = create->ss->create_trace_ctl(create->ss);
         e->place(e,.26,0,.1,.1);
         *((ui_decor*)&e->look) = std_gr_gn_decor;
         b->add_item(b,e);
         b->add_item(b,create_item_label(e,"Orbits"));
      }

      e = create->ss->create_marker_ctl(create->ss);
      e->place(e,.39,0,.1,.1);
      *((ui_decor*)&e->look) = std_gr_gn_decor;
      b->add_item(b,e);
      b->add_item(b,create_item_label(e,"Earth Marker"));

//       e = create->ss->create_wire_frame_ctl(create->ss);
//       e->place(e,.39,0,.1,.1);
//       *((ui_decor*)&e->look) = std_gr_gn_decor;
//       b->add_item(b,e);
//       b->add_item(b,create_item_label(e,"Wireframe"));

//       e = create->ss->create_scan_ctl(create->ss);
//       e->place(e,.39,0,.1,.1);
//       *((ui_decor*)&e->look) = std_gr_gn_decor;
//       b->add_item(b,e);
//       b->add_item(b,create_item_label(e,"Show Scan"));

      e = create->ss->create_scan_map_ctl(create->ss);
      e->place(e,.52,0,.1,.1);
      *((ui_decor*)&e->look) = std_gb_decor;
      b->add_item(b,e);
      b->add_item(b,create_item_label(e,"Scan Map"));

      if (create->allow_skymap_control) {
         e = create->ss->create_star_map_ctl(create->ss);
         e->place(e,.65,0,.1,.1);
         *((ui_decor*)&e->look) = std_gb_decor;
         b->add_item(b,e);
         b->add_item(b,create_item_label(e,"Star Map"));
      }

      if (create->allow_date_control) {
         bt = create_button();
         bt->place(bt,.78,0,.1,.1);
         *((ui_decor*)&bt->look) = std_gr_gn_decor;
         bt->id = ctl_date;
         bt->latches = 1;
         bt->value = create->show_date;
         bt->on_use = &switchboard;
         bt->owner = (void*)create;
         b->add_item(b,bt);
         b->add_item(b,create_item_label(bt,"Date"));
      }


       /* All solar system slider controls... */
      if (create->allow_object_scale_control) {
         e = create->ss->create_object_scale_ctl(create->ss);
         e->place(e,.2,.2,.6,.05);
         *((ui_decor*)&e->look) = std_sl_decor;
         ((ui_slider*)e)->handle_width = .065;
         b->add_item(b,create_item_label(e,"Object Scale"));
         b->add_item(b,create_slider_labels((ui_slider*)e,5,
            "1x","10x","20x","30x","40x"));
         b->add_item(b,e);
      }

      e = create->ss->create_time_scale_ctl(create->ss);
      e->place(e,.2,.3,.6,.05);
      *((ui_decor*)&e->look) = std_sl_decor;
      ((ui_slider*)e)->handle_width = .065;
      b->add_item(b,create_item_label(e,"Speed (days/second)"));
      b->add_item(b,create_slider_labels((ui_slider*)e,5,
         "-10","-5","0","5","10"));
      b->add_item(b,e);

      e = create->ss->create_time_ctl(create->ss);
      e->place(e,.1,.4,.8,.05);
      *((ui_decor*)&e->look) = std_sl_decor;
      ((ui_slider*)e)->handle_width = .05;
      b->add_item(b,create_item_label(e,"Time"));
      b->add_item(b,create_slider_labels((ui_slider*)e,6,
         "Launch","Scan",0,0,"Complete Orbit",0));
      b->add_item(b,e);

      e = create->ss->create_sky_panel(create->ss);
      b->add_item(b,e);

      e = create->ss->create_scan_panel(create->ss);
      b->add_item(b,e);

      /* "More" menu pop up... */
      create->more = create_display_common_controls((scene_api*)create);
      b->add_item(b,create->more);

      /* A recorder_control implements ui_balloon, and so can be treated
      as such... */
      create->rc = create_recorder_controls("launch");
      b->add_item(b,create->rc);


   }
   return (scene_api*)create;
}

static void erase(scene_api *object)
{
   planck_scene *p = (planck_scene*)object;

   if (!p)
      return;

   p->ss->erase(p->ss);

   if (p->viewer)
      p->viewer->erase(p->viewer);

   free(p);
}

static void init(scene_api *object)
{
   planck_scene *p = (planck_scene*)object;

   if (!p)
      return;
}

static void update(scene_api *object)
{
   planck_scene *p = (planck_scene*)object;

   if (!p)
      return;

   p->viewer->update(
      p->viewer,
      frame_state.pointer[0],
      frame_state.pointer[1],
      frame_state.dt,
      frame_state.button,
      frame_state.key);

   p->ss->update(p->ss,frame_state.dt);

   if (ctl_button(4,frame_state.button))
       p->menu->grow(p->menu);

   if (p->snd_init_delay) {
      p->snd_init_delay -= frame_state.dt;
      if (p->snd_init_delay <= 0) {
         p->snd_init_delay = 0;
         if (p->snd_src)
            alSourcePlay(p->snd_src);
      }
   }
}


static void switchboard(void *obj,ui_update *u)
{
   ui_element *e = (ui_element*)obj;
   planck_scene *p;

   if (!e)
      return;

   p = (planck_scene*)e->owner;

   switch (e->id) {
      case ctl_exit:
         p->menu->shrink(p->menu);
         stage_pop_scene();
         break;
      case ctl_menu:
         if (ctl_button(4,u->button))
            p->menu->shrink(p->menu);
         break;
      case ctl_options:
         p->more->grow(p->more);
         break;
      case ctl_rec:
         if (p->rc)
            ((ui_balloon*)p->rc)->grow((ui_balloon*)p->rc);
      case ctl_date:
         p->show_date = ((ui_button*)e)->value;
         if (!p->show_date && p->hud) {
            p->viewer->screen->remove_item(p->viewer->screen,p->hud);
            p->hud = 0;
         }
         if (p->show_date && !p->hud)
            add_ss_hud(p);
         break;
   }
}

static void add_ss_hud(planck_scene *p)
{
   ui_container *h = create_container();
   ui_element *e   = p->ss->create_info_dsp(p->ss,p->date_update_interval);
   float placement;

   if (p->date_text_size > 1)
      p->date_text_size = 1;
   if (p->date_text_size < .05)
      p->date_text_size = .05;

   placement = .95 - .143 * (p->date_text_size - .3);
   e->place(e,0.01,placement,p->date_text_size,p->date_text_size);

   decor_all_colors((ui_decor*)&e->look,.4,.4,0,.8);
   h->add_item(h,e);
   //p->viewer->screen->add_item(p->viewer->screen,h);
   p->viewer->screen->to_list_head(p->viewer->screen,h);
   p->hud = (ui_element*)h;
}

static void on_menu(ui_balloon *b, ui_update *u)
{
   planck_scene *p;

   if (!b)
      return;

   p = (planck_scene*)b->owner;

   p->viewer->pointer = &ui_basic_cursor;
   p->viewer->px = .5;
   p->viewer->py = .17;
   frame_state.rotation = 0;
   frame_state.position = 0;
}

static void no_menu(ui_balloon *b, ui_update *u)
{
   planck_scene *p;

   if (!b)
      return;

   p = (planck_scene*)b->owner;

   p->viewer->pointer   = 0;
   frame_state.rotation = p->rotation;
   frame_state.position = p->position;
}

static void enter(scene_api *object)
{
   planck_scene *p = (planck_scene*)object;

   if (!p)
      return;

   p->snd_src = p->ss->begin(p->ss);

   p->snd_init_delay = MUSIC_START_DELAY;

   if (p->rc)
      p->rc->resume(p->rc);

   p->cmn = retrieve_common_display();
   if (p->cmn)
      p->viewer->screen->add_item(p->viewer->screen,p->cmn);

   display_set_movement_acceleration(.0003);
}

static void leave(scene_api *object)
{
   planck_scene *p = (planck_scene*)object;

   if (!p)
      return;

   p->ss->stop(p->ss);
   p->snd_src = AL_NONE;

   if (p->rc)
      p->rc->pause(p->rc);

   if (p->cmn)
      p->viewer->screen->remove_item(p->viewer->screen,p->cmn);
}

static void audio_level(scene_api *object,float lvl)
{
   planck_scene *p = (planck_scene*)object;

   if (alIsSource(p->snd_src))
      alSourcef(p->snd_src,AL_GAIN,lvl);
}

static void draw(scene_api *object)
{
   planck_scene *p = (planck_scene*)object;
   mat44 tm, mv;

   if (!p)
      return;

   scene_api_gl_defaults();

   mat44_rotation(&mv,p->rotation[0],1,0,0);
   mat44_rotation(&tm,p->rotation[1],0,1,0);
   mat44_mult(&mv,&tm);
   mat44_translation(&tm,p->position[0],p->position[1],p->position[2]);
   mat44_mult(&mv,&tm);

   p->ss->draw(p->ss,p->projection,&mv);

   p->viewer->display(p->viewer, p->projection);
}

#ifdef __cplusplus
}
#endif

