#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include "display.h"
#include "aluf.h"
#include "std_decor.h"
#include "video_recorder.h"
#include "recorder_controls.h"

enum {ctl_width, ctl_height, ctl_fps, ctl_audio, ctl_quality, ctl_run, ctl_done};
enum {vr_stop = 0, vr_record, vr_pause};

typedef struct _recorder_controls _recorder_controls;
struct _recorder_controls {
   IMPLEMENTS_RECORDER_CONTROLS

   /* Container callbacks which we are going to intercept... */
   void (*container_update)(void *obj,ui_update *info);

   video_recorder *vid;
   char *title;
   unsigned width;
   unsigned height;
   unsigned fps;
   unsigned use_audio;
   float quality;
   unsigned run_state;
   ui_slider *sl_width;
   ui_slider *sl_height;

};

static void erase(void *obj);
static void vid_update(void *obj,ui_update *info);
static void vid_pause(recorder_controls *r);
static void vid_resume(recorder_controls *r);

static void switchboard(void *obj,ui_update *u);
static void on_grow(ui_balloon *b, ui_update *u);

static void text_updater(ui_text_value *tx);


recorder_controls *create_recorder_controls(const char* id)
{
   _recorder_controls *rc;
   ui_balloon *uc;
   ui_icon *ic;
   ui_button *bt;
   ui_container *c;
   ui_text_value *tx;
   ui_slider *sl;

   rc = (_recorder_controls*)calloc(1,sizeof(_recorder_controls));
   if (!rc)
      return 0;

   uc = create_balloon();
   if (!uc) {
      free(rc);
      return 0;
   }

   /* Essentially, we "inherit" the properties of the ui_container object
   (via bitwise memory copy); we then dispose of the parent object and
   assume its role... */

   memcpy(rc,uc,sizeof(ui_balloon));
   free(uc);
   uc = (ui_balloon*)rc;

   uc->erase            = &erase;
   rc->container_update = uc->update;
   uc->update           = &vid_update;
   uc->on_grow          = &on_grow;
   rc->pause            = &vid_pause;
   rc->resume           = &vid_resume;
   rc->width            = 800;
   rc->height           = 600;
   rc->fps              = 30;
   rc->quality          = 0;
   rc->use_audio        = 1;

   if (id) {
      rc->title = calloc(strlen(id) + 1, sizeof(char));
      if (rc->title)
         memcpy(rc->title,id, strlen(id));
   }

   /* Create the pop up menu... */

   /* Control background */
   ic = create_custom_icon(.1);
   ic->place(ic,0.,0,1,.75);
   ic->look = material_is(0,0,0,.95,0);
   uc->add_item(uc,ic);

   /* Done button */
   bt = create_button();
   bt->place(bt,.85,.12,.1,.1);
   *((ui_decor*)&bt->look) = std_rb_decor;
   bt->id = ctl_done;
   bt->on_use = &switchboard;
   bt->owner = (void*)uc;
   uc->add_item(uc,bt);
   uc->add_item(uc,create_item_label(bt,"Done"));

   ic = create_icon();
   ic->place(ic,0.05,.4,.6,.3);
   ic->look = material_is(.5,.5,.5,.5,0);
   uc->add_item(uc,ic);
   uc->add_item(uc,create_item_label(ic,"Xvid Encoder Properties"));

   ic = create_icon();
   ic->place(ic,0.0525,.4025,.595,.295);
   ic->look = material_is(.02,.02,.02,1,0);
   uc->add_item(uc,ic);

   c = create_container();
   c->place(c,.05,.05,1,1);

   tx = create_text_value(100,256,&text_updater);
   tx->update_on = .1;
   tx->font = ui_pretty_fnt;
   tx->force_monospace = 45;
   tx->place(tx,.05,.4,1,1);
   tx->owner = (void*)rc;
   decor_all_colors((ui_decor*)&tx->look,.85,.85,.4,1);
   c->add_item(c,tx);

   bt = create_button();
   bt->place(bt,.65,.3,.1,.1);
   *((ui_decor*)&bt->look) = std_gr_gn_decor;
   bt->owner = (void*)rc;
   bt->on_use = &switchboard;
   bt->chain = (int*)&rc->run_state;
   bt->value_down = vr_record;
   bt->id = ctl_run;
   c->add_item(c,bt);
   c->add_item(c,create_item_label(bt,"Start"));

   bt = create_button();
   bt->place(bt,.65,.15,.1,.1);
   *((ui_decor*)&bt->look) = std_gr_gn_decor;
   bt->owner = (void*)rc;
   bt->on_use = &switchboard;
   bt->chain = (int*)&rc->run_state;
   bt->value_down = vr_pause;
   bt->id = ctl_run;
   c->add_item(c,bt);
   c->add_item(c,create_item_label(bt,"Pause"));

   bt = create_button();
   bt->place(bt,.65,0,.1,.1);
   *((ui_decor*)&bt->look) = std_gr_gn_decor;
   bt->owner = (void*)rc;
   bt->on_use = &switchboard;
   bt->chain = (int*)&rc->run_state;
   bt->value_down = vr_stop;
   bt->id = ctl_run;
   c->add_item(c,bt);
   c->add_item(c,create_item_label(bt,"Stop"));

   bt = create_button();
   bt->place(bt,.65,.5,.1,.1);
   *((ui_decor*)&bt->look) = std_gr_gn_decor;
   bt->owner = (void*)rc;
   bt->on_use = &switchboard;
   bt->latches = 1;
   bt->value = rc->use_audio;
   bt->id = ctl_audio;
   c->add_item(c,bt);
   c->add_item(c,create_item_label(bt,"Use Audio"));

   sl = create_slider();
   sl->place(sl,0,0,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)rc;
   sl->id     = ctl_width;
   sl->min    = 640;
   sl->span   = 161;
   sl->value  = 800;
   sl->on_use = &switchboard;
   c->add_item(c,create_slider_labels(sl,3,
      "640","Half","None"));
   c->add_item(c,create_item_label(sl,"Width Scaling"));
   c->add_item(c,sl);
   rc->sl_width = sl;

   sl = create_slider();
   sl->place(sl,0,.075,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)rc;
   sl->id     = ctl_height;
   sl->min    = 480;
   sl->span   = 121;
   sl->value  = 600;
   sl->on_use = &switchboard;
   c->add_item(c,create_slider_labels(sl,3,
      "480","Half","None"));
   c->add_item(c,create_item_label(sl,"Height Scaling"));
   c->add_item(c,sl);
   rc->sl_height = sl;

   sl = create_slider();
   sl->place(sl,0,.15,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)rc;
   sl->id     = ctl_fps;
   sl->min    = 20;
   sl->span   = 41;
   sl->value  = rc->fps;
   sl->on_use = &switchboard;
   c->add_item(c,create_slider_labels(sl,5,
      "20","30","40","50","60"));
   c->add_item(c,create_item_label(sl,"Frames Per Second"));
   c->add_item(c,sl);

   sl = create_slider();
   sl->place(sl,0,.225,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)rc;
   sl->id     = ctl_quality;
   sl->min    = 0;
   sl->span   = 1;
   sl->value  = rc->quality;
   sl->on_use = &switchboard;
   c->add_item(c,create_slider_labels(sl,3,
      "Fast","Slower","Slow"));
   c->add_item(c,create_item_label(sl,"Quality/Speed"));
   c->add_item(c,sl);

   uc->add_item(uc,c);

   /* We won't create the video_recorder object until we need it... */

   return (recorder_controls*) rc;
}

static void erase(void *obj)
{
   _recorder_controls *rc = (_recorder_controls*)obj;
   ui_balloon *uc = (ui_balloon*)rc;
   ui_element *target, *save;

   if (!rc)
      return;

   /* Since we have replaced the ui_container erase function, we must
   do the work ourself manually... */

   target = uc->contains;

   while (target) {
      save = target->more;
      target->erase(target);
      target = save;
   }

   if (rc->vid)
      rc->vid->wait_finish(rc->vid);

   if (rc->title)
      free(rc->title);

   free(rc);
}

static void vid_update(void *obj,ui_update *info)
{
   _recorder_controls *rc = (_recorder_controls*)obj;

   if (!rc)
      return;

   if (rc->vid)
      rc->vid->next_frame(rc->vid,info->dt);

   rc->container_update(rc,info);
}

static void vid_pause(recorder_controls *rc)
{
   _recorder_controls *r = (_recorder_controls*)rc;

   if (!r)
      return;

   if (r->vid)
      r->vid->pause(r->vid);
}

static void vid_resume(recorder_controls *rc)
{
   _recorder_controls *r = (_recorder_controls*)rc;

   if (!r)
      return;

   if (r->vid)
      r->vid->record(r->vid);
}

static void switchboard(void *obj,ui_update *u)
{
   _recorder_controls *r;
   ui_element *e = (ui_element*)obj;
   ui_balloon *bl;
   struct tm *loc_time;
   time_t now;
   char temp[128];
   vr_params p;

   if (!e)
      return;

   r = (_recorder_controls*)e->owner;

   switch (e->id) {
      case ctl_width:
         r->width = ((ui_slider*)e)->value;
         break;
      case ctl_height:
         r->height = ((ui_slider*)e)->value;
         break;
      case ctl_fps:
         r->fps = ((ui_slider*)e)->value;
         break;
      case ctl_audio:
         r->use_audio = ((ui_button*)e)->value;
         break;
      case ctl_quality:
         r->quality = ((ui_slider*)e)->value;
         break;
      case ctl_run:
         switch (r->run_state) {
            case vr_stop:
               if (!r->vid)
                  break;
               r->vid->done(r->vid);
               r->vid = 0;
               r->run_state = vr_stop;
               break;
            case vr_record:
               if (!r->vid) {

                  time(&now);
                  loc_time = localtime(&now);
                  snprintf(
                     temp,
                     128,
                     "%s%d.%02d_%02d.%02d.%02d.avi",
                     r->title,
                     loc_time->tm_mon + 1,
                     loc_time->tm_mday,
                     loc_time->tm_hour,
                     loc_time->tm_min,
                     loc_time->tm_sec);

                  p.in_width   = frame_state.window_pixels[0];
                  p.in_height  = frame_state.window_pixels[1];
                  p.out_width  = r->width;
                  p.out_height = r->height;
                  p.out_fps    = r->fps;
                  p.out_file   = temp;
                  p.use_audio  = r->use_audio;
                  p.quality    = r->quality;

                  r->vid = create_video_recorder(&p);
               }
               if (r->vid) {
                  r->vid->record(r->vid);
                  r->run_state = vr_record;
               }
               break;
            case vr_pause:
               if (r->vid) {
                  r->vid->pause(r->vid);
                  r->run_state = vr_pause;
               }
               else r->run_state = vr_stop;
               break;
         }
         break;
      case ctl_done:
         bl = (ui_balloon*)e->owner;
         bl->shrink(bl);
         break;
   }
}

static void on_grow(ui_balloon *b, ui_update *u)
{
   _recorder_controls *r = (_recorder_controls*)b;
   unsigned w_width  = frame_state.window_pixels[0];
   unsigned w_height = frame_state.window_pixels[1];

   if (!b)
      return;

   r->width  = (w_width < 800)  ? w_width  : 800;
   r->height = (w_height < 600) ? w_height : 600;

   r->sl_width->span   = w_width - 640 + 1;
   r->sl_width->value  = r->width;
   r->sl_height->span  = w_height - 480 + 1;
   r->sl_height->value = r->height;
}

static void text_updater(ui_text_value *tx)
{
   _recorder_controls *r;
   unsigned fps;

   if (!tx)
      return;

   r = (_recorder_controls*)tx->owner;

   if (r->vid) {
      fps = (r->vid->record_time) ? r->vid->frames/r->vid->record_time : 0;
      snprintf(tx->text,tx->length,
         "Width:             %d\n"
         "Height:            %d\n"
         "FPS:               %d\n"
         "Quality:           %s\n"
         "Using Audio:       %s\n"
         "Time:              %.1lf\n"
         "File:              %s\n",
         r->vid->out_width,
         r->vid->out_height,
         fps,
         (r->quality > .01) ? (r->quality > .75) ?
            "High(Slow)": "Enhanced(Slower)" : "Default(Fast)",
         (r->vid->audio_active) ? "Yes" : "No",
         r->vid->record_time,
         (r->vid->out_file) ? r->vid->out_file : "Pending");
   }
   else {
      snprintf(tx->text,tx->length,
         "Width:             %d\n"
         "Height:            %d\n"
         "FPS:               %d\n"
         "Quality:           %s\n"
         "Use Audio:         %s\n\n\n",
         r->width,
         r->height,
         r->fps,
         (r->quality > .01) ? (r->quality > .75) ?
            "High(Slow)": "Enhanced(Slower)" : "Default(Fast)",
         (r->use_audio) ? "Yes" : "No");
   }
}


#ifdef __cplusplus
}
#endif
