#include "gl_extensions.h"
#include "shader_api.h"

#ifdef __cplusplus
"C" {
#endif



unsigned shader_program_compile(shader_program program,unsigned validate)
{
   shader_program_api *p = (shader_program_api*)program;

   if (!p || !p->compile)
      return 0;

   return p->compile(p,validate);
}

unsigned shader_std_compile_fn(shader_program program,unsigned validate)
{
   shader_program_api *p = (shader_program_api*)program;
   GLenum err;
   GLint status = 0;
   const char *src[1];
   char *name, *no_name = "<?>";

   if (!gl_extensions_init() || !p)
      return 0;

   if (glIsShader(p->vert_shader) &&
       glIsShader(p->frag_shader) &&
       glIsProgram(p->program))
      return 1;

	if (p->name)
		name = p->name;
	else
		name = no_name;

   /* Clear error buffer */
   glGetError();

   p->vert_shader = glCreateShader(GL_VERTEX_SHADER);
   if (glGetError() != GL_NO_ERROR) {
      printf("Module '%s', could not create vertex shader.\n",name);
      return 0;
   }

   src[0] = p->vert_src;
   glShaderSource(p->vert_shader,1,src,NULL);
   if (glGetError() != GL_NO_ERROR) {
      printf("Module '%s', could not set vertex shader source.\n",name);
      return 0;
   }

   glCompileShader(p->vert_shader);
   if (glGetError() != GL_NO_ERROR) {
      printf("Module '%s', no compiler for vertex shader source.\n",name);
      return 0;
   }

   glGetShaderiv(p->vert_shader,GL_COMPILE_STATUS,&status);
   if (!status) {
      GLint len;
      char *log;

      printf("Module '%s', errors in vertex shader source...\n",name);
      glGetShaderiv(p->vert_shader,GL_INFO_LOG_LENGTH,&len);
      log = (char*)calloc(len + 1,sizeof(char));
      if (log) {
         glGetShaderInfoLog(p->vert_shader,len,NULL,log);
         printf("\n%s\n",log);
         free(log);
      }

      return 0;
   }

   if (validate) {
      GLint len;
      char *log;

      glGetShaderiv(p->vert_shader,GL_INFO_LOG_LENGTH,&len);
      if (len > 2) {
	      printf("Vertex shader log for module '%s':\n",name);
         log = (char*)calloc(len + 1,sizeof(char));
         if (log) {
            glGetShaderInfoLog(p->vert_shader,len,NULL,log);
            printf("%s\n",log);
            free(log);
         }
      }
   }

   glGetError();

   p->frag_shader = glCreateShader(GL_FRAGMENT_SHADER);
   if (glGetError() != GL_NO_ERROR) {
      printf("Module '%s', could not create fragment shader.\n",name);
      return 0;
   }

   src[0] = p->frag_src;
   glShaderSource(p->frag_shader,1,src,NULL);
   if (glGetError() != GL_NO_ERROR) {
      printf("Module '%s', could not set fragment shader source.\n",name);
      return 0;
   }

   glCompileShader(p->frag_shader);
   if (glGetError() != GL_NO_ERROR) {
      printf("Module '%s', no compiler for fragment shader source.\n",name);
      return 0;
   }

   glGetShaderiv(p->frag_shader,GL_COMPILE_STATUS,&status);
   if (!status) {
      GLint len;
      char *log;

      printf("Module '%s', errors in fragmnet shader source...\n",name);
      glGetShaderiv(p->frag_shader,GL_INFO_LOG_LENGTH,&len);
      log = (char*)calloc(len + 1,sizeof(char));
      if (log) {
         glGetShaderInfoLog(p->frag_shader,len,NULL,log);
         printf("\n%s\n",log);
         free(log);
      }

      return 0;
   }

   if (validate) {
      GLint len;
      char *log;

      glGetShaderiv(p->frag_shader,GL_INFO_LOG_LENGTH,&len);
      if (len > 2) {
	      printf("Fragment shader log for module '%s':\n",name);
         log = (char*)calloc(len + 1,sizeof(char));
         if (log) {
            glGetShaderInfoLog(p->frag_shader,len,NULL,log);
            printf("%s\n",log);
            free(log);
         }
      }
   }

   p->program = glCreateProgram();
   if (glGetError() != GL_NO_ERROR) {
      printf("Module '%s', could not create shader program.\n",name);
      return 0;
   }

   glAttachShader(p->program,p->vert_shader);
   glAttachShader(p->program,p->frag_shader);
   glLinkProgram(p->program);
   if (glGetError() != GL_NO_ERROR) {
      printf("Module '%s', could not link program.\n",name);
      return 0;
   }

   glDetachShader(p->program,p->vert_shader);
   glDetachShader(p->program,p->frag_shader);
   glDeleteShader(p->vert_shader);
   glDeleteShader(p->frag_shader);
   p->vert_shader = 0;
   p->frag_shader = 0;

   if (validate) {
      glValidateProgram(p->program);
      glGetProgramiv(p->program,GL_VALIDATE_STATUS,&status);
      if (!status) {
         GLint len;
         char *log;

         printf("Module '%s', errors in shader program runtime...\n",name);
         glGetProgramiv(p->program,GL_INFO_LOG_LENGTH,&len);
         log = (char*)calloc(len + 1,sizeof(char));
         if (log) {
            glGetProgramInfoLog(p->program,len,NULL,log);
            printf("\n%s\n",log);
            free(log);
         }

         return 0;
      }
   }

   if (validate) {
      GLint len;
      char *log;

      glGetProgramiv(p->program,GL_INFO_LOG_LENGTH,&len);
      if (len > 2) {
	      printf("Program log for module '%s':\n",name);
         log = (char*)calloc(len + 1,sizeof(char));
         if (log) {
            glGetProgramInfoLog(p->program,len,NULL,log);
            printf("%s\n",log);
            free(log);
         }
      }
   }

   if (p->enumerate)
      p->enumerate(p);

   return 1;
}

void shader_program_execute(shader_program program)
{
   shader_program_api *p = (shader_program_api*)program;

   if (!gl_extensions_init() || !p || !p->execute)
      return;

   p->execute(p);
}

void shader_program_free(shader_program program)
{
   shader_program_api *p = (shader_program_api*)program;


   if (!p)
      return;

   if (p->allocated_srcs) {
      if (p->frag_src)
         free(p->frag_src);
      if (p->vert_src)
         free(p->vert_src);
   }

   if (!gl_extensions_init())
      return;

   glUseProgram(0);
   if (glIsShader(p->vert_shader))
   	glDeleteShader(p->vert_shader);
   if (glIsShader(p->frag_shader))
   	glDeleteShader(p->frag_shader);
   glDeleteProgram(p->program);
   p->program = 0;
   p->vert_shader = 0;
   p->frag_shader = 0;
}

#ifdef __cplusplus
}
#endif
