#ifdef __cplusplus
extern "C" {
#endif


#include "platform.h"
#include "renderer_api.h"
#include "display.h"
#include "obj_model.h"
#include "gluf.h"
#include "aluf.h"
#include "planck_functions.h"
#include "std_decor.h"
#include "shader_api.h"
#include "basic_shader.h"
#include "simple_shader.h"
#include "line_shader.h"
#include "geom_object.h"
#include "model_shader.h"
#include "model_shader_nice.h"
#include "model_shader_nicest.h"
#include "gltf.h"

#include "solar_system.h"



/******************************* introduction *********************************

The relationship between the orbit distances and body sizes is described by:

   Relationship = 1 Earth's radius / 1 Earth's orbital radius
                = ER / AU
                = 6378.0 km / 149600000.0 km
                = 4.2634 x 10^-5

We define these conversion macros for general usage...*/
#define AU_TO_ER(ed)                   ((ed) * 23455.0)
#define ER_TO_AU(au)                   ((au) * 4.2634e-5)
#define KM_TO_ER(km)                   ((km) / 6378.0)
#define ER_TO_KM(ed)                   ((ed) * 6378.0)
#define KM_TO_AU(km)                   ((km) / 149600000.0)
#define AU_TO_KM(au)                   ((au) * 149600000.0)


/********************************* constants **********************************/

#define TWO_PI                         6.283185307180

/* Defaults and Planck mission parameters... */
#define EARTH_TIME                     1.0               /* 2 pi = 1 day */
#define EARTH_SIZE                     15.0              /* size multiplier */
#define AU_BASE                        50.0              /* GL scale factor */
#define ORBIT_DIVS                     0.0042            /* radians */
#define L2_RADIUS                      1500000.0         /* km */
#define START_TIME                     3388.25           /* J2K days */
#define START_SCAN                     136.0658          /* mission days */
#define AX_TLT                         0.409             /* Earths axial tilt */

/* "Real World" fudge factors... */
#define LAUNCH_CORRECTION              32.55             /* In days*/
#define LAUNCH_X                       6378.0            /* km */
#define LAUNCH_Y                       1275.6            /* km */
#define LAUNCH_Z                       637.80            /* km */
#define PLANCK_TRAIL_X_ALIGN           1500.0            /* km */


#define CELESTIAL_SPHERE_ZERO          -22.177           /* spring equinox = 0 */
#define STAR_RADIUS                    40.0              /* AUs */
#define SCAN_SHELL_RADIUS              39.0              /* AUs */
#define EARTH_SKIN_SHIFT               297.0             /* Degrees */
#define LUNA_SKIN_SHIFT                -90.0             /* Degrees */
#define SAT_RING_SKIN_SHIFT            200.0             /* Degrees */

/* Other stuff... */
//#define PLANCK_MODEL_FILE              "planck_9_6"
#define PLANCK_MODEL_FILE              "planck2"
#define PLANCK_SCALE_MODEL             0.008
#define ALL_TEXTURES                   19

/* Position types. */
#define BODY                           0
#define ORBIT                          1
#define SUN_SYNC_ORBIT                 2
#define SUN_SYNC_EQ_ORBIT              3

/* All definable body locations... */
#define NUM_OF_LOCATIONS               12

#define PLANETS_DIR                    "./data/images/solar_system/"

/* Ambient audio files... */
static const char *AUDIO_LAUNCH_FILE   = "./data/audio/launch.ogg\0";
static const char *AUDIO_SCAN_FILE     = "./data/audio/scan.ogg\0";

static const char *MAP_DIR             = "./data/images/maps";
static const char *THUMB_DIR           = "./data/images/maps/thumbs";


/**************************** constant tables *********************************/

/* Entities */
enum {
   SUN = 0,
   MERCURY,
   VENUS,
   EARTH,
   MARS,
   JUPITER,
   SATURN,
   URANUS,
   NEPTUNE,
   LUNA,
   LAGRNG2,
   PLANCK,
   SAT_RINGS,
   PLANCK_PATH,
   SCAN_SHELL,
   STARS,
   STAR_MAP,
   STAR_ALT,
   SUN_GAUSS
};

/* Defintion of the table structure. Orbital and axial are velocites positive
with respect to CCW motion. */
typedef struct {
   char name[20];  /* short name of body */
   float size;     /* size of graphical model with respect to Earth's model */
   float s_maj_ax; /* semi-major axis with respect to Earth's orbit */
   float or_ecc;   /* orbial eccentricy factor */
   float or_inc;   /* orbital inclination in radians */
   float ax_tlt;   /* axial tilt in radians */
   float or_v;     /* average orbital velocity in revolutions per 24.0 hours */
   float ax_v;     /* axial velocity in revolutions per 24.0 hours */
   float ini_pos;  /* initial position on April 12,2009 00:00 */
}
BodyProp;

static const BodyProp p_tab[] = {

/*-----------------------------------------------------------------------------
  name     size  s_maj_ax  or_ecc  or_inc  ax_tlt  or_v        ax_v     ini_pos
  -----------------------------------------------------------------------------*/
{"Sun",    109.0,   0.000,   0.000, 0.000, 0.000,  0.000,      0.0394,  0.0000 },
{"Mercury",  0.382, 0.387,   0.206, 0.122, 0.0047, 0.0114,     0.0171,  3.9671 },
{"Venus",    0.949, 0.723,   0.007, 0.059, 3.096,  0.00445,   -0.00412, 1.0142 },
{"Earth",    1.000, 1.000,   0.017, 0.000, AX_TLT, 0.00274,    1.003,   0.0000 },
{"Mars",     0.533, 1.524,   0.093, 0.032, 0.440,  0.00146,    0.975,   2.3433 },
{"Jupiter", 11.19,  5.203,   0.048, 0.023, 0.055,  0.000231,   2.439,   1.9694 },
{"Saturn",   9.46,  9.539,   0.056, 0.043, 0.467,  0.0000929,  2.347,   3.6730 },
{"Uranus",   3.98, 19.19,    0.046, 0.013, 1.706,  0.0000326, -1.392,   2.6545 },
{"Neptune",  3.81, 30.06,    0.010, 0.031, 0.494,  0.0000166,  1.490,   2.1835 },
{"Luna",     0.272, 0.00256, 0.055, 0.140, 0.026,  0.0366,     0.0366,  0.7854 },
{"L2",       0.100, 0.0108,  0.000, 0.000, 0.000,  0.00274,    0.100,   0.0000 },
{"Planck",   0.100, 0.000,   0.000, 0.000, 0.000,  0.000,      0.000,   0.0000 },
{"Sat Rng", 28.38,  0.000,   0.000, 0.000, 0.000,  0.000,      0.000,   0.0000 }
};

static const char month_name[13][20] = {
   "Undefined",
   "January",
   "February",
   "March",
   "April",
   "May",
   "June",
   "July",
   "August",
   "September",
   "October",
   "November",
   "December"
};

/* End-of-month day... */
static const int norm_year[13] = {
   0,31,59,90,120,151,181,212,243,273,304,334,365
};

static const int leap_year[13] = {
   0,31,60,91,121,152,182,213,244,274,305,335,366
};

typedef struct {
   char *name;
   char *desc;
   char *file;
   float zero_align, inclination, final_align;
}
tex_rec;

#define MAX_SKY_TEXTURES 5
static const tex_rec sky_texture_files[MAX_SKY_TEXTURES] = {

   "None",
   "None",
   0,
   0,0,0,

   "Visible Stars",
   "Standard visible star background",
   "starmap.jpg",
   0,DEG(AX_TLT),0,

   "Markers",
   "Colorful visible markers",
   "scales.jpg",
   0,DEG(AX_TLT),0,

   "Star Chart",
   "Star background with constellations",
   "skymap.jpg",
   0,DEG(AX_TLT),0,

   "Visible Sky",
   "Visible sky composite",
   "DSS2.jpg",
   -6.327595,-60.283070,270.126465,
};

#define MAX_SCAN_TEXTURES 10
static const tex_rec scan_texture_files[MAX_SCAN_TEXTURES] = {

   "None",
   "None",
   0,
   0,0,0,

   "Planck Full Sky 2011",
   "First year CMB map",
   "PLANCK_FS.png",
   -6.327595,-60.283070,270.126465,

   "Planck Full Sky 2013",
   "2013 CMB Map",
   "PLANCK_FS_2013.jpg",
   -6.327595,-60.283070,270.126465,

   "Planck Matter",
   "Matter Distribution Map",
   "PLANCK_FS_MATTER.jpg",
   -6.327595,-60.283070,270.126465,

   "Visible Sky",
   "Visible sky composite",
   "DSS2.jpg",
   -6.327595,-60.283070,270.126465,

   "Gamma Ray Sky",
   "Fermi sky composite",
   "fermi.jpg",
   -6.327595,-60.283070,270.126465,

   "H-alpha Sky",
   "Visible sky composite",
   "H-alpha.jpg",
   -6.327595,-60.283070,270.126465,

   "IRAS Sky",
   "Visible sky composite",
   "IRAS.jpg",
   -6.327595,-60.283070,270.126465,

   "Radio Sky",
   "Visible sky composite",
   "Radio_vsmoothed.jpg",
   -6.327595,-60.283070,270.126465,

   "RASS Sky",
   "Visible sky composite",
   "RASS.jpg",
   -6.327595,-60.283070,270.126465

};



/******************************* object definition ****************************/

typedef struct _solar_system _solar_system;

struct _solar_system {

   IMPLEMENTS_SOLAR_SYSTEM

   float          obj_scale;
   float          orb_scale;
   float          t_scale;
   float          t_scale_was;
   double         sys_time;
   double         fx_time;
   double         j2k_time;
   unsigned int   origin;
   unsigned int   trace;
   unsigned int   trace_pulse;
   unsigned int   fx;
   unsigned int   show_map;
   unsigned int   wire_frame;
   unsigned int   show_marker;
   unsigned int   star_map;
   unsigned int   scan_map;
   unsigned int   pending_index;
   gluf_bitmap    *new_skymap;
   gluf_bitmap    *new_scanmap;
   int            context_cycle;
   ALuint         snd_src;
   stream_src     *stream;
   planck_object  *planck;
   obj_model      planck_model;
   obj_model      L2_model;

   ui_balloon     *sky_menu;
   ui_balloon     *scan_menu;
   ui_balloon     *adjustment_menu;
   float          t_alpha,t_origin,t_tilt,t_final;
   ui_cursor_fn   *cursor;
   float          *rotation_array;

   float          pos_of[12][3];
};


/*************************** function prototypes ******************************/

static void update(solar_system *obj, double frame_dt);
static void draw(solar_system *obj, mat44 *projection, mat44 *modelview);
static unsigned begin(solar_system *obj);
static void stop(solar_system *obj);
static void erase(solar_system *obj);

static ui_element *create_object_scale_ctl(solar_system *obj);
static ui_element *create_time_ctl(solar_system *obj);
static ui_element *create_time_scale_ctl(solar_system *obj);
static ui_element *create_location_ctl(solar_system *obj);
static ui_element *create_trace_ctl(solar_system *obj);
static ui_element *create_scan_map_ctl(solar_system *obj);
static ui_element *create_wire_frame_ctl(solar_system *obj);
static ui_element *create_marker_ctl(solar_system *obj);
static ui_element *create_star_map_ctl(solar_system *obj);
static ui_element *create_info_dsp(solar_system *obj, float update_interval);
static ui_element *create_hotkeys(solar_system *obj);
static ui_element *create_scan_ctl(solar_system *obj);
static ui_element *create_sky_panel(solar_system *obj);
static ui_element *create_scan_panel(solar_system *obj);

static void control_action(void *obj, ui_update *u);
static void time_slider_update(void *slider, ui_update *u);

static void sky_switchboard(void *obj,ui_update *u);
static void scan_switchboard(void *obj,ui_update *u);
static void close_panel(void *obj,ui_update *u);

static ui_element *create_adjustment_panel(solar_system *obj);
static void adjustment_switchboard(void *obj,ui_update *u);

static void create_thumb(void *reference, unsigned id);

static void create_skymap(void *owner, unsigned id);
static void create_scanmap(void *owner, unsigned id);
static void orbit_creator(void);

static void generic_context_begin(void* owner);
static void generic_context_end(void* owner);
static void vbo1_context(void *owner, unsigned id);
static void vbo2_context(void *owner, unsigned id);
static void vbo3_context(void *owner, unsigned id);
static void vbo4_context(void *owner, unsigned id);
static void vbo5_context(void *owner, unsigned id);

static void date_update(ui_text_value *tx);
static void update_pos(_solar_system *obj, int body);

static void position_of(_solar_system *obj, int body ,int type,mat44 *modelview);
static void goto_origin(_solar_system *obj, mat44 *modelview);
static void draw_model(_solar_system *obj, int body, mat44 *modelview, mat44 *projection);
static void draw_orbit(_solar_system *obj, int body, mat44 *modelview, mat44 *projection);

/************************ default object prototype ****************************/

static _solar_system solar_system_proto = {
   &update,
   &draw,
   &begin,
   &stop,
   &erase,
   &create_object_scale_ctl,
   &create_time_ctl,
   &create_time_scale_ctl,
   &create_location_ctl,
   &create_trace_ctl,
   &create_scan_map_ctl,
   &create_wire_frame_ctl,
   &create_marker_ctl,
   &create_star_map_ctl,
   &create_info_dsp,
   &create_hotkeys,
   &create_scan_ctl,
   &create_sky_panel,
   &create_scan_panel,

   EARTH_SIZE,             /* obj_scale            */
   AU_BASE,                /* orb_scale            */
   .5,                     /* t_scale              */
   0,                      /* t_scale_was          */
   LAUNCH_CORRECTION,      /* sys_time             */
   0,                      /* fx_time              */
   START_TIME,             /* j2k_time             */
   EARTH,                  /* origin               */
   0,                      /* trace                */
   1,                      /* trace_pulse          */
   0,                      /* fx                   */
   0,                      /* show_map             */
   0,                      /* wire_frame           */
   0,                      /* show_marker          */
   1,                      /* star_map             */
   4,                      /* scan_map             */
   0,                      /* pending_index        */
   0,                      /* new_skymap           */
   0,                      /* new_scanmap          */
   0,                      /* context_cycle        */
   AL_NONE,                /* snd_src              */
   0,                      /* stream               */
   0,                      /* planck               */
   0,                      /* planck_model         */
   0,                      /* L2_model             */
   0,                      /* sky_menu             */
   0,                      /* scan_menu            */
   0,                      /* adjustment_menu */
   .5,-6.327595,-60.283070,270.126465,0
                           /* pos_of: auto_filled  */
};


/******************************* static variables *****************************/

static int textures[ALL_TEXTURES];
static float *or_vtx[9];
static float or_clr[10][4] = {
   { 1, 1, 1, 1 },
   { 1, 1, 1, 1 },
   { 1, 1, 1, 1 },
   { 1, 0, 0, 1 },
   { 1, 1, 1, 1 },
   { 1, 1, 1, 1 },
   { 1, 1, 1, 1 },
   { 1, 1, 1, 1 },
   { 1, 1, 1, 1 },
   { 0, 1, 0, 1 }
};

static unsigned or_vtx_sz;
static unsigned int obj_ref = 0;


static unsigned         vbo, vbo2, vbo3, vbo4, vbo5;
static basic_shader_t   shader1, shader2, shader3;
static simple_shader_t  simple1, simple2, simple3, simple4;
static line_shader_t    line1, line2;
static model_shader_nice_t   model1;
static geom_object      *sphere, *ring, *board;
static geom_list        *l2, *planck;

static mat44 base_matrix;
static gl_tex_font *font, *mono;
static vec4 marker_hue = {0,1,0,1};


/********************** creation and destruction functions ********************/

solar_system *create_solar_system(unsigned skymap_on, unsigned traces_on,
                                        unsigned trace_fade, float object_scale,
                                        ui_cursor_fn *cursor,
                                        float *rotation_array)
{
   _solar_system *create = (_solar_system*)malloc(sizeof(_solar_system));

   if (create) {
      *create              = solar_system_proto;
      //create->star_map     = skymap_on;
      create->trace        = traces_on;
      create->trace_pulse  = trace_fade;
      create->obj_scale    = object_scale;
      create->planck       = create_planck_object(0);
      create->planck_model = obj_model_get(PLANCK_MODEL_FILE,PLANCK_SCALE_MODEL);
      planck = obj_model_to_geom_list(create->planck_model);
      create->L2_model     = obj_model_get("L2",1);
      l2 = obj_model_to_geom_list(create->L2_model);
      create->cursor = cursor;
      create->rotation_array = rotation_array;
      font = create_texture_font("./data/fonts/comfortaa.fnt");
      mono = create_texture_font("./data/fonts/vera_mono.fnt");


      if (!obj_ref) {

         orbit_creator();

         /* Unchanging orbital data... */
         create->pos_of[LAGRNG2][0] = p_tab[LAGRNG2].s_maj_ax;
         create->pos_of[LAGRNG2][1] = create->pos_of[LAGRNG2][2] = 0;

         #define REG_TEX(w,n) textures[w] = renderer_register_texture_file(n)

         textures[STAR_MAP] = renderer_register_texture_function(
            create,
            &create_skymap);

         textures[SCAN_SHELL] = renderer_register_texture_function(
            create,
            &create_scanmap);

         REG_TEX(SUN,      PLANETS_DIR "sunTex.png");
         REG_TEX(MERCURY,  PLANETS_DIR "mercuryTex.png");
         REG_TEX(VENUS,    PLANETS_DIR "venusTex.png");
         REG_TEX(EARTH,    PLANETS_DIR "earthTex.png");
         REG_TEX(LUNA,     PLANETS_DIR "lunaTex.png");
         REG_TEX(MARS,     PLANETS_DIR "marsTex.png");
         REG_TEX(JUPITER,  PLANETS_DIR "jupiterTex.png");
         REG_TEX(SATURN,   PLANETS_DIR "saturnTex.png");
         REG_TEX(SAT_RINGS,PLANETS_DIR "satRingTex.png");
         REG_TEX(URANUS,   PLANETS_DIR "uranusTex.png");
         REG_TEX(NEPTUNE,  PLANETS_DIR "neptuneTex.png");
         REG_TEX(SUN_GAUSS,PLANETS_DIR "sun_gauss.png");

         renderer_register_generic(create,&generic_context_begin,&generic_context_end);
         sphere = geom_object_create_sphere(45,45,0);
         ring   = geom_object_create_washer(0,90,.4743);
         board  = geom_object_create_washer(0,90,.01);

         vbo  = renderer_register_vertex_buffer_function(create,&vbo1_context);
         vbo2 = renderer_register_vertex_buffer_function(create,&vbo2_context);
         vbo3 = renderer_register_vertex_buffer_function(create,&vbo3_context);
         vbo4 = renderer_register_vertex_buffer_function(create,&vbo4_context);
         vbo5 = renderer_register_vertex_buffer_function(create,&vbo5_context);
      }
      obj_ref++;
   }

   return (solar_system*)create;
}


static void erase(solar_system *obj)
{
   _solar_system *s = (_solar_system*)obj;
   unsigned int n = 0;

   if (!s)
      return;

  /*  printf("origin: %f\n",s->t_origin);
      printf("tilt: %f\n",s->t_tilt);
      printf("final: %f\n",s->t_final); */

   if (s->planck)
      s->planck->erase(s->planck);

   if (s->stream)
      erase_stream_source(s->stream);

   obj_model_delete(s->planck_model);
   obj_model_delete(s->L2_model);

   if (s->new_skymap)
      free(s->new_skymap);

   if (s->new_scanmap)
      free(s->new_scanmap);

   free(s);

   obj_ref--;

   if (!obj_ref) {

      while (n < 9)
         free(or_vtx[n++]);

      geom_object_free(sphere);
      geom_object_free(ring);
      geom_object_free(board);

      geom_list_free(l2);
      geom_list_free(planck);

      font->erase(font);
      mono->erase(mono);
   }
}


static void create_skymap(void *owner, unsigned id)
{
   _solar_system *s = (_solar_system*)owner;
   char temp[256];

   if (!s)
      return;

   if (s->new_skymap) {
      gluf_bitmap_flip_vertical(s->new_skymap);
      renderer_map_texture(s->new_skymap,1,id);
   }
   else {
      gluf_bitmap *bm;

      if (!s->star_map)
         return;

      snprintf(temp,256,"%s/%s",MAP_DIR,sky_texture_files[s->star_map].file);
      bm = gluf_get_bitmap(temp);

      if (bm) {
         gluf_bitmap_flip_vertical(bm);
         renderer_map_texture(bm,1,id);
         gluf_erase_bitmap(bm);
      }
   }
}


static void create_scanmap(void *owner, unsigned id)
{
   _solar_system *s = (_solar_system*)owner;
   char temp[256];

   if (!s)
      return;

   if (s->new_scanmap) {
      gluf_bitmap_flip_vertical(s->new_scanmap);
      renderer_map_texture(s->new_scanmap,1,id);
   }
   else {
      gluf_bitmap *bm;

      if (!s->scan_map)
         return;

      snprintf(temp,256,"%s/%s",MAP_DIR,scan_texture_files[s->scan_map].file);
      bm = gluf_get_bitmap(temp);

      if (bm) {
         gluf_bitmap_flip_vertical(bm);
         renderer_map_texture(bm,1,id);
         gluf_erase_bitmap(bm);
      }
   }
}

static void create_thumb(void *reference, unsigned id)
{
   char *file = (char*)reference;
   gluf_bitmap *image, *thumb;
   char temp[256];

   if (!file)
      return;

   snprintf(temp,256,"%s/%s",THUMB_DIR,file);
   if (thumb = gluf_get_bitmap(temp)) {
      renderer_map_texture(thumb,1,id);
      gluf_erase_bitmap(thumb);
      return;
   }

   snprintf(temp,256,"%s/%s",MAP_DIR,file);
   image = gluf_get_bitmap(temp);
   if (image) {
     thumb = gluf_bitmap_extract_thumb(image);
     gluf_erase_bitmap(image);
     if (thumb) {
         renderer_map_texture(thumb,1,id);
         snprintf(temp,256,"%s/%s",THUMB_DIR,file);
         gluf_bitmap_to_file(thumb,temp);
         gluf_erase_bitmap(thumb);
      }
   }
}


static void orbit_creator(void)
{
   double theta, ctr_x, s_minor, s_major, max_dec;
   unsigned n = 0, vtx;

   or_vtx_sz = (unsigned)(TWO_PI/ORBIT_DIVS) + 1;

   while (n < 9) {
      vtx       = 0;
      theta     = 0;
      s_major   = p_tab[n + 1].s_maj_ax;
      ctr_x     = p_tab[n + 1].or_ecc * s_major;
      s_minor   = s_major * sqrt(1 - p_tab[n + 1].or_ecc);
      max_dec   = s_major * tan(p_tab[n + 1].or_inc);
      or_vtx[n] = (float*)malloc(or_vtx_sz * 3 * sizeof(float));

      if (!or_vtx[n]) {
         printf("Orbit path allocation failure.\n");
         exit(1);
      }

      while (vtx < or_vtx_sz) {
         or_vtx[n][3 * vtx]     = ctr_x + s_major * cos(theta);
         or_vtx[n][3 * vtx + 1] = max_dec * cos(theta);
         or_vtx[n][3 * vtx + 2] = s_minor * sin(theta);
         theta += ORBIT_DIVS;
         vtx++;
      }
      n++;
   }
}


static void vbo1_context(void *owner, unsigned id)
{
   renderer_static_buffer_data(id,sphere->array_bytes,sphere->array);
}

static void vbo2_context(void *owner, unsigned id)
{
   renderer_static_buffer_data(id,ring->array_bytes,ring->array);
}

static void vbo3_context(void *owner, unsigned id)
{
   renderer_static_buffer_data(id,board->array_bytes,board->array);
}

static void vbo4_context(void *owner, unsigned id)
{
   renderer_static_buffer_data(id,l2->array_bytes,l2->array);
}

static void vbo5_context(void *owner, unsigned id)
{
   renderer_static_buffer_data(id,planck->array_bytes,planck->array);
}

static void generic_context_begin(void* owner)
{
   vec4 light = {1,1,1,1};

   if (!shader_program_compile(&basic_shader,1))
      exit(0);

   shader1 = shader2 = shader3 = basic_shader;

   shader1.vbo = vbo;
   shader1.av3vertex = (void*)0;
   shader1.av3normal = (void*)sphere->normal_offset;
   shader1.av2texture = (void*)sphere->texture_offset;
   shader1.count = sphere->vertex_count;
   shader1.bias = 0.5;

   shader2.vbo = vbo2;
   shader2.av3vertex = (void*)0;
   shader2.av3normal = (void*)ring->normal_offset;
   shader2.av2texture = (void*)ring->texture_offset;
   shader2.count = ring->vertex_count;
   shader2.bias = 0;

   shader3.vbo = vbo3;
   shader3.av3vertex = (void*)0;
   shader3.av3normal = (void*)board->normal_offset;
   shader3.av2texture = (void*)board->texture_offset;
   shader3.count = board->vertex_count;
   shader2.bias = 0;

   if (!shader_program_compile(&simple_shader,1))
      exit(0);

   simple1 = simple2 = simple3 = simple4 = simple_shader;

   simple1.vbo = vbo;
   simple1.av3vertex = (void*)0;
   simple1.av2texture = (void*)sphere->texture_offset;
   simple1.count = sphere->vertex_count;

   simple2.vbo = vbo2;
   simple2.av3vertex = (void*)0;
   simple2.av2texture = (void*)ring->texture_offset;
   simple2.count = ring->vertex_count;
   simple2.cull_face = 0;

   simple3.vbo = vbo3;
   simple3.av3vertex = (void*)0;
   simple3.av2texture = (void*)board->texture_offset;
   simple3.count = board->vertex_count;
   simple3.cull_face = 0;

   simple4.vbo = vbo4;
   simple4.texture = 0;
   simple4.av3vertex = (void*)l2->part[0].vertex_start;
   simple4.av2texture = (void*)l2->part[0].texture_start;
   simple4.uv4color = l2->part[0].material.Ka.fv;
   simple4.count = l2->part[0].vertex_count;
   simple4.cull_face = 1;

   if (!shader_program_compile(&line_shader,1))
      exit(0);

   line1 = line2 = line_shader;

   if (!shader_program_compile(&model_shader_nice,1))
      exit(0);

   model1 = model_shader_nice;

   model1.vbo  = vbo5;
   model1.list = planck;
   planck->light.cutoff_distance = .5;
   //planck->light.cutoff_distance = .75;
   //planck->light.minimum_level   = .3;
   planck->light.minimum_level   = .5;
   planck->light.La = light;
   planck->light.Ld = light;
   planck->light.Ls = light;
}

static void generic_context_end(void* owner)
{
   shader_program_free(&basic_shader);
   shader_program_free(&simple_shader);
   shader_program_free(&line_shader);
   shader_program_free(&model_shader_nice);
}


/********************************** sound *************************************/

static unsigned begin(solar_system *obj)
{
   _solar_system *s = (_solar_system*)obj;

   if (!s)
      return AL_NONE;

   if (!s->stream) {
      s->stream = create_stream_source(AUDIO_LAUNCH_FILE);
      if (s->stream)
         s->snd_src = s->stream->src;
   }

   s->sys_time = LAUNCH_CORRECTION;
   return s->snd_src;
}


static void stop(solar_system *obj)
{
   _solar_system *s = (_solar_system*)obj;

   if (!s)
       return;

   if (s->stream) {
      erase_stream_source(s->stream);
      s->stream = 0;
   }

   s->snd_src = AL_NONE;
}

/******************** ui control creation and calbacks ************************/

static ui_element *create_sky_panel(solar_system *obj)
{
   _solar_system *s = (_solar_system*)obj;
   ui_balloon *b;
   ui_container *c;
   ui_icon *i;
   ui_button *bt;
   int n;
   float at_x, at_y;

   if(!s)
      return 0;

   if (s->sky_menu)
      return (ui_element*)s->sky_menu;

   /* The pop up menu... */
   b = create_balloon();
   b->owner = (void*)s;

   /* Control board background... */
   i = create_custom_icon(.1);
   i->look = material_is(0,0,0,1,0);
   i->place(i,0,0,1,.5);
   b->add_item(b,i);

   at_x = .025;
   at_y = .315;
   n = 0;

   while (n < MAX_SKY_TEXTURES) {

      i = create_custom_icon(.5);
      decor_all_colors((ui_decor*)&i->look,.2,.2,.2,.2);
      i->place(i,at_x - .00125,at_y - .001,.1525,.1525);
      b->add_item(b,i);

      i = create_custom_icon(.5);
      i->place(i,at_x,at_y,.15,.15);
      if (!n)
         decor_all_colors((ui_decor*)&i->look,0,0,0,1);
      else {
         *((ui_decor*)&i->look) = std_icon_decor;
         decor_all_textures((ui_decor*)&i->look,renderer_register_texture_function(
           sky_texture_files[n].file,
           &create_thumb));
      }
      i->use_look.r = 0.35;
      i->use_look.g = 0.35;
      i->use_look.b = 0.35;
      i->on_use = &sky_switchboard;
      i->owner  = (void*)s;
      i->id = n;
      b->add_item(b,i);
      b->add_item(b,create_item_label(i,sky_texture_files[n].name));

      n++;
      at_x += .15 + .05;
      if (!(n%5)) {
         at_y -= .2;
         at_x = .025;
      }
   }

    /* Exit button... */
   bt = create_button();
   bt->place(bt,.85,0.0,.1,.1);
   *((ui_decor*)&bt->look) = std_rb_decor;
   bt->on_use = &close_panel;
   bt->owner = (void*)b;
   b->add_item(b,bt);
   b->add_item(b,create_item_label(bt,"Done"));

   s->sky_menu = b;
   return (ui_element*)b;
}

enum adj {adj_origin, adj_tilt, adj_final, adj_alpha, adj_show};

static ui_element *create_scan_panel(solar_system *obj)
{
   _solar_system *s = (_solar_system*)obj;
   ui_balloon *b;
   ui_container *c;
   ui_icon *i;
   ui_button *bt;
   ui_slider *sl;
   int n;
   float at_x, at_y;

   if(!s)
      return 0;

   if (s->scan_menu)
      return (ui_element*)s->scan_menu;

   /* The pop up menu... */
   b = create_balloon();
   b->owner = (void*)s;

   /* Control board background... */
   i = create_custom_icon(.1);
   i->look = material_is(0,0,0,1,0);
   i->place(i,0,0,1,.5);
   b->add_item(b,i);

   at_x = .025;
   at_y = .315;
   n = 0;

   while (n < MAX_SCAN_TEXTURES) {

      i = create_custom_icon(.5);
      decor_all_colors((ui_decor*)&i->look,.2,.2,.2,.2);
      i->place(i,at_x - .00125,at_y - .001,.1525,.1525);
      b->add_item(b,i);

      i = create_custom_icon(.5);
      i->place(i,at_x,at_y,.15,.15);
      if (!n)
         decor_all_colors((ui_decor*)&i->look,0,0,0,1);
      else {
         *((ui_decor*)&i->look) = std_icon_decor;
         decor_all_textures((ui_decor*)&i->look,renderer_register_texture_function(
           scan_texture_files[n].file,
           &create_thumb));
      }
      i->on_use = &scan_switchboard;
      i->use_look.r = 0.35;
      i->use_look.g = 0.35;
      i->use_look.b = 0.35;
      i->owner  = (void*)s;
      i->id = n;
      b->add_item(b,i);
      b->add_item(b,create_item_label(i,scan_texture_files[n].name));

      n++;
      at_x += .15 + .05;
      if (!(n%5)) {
         at_y -= .185;
         at_x = .025;
      }
   }

    /* Exit button... */
   bt = create_button();
   bt->place(bt,.85,0.0,.1,.1);
   *((ui_decor*)&bt->look) = std_rb_decor;
   bt->on_use = &close_panel;
   bt->owner = (void*)b;
   b->add_item(b,bt);
   b->add_item(b,create_item_label(bt,"Done"));

   sl = create_slider();
   sl->place(sl,0.15,0.025,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)s;
   sl->id     = adj_alpha;
   sl->min    = 0;
   sl->span   = 1.0;
   sl->value  = s->t_alpha;
   sl->on_use = &adjustment_switchboard;
   b->add_item(b,create_slider_labels(sl,2,
      "Less","More"));
   b->add_item(b,create_item_label(sl,"Visibility"));
   b->add_item(b,sl);

   /* Hidden adjust button... */
   c = create_container();
   c->place(c,0,0.95,.05,.05);
   c->on_use = &adjustment_switchboard;
   c->id = adj_show;
   c->owner = (void*)s;
   b->add_item(b,c);
   b->add_item(b,create_adjustment_panel(obj));

   s->scan_menu = b;
   return (ui_element*)b;
}

ui_element *create_adjustment_panel(solar_system *obj)
{
   _solar_system *s = (_solar_system*)obj;
   ui_balloon *b;
   ui_container *c;
   ui_element *e;
   ui_icon *i;
   ui_button *bt;
   ui_slider *sl;

   if(!s)
      return 0;

   if (s->adjustment_menu)
      return (ui_element*)s->adjustment_menu;

   /* The pop up menu... */
   b = create_balloon();
   b->owner = (void*)s;

   /* Control board background... */
   i = create_custom_icon(.1);

   i->look = material_is(0,0,0,1,0);
   i->place(i,0,0,1,.5);
   b->add_item(b,i);


   sl = create_slider();
   sl->place(sl,.1,0,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)s;
   sl->id     = adj_alpha;
   sl->min    = 0;
   sl->span   = 1.0;
   sl->value  = s->t_alpha;
   sl->on_use = &adjustment_switchboard;
   b->add_item(b,create_slider_labels(sl,3,
      "0",".5","1.0"));
   b->add_item(b,create_item_label(sl,"Alpha"));
   b->add_item(b,sl);


   sl = create_slider();
   sl->place(sl,.1,.075,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)s;
   sl->id     = adj_origin;
   sl->min    = -10;
   sl->span   = 20;
   sl->value  = s->t_origin;
   sl->on_use = &adjustment_switchboard;
   b->add_item(b,create_slider_labels(sl,3,
      "-10","0","10"));
   b->add_item(b,create_item_label(sl,"Offset Angle"));
   b->add_item(b,sl);


   sl = create_slider();
   sl->place(sl,.1,.15,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)s;
   sl->id     = adj_tilt;
   sl->min    = -62.5;
   sl->span   = 5;
   sl->value  = s->t_tilt;
   sl->on_use = &adjustment_switchboard;
   b->add_item(b,create_slider_labels(sl,3,
      "62.5","60","57.5"));
   b->add_item(b,create_item_label(sl,"Plane Angle"));
   b->add_item(b,sl);

   sl = create_slider();
   sl->place(sl,.1,.225,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)s;
   sl->id     = adj_final;
   sl->min    = 267.5;
   sl->span   = 5;
   sl->value  = s->t_final;
   sl->on_use = &adjustment_switchboard;
   b->add_item(b,create_slider_labels(sl,3,
      "267.5","270","275.5"));
   b->add_item(b,create_item_label(sl,"Intersection Angle"));
   b->add_item(b,sl);

    /* Exit button... */
   bt = create_button();
   bt->place(bt,.85,0.2,.1,.1);
   *((ui_decor*)&bt->look) = std_rb_decor;
   bt->on_use = &close_panel;
   bt->owner = (void*)b;
   b->add_item(b,bt);
   b->add_item(b,create_item_label(bt,"Done"));

   s->adjustment_menu = b;
   return (ui_element*)b;
}



static void adjustment_switchboard(void *obj,ui_update *u)
{
   ui_control *c = (ui_control*)obj;
   _solar_system *s;
   float info;


   if (!c)
      return;

   s = (_solar_system*)c->owner;
   info = *((float*)&c->info);

   switch (c->id) {
      case adj_show:
         if (s->adjustment_menu)
            s->adjustment_menu->grow(s->adjustment_menu);
         break;
      case adj_alpha:
         s->t_alpha = info;
         break;
      case adj_origin:
         s->t_origin = info;
         break;
      case adj_tilt:
         s->t_tilt = info;
         break;
      case adj_final:
         s->t_final = info;
         break;
    }
}


static void close_panel(void *obj,ui_update *u)
{
   ui_element *e = (ui_element*)obj;
   ui_balloon *b;


   if (!e)
      return;

   b = (ui_balloon*)e->owner;
   b->shrink(b);
}

static void sky_switchboard(void *obj,ui_update *u)
{
   ui_element *e = (ui_element*)obj;
   _solar_system *s;
   char temp[256];

   if (!e)
      return;

   s = (_solar_system*)e->owner;

   if (s->context_cycle)
      return;

   s->pending_index = e->id;

   if (!e->id) {
      s->star_map = 0;
      return;
   }

   snprintf(temp,256,"%s/%s",MAP_DIR,sky_texture_files[s->pending_index].file);
   *s->cursor = &ui_busy_cursor;
   gluf_get_bitmap_async(
      temp,
      &s->new_skymap);
   s->context_cycle = -1;
}

static void scan_switchboard(void *obj,ui_update *u)
{
   ui_element *e = (ui_element*)obj;
   _solar_system *s;
   char temp[256];

   if (!e)
      return;

   s = (_solar_system*)e->owner;

   if (s->context_cycle)
      return;

   s->pending_index = e->id;

   if (!e->id) {
      s->scan_map = 0;
      return;
   }

   snprintf(temp,256,"%s/%s",MAP_DIR,scan_texture_files[s->pending_index].file);
   *s->cursor = &ui_busy_cursor;
   gluf_get_bitmap_async(
      temp,
      &s->new_scanmap);
   s->context_cycle = -1;
}


enum {ctl_size,ctl_time,ctl_speed,ctl_at,ctl_orbit,ctl_scan,ctl_wire,ctl_map,
      ctl_keys,ctl_show,ctl_marker};


static void control_action(void *obj, ui_update *u)
{
   ui_element *e = (ui_element*)obj;
   _solar_system *s;
   int loc;

   if (!e)
      return;

   s = (_solar_system*)e->owner;

   switch (e->id) {
      case ctl_size:
         s->obj_scale = ((ui_slider*)e)->value;
         break;
      case ctl_time:
         s->sys_time = ((ui_slider*)e)->value;
         break;
      case ctl_speed:
         s->t_scale = ((ui_slider*)e)->value;
         break;
      case ctl_at:
         s->origin++;
         if (s->origin > NEPTUNE)
            s->origin = SUN;
         break;
      case ctl_orbit:
         s->trace = ((ui_button*)e)->value;
         break;
      case ctl_scan:
         if (s->scan_menu)
            s->scan_menu->grow(s->scan_menu);
         break;
      case ctl_wire:
         s->wire_frame = ((ui_button*)e)->value;
         break;
      case ctl_marker:
         s->show_marker = ((ui_button*)e)->value;
         break;
      case ctl_map:
         if (s->sky_menu)
            s->sky_menu->grow(s->sky_menu);
         break;
      case ctl_show:
         s->show_map = ((ui_button*)e)->value;
         break;
      case ctl_keys:
         if (u->key < 97)
            loc = u->key - 49;                 /* base = keyboard '1' */
         else if (u->key > 96 && u->key < 106) /* base = WinAPI VK_NUMPAD1; */
            loc = u->key - 97;                 /* also, ASCII 'a'... */
         else
            loc = u->key - 303;                /* base = GLFW_KEY_KP_1   */

         if (loc >= 0 && loc <= NEPTUNE)
            s->origin = loc;
         break;
   }
}

/* Special slider update for time follower bar... */
static void time_slider_update(void *slider, ui_update *u)
{
   ui_slider *sl = (ui_slider*)slider;
   _solar_system *ss;
   double value;

   if (!sl)
      return;

   ss = (_solar_system*)sl->owner;

   if (!sl->busy) {
      sl->value = ss->sys_time;
      if (sl->value - sl->min > sl->span)
         sl->value = sl->span + sl->min;
      if (sl->value < sl->min)
         sl->value = sl->min;
   }

   if (sl->state == Action) {

      sl->sticky = 1;
      value = (u->px - sl->handle_width/2)/(1 - sl->handle_width);

      if (value >= 0 && value <= 1) {
         if (!sl->busy)
            ss->t_scale_was = ss->t_scale;
         sl->busy    = 1;
         ss->t_scale = 0;
         sl->value   = value * sl->span + sl->min;
         if (sl->value - sl->min > sl->span)
            sl->value = sl->span + sl->min;
         if (sl->value < sl->min)
            sl->value = sl->min;

         if (sl->on_use)
            sl->on_use(sl,u);
      }
   }

   if (sl->state == Hover && sl->on_hover)
      sl->on_hover(sl,u);

   if (sl->state != Action && sl->busy) {
      sl->busy    = 0;
      ss->t_scale = ss->t_scale_was;
   }

   value = (sl->value - sl->min)/sl->span * (1 - sl->handle_width);

  /* Slider channel */
   sl->vtx[0].x = sl->vtx[3].x = sl->vtx[5].x = sl->handle_width/2;
   sl->vtx[1].x = sl->vtx[2].x = sl->vtx[4].x = 1 - sl->handle_width/2;

   /* Slider handle */
   sl->vtx[6].x = sl->vtx[9].x = sl->vtx[11].x = value;
   sl->vtx[7].x = sl->vtx[8].x = sl->vtx[10].x = value + sl->handle_width;
}

static ui_element *create_object_scale_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_slider *create = create_slider();
   if (create) {
      create->owner  = (void*)s;
      create->id     = ctl_size;
      create->min    = 1;
      create->span   = 40;
      create->value  = s->obj_scale;
      create->on_use = &control_action;
   }
   return (ui_element*)create;
}

static ui_element *create_time_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_slider *create = create_slider();
   if (create) {
      create->owner  = (void*)s;
      create->id     = ctl_time;
      create->min    = LAUNCH_CORRECTION;
      create->span   = 500;
      create->value  = s->sys_time;
      create->update = &time_slider_update;
      create->on_use = &control_action;
   }
   return (ui_element*)create;
}

static ui_element *create_time_scale_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_slider *create = create_slider();
   if (create) {
      create->owner        = (void*)s;
      create->id           = ctl_speed;
      create->min          = -10;
      create->span         = 20;
      create->value        = s->t_scale;
      create->on_use       = &control_action;
      create->detents      = 1;
      create->detent_width = .03;
   }
   return (ui_element*)create;
}

static ui_element *create_location_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_button *create = create_button();
   if (create) {
      create->owner  = (void*)s;
      create->id     = ctl_at;
      create->on_use = &control_action;
   }
   return (ui_element*)create;
}

static ui_element *create_trace_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_button *create = create_button();
   if (create) {
      create->owner   = (void*)s;
      create->id      = ctl_orbit;
      create->value   = s->trace;
      create->latches = 1;
      create->on_use  = &control_action;
   }
   return (ui_element*)create;
}

static ui_element *create_scan_map_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_button *create = create_button();
   if (create) {
      create->owner   = (void*)s;
      create->id      = ctl_scan;
      create->value   = s->scan_map;
      create->latches = 0;
      create->on_use  = &control_action;
   }
   return (ui_element*)create;
}

static ui_element *create_wire_frame_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_button *create = create_button();
   if (create) {
      create->owner   = (void*)s;
      create->id      = ctl_wire;
      create->value   = s->wire_frame;
      create->latches = 1;
      create->on_use  = &control_action;
   }
   return (ui_element*)create;
}

static ui_element *create_marker_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_button *create = create_button();
   if (create) {
      create->owner   = (void*)s;
      create->id      = ctl_marker;
      create->value   = s->show_marker;
      create->latches = 1;
      create->on_use  = &control_action;
   }
   return (ui_element*)create;
}

static ui_element *create_star_map_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_button *create = create_button();
   if (create) {
      create->owner   = (void*)s;
      create->id      = ctl_map;
      create->value   = s->star_map;
      create->latches = 0;
      create->on_use  = &control_action;
   }
   return (ui_element*)create;
}

static ui_element *create_hotkeys(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_key *create = create_key();
   if (create) {
      create->owner   = (void*)s;
      create->id      = ctl_keys;
      create->on_use  = &control_action;
   }
   return (ui_element*)create;
}

static ui_element *create_scan_ctl(solar_system *obj)
{
   _solar_system *s  = (_solar_system*)obj;
   ui_button *create = create_button();
   if (create) {
      create->owner   = (void*)s;
      create->id      = ctl_show;
      create->value   = s->show_map;
      create->latches = 1;
      create->on_use  = &control_action;
   }
   return (ui_element*)create;
}


/******************** ui information display **********************************/

static ui_element *create_info_dsp(solar_system *obj, float update_interval)
{
   ui_text_value *create = create_text_value(30,80,&date_update);

   if (create) {
      create->update_on = update_interval;
      create->owner     = (void*)obj;
      create->font      = ui_pretty_fnt;
   }
   return (ui_element*)create;
}

static void date_update(ui_text_value *tx)
{
   _solar_system *s;
   int year, day, month = 1;
   const int *month_table;
   float raw_day, hours, minutes;

   if (!tx)
      return;

   s = (_solar_system*)tx->owner;

   /* Update solar system info */
   year        = (int)(s->j2k_time/365.25);
   raw_day     = s->j2k_time - year * 365.25 + 1;
   day         = (int)raw_day;
   hours       = (raw_day - day) * 24;
   minutes     = (hours - (int)hours) * 60;
   month_table = (!(year%4) && year%100 || !(year%400)) ? leap_year : norm_year;

   while (month < 13 && day > month_table[month])
      month++;

   day -= month_table[month - 1];

   snprintf(
      tx->text,
      tx->length,
      "%s %d, %4d %02d:%02d\nLocation: %s",
      month_name[month],
      day,
      year + 2000,
      (int)hours,
      (int)minutes,
      p_tab[s->origin].name);
}


/**************************** update functions ********************************/

static void update(solar_system *obj, double dt)
{
   _solar_system *s = (_solar_system*)obj;
   int count = SUN;

   if (!s)
      return;

   if (s->context_cycle > 0) {
      s->context_cycle = 0;
      if (s->new_skymap) {
         gluf_erase_bitmap(s->new_skymap);
         s->star_map = s->pending_index;
         s->new_skymap = 0;
         *s->cursor = &ui_basic_cursor;
      }
      if (s->new_scanmap) {
         gluf_erase_bitmap(s->new_scanmap);
         s->scan_map = s->pending_index;
         s->new_scanmap = 0;
         *s->cursor = &ui_basic_cursor;

         s->t_final  = scan_texture_files[s->scan_map].final_align;
         s->t_tilt   = scan_texture_files[s->scan_map].inclination;
         s->t_origin = scan_texture_files[s->scan_map].zero_align;
      }
   }

   if (s->new_skymap)
      s->context_cycle = 1;

   if (s->new_scanmap)
      s->context_cycle = 2;

   s->fx_time  += dt * .45;
   s->sys_time += dt * s->t_scale;
   s->j2k_time  = START_TIME + s->sys_time;

   while (count < LAGRNG2)
      update_pos(s,count++);

   s->planck->update(s->planck,s->sys_time - LAUNCH_CORRECTION, dt);

   /* Planck distances are in km... */
   s->pos_of[PLANCK][0] = KM_TO_AU(s->planck->x + LAUNCH_X * s->obj_scale);
   s->pos_of[PLANCK][1] = KM_TO_AU(s->planck->y + LAUNCH_Y * s->obj_scale);
   s->pos_of[PLANCK][2] = KM_TO_AU(s->planck->z + LAUNCH_Z * s->obj_scale);
}


static void update_pos(_solar_system *s, int sat)
{
   double semi_major, semi_minor, center_x, max_dec, orb_pos;

   if (!s)
      return;

   semi_major = p_tab[sat].s_maj_ax;
   center_x   = p_tab[sat].or_ecc * semi_major;
   semi_minor = semi_major * sqrt(1.0 - p_tab[sat].or_ecc);
   max_dec    = semi_major * tan(p_tab[sat].or_inc);

   /* Orbits are CCW: must negate orb_pos value since OpenGL Z axis orientation
   creates effective clockwise rotation on direct translation of positive
   parametric value. */

   orb_pos = -TWO_PI * s->sys_time * p_tab[sat].or_v - p_tab[sat].ini_pos;

   s->pos_of[sat][0] = center_x + semi_major * cos(orb_pos);
   s->pos_of[sat][1] = max_dec * cos(orb_pos);
   s->pos_of[sat][2] = semi_minor * sin(orb_pos);
}


/*********************** positioning functions ********************************/

static void position_of(_solar_system *s, int sat, int type, mat44 *modelview)
{
   mat44 mt;

   if (!s || !modelview)
      return;

   mat44_translation(&mt,s->pos_of[sat][0],s->pos_of[sat][1],s->pos_of[sat][2]);
   mat44_mult(modelview,&mt);

   switch (type) {
      case SUN_SYNC_EQ_ORBIT:
         mat44_rotation(&mt,p_tab[sat].ax_tlt,1,0,0);
         mat44_mult(modelview,&mt);
         mat44_rotation(&mt,TWO_PI * s->sys_time * p_tab[sat].or_v,0,1,0);
         mat44_mult(modelview,&mt);
         break;
      case SUN_SYNC_ORBIT:
         mat44_rotation(&mt,TWO_PI * s->sys_time * p_tab[sat].or_v,0,1,0);
         mat44_mult(modelview,&mt);
         break;
      case BODY:
         mat44_rotation(&mt,p_tab[sat].ax_tlt,1,0,0);
         mat44_mult(modelview,&mt);
         mat44_rotation(&mt,TWO_PI * s->sys_time * p_tab[sat].ax_v,0,1,0);
         mat44_mult(modelview,&mt);
         break;
      case ORBIT:
         break;
   }
}


static void goto_origin(_solar_system *s, mat44 *modelview)
{
   unsigned int at;
   mat44 mt;

   if (!s || !modelview)
      return;

   at = s->origin;

   /* The following uses the scaled L2 distance to create a consistant initial
   view of planetary locations... */

   if (at == LAGRNG2 || at == LUNA || at == PLANCK || at == EARTH)
      at = EARTH;

   mat44_translation(&mt,0,0,-p_tab[at].size * p_tab[LAGRNG2].s_maj_ax * 1.45);
   mat44_mult(modelview,&mt);
   mat44_rotation(
      &mt,
      -TWO_PI * s->sys_time * p_tab[at].or_v - p_tab[at].ini_pos,
      0,1,0);
   mat44_mult(modelview,&mt);
   mat44_translation(&mt,-s->pos_of[at][0],-s->pos_of[at][1],-s->pos_of[at][2]);
   mat44_mult(modelview,&mt);
}


/****************************** drawing functions *****************************/

static void draw_model(_solar_system *s, int body, mat44 *modelview, mat44 *projection)
{
   float size;
   unsigned texture;
   mat44 tm, mv , mvp;
   vec4 light = {0,0,0,1}, position, color = {1,1,1,1};

   if (!s || !modelview || !projection)
      return;

   if (body == SUN_GAUSS)
      size = ER_TO_AU(p_tab[SUN].size) * s->obj_scale;
   else if (body < PLANCK_PATH)
      size = ER_TO_AU(p_tab[body].size) * s->obj_scale;
   else
      size = KM_TO_AU(1) * s->obj_scale;

   mv  = *modelview;
   mvp = *projection;

   if (body == SUN_GAUSS) {

      float rot = 0;

      if (s->rotation_array)
         rot = s->rotation_array[0] + s->rotation_array[1];

      mv.a11 = mv.a22 = mv.a33 = 1;
      mv.a12 = mv.a13 = mv.a21 = mv.a23 = mv.a31 = mv.a32 = 0;

      mat44_scaling(&tm,s->orb_scale,s->orb_scale,s->orb_scale);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,size,size,size);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,6,6,1);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,10 * rot,0,0,1);
      mat44_mult(&mv,&tm);

      mat44_mult(&mvp,&mv);

      simple3.um44mvp   = &mvp;
      simple3.texture   = textures[SUN_GAUSS];
      simple3.uv4color  = color.fv;
      shader_program_execute(&simple3);
      return;
   }

   if (body == SUN) {

      mat44_rotation(&tm,RAD(-90),1,0,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(90),0,0,1);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,size,size,size);
      mat44_mult(&mv,&tm);

      mat44_mult(&mvp,&mv);

      simple1.um44mvp   = &mvp;
      simple1.texture   = textures[SUN];
      simple1.uv4color  = color.fv;
      simple1.cull_face = 0;
      shader_program_execute(&simple1);
      return;
   }

   if (body == LAGRNG2) {

      mat44_rotation(&tm,RAD(90),0,1,0);
      mat44_mult(&mv,&tm);
      mat44_translation(&tm,-size/1.5,-size/2.5,0);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,size * .6,size * .6,size * .6);
      mat44_mult(&mv,&tm);

      mat44_mult(&mvp,&mv);

      simple4.um44mvp = &mvp;
      shader_program_execute(&simple4);

      mv  = *modelview;
      mvp = *projection;

      mat44_scaling(&tm,size,size,size);
      mat44_mult(&mv,&tm);

      mat44_mult(&mvp,&mv);

      color.x = color.y = color.z = color.w = .3;

      simple1.um44mvp   = &mvp;
      simple1.texture   = 0;
      simple1.uv4color  = color.fv;
      simple1.cull_face = 1;
      shader_program_execute(&simple1);
      return;
   }

   if (body == STARS) {

      if (!s->star_map)
         return;

      mat44_rotation(
         &tm,
         RAD(sky_texture_files[s->star_map].final_align + CELESTIAL_SPHERE_ZERO),
         0,1,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(sky_texture_files[s->star_map].inclination),1,0,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(sky_texture_files[s->star_map].zero_align),0,1,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(90),0,1,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(90),1,0,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(90),0,0,1);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,STAR_RADIUS,STAR_RADIUS,STAR_RADIUS);
      mat44_mult(&mv,&tm);

      mat44_mult(&mvp,&mv);

      simple1.um44mvp   = &mvp;
      simple1.texture   = textures[STAR_MAP];
      simple1.uv4color  = color.fv;
      simple1.cull_face = 0;
      shader_program_execute(&simple1);
      return;
   }

   if (body == PLANCK) {

      mat44_translation(&tm,KM_TO_AU(PLANCK_TRAIL_X_ALIGN) * s->obj_scale,0,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(-90),0,1,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(-90),1,0,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,(s->sys_time - (int)s->sys_time) * TWO_PI,0,1,0);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,size,size,size);
      mat44_mult(&mv,&tm);

      mat44_mult(&mvp,&mv);

      model1.um44mvp = &mvp;
      model1.um44mv  = &mv;
      shader_program_execute(&model1);

      return;
   }

   if (body == PLANCK_PATH) {

      float *vtx, *clr;
      unsigned cnt;

      mat44_translation(
         &tm,
         KM_TO_AU(LAUNCH_X) * s->obj_scale,
         KM_TO_AU(LAUNCH_Y) * s->obj_scale,
         KM_TO_AU(LAUNCH_Z) * s->obj_scale);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,KM_TO_AU(1),KM_TO_AU(1),KM_TO_AU(1));
      mat44_mult(&mv,&tm);

      mat44_mult(&mvp,&mv);

      s->planck->get_path(s->planck,&vtx,&clr,&cnt);

      line1.vbo         = 0;
      line1.av3vertex   = (void*)vtx;
      line1.av4color    = (void*)clr;
      line1.um44mvp     = &mvp;
      line1.uv4color    = 0;
      line1.depth_mask  = 0;
      line1.count       = cnt;
      line1.loop        = 0;
      line1.line_width  = .5;
      shader_program_execute(&line1);

      return;
   }

   if (body == SCAN_SHELL) {

      if (!s->scan_map)
         return;

      color.a = s->t_alpha;

      mat44_rotation(&tm,RAD(s->t_final + CELESTIAL_SPHERE_ZERO),0,1,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(s->t_tilt),1,0,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(s->t_origin),0,1,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(90),0,1,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(90),1,0,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(90),0,0,1);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,SCAN_SHELL_RADIUS,SCAN_SHELL_RADIUS,SCAN_SHELL_RADIUS);
      mat44_mult(&mv,&tm);

      mat44_mult(&mvp,&mv);

      simple1.um44mvp   = &mvp;
      simple1.texture   = textures[SCAN_SHELL];
      simple1.uv4color  = color.fv;
      simple1.cull_face = 0;
      shader_program_execute(&simple1);
      return;
   }

   if (body == SAT_RINGS) {

      mat44_rotation(&tm,RAD(SAT_RING_SKIN_SHIFT),0,1,0);
      mat44_mult(&mv,&tm);
      mat44_rotation(&tm,RAD(90),1,0,0);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,size,size,size);
      mat44_mult(&mv,&tm);

      mat44_mult(&mvp,&mv);

      simple2.um44mvp   = &mvp;
      simple2.texture   = textures[SAT_RINGS];
      simple2.uv4color  = color.fv;
      simple2.cull_face = 0;
      shader_program_execute(&simple2);
      return;
   }

   if (s->show_marker && body == EARTH) {

      mat44 mv_c = mv;
      float mag, sz;
      vec4 hue, earth = {0,1,0,1};
      vec3 loc = {s->pos_of[body][0],s->pos_of[body][1],s->pos_of[body][2]};
      char dist[20];

      mv_c.a11 = mv_c.a22 = mv_c.a33 = 1;
      mv_c.a12 = mv_c.a13 = mv_c.a21 = mv_c.a23 = mv_c.a31 = mv_c.a32 = 0;

      vec3_transform44(&loc,&base_matrix);

      mag = vec3_magnitude(&loc);
      sz  = mag * .15;
      sprintf(dist,"%00.2f",mag/50.0);

      mat44_translation(&tm,-(.5 * sz),55 * size ,0);
      mat44_mult(&mv_c,&tm);
      mat44_scaling(&tm,sz,sz,sz);
      mat44_mult(&mv_c,&tm);

      if (body == EARTH)
         hue = earth;
      else
         hue = marker_hue;

      //hue.a = 1.0 - mag/2200.0;

      *((vec4*)font->color) = hue;
      font->cpl = 10;
      font->centered = 1;

      *((vec4*)mono->color) = hue;
      mono->cpl = 12;
      mono->centered = 1;

      font->draw(font,"v",&mvp,&mv_c);

      mat44_translation(&tm,0,.22 ,0);
      mat44_mult(&mv_c,&tm);
      mono->draw(mono,dist,&mvp,&mv_c);

      mat44_translation(&tm,0,.15 ,0);
      mat44_mult(&mv_c,&tm);
      font->draw(font,p_tab[body].name,&mvp,&mv_c);
   }

   texture = textures[body];

   if (body == EARTH) {
      mat44_rotation(&tm,RAD(EARTH_SKIN_SHIFT),0,1,0);
      mat44_mult(&mv,&tm);
   }

   if (body == LUNA) {
      mat44_rotation(&tm,RAD(LUNA_SKIN_SHIFT),0,1,0);
      mat44_mult(&mv,&tm);
      body = EARTH;
   }

   mat44_rotation(&tm,RAD(-90),1,0,0);
   mat44_mult(&mv,&tm);
   mat44_rotation(&tm,RAD(90),0,0,1);
   mat44_mult(&mv,&tm);
   mat44_scaling(&tm,size,size,size);
   mat44_mult(&mv,&tm);

   vec4_set(&light,s->pos_of[SUN][0],s->pos_of[SUN][1],s->pos_of[SUN][2]);
   vec4_set(&position,s->pos_of[body][0],s->pos_of[body][1],s->pos_of[body][2]);
   vec4_subtract(&light,&position);
   vec4_transform(&light,&base_matrix);
   vec4_normalize(&light);
   vec4_scale(&light,-1);

   mat44_mult(&mvp,&mv);

   shader1.um44mvp  = &mvp;
   shader1.um44mv   = &mv;
   shader1.uv3light = light.fv;
   shader1.uv4color = color.fv;
   shader1.texture  = texture;
   shader_program_execute(&shader1);
}


static void draw_orbit(_solar_system *s, int body, mat44 *modelview, mat44 *projection)
{
   mat44 mvp;

   if (!s || !body || body > LUNA || !modelview || !projection)
      return;

   mvp = *projection;
   mat44_mult(&mvp,modelview);

   line1.vbo         = 0;
   line1.av3vertex   = (void*)or_vtx[body - 1];
   line1.av4color    = 0;
   line1.um44mvp     = &mvp;
   line1.uv4color    = (float*)or_clr[body];
   line1.depth_mask  = 0;
   line1.count       = or_vtx_sz;
   line1.loop        = 1;
   line1.line_width  = .5;
   shader_program_execute(&line1);
}


static void draw(solar_system * obj,mat44 *projection, mat44 *modelview)
{
   _solar_system *s = (_solar_system*)obj;
   int count = SUN;
   mat44 mt, base, mv;

   if (!s)
      return;

   switch (s->context_cycle) {
      case 1:
         create_skymap(obj,textures[STAR_MAP]);
         break;
      case 2:
         create_scanmap(obj,textures[SCAN_SHELL]);
         break;
      case 0:
      default:
         break;
   }

   if (s->wire_frame)
      renderer_polygon_wireframe();

   base = *modelview;
   mat44_scaling(&mt,s->orb_scale,s->orb_scale,s->orb_scale);
   mat44_mult(&base,&mt);

   goto_origin(s,&base);
   mv = base_matrix = base;

   /* General bodies. */
   position_of(s,EARTH,ORBIT,&mv);
   draw_model(s,STARS,&mv,projection);

   mv = base;
   position_of(s,EARTH,ORBIT,&mv);
   draw_model(s,SCAN_SHELL,&mv,projection);

   mv = base;
   position_of(s,SUN,BODY,&mv);
   draw_model(s,SUN,&mv,projection);

   count = MERCURY;
   while (count < LUNA) {
      mv = base;
      position_of(s,count,BODY,&mv);
      draw_model(s,count,&mv,projection);
      count++;
   }

   /* Special bodies... */
   mv = base;
   position_of(s,EARTH,ORBIT,&mv);
   position_of(s,LUNA,BODY,&mv);
   draw_model(s,LUNA,&mv,projection);

   mv = base;
   position_of(s,SATURN,SUN_SYNC_EQ_ORBIT,&mv);
   draw_model(s,SAT_RINGS,&mv,projection);


   if (s->trace) {

      mv = base;
      count = MERCURY;

      while (count < LUNA) {
         draw_orbit(s,count,&mv,projection);
         count++;
      }
      position_of(s,EARTH,ORBIT,&mv);
      draw_orbit(s,LUNA,&mv,projection);
   }

   if (s->sys_time > LAUNCH_CORRECTION) {

      mv = base;
      position_of(s,EARTH,SUN_SYNC_ORBIT,&mv);
      position_of(s,PLANCK,ORBIT,&mv);
      draw_model(s,PLANCK,&mv,projection);

      if (s->trace) {
          mv = base;
         position_of(s,EARTH,SUN_SYNC_ORBIT,&mv);
         draw_model(s,PLANCK_PATH,&mv,projection);
      }
   }

   mv = base;
   draw_model(s,SUN_GAUSS,&mv,projection);

   position_of(s,EARTH,SUN_SYNC_ORBIT,&mv);
   position_of(s,LAGRNG2,ORBIT,&mv);
   draw_model(s,LAGRNG2,&mv,projection);

   renderer_polygon_fill();
}

#ifdef __cplusplus
}
#endif

