#ifdef __cplusplus
extern "C" {
#endif

#define USING_OPENGL
#include "platform.h"
#include "renderer_api.h"

#include "ui_element.h"
#include "std_decor.h"
#include "display.h"
#include "aluf.h"
#include "gluf.h"
#include "root_scene.h"
#include "stage.h"
#include "config_loader.h"
#include "shader_api.h"
#include "ui_2d_shader.h"


#define CHANGE_TIME                 .5  /* seconds */
#define MAX_STACK                   10
#define DEFAULT_CONFIG_NAME         "./data/app.cfg"
#define CONFIG_ID                   "STAGE"

static scene_api *stack[MAX_STACK];
static scene_api **stack_top       = 0;
static scene_api *stack_push       = 0;
static scene_api **stack_max       = 0;
static int changing                = 0;
static int context_cycle           = 0;
static double curtain_time         = 0;
static unsigned dead_time          = 0;
static float  al_gain              = .5;

static config_aggregate *ca;
static config_loader *cl;
static unsigned int initial_view_mode = 0;
static float initial_audio_level = .5;
static float illumination_level = .7;
static char config_name[256];

static unsigned once = 0;
static unsigned context_id;



static void scene_context(void);
static void current_scene_preframe(void);
static void current_scene_display(void);
static void curtain_preframe(void);
static void curtain_display(void);
static void configuration(void);


int main(int argc, char **argv)
{
   unsigned designer = 0, n = 1;

   while (n < argc) {

      if (argv[n] == strstr(argv[n],"designer"))
         designer = 1;

      if (argv[n] == strstr(argv[n],"config="))
         strncpy(config_name,argv[n] + 7,256);

      n++;
   }

   configuration();

   aluf_init();
   context_id = renderer_generate_unique_id();
   renderer_set_context_group(context_id);
   ui_initialize();
   std_decor_initialize();

   display_set_context_renew_function(&scene_context);

   display_prepare(argc,argv);

   if (display_set_stereo_mode)
      display_set_stereo_mode(initial_view_mode);
   stage_set_master_audio_level(initial_audio_level);
   scene_api_set_brightness(illumination_level);

   /* Setup stack, scenes... */
   *stack        = (scene_api*)create_root_scene(designer,ca);
   stack_top     = stack;
   stack_max     = stack + MAX_STACK - 1;
   curtain_time  = CHANGE_TIME/2;
   changing      = 1;
   context_cycle = -1;
   scene_api_enter(*stack_top);

   display_set_preframe_function(&curtain_preframe);
   display_set_display_function(&curtain_display);

   display_run_loop();

   /* Teardown scenes... */
   aggregate_save_config(ca,cl);
   (*stack)->erase(*stack);
   erase_config_aggregate(ca);
   ui_end();
   aluf_destroy();
   //gluf_delete_all();
   renderer_erase_registry();
   ext_libs_unload();

   return 0;
}

static void configuration()
{
   config_record *ret;

   if (!(*config_name))
      strncpy(config_name,DEFAULT_CONFIG_NAME,256);


   ca = create_config_aggregate(config_name);
   config_doc(ca,
      "Planck Application Configuration File");
   config_doc(ca,
      "Unless indicated, all show/allow/use variables take '0' or '1'.");

   cl = create_config_loader(CONFIG_ID);
   config_doc(cl,
      "Settings in the 'stage' block apply to all scenes in the application.");

   aggregate_register_config(ca,cl);

   ret = config_useu(cl,initial_view_mode);
   config_doc(ret,"View Modes are:");
   config_doc(ret,"\t0: Normal Viewing");
   config_doc(ret,"\t1: Anaglyph Stereo");
   config_doc(ret,"\t2: Horizontally Interlaced Stereo");
   config_doc(ret,"\t3: Vertically Interlaced Stereo");
   config_doc(ret,"\t4: Checkerboard Stereo");
   config_doc(ret,"\t5: Stereo Pair");
   config_doc(ret,
      "\t6: Quad Buffered Stereo (if enabled by 'hardware'command line parameter)");

   ret = config_usef(cl,initial_audio_level);
   config_doc(ret,"Sets audio level for all scenes");
   //ret = config_usef(cl,illumination_level);
   //config_doc(ret,"Sets GL lighting properties for all scenes");

   aggregate_load_config(ca,cl);
}


void stage_push_scene(void *scene)
{
   scene_api *next = (scene_api*)scene;

   if (!next || stack_top == stack_max || changing || next == *stack_top)
      return;

   stack_push   = next;
   curtain_time = 0;
   changing     = -1;

   display_set_preframe_function(&curtain_preframe);
   display_set_display_function(&curtain_display);
}

void stage_pop_scene(void)
{
   if (stack_top == stack || changing)
      return;

   stack_push   = 0;
   curtain_time = 0;
   changing     = -1;

   display_set_preframe_function(&curtain_preframe);
   display_set_display_function(&curtain_display);
}

void stage_set_master_audio_level(float level)
{
   if (level < 0)
      level = 0;
   if (level > 1)
      level = 1;

   al_gain = level;
   if (!changing && stack_top && (*stack_top)->audio_level)
      (*stack_top)->audio_level(*stack_top,al_gain);
}

float stage_get_master_audio_level(void)
{
   return al_gain;
}

static void scene_context(void)
{
   if (!once) {
      renderer_init();
      //shader_program_api_init();
      //shader_program_compile(&basic_shader,1);
      //shader_program_free(&basic_shader);
      renderer_activate_context_group(context_id);
      once = 1;
   }

   (*stack)->init(*stack);
}

static void current_scene_preframe(void)
{
   (*stack_top)->update(*stack_top);
}

static void current_scene_display(void)
{
   switch (context_cycle) {
      case -1:
         renderer_suppress_context_group((*stack_top)->context);
         dead_time = 1;
         break;
      case 1:
         renderer_activate_context_group((*stack_top)->context);
         dead_time = 1;
         break;
      case 0:
      default:
         (*stack_top)->draw(*stack_top);
         break;
   }
}

static void curtain_preframe(void)
{
   float al_level = al_gain * fabs(2 * curtain_time - CHANGE_TIME)/CHANGE_TIME;

   if (!dead_time)
      curtain_time += frame_state.dt;

   dead_time = 0;

   switch (context_cycle) {
      case -1:
         context_cycle = 1;
         break;
      case 1:
         context_cycle = 0;
         break;
      case 0:
      default:
         break;
   }

   if ((*stack_top)->audio_level)
      (*stack_top)->audio_level(*stack_top,al_level);

   if (changing < 0) {

      if (curtain_time >= CHANGE_TIME/2) {

         if (!context_cycle)
            context_cycle = -1;
         else {

            scene_api_leave(*stack_top);

            if (stack_push) {
               stack_top++;
               *stack_top  = stack_push;
               stack_push = 0;
            }
            else stack_top--;

            scene_api_enter(*stack_top);
            changing  = 1;
         }
      }
   }

   if (changing > 0) {

      if (!context_cycle && curtain_time >= CHANGE_TIME) {
         printf("new scene...\n");
         changing = 0;
         display_set_preframe_function(&current_scene_preframe);
         display_set_display_function(&current_scene_display);
         if ((*stack_top)->audio_level)
            (*stack_top)->audio_level(*stack_top,al_gain);
      }
   }

   current_scene_preframe();
}

static void curtain_display(void)
{
   mat44 tm, mv = MAT44_IDENTITY, mvp = frame_state.projection_matrix;
   float fade_alpha = 1.0 - fabs(2 * curtain_time - CHANGE_TIME)/CHANGE_TIME;
   float color[4] = {0,0,0,fade_alpha};
   float *at = frame_state.overlay_area;
   static float curtain_v[12] = { 0,0, 1,0, 1,1,   0,0, 1,1, 0,1 };

   if (fade_alpha < .9 || context_cycle)
      current_scene_display();

   mat44_translation(&tm,at[0],at[1],at[4]);
   mat44_mult(&mv,&tm);
   mat44_scaling(&tm,at[2],at[3],1);
   mat44_mult(&mv,&tm);

   mat44_mult(&mvp,&mv);

   ui_2d_shader.um44mvp    = &mvp;
   ui_2d_shader.av2vertex  = curtain_v;
   ui_2d_shader.texture    = 0;
   ui_2d_shader.av2texture = 0;
   ui_2d_shader.uv4color   = color;
   ui_2d_shader.count      = 6;

   shader_program_execute(&ui_2d_shader);
}



#ifdef __cplusplus
}
#endif
