#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include "gluf.h"
#include "gltf.h"
#include "ui_2d_shader.h"
#include "ui_element.h"

#define CURSOR_Z_OFFSET  .00001
#define BILLBOARD_Z     -.00003

#define CORNER_VERTEX_COUNT 20


static void init_shader(void *ref);
static void free_shader(void *shader);


/* prototype ui_list */
typedef struct ui_list ui_list;
struct ui_list {
   ui_list *next;
   ui_list *back;
   ui_list *up;
   void (*erase)(void *self);
};

/* Non-generic cleanup functions. */
static void erase_container(void *which);
static void erase_text(void *which);
static void erase_ui_element(void *which);
static void erase_viewer(void *which);

/* Prototype ui_list functions. */
static void *ui_list_append(void *in, void *what);
static void *ui_list_isolate(void *in, void *what);
void ui_list_isolate_all(void *start);
static void *ui_list_remove(void *in, void *what);
static void  ui_list_remove_all(void *in);

/* Generic element functions */
static void ui_element_to_space(void *obj, float x, float y, float w, float h);
static void ui_element_vertex_display(void *element, mat44 *projection,
                                                              mat44 *modelview);
static void ui_element_update(void *element, ui_update *u);

/* Container functions. */
static void container_display(void *container, mat44 *projection,
                                                              mat44 *modelview);
static void container_update(void *container, ui_update *u);
static void container_add_item(void *self,void *item);
static void container_add_items(void *self,int how_many, ...);
static void container_remove_item(void *self,void *item);
static void container_detach_item(void *obj, void *item);
static void container_detach_all(void *obj);
static void container_to_list_head(void *obj, void *item);

/* Balloon container functions. */
static void balloon_update(void *balloon, ui_update *u);
static void balloon_grow(ui_balloon *self);
static void balloon_shrink(ui_balloon *self);

/* Pager container functions. */
static void pager_display(void *pager, mat44 *projection, mat44 *modelview);
static void pager_update(void *pager, ui_update *u);
static void pager_next(ui_pager *self);
static void pager_previous(ui_pager *self);

/* Icon functions */
static void icon_generate_polygon(ui_icon *i);
static void ui_icon_to_space(void *e, float x, float y, float w, float h);

/* Text functions */
static void text_update(void* printer, ui_update* u);
static void text_value_update(void* printer,ui_update* u);

/* Dial/Selector functions */
static void dial_update(void* dial, ui_update* u);
static void selector_update(void* selector, ui_update* u);

/* Slider functions */
static void slider_update(void* slider, ui_update* u);

/* Progress bar functions */
static void progress_bar_update(void* progress_bar, ui_update* u);

/* Button/Touchpad functions */
static void button_update(void* button, ui_update* u);
static void touch_pad_update(void* pad, ui_update* u);

/* Marquee functions */
static void marquee_update(void* marquee, ui_update* u);

/* Viewport functions */
static void viewer_display(void *viewer, mat44 *projection);
static void viewer_update(void *viewer,float dx,float dy,float dt,
                                                      unsigned button,int key);

static void ui_key_update(void *key, ui_update *u);

/* Controls one-time creation actions... */
static int initialized = 0;

/* Default texture fonts... */
gl_tex_font *ui_mono_fnt;
gl_tex_font *ui_pretty_fnt;
gl_tex_font *ui_pretty_bld_fnt;

/* Pre-generated shapes in 2d vertex arrays */
static vec2 box[6] = {
   0,0, 1,0, 1,1,
   0,0, 1,1, 0,1
};

static vec2 cursor[9] = {
    0,    0,    .008,-.025,  .025,-.008,
   .020,-.025,  .025,-.020,  .019,-.014,
   .020,-.025,  .019,-.014,  .014,-.019
};

static vec2 *busy_cursor;
static unsigned busy_vtx_cnt;

static float cursor_color[4] = { .6,.6,.6,.75 };

static ui_2d_shader_t cursor_draw;
static ui_2d_shader_t busy_draw;


/************************** object prototypes *********************************/

#define UI_MT_0   {0,0,0,0,0}
#define UI_MT_WI9 {.9,.9,.9,1,0}
#define UI_MT_WI8 {.8,.8,.8,1,0}
#define UI_MT_WI7 {.7,.7,.7,1,0}
#define UI_MT_WI6 {.6,.6,.6,1,0}
#define UI_MT_WI5 {.5,.5,.5,1,0}

#define UI_ELEMENT_0(ers,dsp,upd,clr,uclr,hclr) {\
   0,0,0,&ers,       /* list pointers: next, back, parent; delete function...*/\
   &ui_element_to_space,\
   &dsp,&upd,0,0,    /* place,dispaly, udpate, action and hover functions...*/\
   0,0,              /* owner object reference, object id...*/\
   {0,0,1,1},        /* ui_space: x,y,w,h...*/\
   0,1,              /* rotation, aspect ratio...*/\
   NoAct,32,0,0,     /* input state, button mask, busy, sticky state...*/\
   clr,hclr,uclr,    /* normal, hover, and on-use material properties...*/\
   0,0,0,0,0,1       /* vertex and texture array properties... */\
}

#define UI_CONTAINER_0(ers,dsp,upd) {\
   0,0,0,&ers,\
   &ui_element_to_space,\
   &dsp,&upd,0,0,\
   0,0,\
   {0,0,1,1},\
   0,1,\
   NoAct,32,0,0,     /* same as default ui_elemet up to this point...*/\
   UI_MT_0,UI_MT_0,UI_MT_0,\
   0,0,0,0,0,0,\
   &container_add_item,\
   &container_add_items,\
   &container_to_list_head,\
   &container_remove_item,\
   &container_detach_item,\
   &container_detach_all,\
   0,\
   1\
}


/****************************** initializers **********************************/

void ui_initialize(void)
{
   unsigned v_sz;

   if (initialized)
      return;

   /*Create default fonts... */
   ui_mono_fnt   = create_texture_font("./data/fonts/vera_mono.fnt");
   ui_pretty_fnt = create_texture_font("./data/fonts/comfortaa_reg.fnt");
   ui_pretty_bld_fnt = create_texture_font("./data/fonts/comfortaa.fnt");

   renderer_register_generic(&ui_2d_shader,&init_shader,&free_shader);

   v_sz = 30;
   busy_cursor = (vec2*)malloc(v_sz * sizeof(vec2));
   if (busy_cursor) {
      gluf_arc_2d(busy_cursor,v_sz,0,330);
      gluf_scale_2d(busy_cursor,v_sz,.04,.04);
      gluf_tessellate_2d(&busy_cursor,&v_sz,0);
      busy_vtx_cnt = v_sz;
   }
}

void ui_end(void)
{
   if (ui_mono_fnt)
      ui_mono_fnt->erase(ui_mono_fnt);
   if (ui_pretty_fnt)
      ui_pretty_fnt->erase(ui_pretty_fnt);

   free(busy_cursor);

   initialized = 0;
}

static void init_shader(void *shader)
{
   ui_2d_shader_t *s = (ui_2d_shader_t*)shader;

   if (!shader_program_compile(s,1))
      exit(0);

   s->vbo         = 0;
   s->cull_face   = 0;
   s->depth_test  = 0;

   busy_draw = *s;

   busy_draw.av2vertex     = busy_cursor;
   busy_draw.texture       = 0;
   busy_draw.av2texture    = 0;
   busy_draw.uv4color      = cursor_color;
   busy_draw.count         = busy_vtx_cnt;

   cursor_draw = *s;
   cursor_draw.av2vertex   = cursor;
   cursor_draw.texture     = 0;
   cursor_draw.av2texture  = 0;
   cursor_draw.uv4color    = cursor_color;
   cursor_draw.count       = 9;
}

static void free_shader(void *shader)
{
   shader_program_free(shader);
}

/****************************** creators and populators ***********************/

ui_space in_space(float x, float y, float width, float height)
{
   ui_space obj;

   obj.x = x;
   obj.y = y;
   obj.w = width;
   obj.h = height;

   return obj;
}

ui_mat material_is(float r,float g,float b, float a, unsigned texture)
{
   ui_mat obj;

   obj.r = r;
   obj.g = g;
   obj.b = b;
   obj.a = a;
   obj.tex = texture;

   return obj;
}

void decor_all_textures(ui_decor *in,unsigned tex_id)
{
   if (!in)
      return;

   in->look.tex = in->hover.tex = in->use.tex = tex_id;
}

void decor_all_colors(ui_decor *in, float r,float g,float b, float a)
{
   if (!in)
      return;

   in->look.r = in->hover.r = in->use.r = r;
   in->look.g = in->hover.g = in->use.g = g;
   in->look.b = in->hover.b = in->use.b = b;
   in->look.a = in->hover.a = in->use.a = a;
}

void decor_all_materials(ui_decor *d,float r,float g,float b,float a,unsigned t)
{
   ui_mat cpy = {r,g,b,a,t};

   if (!d)
      return;

   d->look = d->hover = d->use = cpy;
}

ui_container *create_container(void)
{
   ui_container *obj = (ui_container*)malloc(sizeof(ui_container));

   if (obj) {

      ui_container proto = UI_CONTAINER_0(
         erase_container, container_display, container_update);

      *obj = proto;
   }
   return obj;
}

ui_balloon *create_balloon(void)
{
   ui_balloon *obj = (ui_balloon*)calloc(1,sizeof(ui_balloon));

   if (obj) {

      ui_container proto = UI_CONTAINER_0(
         erase_container, container_display, balloon_update);

      *((ui_container*)obj) = proto;
      obj->grab_focus = 1;
      obj->at     = in_space(0,0,0,0);
      obj->max    = in_space(0,0,1,1);
      obj->grow   = &balloon_grow;
      obj->shrink = &balloon_shrink;
      obj->speed  = .25;
   }
   return obj;
}

ui_pager *create_pager(void)
{
   ui_pager *obj = (ui_pager*)calloc(1,sizeof(ui_pager));

   if (obj) {

      ui_container proto = UI_CONTAINER_0(
         erase_container, pager_display, pager_update);

      *((ui_container*)obj) = proto;
      obj->next     = &pager_next;
      obj->previous = &pager_previous;
      obj->speed    = .2;
   }
   return obj;
}

ui_text *create_text(unsigned cpl, const char *text)
{
   ui_text *obj = (ui_text*)calloc(1,sizeof(ui_text));

   if (obj) {

      ui_element proto = UI_ELEMENT_0(
         erase_text, ui_element_vertex_display, text_update,
         UI_MT_WI9, UI_MT_0, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->grab_focus = 0;
      obj->font = ui_mono_fnt;
      obj->cpl  = cpl;

      if (text) {
         obj->text = (char*)calloc(strlen(text) + 1,sizeof(char));
         if (obj->text)
            memcpy(obj->text,text,strlen(text));
      }
   }
   return obj;
}

ui_text *create_text_ptr(unsigned cpl, char *text)
{
   ui_text *obj = (ui_text*)calloc(1,sizeof(ui_text));

   if (obj) {

      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, text_update,
         UI_MT_WI9, UI_MT_0, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->grab_focus = 0;
      obj->text = text;
      obj->font = ui_mono_fnt;
      obj->cpl  = cpl;
   }
   return obj;
}

ui_element *to_static_text(ui_text* in)
{
   ui_element *obj = (ui_element*)calloc(1,sizeof(ui_element));

   if (obj) {

      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, ui_element_update,
         UI_MT_WI9, UI_MT_0, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->grab_focus = 0;

      if (in->font->vtx)
         free(in->font->vtx);
      if (in->font->tex)
         free(in->font->tex);
      in->font->vtx = 0;
      in->font->tex = 0;
      in->font->vtx_sz = 0;
      in->font->cpl = in->cpl;
      in->font->justify_top = in->justify_top;
      in->font->centered = in->centered;
      in->font->pseudo_mono = in->force_monospace;
      in->font->use_w = in->force_monospace;

      in->font->render = 0;
      in->font->draw(in->font,in->text,0,0);

      obj->vtx = in->font->vtx;
      obj->tex = in->font->tex;
      obj->vtx_sz = in->font->vtx_sz;
      obj->use_vtx = in->font->use_vtx;
      in->font->vtx = 0;
      in->font->tex = 0;
      in->font->vtx_sz = 0;
      in->font->render = 1;

      decor_all_textures((ui_decor*)&obj->look,in->font->tex_id);
      in->erase(in);
   }
   return obj;
}

ui_element *create_label(float x, float y, float size, const char *text)
{
   unsigned cpl = (unsigned)((float)(size * 100.0));
   ui_text *tx;
   ui_element *e = 0;

   if (!text)
      return 0;

   tx = create_text(cpl,text);
   if (tx) {
      tx->centered = 1;
      tx->font = ui_pretty_bld_fnt;
      e = to_static_text(tx);
      e->place(e,x,y,size,size);
      decor_all_colors((ui_decor*)&e->look,.4,.4,.4,.75);
   }

   return e;
}

ui_element *create_item_label(void *item,const char *text)
{
   float x ,y, sz, diff;
   unsigned tst_cpl, str_sz;
   ui_element *e = (ui_element*)item;

   if (!e || !text)
      return 0;

   x  = e->at.x;
   y  = e->at.y + e->at.h + .005;
   sz = e->at.w;

   tst_cpl = (unsigned) (sz * 100.0);
   str_sz = strlen(text);

   if (tst_cpl < str_sz) {
      diff = (str_sz - tst_cpl)/100.0;
      sz += diff;
      x  -= diff/2;
   }

   return create_label(x,y,sz,text);
}

ui_element *create_item_lower_label(void *item,const char *text)
{
   float x ,y, sz, diff;
   unsigned tst_cpl, str_sz;
   ui_element *e = (ui_element*)item;

   if (!e || !text)
      return 0;

   x  = e->at.x;
   y  = e->at.y - .025;
   sz = e->at.w;

   tst_cpl = (unsigned) (sz * 100.0);
   str_sz = strlen(text);

   if (tst_cpl < str_sz) {
      diff = (str_sz - tst_cpl)/100.0;
      sz += diff;
      x  -= diff/2;
   }

   return create_label(x,y,sz,text);
}

ui_element *create_slider_labels(ui_slider *sl,unsigned number,...)
{
   va_list args;
   char *item = 0;
   double x, y, at_x, inc, sz, hdl_off;
   unsigned count = 0, cpl;
   ui_container *c;
   ui_element *e;
   ui_text *tx;

   if (!sl || !number)
      return 0;

   hdl_off = sl->handle_width * sl->at.w;
   inc     = (number - 1) ? (sl->at.w - hdl_off)/(number - 1) : 0;
   x       = (inc) ? sl->at.x + hdl_off/2 : sl->at.x + sl->at.w/2 + hdl_off/2;
   y       = sl->at.y + sl->at.h * .725;

   c = create_container();
   if (!c)
      return 0;

   va_start(args,number);
   while (count < number) {

      item = va_arg(args,char*);
      if (!item || !(*item)) {
         count++;
         continue;
      }
      cpl  = strlen(item);
      sz   = cpl/130.0;
      at_x = x + count * inc - sz/2;

      tx = create_text(cpl,item);
      if (tx) {
         tx->font = ui_pretty_bld_fnt;
         e = to_static_text(tx);
         e->place(e,at_x,y,sz,sz);
         decor_all_colors((ui_decor*)&e->look,.4,.4,.4,.75);
         c->add_item(c,e);
      }

      count++;
   }
   va_end(args);

   return (ui_element*)c;
}


ui_text *create_text_from_file(unsigned cpl, const char *filename)
{
   ui_text *obj = 0;
   FILE *file;
   struct stat info;
   int ret;
   char *buffer = 0;

   if (filename) {

      ret = stat(filename,&info);

      if (!ret && (file = fopen(filename,"rb"))) {

         buffer = (char*)calloc(info.st_size + 1,sizeof(char));

         if (buffer)
            ret = fread(buffer,sizeof(char),info.st_size,file);
         if (ret != info.st_size)
            printf("ui_text_from_file: fread size does not match...\n");
         fclose(file);
      }
      else printf("ui_text_from_file: could not open file...\n");
   }

   if (buffer) {

      obj = (ui_text*)calloc(1,sizeof(ui_text));

      if (obj) {
         ui_element proto = UI_ELEMENT_0(
            erase_text, ui_element_vertex_display, text_update,
            UI_MT_WI9, UI_MT_0, UI_MT_0);

         *((ui_element*)obj) = proto;
         obj->grab_focus = 0;
         obj->text = buffer;
         obj->font = ui_mono_fnt;
         obj->cpl  = cpl;
      }
   }
   return obj;
}

ui_text_value *create_text_value(unsigned cpl,unsigned length, ui_text_updater *target)
{
   ui_text_value *obj = (ui_text_value*)calloc(1,sizeof(ui_text_value));

   if (obj) {

      ui_element proto = UI_ELEMENT_0(
         erase_text, ui_element_vertex_display, text_value_update,
         UI_MT_WI9, UI_MT_0, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->grab_focus = 0;
      obj->text      = calloc(length, sizeof(char));
      obj->target    = target;
      obj->length    = length;
      obj->update_on = 1;
      obj->font      = ui_mono_fnt;
      obj->cpl       = cpl;
   }
   return obj;
}

static void icon_generate_polygon(ui_icon *i)
{
   float edge_offset;
   unsigned n = CORNER_VERTEX_COUNT;

   if (i) {

      if (i->vtx)
         free(i->vtx);
      if (i->tex)
         free(i->tex);
      i->tex      = 0;
      i->vtx_sz   = 0;
      i->use_vtx  = 0;

      i->vtx = malloc(4 * n * sizeof(vec2));
      edge_offset = (1 - i->edge)/2;

      if (i->vtx) {
         i->vtx_sz   = 4 * n;
         i->use_vtx  = 4 * n;

        /* Generate a square icon with rounded corners... */
         gluf_arc_2d(i->vtx,n,0,90);
         gluf_arc_2d(i->vtx + n,n,90,180);
         gluf_arc_2d(i->vtx + 2 * n,n,180,270);
         gluf_arc_2d(i->vtx + 3 * n,n,270,360);
         gluf_scale_2d(i->vtx,4 * n,i->edge,i->edge);
         gluf_translate_2d(i->vtx,n,edge_offset,edge_offset);
         gluf_translate_2d(i->vtx + n,n,-edge_offset,edge_offset);
         gluf_translate_2d(i->vtx + 2 * n,n,-edge_offset,-edge_offset);
         gluf_translate_2d(i->vtx + 3 * n,n,edge_offset,-edge_offset);
      }
   }
}

ui_icon *create_icon(void)
{
   return create_custom_icon(.2);
}

ui_icon *create_custom_icon(double edge_factor)
{
   ui_icon *obj = (ui_icon*)malloc(sizeof(ui_icon));

   if (obj) {

      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, ui_element_update,
         UI_MT_WI7, UI_MT_0, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->grab_focus = 0;
      obj->place = &ui_icon_to_space;
      obj->edge  = edge_factor;
      icon_generate_polygon(obj);

      if (!obj->vtx) {
         printf("Allocation failure for ui_icon vertex arrays...\n");
      }

      if (obj->vtx) {
         gluf_tessellate_2d(&obj->vtx,&obj->vtx_sz,1);
         obj->tex = (vec2*)malloc(obj->vtx_sz * sizeof(vec2));
         gluf_tex_coords_2d(obj->vtx,obj->tex,obj->vtx_sz);
         gluf_translate_2d(obj->vtx,obj->vtx_sz,.5,.5);
         obj->use_vtx = obj->vtx_sz;
      }
   }
   return obj;
}



ui_placard *create_placard(void)
{
   ui_placard *obj = (ui_placard*)malloc(sizeof(ui_placard));

   if (obj) {

      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, ui_element_update,
         UI_MT_WI7, UI_MT_0, UI_MT_0);

      *obj = proto;
      obj->grab_focus = 0;
      obj->vtx = malloc(6 * sizeof(vec2));
      obj->tex = malloc(6 * sizeof(vec2));

      if (obj->vtx && obj->tex) {
         obj->vtx_sz  = 6;
         obj->use_vtx = 6;
         memcpy(obj->vtx,box,6 * sizeof(vec2));
         memcpy(obj->tex,box,6 * sizeof(vec2));
      }
   }
   return obj;
}

ui_dial *create_dial(void)
{
   ui_dial *obj = (ui_dial*)calloc(1,sizeof(ui_dial));

   if (obj) {

      ui_element proto = UI_ELEMENT_0(
          erase_ui_element, ui_element_vertex_display, dial_update,
         UI_MT_WI7, UI_MT_WI5, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->span  = 1;
      obj->speed = 3;
      obj->vtx = malloc(6 * sizeof(vec2));
      obj->tex = malloc(6 * sizeof(vec2));

      if (obj->vtx && obj->tex) {
         obj->vtx_sz  = 6;
         obj->use_vtx = 6;
         memcpy(obj->vtx,box,6 * sizeof(vec2));
         memcpy(obj->tex,box,6 * sizeof(vec2));
      }
   }
   return obj;
}

ui_selector_dial *create_selector(void)
{
   ui_selector_dial *obj = (ui_selector_dial*)calloc(1,sizeof(ui_selector_dial));

   if (obj) {

      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, selector_update,
         UI_MT_WI7, UI_MT_0, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->positions = 2;
      obj->rotation = 0;
      obj->dr = 0;
      obj->speed = 1;
      obj->vtx = malloc(6 * sizeof(vec2));
      obj->tex = malloc(6 * sizeof(vec2));

      if (obj->vtx && obj->tex) {
         obj->vtx_sz  = 6;
         obj->use_vtx = 6;
         memcpy(obj->vtx,box,6 * sizeof(vec2));
         memcpy(obj->tex,box,6 * sizeof(vec2));
      }
   }
   return obj;
}

ui_slider *create_slider(void)
{
   ui_slider *obj = (ui_slider*)calloc(1,sizeof(ui_slider));

   vec2 chl_v[6] = {.05,.1, .95,.1, .95,.9, .05,.1, .95,.9, .05,.9};
   vec2 chl_t[6] = {0,0, .9,0, .9,1, 0,0, .9,1, 0,1};
   vec2 hdl_v[6] = {0,0, 0,0, 0,1, 0,0, 0,1, 0,1};
   vec2 hdl_t[6] = {.9,0, .99,0, .99,1, .9,0, .99,1, .9,1};


   if (obj) {
      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, slider_update,
         UI_MT_WI7, UI_MT_0, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->span = 1;
      obj->handle_width = .1;
      obj->vtx = malloc(12 * sizeof(vec2));
      obj->tex = malloc(12 * sizeof(vec2));

      if (obj->vtx && obj->tex) {
         obj->vtx_sz  = 12;
         obj->use_vtx = 12;
         memcpy(obj->vtx,chl_v,6 * sizeof(vec2));
         memcpy(obj->tex,chl_t,6 * sizeof(vec2));
         memcpy(obj->vtx + 6,hdl_v,6 * sizeof(vec2));
         memcpy(obj->tex + 6,hdl_t,6 * sizeof(vec2));
      }
   }
   return obj;
}

ui_progress_bar *create_progress_bar(void)
{
   ui_progress_bar *obj = (ui_progress_bar*)calloc(1,sizeof(ui_progress_bar));
   vec2 pg_v[4] = {0,0, 0,0, 0,1, 0,1};

   if (obj) {
      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, progress_bar_update,
         UI_MT_WI7, UI_MT_0, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->grab_focus = 0;
      obj->span = 1;
      obj->vtx = malloc(6 * sizeof(vec2));
      obj->tex = malloc(6 * sizeof(vec2));

      if (obj->vtx && obj->tex) {
         obj->vtx_sz  = 6;
         obj->use_vtx = 6;
         memcpy(obj->vtx,pg_v,6 * sizeof(vec2));
         memcpy(obj->tex,pg_v,6 * sizeof(vec2));
      }
   }
   return obj;
}

ui_button *create_button(void)
{
   ui_button *obj = (ui_button*)calloc(1,sizeof(ui_button));

    if (obj) {
      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, button_update,
         UI_MT_WI7, UI_MT_WI8, UI_MT_WI6);

      *((ui_element*)obj) = proto;
      obj->value_down = 1;
      obj->vtx = malloc(6 * sizeof(vec2));
      obj->tex = malloc(6 * sizeof(vec2));

      if (obj->vtx && obj->tex) {
         obj->vtx_sz  = 6;
         obj->use_vtx = 6;
         memcpy(obj->vtx,box,6 * sizeof(vec2));
         memcpy(obj->tex,box,6 * sizeof(vec2));
      }
   }
   return obj;
}

ui_touch_pad *create_touch_pad(void)
{
   ui_touch_pad *obj = (ui_touch_pad*)calloc(1,sizeof(ui_touch_pad));

   if (obj) {
      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, touch_pad_update,
         UI_MT_WI7, UI_MT_WI8, UI_MT_WI6);

      *((ui_element*)obj) = proto;
      obj->vtx = malloc(6 * sizeof(vec2));
      obj->tex = malloc(6 * sizeof(vec2));

      if (obj->vtx && obj->tex) {
         obj->vtx_sz  = 6;
         obj->use_vtx = 6;
         memcpy(obj->vtx,box,6 * sizeof(vec2));
         memcpy(obj->tex,box,6 * sizeof(vec2));
      }
   }
   return obj;
}

ui_marquee *create_marquee(void)
{
   ui_marquee *obj = (ui_marquee*)calloc(1,sizeof(ui_marquee));

  if (obj) {
      ui_element proto = UI_ELEMENT_0(
         erase_ui_element, ui_element_vertex_display, marquee_update,
         UI_MT_WI7, UI_MT_0, UI_MT_0);

      *((ui_element*)obj) = proto;
      obj->grab_focus = 0;
      obj->aperture     = .25;
      obj->scroll_speed = .02;
      obj->vtx = malloc(6 * sizeof(vec2));
      obj->tex = malloc(6 * sizeof(vec2));

      if (obj->vtx && obj->tex) {
         obj->vtx_sz  = 6;
         obj->use_vtx = 6;
         memcpy(obj->vtx,box,6 * sizeof(vec2));
         memcpy(obj->tex,box,6 * sizeof(vec2));
      }
   }
   return obj;
}

ui_viewer *create_viewer(float *overlay)
{
   ui_viewer proto = {
      &erase_viewer,
      &viewer_display,
      &viewer_update,
      &ui_basic_cursor,
      0,
      0,1,
      0.5,0.5,
      48
   };

   if (overlay) {
      ui_viewer *obj = (ui_viewer*)malloc(sizeof(ui_viewer));
      if (obj) {

         *obj         = proto;
         obj->overlay = overlay;
         obj->screen  = create_container();
         obj->screen->aspect_corrected = 0;
      }
      return obj;
   }
   return 0;
}

ui_key *create_key(void)
{
   ui_key *obj = (ui_key*)calloc(1,sizeof(ui_key));

   if (obj) {

      obj->erase  = &erase_ui_element;
      obj->place  = &ui_element_to_space;
      obj->update = &ui_key_update;
   }
   return obj;
}

/************************** non-generic erase functions************************/

void erase_container(void *which)
{
   ui_container *obj = (ui_container*)which;

   if (!obj)
      return;
   if (obj->contains)
      ui_list_remove_all(obj->contains);
   free(obj);
}

void erase_text(void *which)
{
   ui_text *obj = (ui_text*)which;

   if (!obj)
      return;
   if (obj->text)
      free(obj->text);
   if (obj->vtx)
      free(obj->vtx);
   if (obj->tex)
      free(obj->tex);
   free(obj);
}

void erase_ui_element(void *which)
{
   ui_element *obj = (ui_element*)which;

   if (!obj)
      return;
   if (obj->vtx)
      free(obj->vtx);
   if (obj->tex)
      free(obj->tex);
   free(obj);
}

void erase_viewer(void *which)
{
   ui_viewer *obj = (ui_viewer*)which;

   if (!obj)
      return;
   if (obj->screen)
      obj->screen->erase(obj->screen);
   free(obj);
}

/************************************* ui_list ********************************/

void *ui_list_append(void *start, void *which_one)
{
   ui_list *at = (ui_list*)start, *item = (ui_list*)which_one;

   if (!item || item == at)
      return at;
   if (!at) {
      item->next = 0;
      item->back = 0;
      item->up   = 0;
      return item;
   }

   while (at->next) {
      if (at == item)   /* This item is already here... */
         return start;
      at = at->next;
   }

   at->next   = item;
   item->back = at;
   item->next = 0;
   item->up   = 0;;

   return start;
}

void *ui_list_remove(void *start, void *which_one)
{
   ui_list *head = (ui_list*)start, *item = (ui_list*)which_one;

   if (!head) {
      if (item)
         item->erase(item);
      return 0;
   }
   if (!item)
      return head;
   if (item->back)
      (item->back)->next = item->next;
   if (item->next)
      (item->next)->back = item->back;
   if (item == head)
      head = item->next;

   item->erase(item);

   return head;
}

void *ui_list_isolate(void *start, void *which_one)
{
   ui_list *item = (ui_list*)which_one, *head = (ui_list*)start;

   if (!head) {
      item->next = 0;
      item->back = 0;
      item->up   = 0;
      return 0;
   }
   if (!item)
      return head;

   if (item->back)
      (item->back)->next = item->next;

   if (item->next)
      (item->next)->back = item->back;

   if (item == head)
      head = item->next;

   item->back = 0;
   item->next = 0;
   item->up   = 0;

   return head;
}

void ui_list_isolate_all(void *start)
{
   ui_list *target = (ui_list*)start, *save = 0;

    while (target) {
      save = target->next;
      target->back = 0;
      target->next = 0;
      target->up   = 0;
      target = save;
   }
}


void ui_list_remove_all(void *start)
{
   ui_list *target = (ui_list*)start, *save = 0;

   while (target) {
      save = target->next;
      target->erase(target);
      target = save;
   }
}

/**************************** ui_element shared *******************************/


static void ui_element_to_space(void *obj, float x, float y, float w, float h)
{
   ui_element *e = (ui_element*)obj;
   ui_space at   = in_space(x,y,w,h);

   if (!e)
      return;

   e->at = at;
}

static void ui_icon_to_space(void *obj, float x, float y, float w, float h)
{
   ui_icon *e  = (ui_icon*)obj;
   ui_space at = in_space(x,y,w,h);
   mat33 tm, t1, t2;

   unsigned n;

   if (!e)
      return;

   e->at = at;
   if (e->vtx_sz) {

      icon_generate_polygon(e);
      n = e->vtx_sz/4;

      mat33h_translation(&t1,-.5,-.5);
      t2 = t1;
      mat33h_translation(&tm,0,1 - w/h);
      mat33_mult(&t2,&tm);
      mat33h_scaling(&tm,1,w/h);
      mat33_mult(&t1,&tm);
      mat33_mult(&t2,&tm);
      mat33h_translation(&tm,.5,.5);
      mat33_mult(&t1,&tm);
      mat33_mult(&t2,&tm);

		gluf_transform(e->vtx + 2 * n,2 * n,&t1);
		gluf_transform(e->vtx,2 * n,&t2);

//       gluf_translate_2d(e->vtx,e->vtx_sz,.5,.5);
//       gluf_scale_2d(e->vtx,e->vtx_sz,1,w/h);
//       gluf_translate_2d(e->vtx,2 * n,0, 1 - w/h);
//       gluf_translate_2d(e->vtx,e->vtx_sz,-.5,-.5);

      gluf_tessellate_2d(&e->vtx,&e->vtx_sz,1);
      e->tex =(vec2*) malloc(e->vtx_sz * sizeof(vec2));
      gluf_tex_coords_2d(e->vtx,e->tex,e->vtx_sz);

      //gluf_translate_2d(e->vtx,e->vtx_sz,.5,.5);
      mat33h_translation(&tm,.5,.5);
      gluf_transform(e->vtx,e->vtx_sz,&tm);

      e->use_vtx = e->vtx_sz;
   }
}

static void ui_element_vertex_display(void *obj,mat44 *projection,mat44 *modelview)
{
   ui_element *e = (ui_element*)obj;
   ui_mat *mat, *test;
   mat44 mvp;

   if (!e || !e->vtx || !e->tex || !e->vtx_sz || !projection || !modelview)
      return;

   mat = &e->look;

   switch (e->state) {
      case Hover:
         test = &e->hover_look;
         if (test->a || test->r || test->g || test->b)
            mat = &e->hover_look;
         break;
      case Action:
         test = &e->use_look;
         if (test->a || test->r || test->g || test->b)
            mat = &e->use_look;
         break;
      default:
         break;
   }

   mvp = *projection;
   mat44_mult(&mvp,modelview);

   ui_2d_shader.um44mvp    = &mvp;
   ui_2d_shader.av2vertex  = e->vtx;
   ui_2d_shader.texture    = mat->tex;
   ui_2d_shader.av2texture = e->tex;
   ui_2d_shader.uv4color   = (float*)mat;
   ui_2d_shader.count      = e->use_vtx;

   shader_program_execute(&ui_2d_shader);
}

void ui_element_update(void *element, ui_update *u)
{
   ui_element *e = (ui_element*)element;
   if (!e)
      return;

   if (e->state == Hover && e->on_hover)
      e->on_hover(e,u);
   if (e->state == Action && e->on_use)
      e->on_use(e,u);
}

/******************************* ui_container *********************************/

void container_remove_item(void *obj, void *item)
{
   ui_container *c = (ui_container*)obj;

   if (!c || !item)
      return;

   c->contains = ui_list_remove(c->contains, item);
}

void container_detach_item(void *obj, void *item)
{
   ui_container *c = (ui_container*)obj;

   if (!c || !item)
      return;

   c->contains = ui_list_isolate(c->contains, item);
}

static void container_detach_all(void *obj)
{
   ui_container *c = (ui_container*)obj;

   if (!c || !c->contains)
      return;

   ui_list_isolate_all(c->contains);
   c->contains = 0;
}


void container_add_item(void *obj,void *what)
{
   ui_container *c = (ui_container*)obj;
   ui_element *item = (ui_element*)what;

   if (!c || !item)
      return;

   c->contains = ui_list_append(c->contains,item);
   item->container = (ui_element*)c;
}

void container_add_items(void *obj,int how_many, ...)
{
   ui_container *c = (ui_container*)obj;
   va_list args;
   int count = 0;
   ui_element *item = 0;

   if (!c || !how_many)
      return;

   va_start(args,how_many);
   while (count < how_many) {
      item = va_arg(args,ui_element*);
      c->contains = ui_list_append(c->contains,item);
      item->container = (ui_element*)c;
      count++;
   }
   va_end(args);
}

static void container_to_list_head(void *obj, void *item)
{
   ui_container *c = (ui_container*)obj;
   ui_element *e = (ui_element*)item;

   if (!c || !item)
      return;

   c->contains = ui_list_isolate(c->contains, item);
   e->more = c->contains;
   if (c->contains)
      (c->contains)->back = e;
   e->container = (ui_element*)c;
   c->contains = e;
}

static void balloon_grow(ui_balloon *obj)
{
   if (!obj || obj->change)
      return;

   obj->change = 1;
}

static void balloon_shrink(ui_balloon *obj)
{
   if (!obj || obj->change)
      return;

   obj->change = -1;
}

static void balloon_update(void *balloon, ui_update *u)
{
   ui_balloon *b = (ui_balloon*)balloon;

   if (!b)
      return;

   if (b->change) {
      b->presence += u->dt/b->speed * b->change;

      if (b->presence > 1) {
         if (b->on_grow)
            b->on_grow(b,u);
         b->presence = 1;
         b->change   = 0;
      }

      if (b->presence < 0) {
         if (b->on_shrink)
            b->on_shrink(b,u);
         b->presence = 0;
         b->change   = 0;
      }

      if (b->expand_left)
         b->at.x = (1 - b->presence) * (1 - b->max.x) + b->max.x;
      else b->at.x = b->presence * b->max.x;

      if (b->expand_down)
         b->at.y = (1 - b->presence) * (1 - b->max.y) + b->max.y;
      else b->at.y = b->presence * b->max.y;

      b->at.w = b->presence * b->max.w;
      b->at.h = b->presence * b->max.h;
   }
   else container_update(b,u);
}

static void pager_display(void *obj,mat44 *projection, mat44 *modelview)
{
   ui_pager *pg = (ui_pager*)obj;
   ui_element *s, *n;
   float abs_c = 0, value;
   mat44 tm, mv;

   if (!pg || !pg->show || pg->disabled || !projection || !modelview)
      return;

   s = pg->show;
   n = pg->to;

   if (pg->change > 0) {
      abs_c = pg->change;
      value = 1 - abs_c;

      if (value) {
         mv = *modelview;
         mat44_scaling(&tm,value,value,1);
         mat44_mult(&mv,&tm);

         if (n->on_display)
            n->on_display(n,projection,&mv);
      }

      if (abs_c) {
         mv = *modelview;
         mat44_translation(&tm,value,0,0);
         mat44_mult(&mv,&tm);
         mat44_scaling(&tm,abs_c,abs_c,1);
         mat44_mult(&mv,&tm);

         if (s->on_display)
            s->on_display(s,projection,&mv);
      }
   }
   else if (pg->change < 0) {

      abs_c = -pg->change;
      value = 1 - abs_c;

      if (value) {
         mv = *modelview;
         mat44_translation(&tm,abs_c,0,0);
         mat44_mult(&mv,&tm);
         mat44_scaling(&tm,value,value,1);
         mat44_mult(&mv,&tm);

         if (n->on_display)
            n->on_display(n,projection,&mv);
      }

      if (abs_c) {
         mv = *modelview;
         mat44_scaling(&tm,abs_c,abs_c,1);
         mat44_mult(&mv,&tm);

         if (s->on_display)
            s->on_display(s,projection,&mv);
      }
   }
   else {
      /* Default display; pg->change = 0; element fills the container area... */
      if (s->on_display)
          s->on_display(s,projection,modelview);
   }
}

static void pager_update(void *obj, ui_update *u)
{
   ui_pager *pg = (ui_pager*)obj;

   if (!pg || !pg->contains)
      return;

   if (!pg->show)
      pg->show = pg->contains;

   if (pg->change > 0) {

      pg->change -= u->dt/pg->speed;

      if (pg->change < 0) {
         pg->change = 0;
         pg->show   = pg->to;
         pg->to     = 0;
      }
      return;
   }

   if (pg->change < 0) {

      pg->change += u->dt/pg->speed;

      if (pg->change > 0) {
         pg->change = 0;
         pg->show   = pg->to;
         pg->to     = 0;
      }
      return;
   }

   if (pg->state == Hover && pg->on_hover)
      pg->on_hover(pg,u);

   if (pg->state == Action && pg->button_mask & u->button && pg->on_use)
      pg->on_use(pg,u);

   pg->show->aspect = pg->aspect;
   pg->show->state = pg->state;

   if (pg->show->update)
      pg->show->update(pg->show,u);
}

static void pager_next(ui_pager *pg)
{
   if (!pg || !pg->contains || pg->change)
      return;

   if (!pg->show->more)
      pg->to = pg->contains;
   else pg->to = pg->show->more;

   pg->to->aspect = pg->aspect;
   pg->change = 1;
}

static void pager_previous(ui_pager *pg)
{
   ui_element *last;

   if (!pg || !pg->contains || pg->change)
      return;

   /* Make sure we are not at the first displayable element... */
   if (pg->show == pg->contains) {
      last = pg->contains;
      while (last->more)
         last = last->more;
      pg->to = last;
   }
   else pg->to = pg->show->back;

   pg->to->aspect = pg->aspect;
   pg->change = -1;
}

void container_update(void *which, ui_update *in)
{
   ui_container *c = (ui_container*)which;
   ui_element *e;
   ui_update u;
   ui_space e_at;
   float theta, rot, r;
   unsigned loop_state;

   if (!c || c->disabled)
      return;

   /* Call this containers update functions if needed... */
   if (c->state == Hover && c->on_hover)
      c->on_hover(c,in);

   if (c->state == Action && c->button_mask & in->button && c->on_use)
      c->on_use(c,in);

   /* Now, we begin to process the linked list of child elements in this
   container in reverse order (so that foremost objects are processed
   first, and may steal focus)... */

   if (!c->contains)
      return;

   e = c->contains;
   loop_state = c->state;

   while (e->more)
      e = e->more;


   while (e) {

      if (e->disabled) {
         e = e->back;
         continue;
      }

      e->state = NoAct;

      if (loop_state != Action)
         e->sticky = 0;

      /* We need to insure that the aspect correction is done only ONCE, and
      as containers may be recursively called; we have to stop the chain
      of over adjustments... */

      if (c->aspect_corrected)
         e->aspect = 1;
      else e->aspect = c->aspect;

      /* Short circuit for when the parent is not active,
      or focus has been grabbed... */
      if (loop_state == NoAct) {
          if (e->update)
             e->update(e,in);
          e = e->back;
          continue;
      }

      e_at = e->at;
      u    = *in;

      if (c->aspect_corrected && c->aspect != 1) {
         e_at.w *= c->aspect;
         e_at.x *= c->aspect;
         e_at.x += (1 - c->aspect)/2;
      }

      u.px -= e_at.x;
      u.py -= e_at.y;

      if (e->orientation) {
         rot   = e->orientation * .0174533;
         r     = sqrt(u.px * u.px + u.py * u.py);
         theta = atan2(u.py,u.px) - rot;
         u.px  = r * cos(theta);
         u.py  = r * sin(theta);
      }

      u.px = u.px/e_at.w;
      u.py = u.py/e_at.h;

      if (e->sticky || u.px > 0 && u.px < 1 && u.py > 0 && u.py < 1) {
         if (loop_state == Action && e->button_mask & u.button)
            e->state = Action;
         else e->state = Hover;
      }

      if (e->update)
         e->update(e,&u);

      /* Short circuit: creates a foremost focus effect... */
      if (e->state != NoAct && e->grab_focus)
         loop_state = NoAct;

     e = e->back;
   }
}

void container_display(void *container,mat44 *projection, mat44 *modelview)
{
   /* Container displays nothing in itself, only its children. */
   ui_container *c = (ui_container*)container;
   ui_element *e;
   mat44 base, mv, tm;

   if (!c || !c->contains || c->disabled || !projection || !modelview)
      return;

   base = *modelview;

   if (c->aspect_corrected && c->aspect != 1) {
      mat44_translation(&tm,(1 - c->aspect)/2,0,0);
      mat44_mult(&base,&tm);
      mat44_scaling(&tm,c->aspect,1,1);
      mat44_mult(&base,&tm);
   }

   e = c->contains;

   while (e) {

      if (e->at.w <= 0 || e->at.h <= 0 || e->disabled) {
         e = e->more;
         continue;
      }

      mv = base;
      mat44_translation(&tm,e->at.x,e->at.y,0);
      mat44_mult(&mv,&tm);

      if (e->orientation) {
         mat44_rotation(&tm,RAD(e->orientation),0,0,1);
         mat44_mult(&mv,&tm);
      }

      mat44_scaling(&tm,e->at.w,e->at.h,1);
      mat44_mult(&mv,&tm);

      if (e->on_display)
         e->on_display(e,projection,&mv);

      e = e->more;
   }
}

/***************************** ui_printer *************************************/


void text_update(void *printer, ui_update* u)
{
   ui_text *ut = (ui_text*)printer;
   vec2 *vtx, *tex;
   unsigned vtx_sz;

   if (!ut || !ut->font || !ut->text || !ut->text[0])
      return;

   vtx = ut->font->vtx;
   ut->font->vtx = ut->vtx;
   tex = ut->font->tex;
   ut->font->tex = ut->tex;
   vtx_sz = ut->font->vtx_sz;
   ut->font->vtx_sz = ut->vtx_sz;

   ut->font->cpl = ut->cpl;
   ut->font->justify_top = ut->justify_top;
   ut->font->centered = ut->centered;
   ut->font->pseudo_mono = ut->force_monospace;
   ut->font->use_w = ut->force_monospace;
   ut->font->render = 0;
   ut->font->draw(ut->font,ut->text,0,0);

   ut->vtx_sz = ut->font->vtx_sz;
   ut->font->vtx_sz = vtx_sz;
   ut->vtx = ut->font->vtx;
   ut->font->vtx = vtx;
   ut->tex = ut->font->tex;
   ut->font->tex = tex;
   ut->font->render = 1;
   ut->use_vtx = ut->font->use_vtx;

   decor_all_textures((ui_decor*)&ut->look,ut->font->tex_id);
}

void text_value_update(void *obj, ui_update *u)
{
   ui_text_value *tv = (ui_text_value*)obj;

   if (!tv)
      return;

   tv->time += u->dt;

   if (tv->time >= tv->update_on) {
      if(tv->target)
         tv->target(tv);
      tv->time = 0;
      text_update(tv,u);
   }
}

/********************** ui_dial / ui_selector *********************************/

void dial_update(void *element, ui_update *u)
{
   ui_dial *d = (ui_dial*)element;

   if (!d)
      return;

   if (d->state == Action) {
      d->busy = 1;
      if (u->px > .6)
         d->value += u->dt * d->speed * (u->px - 0.6) / .4;
      if (d->value > d->min + d->span)
         d->value = d->min + d->span;
      if (u->px < .4)
         d->value -= u->dt * d->speed * (1.0 - u->px / .4);
      if (d->value < d->min)
         d->value = d->min;
   }

   if (d->state == Hover && d->on_hover)
      d->on_hover(d,u);
   if (d->state != Action && d->busy) {
      d->busy = 0;
      if (d->on_use)
         d->on_use(d,u);
   }
   /* This is not great: ideally the dial is turned only when action happens.
   However, the dial must always reflect the current dial value (such as
   when the application is initialized), so this is why an "identity" box
   is loaded and then rotated each frame. */
   memcpy(d->vtx,box,6 * sizeof(vec2));
   gluf_translate_2d(d->vtx,6,-.5,-.5);
   gluf_rotate_2d(d->vtx,6,(d->value - d->min) / d->span * -360);
   gluf_translate_2d(d->vtx,6,.5,.5);
}

void selector_update(void *element, ui_update *u)
{
   ui_selector_dial *d = (ui_selector_dial*)element;
   float inc;

   if (!d) return;

   if (d->rotation >= 360)
      d->rotation = d->rotation - 360;

   if (d->dr) {
      inc = 360/d->positions;
      d->rotation = d->value * inc + (1 - d->dr) * inc;
      d->dr -= u->dt * 10 * d->speed;

      if (d->dr <= 0) {
         d->dr = 0;
         d->value++;
         if (d->value == d->positions)
            d->value = 0;
         d->rotation = d->value * inc;
         if (d->on_use)
            d->on_use(d,u);
      }
   }

   if (d->state == Action && !d->busy) {
       d->dr   = 1;
       d->busy = 1;
   }

   if (d->busy && d->state != Action)
       d->busy = 0;

   if (d->state == Hover && d->on_hover)
      d->on_hover(d,u);

   memcpy(d->vtx,box,6 * sizeof(vec2));
   gluf_translate_2d(d->vtx,6,-.5,-.5);
   gluf_rotate_2d(d->vtx,6,-d->rotation);
   gluf_translate_2d(d->vtx,6,.5,.5);
}

/********************************* ui_slider **********************************/

void slider_update(void *slider, ui_update *u)
{
   ui_slider *s = (ui_slider*)slider;
   double value, d_inc, d_span, n;

   if (!s)
      return;

   if (s->state == Action) {

      s->sticky = 1;
      value = (u->px - s->handle_width/2)/(1 - s->handle_width);

      if (value >= 0 && value <= 1) {
         s->busy = 1;

         s->value = value * s->span + s->min;

         if (s->value - s->min > s->span)
            s->value = s->span + s->min;
         if (s->value < s->min)
            s->value = s->min;

         if (s->detents) {
            d_inc  = s->span/(s->detents + 1);
            d_span = s->span * s->detent_width/2;

            n = d_inc + s->min;
            while (n <= s->span - d_inc) {
               if (s->value <= n + d_span && s->value >= n - d_span)
                  s->value = n;
               n += d_inc;
            }
         }

         if (s->on_use)
            s->on_use(s,u);
      }
   }

   if (s->state == Hover && s->on_hover)
      s->on_hover(s,u);

   /* A last call to the update function with a definable state (busy = 0) that
   indicates the slider value has stablized... */
   if (s->state != Action && s->busy) {
      s->busy = 0;
      if (s->on_use)
         s->on_use(s,u);
   }

   value = (s->value - s->min)/s->span * (1 - s->handle_width);

   /* Slider channel */
   s->vtx[0].x = s->vtx[3].x = s->vtx[5].x = s->handle_width/2;
   s->vtx[1].x = s->vtx[2].x = s->vtx[4].x = 1 - s->handle_width/2;

   /* Slider handle */
   s->vtx[6].x = s->vtx[9].x = s->vtx[11].x = value;
   s->vtx[7].x = s->vtx[8].x = s->vtx[10].x = value + s->handle_width;
}


/******************************** ui_progress_bar *****************************/

static void progress_bar_update(void *progress_bar, ui_update *u)
{
   ui_progress_bar *p = (ui_progress_bar*)progress_bar;

   if (!p || !p->span)
      return;

   p->vtx[1].x = p->vtx[2].x = p->vtx[4].x = (p->value - p->min)/p->span;
}

/************************ ui_button / ui_touch_pad ****************************/



void button_update(void *element, ui_update *u)
{
   static float chain_sleep = 0;
   static ui_button *sleeper;
   ui_button *b = (ui_button*)element;
   ui_mat swap;

   if (!b)
      return;

   /* Variable "chain_sleep" prevents cycling when view mode selection changes
   viewport and active cursor lands on some other chained (a.k.a radio button)
   mode selector (which then reverts back to previous viewport
   configuration). */
   if (chain_sleep && b == sleeper) {
      chain_sleep -= u->dt;
      if (chain_sleep < 0)
         chain_sleep = 0;
   }

   if (!b->latches && !b->chain)
      b->value = 0;
   if (b->state != Action)
      b->busy = 0;

   if (b->state == Hover && b->on_hover)
      b->on_hover(b,u);

   if (b->state == Action && !b->busy && !chain_sleep) {
      b->busy = 1;

      if (b->latches)
         b->value = (b->value) ? 0 : b->value_down;
      else if (b->chain) {
         *b->chain = b->value_down;
         chain_sleep = .5;
         sleeper = b;
      }
      else b->value = b->value_down;

      if (b->on_use)
         b->on_use(b,u);
   }

   if (b->latches && b->value == b->value_down && !b->mtl_flag) {
      swap = b->look;
      b->look = b->use_look;
      b->hover_look.tex = b->use_look.tex;
      b->use_look = swap;
      b->mtl_flag = 1;
   }

   if (b->latches && b->value != b->value_down && b->mtl_flag) {
      swap = b->look;
      b->look = b->use_look;
      b->hover_look.tex = b->use_look.tex;
      b->use_look = swap;
      b->mtl_flag = 0;
   }

   if (b->chain && *b->chain == b->value_down && !b->mtl_flag) {
      swap = b->look;
      b->look = b->use_look;
      b->hover_look.tex = b->use_look.tex;
      b->use_look = swap;
      b->mtl_flag = 1;
   }

   if (b->chain && *b->chain != b->value_down && b->mtl_flag) {
      swap = b->look;
      b->look = b->use_look;
      b->hover_look.tex = b->use_look.tex;
      b->use_look = swap;
      b->mtl_flag = 0;
   }
}

void touch_pad_update(void *element, ui_update *u)
{
   ui_touch_pad *t = (ui_touch_pad*)element;
   float x, y, abs_x, abs_y;

   if (!t)
      return;

   if (t->state != NoAct) {
      if (t->state == Action)
         t->sticky = 1;
      else t->sticky = 0;

      x = abs_x = u->px - .5;
      if (abs_x < 0)
         abs_x = -abs_x;
      y = abs_y = u->py - .5;
      if (abs_y < 0)
         abs_y = -abs_y;

      if (abs_x > t->dead_zone/2)
         t->x = x;
      else t->x = 0;

      if (abs_y > t->dead_zone/2)
         t->y = y;
      else t->y = 0;
   }

   if (t->state == Hover && t->on_hover)
      t->on_hover(t,u);
   if (t->state == Action && t->on_use)
      t->on_use(t,u);
}

/******************************** ui_marquee **********************************/

void marquee_update(void *element, ui_update *u)
{
   ui_marquee *m = (ui_marquee*)element;
   float inv_value, slide_const;

   if (!m)
      return;

   m->position += m->scroll_speed * u->dt;
   if (m->position > 1)
      m->position = 0;

   if (m->state == Hover && m->on_hover)
      m->on_hover(m,u);
   if (m->state == Action && m->on_use)
      m->on_use(m,u);

   inv_value   = 1 - m->position;
   slide_const = 1 - m->aperture;
   m->tex[0].y = m->tex[1].y = m->tex[3].y = inv_value * slide_const;
   m->tex[2].y = m->tex[4].y = m->tex[5].y = m->aperture + inv_value * slide_const;
}

/******************************** ui_viewport *********************************/

void ui_basic_cursor(mat44 *projection,mat44 *modelview)
{
   mat44 mvp = *projection;
   mat44_mult(&mvp,modelview);
   cursor_draw.um44mvp = &mvp;
   shader_program_execute(&cursor_draw);
}

void ui_busy_cursor(mat44 *projection,mat44 *modelview)
{
   mat44 tm, mv = *modelview, mvp = *projection;
   static float rotation = 1;

   mat44_rotation(&tm,RAD(rotation),0,0,1);
   mat44_mult(&mv,&tm);
   mat44_mult(&mvp,&mv);
   busy_draw.um44mvp = &mvp;
   shader_program_execute(&busy_draw);
   rotation -= 1.75;
}

void viewer_display(void *viewer, mat44 *projection)
{
   ui_viewer *vw = (ui_viewer*)viewer;
   mat44 tm, mv = MAT44_IDENTITY;

   if (!vw || !projection)
      return;

   vw->aspect = (vw->overlay[2]) ? vw->overlay[3]/vw->overlay[2] : 1000000;

   mat44_translation(&tm,vw->overlay[0],vw->overlay[1],vw->overlay[4] + BILLBOARD_Z);
   mat44_mult(&mv,&tm);
   mat44_scaling(&tm,vw->overlay[2],vw->overlay[3],1);
   mat44_mult(&mv,&tm);

   if (vw->screen && vw->screen->on_display)
      vw->screen->on_display(vw->screen,projection,&mv);

   if (vw->pointer) {
      mat44_translation(&tm,vw->px,vw->py,CURSOR_Z_OFFSET);
      mat44_mult(&mv,&tm);
      mat44_scaling(&tm,vw->aspect,1,1);
      mat44_mult(&mv,&tm);
      vw->pointer(projection,&mv);
   }
}

void viewer_update(void *viewer,float dx,float dy,float dt,unsigned button,int key)
{
   ui_viewer *vw = (ui_viewer*)viewer;
   ui_update in;
   input_state action_state = NoAct;

   if (!vw)
      return;

   vw->px += .0005 * dx;
   vw->py += .0005 * dy;

   if (vw->px > 1)
      vw->px = 1;
   if (vw->px < 0)
      vw->px = 0;
   if (vw->py > 1)
      vw->py = 1;
   if (vw->py < 0)
      vw->py = 0;

   if (vw->pointer)
      action_state = Hover;
   if (button & vw->action_mask)
      action_state = Action;

   in.px     = vw->px;
   in.py     = vw->py;
   in.dt     = dt;
   in.button = button;
   in.key    = key;

   if (vw->screen) {
      vw->screen->aspect = vw->aspect;
      vw->screen->state = action_state;
      vw->screen->update(vw->screen,&in);
   }
}

/************************************ ui_key **********************************/

static void ui_key_update(void *key, ui_update *u)
{
   ui_key *k = (ui_key*)key;

   if (!k || !u)
      return;

   if (u->key && k->on_use) {
      //printf("%c ",u->key);
      k->on_use(k,u);
   }
}

#ifdef __cplusplus
}
#endif
