#ifdef __cplusplus
extern "C" {
#endif


/*  Implementaion of a "fast" multi-threaded OpenGL video capture system.
This system (found in the video_recorder.h and video_recorder.c files) is
capable of near realtime XVid encoding of fullscreen OpenGL framebuffer grabs.
An unsynchronized audio stream from the default system capture device is also
added to the output avi file.

Performance is limited by the encoder thread. A reasonably fast multi-core
processor should be able to do full screen recording with audio at 30 frames
per second.

Tested:
   Core2Duo 1.8 Ghz, Nvidia GeForce GTS 250, 4 GB RAM, Windows 7 32 bit.
      - Good perfomance on windowed modes (typical, 1000 x 1000 pixels)
      - Noticable frame drops on fullcreen (1920 x 1200 pixels); appears
        encoder can't keep up.
   Phenom II x3 2.1 Ghz, ATI Radeon 5650 mobile, 4 GB RAM, Windows 7 64 bit.
      - Excellent performance all around, including full screen mode
        (1600 x 900).

Portions of this code are based on the "pboUnpack" example designed
by Song Ho Ahn; found at:
    http://www.songho.ca/opengl/gl_pbo.html

Also, see the "Usage Examples" section of the ARB registry specification for
pixel buffer object; found at:
   http://www.opengl.org/registry/specs/ARB/pixel_buffer_object.txt

The OpenAL capture setup was based on information related at:
   http://www.blitzbasic.com/Community/posts.php?topic=72909
in conjunction with the OpenAL Programmer's Guide

Finally, the key component is the Revel library, found at:
   http://revel.sourceforge.net/

This library is rather old (2004), but worked well after a few modifications
to the header file and a complete recomplation (including xvidcore) for
the VC 9 runtime... */

#define USING_OPENGL
#include "platform.h"
#define USING_OPENAL_RUNTIME_LD
#define USING_FREEIMAGE_RUNTIME_LD
#define USING_REVEL_RUNTIME_LD
#include "ext_libs.h"
#include "video_recorder.h"

enum {Inactive, Recording, Paused}; /* video_recorder state */

#define COLOR_BYTES        4        /* BGRA format */
#define BUFFERED_FRAMES    60
#define AUDIO_UPDATE       1.0      /* seconds */
#define MAX_THREADS        100

typedef struct enc_buff enc_buff;
struct enc_buff {
   unsigned char *data;
   unsigned id;
};


typedef struct _video_recorder _video_recorder;
struct _video_recorder {
   IMPLEMENTS_VIDEO_RECORDER

   GLuint pbo[2];
   char *px_buff;
   enc_buff *stream[BUFFERED_FRAMES];

   unsigned buffer_sz;
   unsigned width;
   unsigned height;
   unsigned fps;
   unsigned state;

   double frame_dt;
   double video_accum;
   double audio_accum;

   unsigned rescale;
   double poll_time;
   unsigned sync_end;
   unsigned thread_done;

   int encoder_handle;
   Revel_VideoFrame frame;
   void *tid;
   unsigned encode_run;
   int serial_no;

   ALCdevice *capture;
};


#define glGenBuffersARB             _glGenBuffersARB
#define glBindBufferARB             _glBindBufferARB
#define glBufferDataARB             _glBufferDataARB
#define glBufferSubDataARB          _glBufferSubDataARB
#define glDeleteBuffersARB          _glDeleteBuffersARB
#define glGetBufferParameterivARB   _glGetBufferParameterivARB
#define glMapBufferARB              _glMapBufferARB
#define glUnmapBufferARB            _glUnmapBufferARB

PFNGLGENBUFFERSARBPROC              glGenBuffersARB;
PFNGLBINDBUFFERARBPROC              glBindBufferARB;
PFNGLBUFFERDATAARBPROC              glBufferDataARB;
PFNGLBUFFERSUBDATAARBPROC           glBufferSubDataARB;
PFNGLDELETEBUFFERSARBPROC           glDeleteBuffersARB;
PFNGLGETBUFFERPARAMETERIVARBPROC    glGetBufferParameterivARB;
PFNGLMAPBUFFERARBPROC               glMapBufferARB;
PFNGLUNMAPBUFFERARBPROC             glUnmapBufferARB;



static unsigned populate_functions(void);
static void invert_copy(char *in, char *out, unsigned width, unsigned sz);
static void rescale(char *in, char *out, unsigned in_width,  unsigned in_height,
                                       unsigned out_width, unsigned out_height);
static t_func_t encoder_run(void *obj);

static void erase(video_recorder *obj);
static void wait_finish(video_recorder *obj);
static void done(video_recorder *obj);
static void doneARB(video_recorder *obj);
static void record(video_recorder *obj);
static void pause_vid(video_recorder *obj);
static void frame(video_recorder *obj, double dt);
static void frameARB(video_recorder *obj, double dt);
static int  register_active_thread(void);
static void unregister_active_thread(int t);

static unsigned initialized = 0;
static int revel_lib_loaded  = -1;
static unsigned active_thread[MAX_THREADS];


video_recorder *create_video_recorder(vr_params *p)
{
   _video_recorder *vr;
   unsigned use_ARB, buffer_sz, rescale = 0;
   int encoder_handle;
   Revel_Error err;
   Revel_Params params;
   const char *dev_name;
   char *formatted_name, *find;

   if (!p)
      return 0;

   #ifdef USING_REVEL_RUNTIME_LD
      if (revel_lib_loaded < 0)
         revel_lib_loaded = ext_load_revel();
      if (!revel_lib_loaded)
         return 0;
   #endif

   if (p->out_width || p->out_height) {
      if (p->out_width != p->in_width && p->out_height != p->in_height)
         rescale = 1;
   }
   if (!p->out_width)
      p->out_width = p->in_width;
   if (!p->out_height)
      p->out_height = p->in_height;

   /* Apparent Revel/Xvid limitation: dimensions must be evenly divisible... */
   if (p->out_width%2)
      p->out_width = p->out_width - 1;
   if (p->out_height%2)
      p->out_height = p->out_height - 1;

   buffer_sz = p->in_width * p->in_height * COLOR_BYTES;

   if (!buffer_sz)
      return 0;

   if (p->quality > 1)
      p->quality = 1;
   if (p->quality < 0)
      p->quality = 0;

   if (!initialized) {
      initialized = 1;
      use_ARB     = populate_functions();
      /* use_ARB = 0; */
   }

   if (REVEL_API_VERSION != Revel_GetApiVersion()) {
      printf("Revel version mismatch!\n");
      return 0;
   }

   err = Revel_CreateEncoder(&encoder_handle);
   if (err != REVEL_ERR_NONE) {
      printf("Error while creating encoder: %d\n",err);
      return 0;
   }

   Revel_InitializeParams(&params);

   vr = (_video_recorder*)calloc(1,sizeof(_video_recorder));

   if (vr) {

      glPixelStorei(GL_UNPACK_ALIGNMENT,COLOR_BYTES);

      if (use_ARB) {
         printf("GL PBO extensions are supported on this hardware.\n");
         glGenBuffersARB(2,vr->pbo);
         glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB,vr->pbo[0]);
         glBufferDataARB(
            GL_PIXEL_UNPACK_BUFFER_ARB,
            buffer_sz,0,
            GL_STREAM_READ_ARB);
         glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB,vr->pbo[1]);
         glBufferDataARB(
            GL_PIXEL_UNPACK_BUFFER_ARB,
            buffer_sz,0,
            GL_STREAM_READ_ARB);
         glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB,0);

         vr->erase      = &erase;
         vr->done       = &doneARB;
         vr->next_frame = &frameARB;
      }
      else {
         printf("GL PBO extensions not supported, using glReadPixels.\n");
         vr->erase      = &erase;
         vr->done       = &done;
         vr->next_frame = &frame;
         vr->px_buff    = (char*)malloc(buffer_sz * sizeof(char));

         if (!vr->px_buff) {
            vr->erase((video_recorder*)vr);
            return 0;
         }
      }

      vr->out_file = (char*)calloc(strlen(p->out_file) + 1,sizeof(char));
      if (vr->out_file)
         strncpy(vr->out_file,p->out_file,strlen(p->out_file));

      vr->record         = &record;
      vr->pause          = &pause_vid;
      vr->wait_finish    = &wait_finish;
      vr->buffer_sz      = buffer_sz;
      vr->width          = p->in_width;
      vr->height         = p->in_height;
      vr->out_width      = p->out_width;
      vr->out_height     = p->out_height;
      vr->rescale        = rescale;
      vr->fps            = p->out_fps;
      vr->quality        = p->quality;
      vr->state          = Inactive;
      vr->record_time    = 0;
      vr->frames         = 0;
      vr->frame_dt       = 1.0/(vr->fps * 1.0275);
      vr->video_accum    = vr->frame_dt;
      vr->audio_accum    = 0;
      vr->encoder_handle = encoder_handle;
      vr->poll_time      = vr->frame_dt * 1000 * .9;

      /* Encoder video parameters... */
      params.width       = vr->out_width;
      params.height      = vr->out_height;
      params.frameRate   = vr->fps;
      params.quality     = vr->quality;
      params.codec       = REVEL_CD_XVID;

      /* Encoder audio parameters... */
      if (p->use_audio) {

         dev_name = alcGetString(0,ALC_CAPTURE_DEFAULT_DEVICE_SPECIFIER);
         if (dev_name) {
            vr->capture = alcCaptureOpenDevice(
               dev_name,
               22050,
               AL_FORMAT_MONO16,
               22050 * 2 * 2); /* 2 second of audio in ring buffer... */

            if (vr->capture) {
               formatted_name = (char*)calloc(strlen(dev_name) + 1,sizeof(char));
               strncpy(formatted_name,dev_name,strlen(dev_name));
               if (find = strstr(formatted_name," ("))
                  *find = 0;
               printf("Using audio recording device: %s\n\n",formatted_name);
               free(formatted_name);

               params.hasAudio          = 1;
               params.audioChannels     = 1;
               params.audioRate         = 22050;
               params.audioBits         = 16;
               params.audioSampleFormat = REVEL_ASF_PCM;
               vr->audio_active         = 1;
            }
         }
      }
      else {
         params.hasAudio          = 0;
         params.audioChannels     = 0;
         params.audioRate         = 0;
         params.audioBits         = 0;
         params.audioSampleFormat = 0;
      }

      err = Revel_EncodeStart(vr->encoder_handle,vr->out_file,&params);
      if (err != REVEL_ERR_NONE) {
         printf("Error while starting encoding: %d\n", err);
         vr->erase((video_recorder*)vr);
         return 0;
      }

      /* Revel frame struct setup... */
      vr->frame.width         = vr->out_width;
      vr->frame.height        = vr->out_height;
      vr->frame.bytesPerPixel = COLOR_BYTES;
      vr->frame.pixelFormat   = REVEL_PF_BGRA;
      vr->frame.pixels        = 0;

      vr->encode_run = 1;
      platform_new_thread(&encoder_run,vr,&vr->tid);
      vr->serial_no = register_active_thread();
   }
   return (video_recorder*)vr;
}

static unsigned populate_functions(void)
{
   unsigned ok = 1;

   #define GB_GL(t,v) ok &= (v =(t)platform_ARB_load(#v))? 1 : 0;
   GB_GL(PFNGLGENBUFFERSARBPROC,glGenBuffersARB);
   GB_GL(PFNGLBINDBUFFERARBPROC,glBindBufferARB);
   GB_GL(PFNGLBUFFERDATAARBPROC,glBufferDataARB);
   GB_GL(PFNGLBUFFERSUBDATAARBPROC,glBufferSubDataARB);
   GB_GL(PFNGLDELETEBUFFERSARBPROC,glDeleteBuffersARB);
   GB_GL(PFNGLGETBUFFERPARAMETERIVARBPROC,glGetBufferParameterivARB);
   GB_GL(PFNGLMAPBUFFERARBPROC,glMapBufferARB);
   GB_GL(PFNGLUNMAPBUFFERARBPROC,glUnmapBufferARB);

   return ok;
}

static int register_active_thread(void)
{
   unsigned n = 0;

   while (n < MAX_THREADS) {
      if (!active_thread[n]) {
         active_thread[n] = 1;
         break;
      }
      n++;
   }

   if (n == MAX_THREADS)
      return -1;
   return n;
}

static void unregister_active_thread(int t)
{
   if (t < 0) {
      printf("Ecoder: attempt to unregister undefined thread...\n");
      return;
   }

   active_thread[t] = 0;
}

void video_recorder_wait_process_end(void)
{
   unsigned done = 0, n;

    /* Be sure all of our registered encoder threads have terminated... */
   while (!done) {
      n = 0;
      while (n < MAX_THREADS && !active_thread[n])
         n++;

      if (n == MAX_THREADS)
         done = 1;
      else Sleep(10);
   }
}

static void invert_copy(char *in, char *out, unsigned width, unsigned sz)
{
   unsigned cnt = 0,ttl = sz/4,  col;
   unsigned *to = (unsigned*)out, *from = (unsigned*)in;

   while (cnt < ttl) {
      col = 0;
      while (col < width) {
         *(to + cnt + width - col - 1) = *(from + ttl - cnt - col - 1);
         col++;
      }
      cnt+= width;
   }
}

static void rescale(char *in, char *out, unsigned in_width, unsigned in_height,
                                        unsigned out_width, unsigned out_height)
{
   unsigned col_n, row_n = 0, *in_row, *out_row, *targ_out, *targ_in;
   float col_k = (float)in_width/out_width, row_k = (float)in_height/out_height;

   while (row_n < out_height) {

      in_row  = (unsigned*)in + in_width * (unsigned)(row_n * row_k);
      out_row = (unsigned*)out + out_width * row_n;
      col_n   = 0;

      while (col_n < out_width) {

         targ_out = out_row + col_n;
         targ_in  = in_row + (unsigned)(col_n * col_k);

         /* Simple pixel averaging... */
         if (*targ_out) {
            *((char*)targ_out)     += *((char*)targ_in);
            *((char*)targ_out)     /= 2;
            *((char*)targ_out + 1) += *((char*)targ_in + 1);
            *((char*)targ_out + 1) /= 2;
            *((char*)targ_out + 2) += *((char*)targ_in + 2);
            *((char*)targ_out + 2) /= 2;
            *((char*)targ_out + 3) += *((char*)targ_in + 3);
            *((char*)targ_out + 3) /= 2;
         }
         else *targ_out = *targ_in;

         col_n++;
      }
      row_n++;
   }
}

static t_func_t encoder_run(void *obj)
{
   _video_recorder *vr = (_video_recorder*)obj;
   unsigned at_frame = 0, buff_cnt, done = 0, audio_samples;
   unsigned char *audio_buffer;
   enc_buff *mine;
   Revel_Error err;
   int frame_size, total_size;

   if (!vr)
      platform_end_thread(0);

   while (vr->encode_run || !done) {

      buff_cnt = 0;
      while (buff_cnt < BUFFERED_FRAMES) {

         if (vr->stream[buff_cnt] && vr->stream[buff_cnt]->id == at_frame) {

            mine = vr->stream[buff_cnt];
            vr->stream[buff_cnt] = 0;

            if (mine->data) {
               vr->frame.pixels = mine->data;

               err = Revel_EncodeFrame(
                  vr->encoder_handle,
                  &vr->frame,
                  &frame_size);

               if (err != REVEL_ERR_NONE)
                  printf("Error while writing frame: %d\n", err);

               at_frame++;

               free(mine->data);
            }
            free(mine);

            if (!vr->encode_run) {
               buff_cnt = 0;
               continue;
            }
         }
         buff_cnt++;
      }

      if (!vr->encode_run && buff_cnt == BUFFERED_FRAMES)
         done = 1;

      if (vr->capture && (vr->audio_accum > AUDIO_UPDATE || done)) {
         alcGetIntegerv(vr->capture,ALC_CAPTURE_SAMPLES,4,(ALCint*)&audio_samples);
         if (audio_samples) {
            audio_buffer = (unsigned char*)malloc(audio_samples * sizeof(short));
            if (audio_buffer) {
               alcCaptureSamples(vr->capture,audio_buffer,audio_samples);
               err = Revel_EncodeAudio(
                  vr->encoder_handle,
                  audio_buffer,
                  audio_samples * sizeof(short),
                  &total_size);
               if (err != REVEL_ERR_NONE)
                  printf("Error while encoding audio: %d\n", err);
               vr->audio_accum = 0;
               free(audio_buffer);
            }
         }
      }

      Sleep(vr->poll_time);
   }

   /* Finally, finish encode the avi file and clean up. */
   err = Revel_EncodeEnd(vr->encoder_handle, &total_size);
   if (err != REVEL_ERR_NONE)
      printf("Error while ending encoding: %d\n", err);

   printf(
      "\nXvid Encoder Report\n"
      "----------------------------\n"
      "File out:         %s\n"
      "Total size:       %.2f MB\n"
      "Resolution:       %d x %d\n"
      "Recording time:   %.2f s\n"
      "Encoded frames:   %d\n"
      "Average fps:      %.2f\n\n",
      vr->out_file,
      total_size/1000000.0,
      vr->out_width,
      vr->out_height,
      vr->record_time,
      at_frame,
      at_frame/vr->record_time);

   vr->thread_done = 1;
   while (vr->sync_end)
      Sleep(30);

   vr->erase((video_recorder*)vr);

   printf("Encoder thread exiting...\n");

   platform_end_thread(0);
}

static void doneARB(video_recorder *obj)
{
   _video_recorder *vr = (_video_recorder*)obj;

   glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, vr->pbo[0]);
   glBufferDataARB(GL_PIXEL_PACK_BUFFER_ARB,vr->buffer_sz,0,GL_STREAM_READ_ARB);
   glUnmapBufferARB(GL_PIXEL_PACK_BUFFER_ARB);
   glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB, vr->pbo[1]);
   glBufferDataARB(GL_PIXEL_PACK_BUFFER_ARB,vr->buffer_sz,0,GL_STREAM_READ_ARB);
   glUnmapBufferARB(GL_PIXEL_PACK_BUFFER_ARB);
   glDeleteBuffersARB(2,vr->pbo);

   if (vr->capture)
      alcCaptureStop(vr->capture);

   vr->encode_run = 0;
}

static void done(video_recorder *obj)
{
   _video_recorder *vr = (_video_recorder*)obj;

   if (vr->capture)
      alcCaptureStop(vr->capture);

   vr->encode_run = 0;
}

static void wait_finish(video_recorder *obj)
{
   _video_recorder *vr = (_video_recorder*)obj;

   if (vr->capture)
      alcCaptureStop(vr->capture);

   vr->sync_end   = 1;
   vr->encode_run = 0;

   while (!vr->thread_done)
      Sleep(30);

   vr->sync_end = 0;
}

static void erase(video_recorder *obj)
{
   _video_recorder *vr = (_video_recorder*)obj;

   if (!vr)
      return;

   Revel_DestroyEncoder(vr->encoder_handle);

   if (vr->capture)
      alcCaptureCloseDevice(vr->capture);

   if (vr->out_file)
      free(vr->out_file);
   if (vr->px_buff)
      free(vr->px_buff);

   unregister_active_thread(vr->serial_no);

   free(vr);
}

static void record(video_recorder *obj)
{
   _video_recorder *vr = (_video_recorder*)obj;

   if (!vr)
      return;

   if (vr->capture)
      alcCaptureStart(vr->capture);

   vr->state = Recording;
}

static void pause_vid(video_recorder *obj)
{
   _video_recorder *vr = (_video_recorder*)obj;

   if (!vr || vr->state != Recording)
      return;

   if (vr->capture)
      alcCaptureStop(vr->capture);

   vr->state = Paused;
}

static void frameARB(video_recorder *obj,double dt)
{
   _video_recorder *vr = (_video_recorder*)obj;
   GLuint *get, *enc;
   GLubyte *data, *data_out;
   enc_buff *out;
   unsigned find_slot, sz;

   if (!vr || vr->state != Recording)
      return;

   vr->video_accum += dt;
   vr->audio_accum += dt;
   vr->record_time += dt;

   if (vr->video_accum < vr->frame_dt)
      return;

   vr->video_accum = 0;

   if (!(vr->frames%2)) {
      get = &vr->pbo[0];
      enc = &vr->pbo[1];
   }
   else {
      get = &vr->pbo[1];
      enc = &vr->pbo[0];
   }

   /* Setup for the n + 1 frame */
   glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB,*get);
   glBufferDataARB(GL_PIXEL_PACK_BUFFER_ARB,vr->buffer_sz,0,GL_STREAM_READ_ARB);
   glReadPixels(0,0,vr->width,vr->height,GL_BGRA,GL_UNSIGNED_BYTE,0);

   /*Process n frame */
   glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB,*enc);
   data = glMapBufferARB(GL_PIXEL_PACK_BUFFER_ARB,GL_READ_ONLY);

   if (data) {

      find_slot = 0;

      while (find_slot < BUFFERED_FRAMES) {

         if (!vr->stream[find_slot]) {
            sz  = vr->buffer_sz * sizeof(char);
            out = (enc_buff*)malloc(sizeof(enc_buff));

            if (out) {

               if (vr->rescale) {
                  sz = vr->out_width * vr->out_height * COLOR_BYTES;
                  data_out = (GLubyte*)calloc(sz,sizeof(GLubyte));
                  if (data_out) {

                     rescale(
                        data,data_out,
                        vr->width,vr->height,
                        vr->out_width,vr->out_height);

                     data = data_out;
                  }
                  else data = 0;
               }

               out->data = (char*)malloc(sz);
               if (out->data && data) {
                  invert_copy(data,out->data,vr->out_width,sz);
                  if (vr->rescale)
                     free(data);
               }

               out->id = vr->frames;
               vr->frames++;
            }
            vr->stream[find_slot] = out;
            break;
         }
         find_slot++;
      }
   }

   glUnmapBufferARB(GL_PIXEL_PACK_BUFFER_ARB);
   glBindBufferARB(GL_PIXEL_PACK_BUFFER_ARB,0);
}

static void frame(video_recorder *obj,double dt)
{
   _video_recorder *vr = (_video_recorder*)obj;
   enc_buff *out;
   unsigned find_slot, sz;
   char *data, *data_out;

   if (!vr || vr->state != Recording)
      return;

   vr->video_accum += dt;
   vr->audio_accum += dt;
   vr->record_time += dt;

   if (vr->video_accum < vr->frame_dt)
      return;

   vr->video_accum = 0;

   glReadPixels(0,0,vr->width,vr->height,GL_BGRA,GL_UNSIGNED_BYTE,vr->px_buff);

   data = vr->px_buff;
   find_slot = 0;

   while (find_slot < BUFFERED_FRAMES) {

      if (!vr->stream[find_slot]) {
         sz = vr->buffer_sz * sizeof(char);
         out = (enc_buff*)malloc(sizeof(enc_buff));

         if (out) {

            if (vr->rescale) {
               sz = vr->out_width * vr->out_height * COLOR_BYTES;
               data_out = (char*)calloc(sz,sizeof(char));
               if (data_out) {

                  rescale(
                     data,data_out,
                     vr->width,vr->height,
                     vr->out_width,vr->out_height);

                  data = data_out;
               }
               else data = 0;
            }

            out->data = (char*)malloc(sz);
            if (out->data && data) {
               invert_copy(data,out->data,vr->out_width,sz);
               if (vr->rescale)
                  free(data);
            }
            out->id = vr->frames;
            vr->frames++;
         }

         vr->stream[find_slot] = out;
         break;
      }
      find_slot++;
   }
}


#ifdef __cplusplus
}
#endif
