/* Implementation of VRJuggler front end for Planck II renderer API... */

#include <gmtl/Matrix.h>
#include <gmtl/Generate.h>
#include <gmtl/Vec.h>
#include <vrj/Kernel/Kernel.h>
#include <vrj/vrjConfig.h>

#if defined(__APPLE__)
#  include <OpenGL/gl.h>
#else
#  include <GL/gl.h>
#endif

#include <vrj/Draw/OGL/GlApp.h>
#include <gadget/Type/PositionInterface.h>
#include <gadget/Type/DigitalInterface.h>

#include <math.h>
#include <errno.h>
#include <iostream>
#include <iomanip>

#ifdef __cplusplus
extern "C" {
#endif

#include "renderer.h"
#include "platform.h"

#ifdef __cplusplus
}
#endif

/* Trig definitions */
#ifndef HALF_PI
   #define HALF_PI            1.5707963f
#endif
#ifndef PI
   #define PI                 3.1415927f
#endif
#ifndef THREE_HALVES_PI
   #define THREE_HALVES_PI    4.7123890f
#endif
#ifndef TWO_PI
   #define TWO_PI             6.2831853f
#endif

//#define INTENSE_ERROR_CHECKING
#define FPS_UPDATE            1.0  /* seconds */

using namespace gmtl;
using namespace vrj;

/************************* class vrj_rndr prototype ***************************/

class vrj_rndr : public GlApp {
   public:
      vrj_rndr() {};
      virtual ~vrj_rndr(void) {};
      virtual void init();
      virtual void apiInit() {};
      virtual void bufferPreDraw();
      virtual void preFrame();
      virtual void intraFrame() {};
      virtual void postFrame();
      virtual void contextInit();
      virtual void draw();
   private:
      gadget::PositionInterface        mWand;
      gadget::PositionInterface        mHead;
      gadget::DigitalInterface         mButton0;
      gadget::DigitalInterface         mButton1;
      gadget::DigitalInterface         mButton2;
      gadget::DigitalInterface         mButton3;
      gadget::DigitalInterface         mButton4;
      gadget::DigitalInterface         mButton5;
      double                           lastXRot;
      double                           lastYRot;
      float                            waitTimer;
      int                              waitOnPress;
      int                              errnoCheck;
      int                              errorCount;
};
/********************* end class vrj_rndr prototype ***************************/



/*********************** begin renderer.h definitions *************************/

renderer_properties frame_state;

/* Unsupported feature functions set to null... */
void (*renderer_set_stereo_mode)(unsigned int mode_number) = 0;
void (*renderer_set_clip_plane)(double gl_z_value)         = 0;
void (*renderer_set_eye_seperation)(double value)          = 0;
void (*renderer_set_focal_plane)(double gl_z_value)        = 0;

/* Static function pointers... */
static void_callback *context_fn  = 0;
static void_callback *preframe_fn = 0;
static void_callback *display_fn  = 0;

static vrj::Kernel   *kernel;
static App           *application;
static unsigned int  frame_count   = 0;
static double        fps_dt_accum  = 0;

static char mode_name[1][30] = {"Normal"};

void renderer_set_context_renew_function(void_callback *ptr)
{
   context_fn = ptr;
}
void renderer_set_preframe_function(void_callback *ptr)
{
   preframe_fn = ptr;
}
void renderer_set_display_function(void_callback *ptr)
{
   display_fn = ptr;
}

void renderer_prepare(int argc, char **argv)
{
   unsigned found_config = 0;

   putenv("VPR_DEBUG_ENABLE=0");
   kernel      = Kernel::instance();
   application = new vrj_rndr();

   /* Load any config files specified on the command line... */
   for (int i = 1; i < argc; ++i) {
      if (strstr(argv[i],".jconf")) {
         kernel->loadConfigFile(argv[i]);
         found_config = 1;
      }
   }

   if (!found_config)
      kernel->loadConfigFile("./data/standalone.jconf");

   kernel->start();

   frame_state.stereo_modes = 0;
   frame_state.current_mode = 0;
   frame_state.stereo_mode_names = mode_name;
   /* This is for the video recorder...don't know what to do here yet. */
   frame_state.window_pixels[0] = 2800;
   frame_state.window_pixels[1] = 2100;

   /* From old Planck application calibration parameters:
      [-4.604000000000, -7.208000000000, 4.608000000000, -0.492000000000] */
   frame_state.overlay_area[0] = -4.604;
   frame_state.overlay_area[1] = -7.208;
   frame_state.overlay_area[2] = 9.212;
   frame_state.overlay_area[3] = 6.716;
   frame_state.overlay_area[4] = 0;

}

void renderer_run_loop(void)
{
   kernel->setApplication(application);
   kernel->waitForKernelStop();

   delete application;
}

void renderer_shutdown(void)
{
   kernel->stop();
}
/************************ end renderer.h definitions **************************/



/************************ begin vrj_rndr definitions **************************/

void vrj_rndr::init()
{
   if (errno > 0) {
      perror("\t\t-----> Startup errno cleanup; setting to zero, was");
      errno = 0;
   }

   printf("---------- App:init() ---------------\n");

   /* Device initializers... */
   mWand.init("VJWand");
   mHead.init("VJHead");
   mButton0.init("VJButton0");
   mButton1.init("VJButton1");
   mButton2.init("VJButton2");
   mButton3.init("VJButton3");
   mButton4.init("VJButton4");
   mButton5.init("VJButton5");

   errnoCheck = 0;
   errorCount = 0;

   printf("---------- Finished App:init() ---------------\n");
}

void vrj_rndr::bufferPreDraw()
{
   glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
   glClear(GL_COLOR_BUFFER_BIT);
   glClear(GL_DEPTH_BUFFER_BIT);
}

void vrj_rndr::preFrame()
{
   Matrix44f wand_mtx = mWand->getData();
   double wandXRot = (gmtl::makeRot<gmtl::EulerAngleXYZf>(wand_mtx))[0];
   double wandYRot = (gmtl::makeRot<gmtl::EulerAngleXYZf>(wand_mtx))[1];
   double transXLocal = 0, transZLocal = 0, moveMult = .1;
   static double dt, *rot, *trans;
   unsigned int input = 0;

   /* Time, metrics, and synchronization... */
   dt = frame_state.dt = platform_delta_timer();

   fps_dt_accum += frame_state.dt;
   frame_count++;
   if (fps_dt_accum >= FPS_UPDATE) {
      frame_state.fps = frame_count/fps_dt_accum;
      frame_count     = 0;
      fps_dt_accum    = 0;
   }

   if (mButton0->getData())
      ctl_button_set(0,input);
   if (mButton1->getData())
      ctl_button_set(1,input);
   if (mButton2->getData())
      ctl_button_set(2,input);
   if (mButton3->getData())
      ctl_button_set(3,input);
   if (mButton4->getData())
      ctl_button_set(4,input);
   if (mButton5->getData())
      ctl_button_set(5,input);

   frame_state.pointer[0] = (int)((lastYRot - wandYRot) * 1200.0);
   frame_state.pointer[1] = (int)(-(lastXRot - wandXRot) * 1200.0);
   frame_state.button     = input;
   rot                    = frame_state.rotation;
   trans                  = frame_state.position;

   if (rot && trans) {

      if (ctl_button(2,input))
         rot[1] -= sin(wandYRot) * .017;

      if (ctl_button(0,input)) {

         switch (waitOnPress) {
            case -1:
               transXLocal = cos(wandYRot) * cos(wandXRot);
               trans[1]   -= sin(wandXRot);
               transZLocal = -sin(wandYRot);
               break;
            case  0:
               waitOnPress = 1;
            default:
               moveMult    = waitOnPress / 1000.0;
               transXLocal = cos(wandYRot) * cos(wandXRot) * moveMult;
               trans[1]   -= sin(wandXRot) * moveMult;
               transZLocal = -sin(wandYRot) * moveMult;

               waitOnPress++;
               if (waitOnPress == 1000)
                  waitOnPress = -1;
               break;
         }

         trans[0] -= transXLocal * sin(rot[1]) + transZLocal * cos(rot[1]);
         trans[2] += transXLocal * cos(rot[1]) - transZLocal * sin(rot[1]);
      }

      if (!input)
         waitOnPress = 0;
   }

   lastXRot = wandXRot;
   lastYRot = wandYRot;

   if (preframe_fn)
      preframe_fn();
}

void vrj_rndr::postFrame()
{
   #ifdef INTENSE_ERROR_CHECKING
      int errorCheck = glGetError();

      errorCount = (errorCheck == errnoCheck) ? errorCount + 1: 0;
      errnoCheck = errorCheck;

      if (errorCount == 10)
         printf("\t\t----->Repeating error...masking.\n");
      if (errorCount >= 10)
         errorCheck = 35345;

      switch (errorCheck) {
         case 35345:
            break;
         case GL_NO_ERROR:
            printf("\t\t!!!!! GL_ERROR: state change-> no error !!!!!\n");
            errorCount = 11;
            break;
         case GL_INVALID_ENUM:
            printf("\t\t!!!!! GL_ERROR: invalid enumerator !!!!!\n");
            break;
         case GL_INVALID_VALUE:
            printf("\t\t!!!!! GL_ERROR: invalid value !!!!!\n");
            break;
         case GL_INVALID_OPERATION:
            printf("\t\t!!!!! GL_ERROR: invalid operation !!!!!\n");
            break;
         case GL_STACK_OVERFLOW:
            printf("\t\t!!!!! GL_ERROR: stack overflow !!!!!\n");
            break;
         case GL_STACK_UNDERFLOW:
            printf("\t\t!!!!! GL_ERROR: stack underflow !!!!!\n");
            break;
         case GL_OUT_OF_MEMORY:
            printf("\t\t!!!!! GL_ERROR: out of memory !!!!!\n");
            break;
         default:
            printf("\t\t!!!!! GL_ERROR: unknown !!!!!\n");
      }

      if (errno > 0) {
         perror("\t\t!!!!! C runtime error detected !!!!!");
         errno = 0;
      }
   #endif

   return;
}

void vrj_rndr::contextInit()
{
   printf("\t\t----->beginning context init.\n");
   if (context_fn)
      context_fn();
   printf("\t\t----->done in context init.\n");
}

void vrj_rndr::draw()
{
   if (display_fn)
      display_fn();
}
/************************* end vrj_rndr definitions ***************************/
