/* Jerry Dekker
   CIVS

   Major rewrite of entire system, February 2013.
   Transition from fixed pipeline to OpenGLES 2.0 compliant model.


Model viewer with mouse/keyboard roaming.
The keyboard arrow keys control eye position.
Mouse position controls eye XZ and XY rotation angles. */

#ifdef __cplusplus
extern "C" {
#endif


#include "platform.h"
#include "renderer_api.h"
#include <errno.h>
#include "display.h"

//#define DEBUG_ALL
#

#define FPS_UPDATE            1.0  /* seconds */

/* Trig definitions */
#define HALF_PI               1.570796326795
#define PI                    3.141592653590
#define THREE_HALVES_PI       4.712388980385
#define TWO_PI                6.283185307180

#ifndef DEG
   #define DEG(x)            ((x) * 57.29577951300)
#endif
#ifndef RAD
   #define RAD(x)            ((x) * 0.017453292519)
#endif

/* Screen modes */
enum { WINDOWED_MODE, FULLSCREEN_MODE };

/* View modes */
enum {
   NORMAL_VIEWING,
   ANAGLYPH,
   HORIZ_INTERLACED,
   VERT_INTERLACED,
   CHECKERBOARD,
   STEREO_PAIR,
   HARDWARE_STEREO,
   HARDWARE_QUAD,
   STEREO_QUAD
};

static char mode_names[9][30] = {
   "None",
   "Anaglyph",
   "      Row\nInterlaced",
   "  Column\nInterlaced",
   "Checker",
   "Stereo Pair",
   "Hardware\n   Stereo",
   "Hardware Quad",
   "Stereo Quad"
};

/* Viewport designators */
enum {
   VIEW_FRONT_CENTER,
   VIEW_FRONT_RIGHT,
   VIEW_FRONT_LEFT,
   VIEW_BACK_RIGHT,
   VIEW_BACK_LEFT
};

/* Rendering system functions. */
static void reshape(int ,int);
static void set_stencil(int ,int);
static void keyboard (int, int);
static void toggle_back_buffers(void);
static void toggle_front_buffers(void);
static void pre_render(void);
static void navigation(void);
static void update_view(void);
static void render(void);

/* Window environment */
static int app_run                  = 0;
static int window_width             = 0;
static int window_height            = 0;
static int framed_width             = 800;
static int framed_height            = 600;
static int full_width               = 1920;
static int full_height              = 1080;
static int origin_x                 = 0;
static int origin_y                 = 0;
static int window_mode              = FULLSCREEN_MODE;
static unsigned fullscreen_discover = 1;

/* Buffer targets */
static unsigned int left_front      = RENDERER_LEFT;
static unsigned int right_front     = RENDERER_RIGHT;
static unsigned int left_bottom     = RENDERER_RIGHT;
static unsigned int right_bottom    = RENDERER_LEFT;

/* Viewport info */
static int view_width               = 0;
static int view_height              = 0;
static int view_mode                = NORMAL_VIEWING;

/* View frustum info */
static float near_clip              = 0.05;
static float far_clip               = 20000;
static float fov                    = RAD(50);
static float app_eye_seperation     = -0.001800;
static float app_focal_length       = 30.0;

static mat44 left_frustum;
static mat44 center_frustum;
static mat44 right_frustum;
static mat44 right_floor_frustum;
static mat44 left_floor_frustum;
static mat44 identity_matrix = {
   1,0,0,0,
   0,1,0,0,
   0,0,1,0,
   0,0,0,1
};

static int max_viewport;
static int min_viewport;
static unsigned int app_mode_limit  = STEREO_PAIR;

/* Control configuration */
static int set_mouse                = 1;
static int last_x                   = 0;
static int last_y                   = 0;
static float move_scale             = 0.0001;
static float look_scale             = 0.001;
static float move_accel             = 0.1;

/***** New additions according to display.h... *****/
display_properties frame_state;

/* Application function pointers. */
static void (*app_preframe)(void)   = 0;
static void (*app_context)(void)    = 0;
static void (*app_draw)(void)       = 0;

static unsigned int frame_count     = 0;
static float        fps_dt_accum    = 0;
static float        dt              = 0;

/***** New additions for wgl... *****/
static HWND       main_wnd;
static HWND       second_wnd;
static char       *wnd_title        = "Planck Simulation";
static char       *wc_name          = "OpenGL";
static unsigned   locked_mouse;
static unsigned   keys[256];
static unsigned   active            = 0;
static int        mouse_x           = 0;
static int        mouse_y           = 0;
static unsigned   mouse_last_x      = 0;
static unsigned   mouse_last_y      = 0;
static unsigned   mouse_left        = 0;
static unsigned   mouse_right       = 0;
static unsigned   mouse_center      = 0;
static int        mouse_wheel       = 0;
static unsigned   mouse_calibrate   = 0;
static unsigned   mouse_moved       = 0;
static float      control_wait      = 0;

/* New for display.h: exposed function pointers. */
void (*display_set_stereo_mode)(unsigned int mode_number);
void (*display_set_clip_plane)(float gl_z_value);
void (*display_set_eye_seperation)(float value);
void (*display_set_focal_plane)(float gl_z_value);
void (*display_set_movement_acceleration)(float value);

/* Internal implemenation of above rendere.h pointers... */
static void set_stereo_mode(unsigned int mode_number);
static void set_clip_plane(float gl_z_value);
static void set_eye_seperation(float value);
static void set_focal_plane(float gl_z_value);
static void set_movement_acceleration(float value);


/* New for wgl: window management functions */
static LRESULT CALLBACK WndProc(HWND,UINT,WPARAM,LPARAM);
static HWND open_display(unsigned GL_stereo, char *name);
static void show_window(HWND win,int x, int y, int w, int h, unsigned border);
static void close_window(HWND);
static void show_error(char *msg);
static void lock_cursor(unsigned yes);
static void handle_events(void);



void display_prepare(int argc, char **argv)
{
   unsigned n = 1;

   mlInit();

   if (errno > 0) {
      perror("Startup errno cleanup; setting to zero, was");
      errno = 0;
   }

   while (n < argc) {
      if (argv[n] == strstr(argv[n],"hardware")) {
         app_mode_limit = HARDWARE_STEREO;
         break;
      }
      n++;
   }

   if (app_mode_limit == HARDWARE_STEREO) {
      if (!(main_wnd = open_display(1,wnd_title))) {
         printf("Could not create GL stereo window.\n");
         app_mode_limit = STEREO_PAIR;
      }
   }

   if (!main_wnd) {
      if (!(main_wnd = open_display(0,wnd_title))) {
         printf("Could not create window.\n");
         exit(1);
      }
   }

//    if (!(second_wnd = open_display(0,wnd_title))) {
//       printf("Could not create second window.\n");
//    }

   display_set_stereo_mode             = &set_stereo_mode;
   display_set_clip_plane              = &set_clip_plane;
   display_set_eye_seperation          = &set_eye_seperation;
   display_set_focal_plane             = &set_focal_plane;
   display_set_movement_acceleration   = &set_movement_acceleration;

   frame_state.stereo_modes            = app_mode_limit;
}

void display_set_context_renew_function(void_callback *ptr)
{
   app_context = ptr;
}

void display_set_preframe_function(void_callback *ptr)
{
   app_preframe = ptr;
}

void display_set_display_function(void_callback *ptr)
{
   app_draw = ptr;
}

static void set_stereo_mode(unsigned int mode_number)
{
   if (mode_number > app_mode_limit)
      return;

   view_mode = mode_number;
   if (view_mode == HORIZ_INTERLACED || view_mode == VERT_INTERLACED
      || view_mode == CHECKERBOARD)
      set_stencil(window_width,window_height);
   update_view();
}

static void set_clip_plane(float gl_z_value)
{
   if (gl_z_value <= 0)
      gl_z_value = .000001;

   near_clip = gl_z_value;
   update_view();
}

static void set_eye_seperation(float value)
{
   app_eye_seperation = value;
   update_view();
}

static void set_focal_plane(float gl_z_value)
{
   app_focal_length = gl_z_value;
   update_view();
}

static void set_movement_acceleration(float value)
{
   move_accel = value;
}

static void show_error(char *msg)
{
   printf("Win32 Error: %s\n",msg);
   /* MessageBox(NULL,msg,"RUNTIME ERROR",MB_OK|MB_ICONEXCLAMATION); */
}

static void lock_cursor(unsigned yes)
{
   RECT w;

   if (yes) {
      if (locked_mouse)
         return;

      if (GetWindowRect(main_wnd,&w)) {
         ClipCursor(&w);
         SetCapture(main_wnd);
         SetForegroundWindow(main_wnd);
         SetFocus(main_wnd);
         ShowCursor(0);
         locked_mouse = 1;
         mouse_moved = 1;
      }
   }
   else {
      if (!locked_mouse)
         return;

      ShowCursor(1);
      ReleaseCapture();
      ClipCursor(0);
      locked_mouse = 0;
      mouse_moved = 1;
   }
}


static LRESULT CALLBACK WndProc(HWND hWnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
   int x,y;
   RECT w;

   switch (uMsg) {

      case WM_ACTIVATE:
         if (!HIWORD(wParam))
            active = TRUE;
         else
            active = FALSE;
         return 0;

      case WM_SYSCOMMAND:
         switch (wParam) {
            case SC_SCREENSAVE:
            case SC_MONITORPOWER:
               return 0;
         }
         break;

      case WM_CLOSE:
         PostQuitMessage(0);
         return 0;

      case WM_KEYDOWN:
         frame_state.key = (int)wParam;
         keys[wParam] = 1;
         return 0;

      case WM_KEYUP:
         frame_state.key = 0;
         keys[wParam] = 0;
         return 0;

      case WM_MOUSEMOVE:
         x = (short int)LOWORD(lParam);
         y = (short int)HIWORD(lParam);

         if (x == mouse_last_x && y == mouse_last_y)
            return 0;

         mouse_x += x - mouse_last_x;
         mouse_y += y - mouse_last_y;
         mouse_last_x = x;
         mouse_last_y = y;
         mouse_moved  = 1;
         return 0;

      case WM_LBUTTONDOWN:
         mouse_left = 1;
         return 0;

      case WM_LBUTTONUP:
         mouse_left = 0;
         return 0;

      case WM_RBUTTONDOWN:
         mouse_right = 1;
         return 0;

      case WM_RBUTTONUP:
         mouse_right = 0;
         return 0;

      case WM_MBUTTONDOWN:
         mouse_center = 1;
         return 0;

      case WM_MBUTTONUP:
         mouse_center = 0;
         return 0;

      case WM_MOUSEWHEEL:
         mouse_wheel += ((short int)HIWORD(wParam) > 0) ? 1 : -1;
         return 0;

      case WM_SIZE:
         reshape(LOWORD(lParam),HIWORD(lParam));
      case WM_MOVE:
         if (locked_mouse && hWnd == main_wnd) {
            GetWindowRect(hWnd,&w);
            ClipCursor(&w);
         }
         return 0;
   }

   return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

static void handle_events(void)
{
   MSG msg;
   POINT p;

   while (PeekMessage(&msg,NULL,0,0,PM_REMOVE)) {

      if (msg.message == WM_QUIT) {
         app_run = 0;
         return;
      }
      else {
         TranslateMessage(&msg);
         DispatchMessage(&msg);
      }
   }

   if (active && keys[VK_ESCAPE]) {
      app_run = 0;
      return;
   }

   if (keys[VK_F1]) {
      keys[VK_F1] = 0;

      if (window_mode == FULLSCREEN_MODE) {
         window_mode = WINDOWED_MODE;
         show_window(
            main_wnd,
            origin_x + 30,
            origin_y + 30,
            framed_width,
            framed_height,
            1);
      }
      else {
         window_mode   = FULLSCREEN_MODE;
         framed_width  = window_width;
         framed_height = window_height;
         show_window(main_wnd,origin_x,origin_y,full_width,full_height,0);
      }
   }

   if (keys[VK_DELETE]) {
      keys[VK_DELETE] = 0;
      if (window_mode == WINDOWED_MODE) {
         set_mouse = !set_mouse;
         lock_cursor(set_mouse);
      }
   }

   if (mouse_moved && locked_mouse) {
      p.x = window_width/2;
      p.y = window_height/2;
      if (ClientToScreen(main_wnd,&p)) {
         SetCursorPos(p.x,p.y);
         mouse_last_x = window_width/2;
         mouse_last_y = window_height/2;
         mouse_moved = 0;
      }
   }

   if (control_wait) {
      control_wait -= dt;
      if (control_wait < 0)
         control_wait = 0;
   }
}


void display_run_loop(void)
{
   HDC  dc    = GetDC(main_wnd);
   HGLRC glrc = wglGetCurrentContext();

   app_run = 0;

   if (dc && glrc)
      app_run = 1;
   else
      printf("Error: no hardware display context.\n");

   show_window(main_wnd,origin_x,origin_y,full_width,full_height,0);
//   show_window(second_wnd,1600,0,full_width,full_height,0);

   /* Main rendering loop. */
   while (app_run) {

      handle_events();

      if (active) {
         pre_render();

//          if ((dc = GetDC(main_wnd)) && wglMakeCurrent(dc,glrc))   {
//             render();
//             SwapBuffers(dc);
//          }
//
//          if ((dc = GetDC(second_wnd)) && wglMakeCurrent(dc,glrc)) {
//             render();
//             SwapBuffers(dc);
//          }

         render();
         SwapBuffers(dc);
      }
   }
   /* Ending...clean up. */

   close_window(main_wnd);
//   close_window(second_wnd);
}

void display_shutdown(void)
{
   app_run = 0;
}

static void show_window(HWND win,int x, int y, int w, int h, unsigned border)
{
   DWORD style, ex_style;
   RECT wr;
   int wdt, hgt;

   if (border) {
      ex_style = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
      style    = WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
   }
   else {
      ex_style = WS_EX_APPWINDOW;
      style    = WS_POPUP;
   }

   wr.left   = x;
   wr.right  = x + w;
   wr.top    = y;
   wr.bottom = y + h;

   SetWindowLong(win,GWL_STYLE,style);
   SetWindowLong(win,GWL_EXSTYLE,ex_style);
   AdjustWindowRectEx(&wr,style,FALSE,ex_style);

   wdt = wr.right - wr.left;
   hgt = wr.bottom - wr.top;

   if (!SetWindowPos(win,0,wr.left,wr.top,wdt,hgt,(border)? SWP_DRAWFRAME: 0)) {
      printf("Window configuration not supported...\n");
      app_run = 0;
      return;
   }

   RedrawWindow(0,0,0,RDW_INVALIDATE | RDW_ERASENOW | RDW_ALLCHILDREN);

   lock_cursor(1);
   ShowWindow(win,SW_SHOW);
   SetForegroundWindow(win);
   SetFocus(win);

   if (app_context)
      app_context();
}

static HWND open_display(unsigned GL_stereo, char *name)
{
   RECT                    wr;
   HDC                     hDC;
   HGLRC                   hRC;
   HWND                    hWnd;
   HINSTANCE               hInstance = GetModuleHandle(NULL);
   WNDCLASS                wc;
   PIXELFORMATDESCRIPTOR   pfd;
   int                     pfd_id;
   unsigned                primary;

   if (fullscreen_discover && GetWindowRect(GetDesktopWindow(),&wr)) {
      full_width  = wr.right;
      full_height = wr.bottom;
   }

   memset(&pfd,0,sizeof(PIXELFORMATDESCRIPTOR));
   pfd.nSize        = sizeof(PIXELFORMATDESCRIPTOR);
   pfd.nVersion     = 1;
   pfd.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
   pfd.iPixelType   = PFD_TYPE_RGBA;
   pfd.cColorBits   = 32;
   pfd.cDepthBits   = 32;
   pfd.cStencilBits = 1;
   pfd.iLayerType   = PFD_MAIN_PLANE;

   if (GL_stereo)
      pfd.dwFlags |= PFD_STEREO;

   if (!GetClassInfo(hInstance,wc_name,&wc)) {

      memset(&wc,0,sizeof(WNDCLASS));
      wc.style          = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
      wc.lpfnWndProc    = (WNDPROC)WndProc;
      wc.cbClsExtra     = 0;
      wc.cbWndExtra     = 0;
      wc.hInstance      = hInstance;
      wc.hIcon          = LoadIcon(NULL, IDI_WINLOGO);
      wc.hCursor        = LoadCursor(NULL, IDC_ARROW);
      wc.hbrBackground  = NULL;
      wc.lpszMenuName   = NULL;
      wc.lpszClassName  = wc_name;

      if (!RegisterClass(&wc)) {
         show_error("Failed To Register The Window Class.");
         return 0;
      }
      primary = 1;
   }
   else
      primary = 0;

   hWnd = CreateWindowEx(0,wc_name,name,0,0,0,0,0,0,0,hInstance,0);
   if (!hWnd) {
      show_error("Window Creation Error.");
      return 0;
   }

   if (!(hDC = GetDC(hWnd)))  {
      show_error("Can't Create A GL Device Context.");
      return 0;
   }

   if (!(pfd_id = ChoosePixelFormat(hDC,&pfd))) {
      show_error("Can't Find A Suitable PixelFormat.");
      return 0;
   }

   if (!SetPixelFormat(hDC,pfd_id,&pfd))  {
      show_error("Can't Set The PixelFormat.");
      return 0;
   }

   if (primary) {

      if (!(hRC = wglCreateContext(hDC))) {
         show_error("Can't Create A GL Rendering Context.");
         return 0;
      }

      if (!wglMakeCurrent(hDC,hRC)) {
         show_error("Can't Activate The GL Rendering Context.");
         return 0;
      }
   }

   return hWnd;
}

static void close_window(HWND win)
{
   HDC         hDC;
   HGLRC       hRC;
   WNDCLASS    wc;
   HINSTANCE   hInstance = GetModuleHandle(NULL);

   if (!hInstance)
      return;

   lock_cursor(0);

   if (hRC = wglGetCurrentContext()) {

      if (!wglMakeCurrent(NULL,NULL))
         show_error("Release Of DC And RC Failed.");
      if (!wglDeleteContext(hRC))
         show_error("Release Rendering Context Failed.");
   }

   if (win && (hDC = GetDC(win))) {

      if (hDC && !ReleaseDC(win,hDC))
         show_error("Release Device Context Failed.");
      if (!DestroyWindow(win))
         show_error("Could Not Release hWnd.");
   }

   if (GetClassInfo(hInstance,wc_name,&wc)) {

      if (!UnregisterClass(wc_name,hInstance))
         show_error("Could Not Unregister Class.");
   }

   active = 0;
}

static void reshape(int width,int height)
{
   /* Resize display related objects. */
   window_width  = width;
   window_height = height;
   set_stencil(window_width,window_height);
   update_view();
}

static void set_stencil(int width,int height)
{
   switch (view_mode) {
      case HORIZ_INTERLACED:
         renderer_generate_stencil(width,height,RENDERER_HORIZONTAL);
         break;
      case VERT_INTERLACED:
         renderer_generate_stencil(width,height,RENDERER_VERTICAL);
         break;
      case CHECKERBOARD:
         renderer_generate_stencil(width,height,RENDERER_CHECKER);
         break;
      default:
         break;
   }
}


/* Functions which reverse left and right buffers for hardware stereo
rendering. This has the effect of reversing eye pairs. */

static void toggle_back_buffers()
{
   unsigned int temp = left_bottom;

   left_bottom       = right_bottom;
   right_bottom      = temp;
}

static void toggle_front_buffers()
{
   unsigned int temp = left_front;

   left_front        = right_front;
   right_front       = temp;
}

static void navigation(void)
{
   static int dx, dy;
   float up_down = 0, left_right = 0, move = 0;
   float coeff = (dt < .01) ? dt : dt/.3;
   static float *rot, *trans, rev = 0;
   unsigned int input = 0;

   dx     = last_x - mouse_x;
   last_x = mouse_x;
   dy     = last_y - mouse_y;
   last_y = mouse_y;

   if (mouse_left)
      ctl_button_set(5,input);
   if (mouse_right)
      ctl_button_set(4,input);

   frame_state.pointer[0] = -dx;
   frame_state.pointer[1] = dy;
   frame_state.button     = input;

   if (keys[VK_UP])
      up_down = move = move_scale;
   if (keys[VK_DOWN])
      up_down = move = -move_scale;
   if (keys[VK_RIGHT])
      left_right = move = move_scale;
   if (keys[VK_LEFT])
      left_right = move = -move_scale;

   rot   = frame_state.rotation;
   trans = frame_state.position;

   if (rot && trans) {

      mat33 tm, final;
      vec3 front      = {0,0,up_down};
      vec3 right_side = {-left_right,0,0};

      rev = (rot[1] - dx * look_scale)/TWO_PI;
      rot[1] = TWO_PI * ((rev >= 0) ? rev - (int)rev : 1 + rev - (int)rev);

      rev = (rot[0] - dy * look_scale)/TWO_PI;
      rot[0] = TWO_PI * ((rev >= 0) ? rev - (int)rev : 1 + rev - (int)rev);

      mat33_rotation(&final,-rot[1],0,1,0);
      mat33_rotation(&tm,-rot[0],1,0,0);
      mat33_mult(&final,&tm);

      vec3_transform(&front,&final);
      vec3_add((vec3*)trans,&front);

      vec3_transform(&right_side,&final);
      vec3_add((vec3*)trans,&right_side);

      move_scale = (move) ? (move_scale += move_accel * coeff) : 0.0001;
   }
}

static void pre_render(void)
{
   /* Time, metrics, and synchronization... */
   dt = frame_state.dt = platform_delta_timer();

   fps_dt_accum += frame_state.dt;
   frame_count++;
   if (fps_dt_accum >= FPS_UPDATE) {
      frame_state.fps = ceil(frame_count/fps_dt_accum);
      frame_count     = 0;
      fps_dt_accum    = 0;
   }

   navigation();

   if (app_preframe)
      app_preframe();

   #ifdef DEBUG_ALL
   {
      renderer_print_error();

      if (errno > 0) {
         perror("C runtime error detected");
         errno = 0;
      }

      printf(
         "eye seperation: %f  focal length: %f            \r",
         app_eye_seperation,
         app_focal_length);
   }
   #endif
}

/* This is called whenever any variable which affects the window / viewport
configuration changes. This can be a window resize, a stereo mode change,
an eye seperation modification, etc. */

static void update_view(void)
{
   /* The "lowview_rotate" rotates the low view (floor) cameras up
   so that the edge of the frustum meets the bottom of the top views.
   This is not complete--additional lower frustum warping is necessary.

   This partitions the window into view port areas based on the viewing
   configuration. */

   float aspect_ratio, frust_shift, lin_eye_sep;
   float flat_clip, clip_left, clip_right, clip_span, lowview_rotate;
   mat44 transform;
   vec4 lower, upper;

   renderer_use_stencil(RENDERER_NONE);

   switch (view_mode) {
      case HORIZ_INTERLACED:
      case VERT_INTERLACED:
      case CHECKERBOARD:
      case NORMAL_VIEWING:
      case HARDWARE_STEREO:
      case ANAGLYPH:
         view_width = window_width;
         view_height = window_height;
         break;
      case STEREO_QUAD:
         view_width = window_width/2;
         view_height = window_height/2;
         break;
      case STEREO_PAIR:
      case HARDWARE_QUAD:
         view_width = window_width/2;
         view_height = window_height;
         break;
   }

   aspect_ratio   = (float)view_height/view_width;
   lin_eye_sep    = -0.5 * app_eye_seperation;
   frust_shift    = lin_eye_sep * near_clip / app_focal_length;

   flat_clip      = tan(fov/2) * near_clip;
   clip_left      = -flat_clip + frust_shift;
   clip_right     = flat_clip + frust_shift;
   clip_span      = flat_clip * aspect_ratio;
   lowview_rotate = 2 * atan(clip_span/near_clip);


   mat44_perspective(
      &right_frustum,
      clip_left,clip_right,-clip_span,clip_span,near_clip,far_clip);
   mat44_translation(&transform,lin_eye_sep,0,0);
   mat44_mult(&right_frustum,&transform);

   right_floor_frustum = right_frustum;
   mat44_rotation(&transform,lowview_rotate,1,0,0);
   mat44_mult(&right_floor_frustum,&transform);

   mat44_perspective(
      &left_frustum,
      -clip_right,-clip_left,-clip_span,clip_span,near_clip,far_clip);
   mat44_translation(&transform,-lin_eye_sep,0,0);
   mat44_mult(&left_frustum,&transform);

   left_floor_frustum = left_frustum;
   mat44_rotation(&transform,lowview_rotate,1,0,0);
   mat44_mult(&left_floor_frustum,&transform);

   mat44_perspective(
      &center_frustum,
      -flat_clip,flat_clip,-clip_span,clip_span,near_clip,far_clip);


   switch (view_mode) {
      case NORMAL_VIEWING:
         min_viewport = VIEW_FRONT_CENTER;
         max_viewport = VIEW_FRONT_CENTER;
         break;
      case ANAGLYPH:
      case HORIZ_INTERLACED:
      case VERT_INTERLACED:
      case CHECKERBOARD:
         min_viewport = VIEW_FRONT_RIGHT;
         max_viewport =  VIEW_FRONT_LEFT;
         break;
      case HARDWARE_STEREO:
      case STEREO_PAIR:
         min_viewport = VIEW_FRONT_RIGHT;
         max_viewport = VIEW_FRONT_LEFT;
         break;
      case HARDWARE_QUAD:
      case STEREO_QUAD:
         min_viewport = VIEW_FRONT_RIGHT;
         max_viewport =  VIEW_BACK_LEFT;
         break;
   }

   /* Calculate the screen overlay configuration... */
   transform = center_frustum;
   mat44_invert(&transform);

   vec4_set(&lower,-1,-1, .998999);
   vec4_set(&upper, 1, 1, .998999);

   if (view_mode == STEREO_PAIR) {
      lower.y *= 1/aspect_ratio;
      upper.y *= 1/aspect_ratio;
   }

   vec4_transform(&lower,&transform);
   vec4_transform(&upper,&transform);
   vec4_scale(&lower,1/lower.w);
   vec4_scale(&upper,1/upper.w);

   /* Export  external properties... */
   frame_state.window_pixels[0]  = window_width;
   frame_state.window_pixels[1]  = window_height;
   frame_state.overlay_area[0]   = lower.x;
   frame_state.overlay_area[1]   = lower.y;
   frame_state.overlay_area[2]   = upper.x - lower.x;
   frame_state.overlay_area[3]   = upper.y - lower.y;
   frame_state.overlay_area[4]   = lower.z;
   frame_state.stereo_modes      = app_mode_limit;
   frame_state.current_mode      = view_mode;
   frame_state.stereo_mode_names = mode_names;
   frame_state.eye_seperation    = app_eye_seperation;
   frame_state.focal_plane       = app_focal_length;
   frame_state.clipping_plane    = near_clip;
}

static void render(void)
{
   int viewport = min_viewport;

   /* Begin main drawing section. */

   renderer_use_buffer(RENDERER_BACK);
   renderer_clear_buffer(RENDERER_COLOR);

   while (viewport <= max_viewport) {

      renderer_clear_buffer(RENDERER_DEPTH);

      /* Viewport and buffer setup */
      switch (viewport) {
         case VIEW_FRONT_CENTER:
         case VIEW_FRONT_RIGHT:
            switch (view_mode) {
               case HARDWARE_QUAD:
               case HARDWARE_STEREO:
                  renderer_use_buffer(left_front);
                  break;
            }
            renderer_viewport(0,0,view_width,view_height);
            break;
         case VIEW_BACK_RIGHT:
            switch (view_mode) {
               case HARDWARE_QUAD:
                  renderer_use_buffer(left_bottom);
                  renderer_viewport(view_width,0,view_width,view_height);
                  break;
               case STEREO_QUAD:
                  renderer_viewport(0,view_height,view_width,view_height);
                  break;
            }
            break;
         case VIEW_FRONT_LEFT:
            switch (view_mode) {
               case STEREO_PAIR:
               case STEREO_QUAD:
                  renderer_viewport(view_width,0,view_width,view_height);
                  break;
               case HARDWARE_STEREO:
               case HARDWARE_QUAD:
                  renderer_use_buffer(right_front);
                  renderer_viewport(0,0,view_width,view_height);
                  break;
            }
            break;
         case VIEW_BACK_LEFT:
            switch (view_mode) {
               case HARDWARE_QUAD:
                  renderer_use_buffer(right_bottom);
                  renderer_viewport(view_width,0,view_width,view_height);
                  break;
               case STEREO_QUAD:
                  renderer_viewport(
                     view_width,view_height,view_width,view_height);
                  break;
            }
            break;
      }
      /* End viewport and buffer setup. */

      /* Frustum setup */
      switch (viewport) {
         case VIEW_FRONT_CENTER:
            frame_state.projection_matrix = center_frustum;
            break;
         case VIEW_FRONT_RIGHT:
            frame_state.projection_matrix = right_frustum;
            break;
         case VIEW_BACK_RIGHT:
            frame_state.projection_matrix = right_floor_frustum;
            break;
         case VIEW_FRONT_LEFT:
            frame_state.projection_matrix = left_frustum;
            break;
         case VIEW_BACK_LEFT:
            frame_state.projection_matrix = left_floor_frustum;
            break;
      }
      /* End frustum setup. */

      /* Colormask and stencil setup */
      switch (viewport) {
         case VIEW_FRONT_CENTER:
            break;
         case VIEW_FRONT_RIGHT:
         case VIEW_BACK_RIGHT:
            if (view_mode == ANAGLYPH)
               renderer_use_colormask(RENDERER_RIGHT);
            if (view_mode == HORIZ_INTERLACED || view_mode == VERT_INTERLACED
               || view_mode == CHECKERBOARD)
               renderer_use_stencil(RENDERER_RIGHT);
            break;
         case VIEW_FRONT_LEFT:
         case VIEW_BACK_LEFT:
            if (view_mode == ANAGLYPH)
               renderer_use_colormask(RENDERER_LEFT);
            if (view_mode == HORIZ_INTERLACED || view_mode == VERT_INTERLACED
               || view_mode == CHECKERBOARD)
               renderer_use_stencil(RENDERER_LEFT);
            break;
      }
      /* End of colormask and stencil setup */

      renderer_set_projection(&frame_state.projection_matrix);
      renderer_set_modelview(&identity_matrix);


      if (app_draw)
         app_draw();

      viewport++;
   }
}

#ifdef __cplusplus
}
#endif
